/* service_monitor.c - Monitor sipd, and keep starting it, if it is exited.
 * We will run as a Win32 service.
 * (c) 2001. Columbia University.
 */

/* Parts of this code come under the following license:
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (C) 1993-1997  Microsoft Corporation.  All Rights Reserved.
*/
#include <windows.h>      /* basic Win32 includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <direct.h>       /* getcwd etc; */
#include <tchar.h>
#include <winreg.h>       /* Registry functions */
#include "service.h"

// when this event is signalled service should end.
HANDLE  hServerStopEvent = NULL;

static char sipd_key[] = /* Base Key for SIPD dir relative to HKLM */
                      "SOFTWARE\\Columbia University\\IRT\\sipd";

static char sipd_location[] = "SIPD_LOCATION"; /* relative to sipd_key */
  
static char sipd_command_line[128]  /* command line to pass to sipd 
                                     * For win32, we elect to keep a console
                                     * so, use -v and -X.
                                     */
          = "sipd.exe -v -X -s.";

static char sipd_path[512];     /* Full path of sipd.exe */
static char sipd_dir[512];      /* Dir of sipd i.e, sipd_path - sipd.exe */
static int sipd_startable = 1;
static PROCESS_INFORMATION pinfo; /* Win32 process information */
static STARTUPINFO sinfo;         /* Win32 console information */

static int InitSipdPath();

/* RunSipd
 *  Called during startup as a separate thread. Runs continously.
 *  Monitor sipd, and keep starting it for upto 20 times.
 *  Quit when Win32 SCM tells us to windup.
 */
void RunSipd() {
    HANDLE                  hEvents[1] = {NULL};
    DWORD                   dwWait;
    char *p = NULL;
    int retval, start_count = 0;
    DWORD exitCode = 0;           /* sipd's exit code */

    sinfo.cb = sizeof(STARTUPINFO);
    sinfo.lpTitle = "sipd";       /* sipd console's title */
  
    /* Initialize all parameters needed to run sipd such as directory,
     * path etc; While doing we should send SERVICE_START_PENDING to 
     * the SCM so that it will know we are booting.
     */
    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;

    // create the event object. The control handler function signals
    // this event when it receives the "stop" control code.
    //
    hServerStopEvent = CreateEvent(
        NULL,    // no security attributes
        TRUE,    // manual reset event
        FALSE,   // not-signalled
        NULL);   // no name

    if (hServerStopEvent == NULL)
        goto cleanup;

    hEvents[0] = hServerStopEvent;

    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;

    /* Get sipd path from registry, and see whether sipd is startable. */
    retval = InitSipdPath();
    if (retval != ERROR_SUCCESS) { /* ERROR_SUCCESS is 0 */
        AddToMessageLog("Unable to get sipd's location from registry...");
        sipd_startable = 0;
    }


    if (!ReportStatusToSCMgr( /* We are ready */
        SERVICE_RUNNING,    // service state
        NO_ERROR,           // exit code
        0))                 // wait hint
        goto cleanup;

    /* get sipd's base directory and chdir to that dir */
    strcpy(sipd_dir, sipd_path);
    p = strrchr(sipd_dir, '\\');
    if (p) 
      *p = '\0';


    retval = _chdir(sipd_dir);
    if (retval) {
      char buf[512];
      sprintf(buf, "Unable to chdir to %s %s", sipd_dir, strerror(errno));
      AddToMessageLog(buf);
      sipd_startable = 0;
    }

    /* Give time for Apache and MySQL services to start.
     * This done by sleeping for 3 seconds
     */
    dwWait = WaitForMultipleObjects(1,hEvents,FALSE,3000L);

    if (dwWait != WAIT_TIMEOUT) goto cleanup;

    while (1)
    {
         /* test whether sipd is running. if not spawn it */
         if (sipd_startable &&
             pinfo.dwProcessId == 0 && 
             start_count <20) { /* sipd is not running, try to start it */
           
            ++start_count;
           
            if (CreateProcess((LPCTSTR)sipd_path, sipd_command_line,
                NULL, NULL, FALSE, CREATE_NEW_CONSOLE, NULL, sipd_dir, &sinfo,
                &pinfo) == 0) {
                  DWORD err = GetLastError();
                  char buf[512];
                  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, err, 0, buf, sizeof(buf), NULL);
                  AddToMessageLog(buf);
            }
        }

        /* check every 1 minute and sleep */
        dwWait = WaitForMultipleObjects(1, hEvents, FALSE, 60000L);

        if (dwWait != WAIT_TIMEOUT) break; /* Win32 is asking us to stop */

        /* check whether sipd is running...if it is not, reset the pinfo struct */
        GetExitCodeProcess(pinfo.hProcess, &exitCode);
        if (exitCode == STILL_ACTIVE) { /* sipd is still running, dont bother */
          continue;
        }
        else { /* reset the pinfo struct, so that we can try starting it */
          memset(&pinfo, 0, sizeof(pinfo));
        }
    }

  cleanup:

    if (hServerStopEvent)
        CloseHandle(hServerStopEvent);

    /* if sipd is running, kill it */
    if (pinfo.hProcess != 0)
      TerminateProcess(pinfo.hProcess, 0);

    return;
}


/* ServiceStop 
 *  The main thread calls this when Win32 SCM asks us to stop.
 *  Signal the global event so that our sipd monitor thread exits.
*/
void ServiceStop() {
    if (hServerStopEvent)
        SetEvent(hServerStopEvent);
}

/* InitSipdPath
 *  Query the registry and find out where sipd is. 
 *  Then set it in the global variable.
*/
static int InitSipdPath() {
 
  char data[512];
  
  DWORD datalen = sizeof(data) - 1;
  DWORD type;
  
  int retval = -1;
  HKEY hKey;
     
  retval = RegOpenKeyEx (HKEY_LOCAL_MACHINE,    // Key handle at root level.
                         sipd_key,              // Path name of child key.
                         0,                     // Reserved.
                         KEY_EXECUTE,           // Requesting read access.
                         &hKey);                // Address of key to be returned.

  if (retval == ERROR_SUCCESS) {
    if ((retval = RegQueryValueEx(hKey, sipd_location, NULL, &type, 
       (unsigned char*)data, &datalen)) == ERROR_SUCCESS) {
      if (type == REG_SZ) {
        strncpy(sipd_path, data, sizeof(sipd_path));
        retval = 0;
      }
    }
    RegCloseKey(hKey);
  }
  return retval;
} 
    