/* afi.c : AF independent functions.
 * Copyright 2001 Columbia University. All rights reserved.
 *  Parts of the code were developed by Timo Ohtonen/Summer 2001.
*/

/* For Windows, print a message prompting the user to test
 * the size of (struct sockaddr_in6). This is because pre-1998
 * versions have a wrong size. XXX: do auto-test
*/
#if defined(WIN32) && !defined(NO_INET6)
#pragma message ("Not testing sizeof(struct sockaddr_in6) in ws2tcpip.h")
#pragma message ("Make sure it is up to date and equals 28")
#endif

#include <string.h>     /* strlen, ... */
#include <unistd.h>     /* close, ... */
#include <memory.h>     /* memset(), ... */
#include <stdlib.h>     /* free(), ... */
#include <stdio.h>      /* NULL, ... */
#include <netinet/in.h> /* sockaddr_inX, ... */
#include <arpa/inet.h>  
#include <sys/socket.h>
#include <netdb.h>      /* inet_ntoa etc; */
#include "error.h"      /* error() */
#include "afi.h"       

/* windows doesnt have in6addr_any.
 * so is any system that doesnt have ipv6, so that autoconf
 * defined NO_INET6
*/
#if defined(WIN32) || defined (NO_INET6)
const struct in6_addr in6addr_any;
#endif

static char* AddrToHost(const struct sockaddr* sin, int numeric);
static int HostToAddr(struct sockaddr_storage *sin, const char* hostname);

static int hasv6 = 0;          /* does this machine have v6 */
static int usev6 = 1;          /* does the user wants v6 */

/* startupSocket
 *   Check whether we have configured IPv6 interfaces that can be used.
 *   + do platform specific socket initialization. 
*/
void startupSocket() {
  int s;
#ifdef WIN32
  WSADATA wsaData;  
  if (WSAStartup(0x0202, &wsaData))  {
    printf("Could not open start up windows socket.\n"); 
    return;
  }
#endif

  s = socket(AF_INET6, SOCK_DGRAM, 0);
  if (s>=0) {
    hasv6 = 1;
    closesocket(s);
  }

  return;
}

/* AFIIPv6Active
 *  v6 is active if host is capable and user wants it.
*/
int AFIIPv6Active() {
  return (hasv6 && usev6);
}

/* AFIUseIPv6
 *  Allow user to turn on/off v6.
*/
void AFIUseIPv6(int use) {
 usev6 = use;  
}

/* AFIGetAddrByName
 *  Return a usable ip address for this host. fill the 
 *  sin structure with that address and port. while calling
 *  fill the sin->sa_family to AF_INET or AF_INET6 to get 
 *  the appropriate address. A value of 0 for family will
 *  cause this function to first search for AF_INET and then
 *  AF_INET6.
*/
int AFIGetAddrByName(struct sockaddr_storage* sin,
                     const char* hostname) {

  if (!sin || !hostname) {
    error("AFIGetAddrByName", "Bad call...params must be non null\n");
    return -1;
  }

  return HostToAddr(sin, hostname);
} /* AFIGetAddrByName */


/* 
 * AFIInitAny
 *  Given an empty sin struct, initialize the sin struct to 'any'
 *  defaults. Also, set the port number in sin to htons(port). 
 *  
 * Parameters
 *  address_family: AF_INET or AF_INET6. 0 will default to AF_INET.
 *  sin           : the structure to initialize.
 *  port          : port number to use (in host byte order)
 */
void AFIInitAny(int address_family, struct sockaddr_storage* sin, 
                in_port_t port) {

  if (!sin) {
    error("InitAnyHost", "Bad call...sin must be non null\n");
    return;
  }
  
  memset(sin, 0, sizeof(struct sockaddr_storage));

  if (!address_family) {
    address_family = AF_INET;
  }

  sin->ss_family = address_family;
    
  if (address_family==AF_INET6) {
    ((struct sockaddr_in6 *)sin)->sin6_addr = in6addr_any;
    ((struct sockaddr_in6 *)sin)->sin6_port = htons(port);
  }
  else if (address_family==AF_INET) {
      ((struct sockaddr_in *)sin)->sin_addr.s_addr = htonl(INADDR_ANY);
      ((struct sockaddr_in *)sin)->sin_port = htons(port);
  }

  return;
  
} /* AFIInitAny */


/* AFIGetOurAddr
 *  Return a usable ip address for this host. fill the sin
 *  structure with that address.
*/
int AFIGetOurAddr(struct sockaddr_storage* sin) {
  
  char hostname[1024];
  int i = 0;
  
  gethostname(hostname, sizeof(hostname));

  if (sin->ss_family == AF_INET6 &&
      !hasv6) /* for nonv6 machines, dont even try it */
    return -1;

  i = AFIGetAddrByName(sin, hostname);
  
  return i;
  
  /* if i <0, the hostname was not resolvable into a usable ip.
   * this may be because: 
   *  1) the caller asked for a address_family say ipv6 
   *     that this host does not have.
   *  2) this host has a different hostname for the ipv6
   *  interface. for example, thalys.cs.columbia.edu for v4
   *  and thalys.cs.ip6.columbia.edu for v6. XXX: find
   *  an elegant solution for this problem without introducing
   *  two hostnames.
  */
  
} /* AFIGetOurAddr */


/* AFIGetOurName
 *  Return the canonical host name for this host. 
 *
 *  The first parameter is the address family. If it is 
 *  zero, this function will assume AF_INET. The parameter
 *  is needed, because, IPv4 and IPv6 names may be
 *  different. 
 *
 *  Note that if af is AF_INET6 the function may return
 *  NULL if no IPv6 interfaces are configured, and so
 *  no name is obtainable.
*/
char* AFIGetOurName(int af) {

 struct sockaddr_storage sin;

 memset(&sin, 0, sizeof(sin));
 sin.ss_family = af;

 if (!af || !hasv6)  /* here, we can default to v4 if v6 is not there */
   sin.ss_family = AF_INET;

 if (AFIGetOurAddr(&sin) == 0) {
   return AFIGetCanonicalHostName((struct sockaddr*)&sin);
 }

 return NULL;
} 


/* AFIGetOurDottedAddress
 *  Return the dotted address for this host. 
 *
 *  The first parameter is the address family. If it is 
 *  zero, this function will assume AF_INET. The parameter
 *  is needed, because, IPv4 and IPv6 names may be
 *  different. 
 *
 *  Note that if af is AF_INET6 the function may return
 *  NULL if no IPv6 interfaces are configured, and so
 *  no name is obtainable.
*/
char* AFIGetOurDottedAddress(int af) {
 struct sockaddr_storage sin;

 sin.ss_family = af;

 if (!hasv6 && af == AF_INET6)
   return NULL;

 if (!af) 
   sin.ss_family = AF_INET;

 if (AFIGetOurAddr(&sin) == 0) {
   return AFIGetDottedAddress((struct sockaddr*)&sin);
 }

 return NULL;
} 


/* AFIIsAnyHost
 *  Return 1 if sin->sinX_addr == inaddrX_any. while
 * calling sa_family must be set.
*/
int AFIIsAnyHost(const struct sockaddr* sin) {

  if (((sin->sa_family == AF_INET || sin->sa_family == 0) && 
     ((struct sockaddr_in *)sin)->sin_addr.s_addr == INADDR_ANY)
    ||
      (sin->sa_family == AF_INET6 &&
       memcmp(&(((struct sockaddr_in6 *)sin)->sin6_addr), 
        &in6addr_any, sizeof(in6addr_any)) == 0) ) {
    return 1;
  }

  return 0;
}

/* AFICompareAddr
 *  Compare two struct sockaddr.sX_addr's and return 1 if the addresses
 *  are equal.
*/
int AFICompareAddr(const struct sockaddr * s1,
                   const struct sockaddr * s2) {

  if (s1->sa_family != s2->sa_family ) { 
    return 0; 
  }

  if (s1->sa_family == AF_INET) {
    if (((struct sockaddr_in *)s1)->sin_addr.s_addr ==
        ((struct sockaddr_in *)s2)->sin_addr.s_addr ) {
      return 1;
    }
  }
  else if (s1->sa_family == AF_INET6) {
    if (memcmp(&((struct sockaddr_in6 *)s1)->sin6_addr,
               &((struct sockaddr_in6 *)s2)->sin6_addr,
         sizeof(((struct sockaddr_in6 *)s1)->sin6_addr)) == 0) {
      return 1;
    }
  }
  return 0;
} /* AFICompareAddr */


/* AFICompareSockAddr
 *  Compare two struct sockaddr's and return 1 if their addresses and
 *   ports match.
*/
int AFICompareSockAddr(const struct sockaddr * s1,
                       const struct sockaddr * s2) {
  
  int val = AFICompareAddr(s1, s2);
  if (!val)
    return val;

  if (s1->sa_family == AF_INET) {
    if (((struct sockaddr_in *)s1)->sin_port ==
        ((struct sockaddr_in *)s2)->sin_port ) {
      return 1;
    }
  }
  else if (s1->sa_family == AF_INET6) {
    if (((struct sockaddr_in6 *)s1)->sin6_port ==
        ((struct sockaddr_in6 *)s1)->sin6_port) {
      return 1;
    }
  }

  return 0;
} /* AFICompareSockAddr */


/* AFICopyInAddr
 *  Copy the sinX_addr part of src to dst { X in (4,6) }.
 *
 * Preconditions
 *  1) the address family for both should be set on call.
 *  2) both parameters must be non-null.
 *  3) src and dst can be either {v4|v6}. they need not belong to same 
 *  family. if they are of different families, destination gets only 
 *  32bits of address (LSB).
 */
void AFICopyInAddr(struct sockaddr* dst, const struct sockaddr* src) {

  void *copyTo;
  void *copyFrom;
  size_t copySize;
  int ip6 = 0;      /* Are both ip6 addresses? */

  if (dst->sa_family == AF_INET6) {
    copyTo = (void*) &((struct sockaddr_in6*)dst)->sin6_addr;
    ip6 = 1;  /* assume both are ip6 addresses */
  }
  else {
    copyTo = (void*)&(((struct sockaddr_in*)dst)->sin_addr);
  }
  
  if (src->sa_family == AF_INET6) {
    copyFrom = (void*)&(((struct sockaddr_in6 *)src)->sin6_addr);
  }
  else {
    copyFrom = (void*)&(((struct sockaddr_in *)src)->sin_addr);
    ip6 = 0;  /* atleast one is not an ip6 address */
  }

  if (ip6) 
    copySize = sizeof(struct in6_addr);
  else
    copySize = sizeof(struct in_addr);
  
  memcpy(copyTo, copyFrom, copySize);
  
  return;
} /* AFICopyInAddr */


/*
 * AFICopyPort
 *  Copy src {v4|v6} port to dst {v4|v6} port.
 *  src and dst can be either of v4 or v6 (not necessarily the same).
 *  
 *  Precondition: dst and src non null.
*/ 
void AFICopyPort(struct sockaddr* dst, const struct sockaddr* src) {

  /* This can be used directly, since the port variable
   * is at the same location in IPv6 and IPv4 structures.
   * If other address families are introduced, this needs to
   * be changed. XXX: this is not clean.
  */

  ((struct sockaddr_in *)dst)->sin_port =
    ((struct sockaddr_in *)src)->sin_port;
} /* AFICopyPort */


/*
 * AFICopySockAddr
 *  Copies the src->sinX_addr to dst->sinX_addr and 
 *  src->sinX_port to dst->sinX_port.
 *
 * Precondition:
 *  Both src and dst are non null.
 *  Both are in {v4 | v6} INET family. can be different in which case
 *  only the lower most 32bits of src ip address will be copied to the 
 *  dest ip.
*/
void AFICopySockAddr(struct sockaddr* dst, const struct sockaddr* src) {
  AFICopyInAddr(dst, src);
  AFICopyPort(dst, src);
} /* AFICopySockAddr */


/*
 * AFIGetCanonicalHostName
 *  Return a hostname for the ipv4 or ipv6 address.
 *
 * Parameters
 *  sin is a parameter that contains a sin4 or sin6 address.
 *  the address family parameter should be set to AF_INET or AF_INET6.
 * 
 * Return values
 *  Valid host name or NULL. Once usage is finished, call the FreeHostName
 *  function with the two parameters to cleanup resources.
*/
char* AFIGetCanonicalHostName(const struct sockaddr* sin) {
  return AddrToHost(sin, 0);
} /* AFIGetCanonicalHostName */


/* AFITranslateNumeric
 *  Convert the dotted address into sockaddr_storage. Will
 * only try to convert the address. Returns -1 if it is
 * not a valid IP v4/v6 address. On return, the family
 * will also be set (AF_INET/AF_INET6).
 */
int AFITranslateNumeric(struct sockaddr_storage* sin,
                        const char* dotaddr) {

  struct sockaddr_in *p4;
  
  if (strchr(dotaddr, ':') != NULL) { /* may be a ipv6 address */
    struct sockaddr_in6 *p6 = (struct sockaddr_in6*)sin;
    int retval = inet_pton(AF_INET6, dotaddr, &p6->sin6_addr);
    p6->sin6_family = AF_INET6;
    return (retval == 1) ? 0 : -1;
  }

  /* test for ipv4 */
  p4 = (struct sockaddr_in*) sin;
  p4->sin_addr.s_addr = inet_addr(dotaddr);
  if (p4->sin_addr.s_addr == -1)
    return -1;
  else {
    p4->sin_family = AF_INET;
    return 0;
  }
} /* AFITranslateNumeric */


/*
 * AFIGetDottedAddress
 *  Return a dotted decimal (. or :) address for the ipv4 or ipv6 address.
 *
 * Parameters
 *  sin is a parameter that contains a sin4 or sin6 address.
 *  the address family parameter should be set to AF_INET or AF_INET6.
 * 
 * Return values
 *  Valid dd address string or null. Use free to free the string
 *  after use.
*/
char* AFIGetDottedAddress(const struct sockaddr* sin) {
  return AddrToHost(sin, 1);
} /* AFIGetDottedAddress */


/* AddrToHost
 *  Get canonical host name or ip address depending on numeric is 0 or not.
*/
static char* AddrToHost(const struct sockaddr* sin, int numeric) {

  int sinlen;
  int af;
  int retval = -1;
  char retbuf[NI_MAXHOST];
  
  if (!sin) {
    error("AddrToHost", "Bad call...param null\n");
    return NULL;
  }

  if ((af = sin->sa_family) == AF_INET6) { 
    sinlen = sizeof(struct sockaddr_in6); 
  }
  else if (af == AF_INET) { 
    sinlen = sizeof(struct sockaddr_in); 
  }
  else { 
    error("AddrToHost", "Called with bad addr family %d\n", sin->sa_family);
    return NULL; 
  }

#ifdef HAVE_GETNAMEINFO
  {
   int flags = 0;
   
   if (numeric)
     flags = NI_NUMERICHOST;
  
#ifdef HAVE_SA_LEN /* FreeBSD's getnameinfo will work only 
                      if sin->salen is set. 
                   */
   sin->sa_len = sinlen;
#endif

   retval = getnameinfo(sin, sinlen, retbuf, NI_MAXHOST, NULL, 0, flags);
  }
#endif

  /* here if retval is 0, then getnameinfo acted on it, and returned
   * a usable result. if retval != 0, then either getnameinfo is not
   * there, or it failed (FreeBSD getnameinfo fails, mostly
   * because i think servname is NULL). use inet_ntop.
  */
  if (retval != 0) {
    void *addr;
   
    if (af == AF_INET) {
      struct sockaddr_in *p4 = (struct sockaddr_in*)sin;
      addr = &p4->sin_addr;
    }
    else {
      struct sockaddr_in6 *p6 = (struct sockaddr_in6*)sin;
      addr = &p6->sin6_addr;
    }

    retval = (inet_ntop(af, addr, retbuf, NI_MAXHOST) != NULL) ? 0 : -1;
  }

  if (!retval) 
    return strdup(retbuf);
  else
    return NULL;
}

static int HostToAddr(struct sockaddr_storage *sin, const char* hostname) {

#ifdef HAVE_GETADDRINFO /* Win32 also has getaddrinfo for ip6*/
  {
    struct addrinfo hints, *result;
    int retval;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = sin->ss_family;
    if (hints.ai_family == 0) { /* try A, AAAA - simple by recursion */
      int retval;
      sin->ss_family = AF_INET;
      retval = HostToAddr(sin, hostname);
      if (retval == 0)
        return retval;
      else {
        sin->ss_family = AF_INET6;
        return HostToAddr(sin, hostname);
      }
    }
    if ((retval = getaddrinfo(hostname, NULL, &hints, &result)) == 0) {
      int port = AFIGetPort(sin);
      memcpy(sin, result->ai_addr, result->ai_addrlen);
      if (AFIGetPort(sin) != port) {
        AFISetPort(sin, port);
      } 
      freeaddrinfo(result);
      return 0;
    }
    else {
      if (retval == EAI_FAMILY && hasv6) { /* this may be because ipnodes
                                            * was not configured 
                                            */
        debug("HostToAddr", "Warning: Machine has IPv6 but i am unable"
                            "to resolve %s\n", hostname);
       /* try resparse and contact the DNS for it ... this is a stop
         * gap arrangement
        */
      }
    
      /* XXX: print error message here by calling gai_strerror */
      return -1;
    }
  }
#else /* older systems that dont have getaddrinfo */
  struct hostent * hent = gethostbyname(hostname);

  if (sin->ss_family == AF_INET6)
    return -1; /* gethostbyname cant resolve AF_INET6 */
 
  if (hent != NULL) { 
    memcpy(&((struct sockaddr_in *)sin)->sin_addr, 
         hent->h_addr, hent->h_length);
  }
  else {
    ((struct sockaddr_in *)sin)->sin_addr.s_addr = inet_addr(hostname);
    sin->ss_family = AF_INET; /* just to be safe */
  }
  
  return 0;
#endif /* !HAVE_GETADDRINFO */
}


/* AFIGetUDPServerSocket
 *  Given a sin { v4 or v6 } structure, create a datagram socket and 
 *  bind this struct to it. return the socket descriptor.
 */
int AFIGetUDPServerSocket(struct sockaddr_storage* sin) {

  int s, sinlen, on = 1;

  if ((s=socket(sin->ss_family, SOCK_DGRAM, IPPROTO_UDP)) <0)  {
    return -1;
  }

#ifdef SO_REUSEADDR
  setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(const char*)&on,sizeof(on));
#endif

#ifdef SO_REUSEPORT
  setsockopt(s,SOL_SOCKET,SO_REUSEPORT,(const char*)&on,sizeof(on));
#endif
    
  sinlen = sizeof(struct sockaddr_in);

  if (sin->ss_family == AF_INET6)
    sinlen = sizeof(struct sockaddr_in6);
  
  if (bind(s, (struct sockaddr*)sin, sinlen) != 0) {     
    closesocket(s);    
    return -1;
  }

  return s;
}


/* AFIGetTCPServerSocket
 *  Given a sin { v4 or v6 } structure create a tcp socket and bind this
 *  struct to it. return the socket descriptor. will also call listen on
 *  this socket.
*/
int AFIGetTCPServerSocket(struct sockaddr_storage* sin) {

  int s, sinlen, on = 1;

  if ((s=socket(sin->ss_family, SOCK_STREAM, IPPROTO_TCP)) < 0)  {
    return -1;
  }

#ifdef SO_REUSEADDR
  setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(const char*)&on,sizeof(on));
#endif

#ifdef SO_REUSEPORT
  setsockopt(s,SOL_SOCKET,SO_REUSEPORT,(const char*)&on,sizeof(on));
#endif
    
  sinlen = sizeof(struct sockaddr_in);

  if (sin->ss_family == AF_INET6)
    sinlen = sizeof(struct sockaddr_in6);
  
  if (bind(s, (struct sockaddr*)sin, sinlen) != 0) {     
    closesocket(s);    
    return -1;
  }

  listen(s, 1000);

  return s;
}

int AFISetTTL(int sockid, unsigned char ttl) {
 return setsockopt(sockid, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));
}

int AFIMulticastLoopBack(int sockid, unsigned char on) {

  return setsockopt(sockid, IPPROTO_IP, 
                    IP_MULTICAST_LOOP, &on, sizeof(on));
}

int AFIIsMulticast(struct sockaddr_storage *sin) {

  if (!sin)
    return -1; /* this will cause caller to segfault mostly!!! */

  if (sin->ss_family == AF_INET) {
    struct sockaddr_in* p4 = (struct sockaddr_in*)sin;
    return IN_MULTICAST(ntohl(p4->sin_addr.s_addr));
  }
  else {
    struct sockaddr_in6* p6 = (struct sockaddr_in6*)sin;
    return IN6_IS_ADDR_MULTICAST(&(p6->sin6_addr));
  }
}

int AFIReuseAddr(int sockid) {

  int one = 1;

#ifdef SO_REUSEADDR
  return setsockopt(sockid, SOL_SOCKET, SO_REUSEADDR, 
                    (char*)&one, sizeof(one));
#else
  return 0;
#endif

}

int AFIReusePort(int sockid) {
  int one = 1;

#ifdef SO_REUSEPORT
  return setsockopt(sockid, SOL_SOCKET, SO_REUSEPORT, 
                    (char*)&one, sizeof(one));
#else
  return 0;
#endif

}

/*
 * AFIJoinMulticastGroup
 *  Joins the multicast group given in mcast_addr for the socket sockid.
 *  Returns 0 on success, -1 on error.
 *   Currently implemented for in_addr only - set the sockaddr_storage.
 *   ss_family to AF_INET, and sin_addr should contain the IP address.
 */
int AFIJoinMulticastGroup(int sockid, struct sockaddr_storage* mcast_addr) {

  struct ip_mreq mreq;
  struct sockaddr_in *p4;

  if (!mcast_addr) return -1;
  
  if (mcast_addr->ss_family != AF_INET) {
    error("AFIJoinMulticastGroup", "Unsupported address family %d\n",
          mcast_addr->ss_family);
    return -1;
  }

#ifdef IP_ADD_MEMBERSHIP
  p4 = (struct sockaddr_in*) mcast_addr;

  memset(&mreq, 0, sizeof(mreq));

  mreq.imr_multiaddr.s_addr =  p4->sin_addr.s_addr;
  mreq.imr_interface.s_addr = htonl(INADDR_ANY);

  return setsockopt(sockid, IPPROTO_IP, IP_ADD_MEMBERSHIP, 
                    (const char*)&mreq, sizeof(mreq));
#else
  return 0;
#endif
}

