/* afi.h - AF Implementation of network i/o functions.
 * 
 * Includes IPv6 headers and defines functions for address
 * family independent manipulation of socket addresses.
 *
 *  Part of the code developed by Timo Ohtonen/Summer 2001.
 *  
 * Copyright 2001 by Columbia University; all rights reserved.
 */
#ifndef __IPV6HDRS_H__
#define __IPV6HDRS_H__

#include <sys/socket.h>       /* socket, ... */
#include <netdb.h>            /* AF_INET, ... */
#include <netinet/in.h>       /* sockaddr_in, ... */
#include "sysdep.h"

/* For Windows, if the user defines NO_INET6, then compile
 * without ipv6 support.
*/
#if defined(WIN32) && !defined(NO_INET6)
#include <ws2tcpip.h>   
#include <tpipv6.h>  
#define HAVE_GETADDRINFO 1
#define HAVE_GETNAMEINFO 1
typedef unsigned short in_port_t;
#endif

#if defined(WIN32) && defined(NO_INET6)
#pragma message ("NO_INET6 defined...compiling without IPv6 support")
#undef HAVE_GETADDRINFO
#undef HAVE_GETNAMEINFO
#endif /* user doesnt want inet6 */

#ifdef NO_INET6
#include "types.h"
/* Non ipv6 machine...but define it for compatibility */

#define AF_INET6 23

typedef struct sockaddr_storage {
  int16 ss_family;
  char padding[126];
} sockaddr_storage;

typedef struct in6_addr {
  union {
    u_char _S6_u8[16];
    u_long _S6_u32[4];
    int64 _S6_u64[2];
  } _S6_un;
} in6_addr;

#define s6_addr _S6_un._S6_u8

typedef struct sockaddr_in6 {
  u_short sin6_family;
  u_short sin6_port;
  u_long sin6_flowinfo;
  struct in6_addr sin6_addr;
  u_long sin6_scope_id;
} sockaddr_in6;

#endif /* NO_INET6 */

extern const struct in6_addr in6addr_any;

#ifndef IN6_IS_ADDR_MULTICAST
#ifdef WORDS_BIGENDIAN
#define IN6_IS_ADDR_MULTICAST(addr) \
   (((addr)->_S6_un._S6_u32[0] & 0xff000000) == 0xff000000)
#else /* BIG_ENDIAN */
#define IN6_IS_ADDR_MULTICAST(addr) \
   (((addr)->_S6_un._S6_u32[0] & 0x000000ff) == 0x000000ff)
#endif /* BIG_ENDIAN */
#endif

#ifndef IN6_IS_ADDR_UNSPECIFIED
#define IN6_IS_ADDR_UNSPECIFIED(addr) \
         (((addr)->_S6_un._S6_u32[3] == 0) && \
         ((addr)->_S6_un._S6_u32[2] == 0) && \
         ((addr)->_S6_un._S6_u32[1] == 0) && \
         ((addr)->_S6_un._S6_u32[0] == 0))
#endif
 
#ifndef NI_MAXHOST
#define NI_MAXHOST 1025
#endif

/* startupSocket
 *  Call this function during program startup. This function
 *  detects IPv6 capabilities of the running system. In addition,
 *  it initializes Winsock for windows.
*/
extern void startupSocket();

/* AFIIPv6Active
 *  Returns 1 if IPv6 is usable.
*/
extern int AFIIPv6Active();

/* AFIUseIPv6
 *  Turn on/off use of IPv6. To be called during config
 *  time after startupSocket. This function affects the
 *  value returned by AFIIPv6Active.
 *
 * Sometimes, a machine will have v6 installed, but there
 * will be no usable addresses. Higher level modules
 * should detect this condition, and use this function
 * to turn v6 off.
*/
extern void AFIUseIPv6(int use);

/*
 * AFIGetCanonicalHostName
 *  Return a hostname for the ipv4 or ipv6 address.
 * (XXX: this is different from old style inet_ntoa).
 *
 * Parameters
 *  sin is a parameter that contains a sin4 or sin6 address.
 *  the address family parameter should be set to AF_INET or AF_INET6.
 * 
 * Return values
 *  Valid host name or NULL. Free the string after use.
*/
extern char* AFIGetCanonicalHostName(const struct sockaddr* sin);


/* AFIGetOurDottedAddress
 *  Return the dotted address for this host. 
 *
 *  The first parameter is the address family. If it is 
 *  zero, this function will assume AF_INET. The parameter
 *  is needed, because, IPv4 and IPv6 names may be
 *  different. 
 *
 *  Note that if af is AF_INET6 the function may return
 *  NULL if no IPv6 interfaces are configured, and so
 *  no name is obtainable.
*/
extern char* AFIGetOurDottedAddress(int af);


/*
 * AFIGetDottedAddress
 *  Return a dotted decimal (. or :) address for the ipv4 or ipv6 address.
 * (XXX: this is different from old style inet_ntoa).
 *
 * Parameters
 *  sin is a parameter that contains a sin4 or sin6 address.
 *  the address family parameter should be set to AF_INET or AF_INET6.
 * 
 * Return values
 *  Valid dd address string or null. Free after using.
*/
extern char* AFIGetDottedAddress(const struct sockaddr* sin);


/* AFIGetOurName
 *  Return the canonical host name for this host. Free
 *  after using (XXX: this is different from old style
 *  inet_ntoa).
 *
 *  The first parameter is the address family. If it is 
 *  zero, this function will assume AF_INET. The parameter
 *  is needed, because, IPv4 and IPv6 names may be
 *  different. 
 *
 *  Note that if af is AF_INET6 the function may return
 *  NULL if no IPv6 interfaces are configured, and so
 *  no name is obtainable.
*/
extern char* AFIGetOurName(int af);


/* AFIGetOurAddr
 *  Return a usable ip address for this host. fill the sin
 *  structure with that address and the port. Address
 * family can be optionally filled with AF_INET or AF_INET6
 *
*/
extern int AFIGetOurAddr(struct sockaddr_storage* sin);


/* AFIGetAddrByName
 *  Return a usable ip address for this host (host name or
 *  dotted decimal address). fill the 
 *  sin structure with that address and port. while calling
 *  fill the sin->sa_family to AF_INET or AF_INET6 to get 
 *  the appropriate address. A value of 0 for family will
 *  cause this function to default to search for any usable
 *  IP (first tries v4 and then v6).
*/
extern int AFIGetAddrByName(struct sockaddr_storage* sin,
                            const char* hostname);


/* AFITranslateNumeric
 *  Convert the dotted address into sockaddr_storage. Will
 * only try to convert the address. Returns -1 if it is
 * not a valid IP v4/v6 dotted address. On return, the family
 * will also be set (AF_INET/AF_INET6). This function replaces
 * calls to inet_addr.
 */
extern int AFITranslateNumeric(struct sockaddr_storage* sin,
                               const char* dotaddr);

/* AFIIsAnyHost
 *  Return 1 if sin->sinX_addr == inX_addrany. while calling 
 * sa_family must be set (AF_INET4|6).
*/
extern int AFIIsAnyHost(const struct sockaddr* sin);


/* AFICompareAddr
 *  Compare two struct  sockaddr.sX_addr's and return 1 if they are equal.
*/
extern int AFICompareAddr(const struct sockaddr * s1,
                          const struct sockaddr * s2);

/* AFICompareSockAddr
 *  Compare two struct sockaddr's and return 1 if their addresses and
 *   ports match.
*/
extern int AFICompareSockAddr(const struct sockaddr * s1,
                              const struct sockaddr * s2);
  

/* AFICopyInAddr
 *  Copy the sinX_addr part of src to dst { X in (4,6) }.
 *
 * Preconditions
 *  1) the address family for both should be set on call.
 *  2) both parameters must be non-null.
 *  3) src and dst can be either {v4|v6}. they need not belong to same 
 *  family. if they are of different families, destination gets only 
 *  32bits of address (LSB).
 */
extern void AFICopyInAddr(struct sockaddr* dst, const struct sockaddr* src);

/*
 * AFICopySockAddr
 *  Copies the src->sinX_addr to dst->sinX_addr and 
 *  src->sinX_port to dst->sinX_port.
 *
 * Precondition:
 *  Both src and dst are non null.
 *  Both are in {v4 | v6} INET family. can be different in which case
 *  only the lower most 32bits of src ip address will be copied to the 
 *  dest ip.
*/
extern void AFICopySockAddr(struct sockaddr* dst, const struct sockaddr* src);

/*
 * AFICopyPort
 *  Copy src {v4|v6} port to dst {v4|v6} port.
 *  src and dst can be either of v4 or v6 (not necessarily the same).
 *  
 *  Precondition: dst and src non null.
*/ 
extern void AFICopyPort(struct sockaddr* dst, const struct sockaddr* src);


/* AFISetPort
 *  Assign port to p_sin->sinX_port. sin should be one of
 *  sockaddr_XXX types.
 */
#define AFISetPort(p_sa, port) \
  ((struct sockaddr_in *)(p_sa))->sin_port = htons(port)

/*
 * AFIGetPort
 *  Return the port number contained in this 'sin' structure (struct
 *  sockaddr*, or any of its variants) in HBO.
 *  p_sin must be non null.
 */
#define AFIGetPort(p_sa) \
  ntohs(((struct sockaddr_in *)p_sa)->sin_port)
  

/*
 * AFIComparePort
 *  Return TRUE (1) if the port numbers contained in p_sa and p_sa1
 *  are equal. (p_sa and p_sa1 are pointers to struct sockaddr*, 
 *  or any of its variants) and must be non null. Returns FALSE (0)
 *  if they are not equal.
 */
#define AFIComparePort(p_sa, p_sa1) \
  (ntohs(((struct sockaddr_in *)(p_sa))->sin_port) == \
   ntohs(((struct sockaddr_in *)(p_sa1))->sin_port))
 
/* 
 * AFIInitAny
 *  Given sin in {sinv4 or sinv6}, initialize the sin struct to 'any'
 *  defaults. Also, set the port number in sin to htons(port). 
 *  
 * Parameters
 *  address_family: AF_INET or AF_INET6. 0 will default to AF_INET.
 *  sin           : the structure to initialize.
 *  port          : port number to use (in host byte order)
 */
extern void AFIInitAny(int address_family, 
                       struct sockaddr_storage* sin, in_port_t port);


/* AFIGetUDPServerSocket
 *  Given a sin { v4 or v6 } structure, create a datagram socket and 
 *  bind this struct to it. return the socket descriptor or -1 on
 *  error.
 */
extern int AFIGetUDPServerSocket(struct sockaddr_storage* sin);

/* AFIGetTCPServerSocket
 *  Given a sin { v4 or v6 } structure create a tcp socket and bind this
 *  struct to it. return the socket descriptor. will also call listen on
 *  this socket if one could be obtained. 
*/
extern int AFIGetTCPServerSocket(struct sockaddr_storage* sin);

/* AFIGetUDPSocket
 *  Return a UDP Socket for the specified address family. 
 */
#define AFIGetUDPSocket(x) \
  socket((x), SOCK_DGRAM, 0)

/* AFIGetTCPSocket
 *  Return a TCP Socket for the specified address family.
*/
#define AFIGetTCPSocket(x) \
  socket((x), SOCK_STREAM, 0)


/* AFISetTTL
 *  set ttl for a multicast socket.
*/
extern int AFISetTTL(int sockid, unsigned char ttl);

/* AFIMulticastLoopBack
 *  set the multicast loop back option (1 = on, 0 = off)
*/
extern int AFIMulticastLoopBack(int sockid, unsigned char on);


/* AFIIsMulticast
 * tests the in_addr or in6_addr component based on
 * sin->ss_family and returns 1 if it is a multicast address.
 */
extern int AFIIsMulticast(struct sockaddr_storage *sin);

/* enables SO_REUSEADDR if supported by OS */
extern int AFIReuseAddr(int sockid);

/* enables SO_REUSEPORT if supported by OS */
extern int AFIReusePort(int sockid);

/*
 * AFIJoinMulticastGroup
 *  Joins the multicast group given in mcast_addr for the socket sockid.
 *  Returns 0 on success, -1 on error.
 *   Currently implemented for in_addr only - set the sockaddr_storage.
 *   ss_family to AF_INET, and sin_addr should contain the IP address.
 */
extern int AFIJoinMulticastGroup(int sockid, 
                                 struct sockaddr_storage* mcast_addr);

#endif

