/* conf-util.c - Common utility functions for configuration etc;
 * (c) 2001 Columbia University. All rights reserved.
*/
#include "sysdep.h"
#include <stdio.h>    /* fopen, etc; */
#include <string.h>   /* strcmp etc; */
#include <stdlib.h>

#ifdef WIN32
#include <windows.h>
#include <winreg.h>   /* Registry functions */
#endif

static char* conf_file = "cinema_db.conf";
static char* global_conf_file = "/etc/cinema_db.conf";

#ifdef WIN32
static char* win32_cinema_root = /* relative to HKEY_LOCAL_MACHINE */
             "SOFTWARE\\Columbia University\\IRT\\CINEMA";

static char* win32_dbkey =  /* relative to win32_cinema_root */
             "CINEMA_DB";   
#endif

static char* CheckFile(const char* filename);
static char* CheckRegistry();

/* 
 * GetDBURI
 *
 *  Try to get a usable database URI string. Return NULL if none
 *  is locatable.
*/
char* GetDBURI(const char* filename) {
 
  char *result = NULL;
  if (filename && strlen(filename) > 0) { /* no need to check valid buffers 
					   * having no data */
    result = CheckFile(filename);
  }

  if (result == NULL) {
    result = CheckFile(conf_file);
  }

  if (result == NULL) {
    result = CheckRegistry();
  }

  return result;
} /* GetDBURI */


static char* CheckFile(const char* filename) {

    FILE *fp;
    char dburi[256];

    if (!(fp = fopen(filename, "r"))) {
      fprintf(stderr, "GetDBURI : Unable to open file %s\n", filename);
      return NULL;
    }
    
    fgets(dburi, sizeof(dburi), fp);
    fclose(fp);
    
    if (strstr(dburi, "sql://") == dburi || 
	strstr(dburi, "fastsql://") == dburi) {
      fprintf(stderr, "GetDBURI: found a usable uri in file %s\n", filename);
      return strdup(dburi); 
    }
    else {
      fprintf(stderr, "No valid database config found in %s\n", filename);
      return NULL;
    }
} 

#ifndef WIN32 /* Unix systems */
/*
 * CheckRegistry
 *  For Unix systems, registry is /etc/cinema_db.conf.
 */
static char* CheckRegistry() {

  return CheckFile(global_conf_file);

} /* CheckRegistry */

#else /* WIN32 */
/*
 * CheckRegistry
 *  For WIN32 systems, registry is as given in win32_regkey, relative to
 *  HKEY_LOCAL_MACHINE.
*/
static char* CheckRegistry() {

  char data[256];
  DWORD datalen = sizeof(data) - 1;
  DWORD type;
  int retval;
  HKEY hKey;
     
  retval = RegOpenKeyEx (HKEY_LOCAL_MACHINE,  /*Key handle at root level.*/
                         win32_cinema_root,   /*Path name of child key. */
                         0,                   /*Reserved.*/
                         KEY_EXECUTE,         /*Requesting read access.*/
                         &hKey);              /*Addr of key to be returned. */

 if (retval == ERROR_SUCCESS) {
   if ((retval = RegQueryValueEx(hKey, win32_dbkey, NULL, &type,
		 (unsigned char*) data, &datalen)) == ERROR_SUCCESS) {
     if (type == REG_SZ) {
       RegCloseKey(hKey);
       fprintf(stderr, "GetDBURI: found usable database uri in registry\n");
       return strdup(data); /* null byte will be there (SZ) */ 
     }
   }
 }

 RegCloseKey(hKey);
 return NULL;   
} /* CheckRegistry */

#endif
     





