/*
* Error handling and debugging functions.
*/
#include <stdio.h>     /* setvbuf() */
#include <stdarg.h>    /* vsnprintf */
#include <stdlib.h>    /* abort() */
#include <string.h>    /* strerror() */
#include <sys/time.h>  /* gettimeofday() */
#include <errno.h>     /* errno */
#include <syslog.h>    /* system log, unix */
#include <time.h>      /* strftime() */
#include "sysdep.h"    /* OS specific definitions */
#include "afi.h"       /* AFIGetDottedAddress */

int debug_flag = 0;

FILE *logfile   = NULL; /* Name of logfile */
FILE *debugfile = NULL; /* Name of debug/trace output file */

/*
 * debug()
 * Print debug message to stdout or log file.
*/
void debug(const char *function_name, const char *format, ...)
{
  static int init_flag = 0;
  struct timeval tv;
  va_list ap;

  if (!debug_flag)
    return;

  /* Default debug trace goes to stdout */
  if (debugfile == NULL) {
    debugfile = stdout;
  }

  va_start(ap, format);
  
  if (!init_flag) {
    setvbuf(debugfile, NULL, _IOLBF, 0);
    init_flag = 1;
  }

  gettimeofday(&tv, NULL);

  /* print out name of function causing error */
  if (function_name != NULL) {
    (void) fprintf(debugfile, "%9.3f %s(): ", 
      tv.tv_sec + tv.tv_usec/1e6, function_name);
  }

  /* print out remainder of message */
  (void) vfprintf(debugfile, format, ap);

  if (debugfile != stdout) {
    fflush(debugfile);
  }
  
  va_end(ap);
} /* debug */


/*
 * debug2()
 * Print debug message to stdout or log file.
 * String representation of sin will also be added to the debug log.
*/
void debug2(const char *function_name, const struct sockaddr* sin,
            const char *format, ...)
{
  static int init_flag = 0;
  struct timeval tv;
  char *hostname;
  int port;

  va_list ap;

  if (!debug_flag)
    return;

  /* Default debug trace goes to stdout */
  if (debugfile == NULL) {
    debugfile = stdout;
  }

  va_start(ap, format);
  
  if (!init_flag) {
    setvbuf(debugfile, NULL, _IOLBF, 0);
    init_flag = 1;
  }

  gettimeofday(&tv, NULL);

  /* print out name of function causing error */
  if (function_name != NULL) {
    (void) fprintf(debugfile, "%9.3f %s(): ", 
      tv.tv_sec + tv.tv_usec/1e6, function_name);
  }

  if (sin != NULL) {
    hostname = AFIGetDottedAddress(sin);
    if (hostname) {
      port = AFIGetPort(sin);
      (void) fprintf(debugfile, "[%s/%d] ", hostname, port);
      free(hostname);
    }
  }

  /* print out remainder of message */
  (void) vfprintf(debugfile, format, ap);

  if (debugfile != stdout) {
    fflush(debugfile);
  }
  
  va_end(ap);
} /* debug2 */


/*
 * fatal()
 * Print error message, using the format 'format', identified by the
 * function name 'function_name' and abort process.
*/
void fatal(const char *function_name, const char *format, ...)
{
  char tmp[1024], *s = tmp;
  int e = errno;
  va_list ap;
  va_start(ap, format);

  /* timestamp */
  tmp[0] = '\0'; 
  if (logfile) {
    time_t t = time(0);

    strftime(s, sizeof(tmp), "%Y-%m-%d %H:%M:%S: ", localtime(&t));
    s += strlen(s);
  }

  /* print out name of function causing error */
  sprintf(s, "Fatal error in %s: ", function_name);
  s += strlen(s);

  if (e > 0) {
    sprintf(s, "%s (%d): ", strerror(e), e);
    s += strlen(s);
  }

  /* print out remainder of message */
  vsprintf(s, format, ap);
  s += strlen(s);
  va_end(ap);

  if (logfile) {
    if (tmp[strlen(tmp)-1] != '\n') {
      sprintf(s, "\n");
      s += strlen(s);
    }
    fputs(tmp, logfile);
  }
  else {
    syslog(LOG_ERR, "%s", tmp);
  }

  abort();
} /* fatal */


/*
 * error()
 * Print error message to stderr.
*/
void error(const char *function_name, const char *format, ...)
{
  char tmp[1024], *s = tmp;
  int e = errno;
  va_list ap;
  va_start(ap, format);

  /* timestamp */
  tmp[0] = '\0'; 
  if (logfile) {
    time_t t = time(0);

    strftime(s, sizeof(tmp), "%Y-%m-%d %H:%M:%S: ", localtime(&t));
    s += strlen(s);
  }

  /* print out name of function causing error */
  sprintf(s, "Error in %s: ", function_name);
  s += strlen(s);

  if (e > 0) {
    sprintf(s, "%s (%d): ", strerror(e), e);
    s += strlen(s);
  }

  /* print out remainder of message */
  vsprintf(s, format, ap);
  s += strlen(s);

  va_end(ap);

  if (logfile) {
    if (tmp[strlen(tmp)-1] != '\n') {
      sprintf(s, "\n");
      s += strlen(s);
    }
    fputs(tmp, logfile);
  }
  else {
    syslog(LOG_ERR, "%s", tmp);
  }

  va_end(ap);
} /* error */


/*
 * ErrorInit()
 * Initialize error log to send output to file 'filename'.  If 'file' is
 * 0, use syslog.  If 'file' is "stderr", use that.
*/
void ErrorInit(const char *filename)
{
  if (!filename) {
    openlog("sipd", LOG_PID, LOG_DAEMON);
    logfile = NULL;
  }
  else if (strcmp(filename, "stderr") == 0) {
    logfile = stderr;
  }
  else if ((logfile = fopen(filename, "a+")) == NULL) {
    logfile = stderr;
    fatal("ErrorInit", "Couldn't open error log %s", filename);
  }
} /* ErrorInit */


/*
 * Close error log file or syslog.
*/
void ErrorClose(void)
{
  if (logfile) fclose(logfile);
  else closelog();
} /* ErrorClose */


/*
 * Initialize debugging log to send output to 'filename'.  If 'file' is
 * 0, use stdout.
*/
void DebugInit(const char *filename)
{
  if (!filename) {
    debugfile = stdout;
  }
  else if (strcmp(filename, "stdout") == 0) {
    debugfile = stdout;
  }
  else if (strcmp(filename, "stderr") == 0) {
    debugfile = stderr;
  }
  else if ((debugfile = fopen(filename, "w+")) == NULL) {
    debugfile = stdout;
    fatal("ErrorInit", "Couldn't open debug log %s", filename);
  }
} /* DebugInit */


/*
 * Close log file or syslog.
*/
void DebugClose(void)
{
  if (debugfile) 
    fclose(debugfile);
} /* DebugClose */
