/*
exitthread.c - wait for signals, and exit when appropriate.

Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#include <stdlib.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <errno.h>

#include "pthread-util.h"
#include "error.h"
#include "sysdep.h"
#include "timespec.h"


/* The number of a received signal. */
static volatile int finish_signal = 0;


/* The time to wait between checking finish_signal.  Should this
 * be a config option? */
static const float exit_delta = 1.0;


/*
 * signal_handler()
 * Handle signals.  Just set finish_signal; the exit thread will catch
 * things, and clean up.
 * Note: triggering a condition variable from a signal handler is *not*
 * safe, which is why we don't just raise cont.
 */
static void signal_handler(int sig)
{
  finish_signal = sig;
} /* signal_handler */


/*
 * exitThread()
 * Check finish_signal once per exit_delta.  When it becomes non-zero,
 * exit().  XXX: it's not entirely clear to me that exit(3) does the Right
 * Thing in a threaded environment?
 */
static void *exitThread(void *arg)
{
  float delay = exit_delta;
  struct timeval start;
  struct timespec next;
  pthread_mutex_t mutex;
  pthread_cond_t cond;
  int e;

  if ((e = pthread_mutex_init(&mutex, NULL)) != 0) {
    fatal("exitThread", "pthread_mutex_init: %s (%d)\n", strerror(e), e);
  }

  if ((e = pthread_cond_init(&cond, NULL)) != 0) {
    fatal("exitThread", "pthread_cond_init: %s (%d)\n", strerror(e), e);
  }

  if ((e = pthread_mutex_lock(&mutex)) != 0) {
    fatal("exitThread", "pthread_mutex_lock: %s (%d)\n", strerror(e), e);
  }

  gettimeofday(&start, NULL);

  while (finish_signal == 0) {
    next = timespec_delay(start, delay);
    /* sleep */
    if ((e = pthread_cond_timedwait(&cond, &mutex, &next)) != 0 &&
        e != ETIMEDOUT) {
      error("exitThread", "pthread_cond_timedwait: %s (%d)\n", strerror(e), e);
    }
      
    delay += exit_delta;
  }
  debug("exitThread", "signal %d received; exiting\n", finish_signal);

  if ((e = pthread_mutex_unlock(&mutex)) != 0) {
    error("exitThread", "pthread_mutex_unlock: %s (%d)\n", strerror(e), e);
  }

  if ((e = pthread_cond_destroy(&cond)) != 0) {
    error("exitThread", "pthread_cond_destroy: %s (%d)\n", strerror(e), e);
  }

  if ((e = pthread_mutex_destroy(&mutex)) != 0) {
    error("exitThread", "pthread_mutex_destroy: %s (%d)\n", strerror(e), e);
  }

  exit(0);
} /* exitThread */


/*
 * setupExitThread()
 * Sets up signal handlers, then starts the exit thread which will call
 * exit() when we receive a terminating signal.
 */
void setupExitThread(void)
{
  pthread_t tid;
  int e;                          /* thread error condition */

#ifndef WIN32
  struct sigaction act;
  
  memset(&act, '0', sizeof(act));
  
  act.sa_handler = signal_handler;
  
  if (sigaction(SIGTERM, &act, NULL) < 0) {
    error("setupExitThread", "sigaction(SIGTERM)");
  }
  if (sigaction(SIGINT, &act, NULL) < 0) {
    error("setupExitThread", "sigaction(SIGINT)");
  }
  if (sigaction(SIGQUIT, &act, NULL) < 0) {
    error("setupExitThread", "sigaction(SIGQUIT)");
  }
#else
  signal(SIGINT, signal_handler);
#endif

  if ((e = pthread_create_detached(&tid, NULL, &exitThread, NULL)) != 0) { 
    fatal("setupExitThread", "Cannot create Exit thread: %s\n", strerror(e));
  }
} /* setupExitThread */
