/* Thread safe gethostbyname_r.c */
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <resolv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <string.h>
#include "gethostbyname_r.h"

/* Debugging for functions 
 * keep the resolver's functions style
 *
 * #define DEBUG
 */

/*
 * Main gethostbyname_r function:
 *
 * - creates the query res_mkquery
 * - sends it to the DNS servers sendquery
 * - decode the answer decode
 */
struct hostent *gethostbyname_r (const char *name, struct hostent *result,
      char *buffer, int buflen, int *h_errnop)
{
  int class, type;	/* class and type of query */
  int anslen = MAXPACKET; /* size of answer buffer */
  u_char *answer=malloc(MAXPACKET); /* buffer to put answer */
  HEADER *hp = (HEADER *) answer;
  int n;
  int *length=malloc(sizeof(int));
  int status;
  struct hostent *hostinfo = malloc(sizeof(struct hostent));

  hp->rcode = NOERROR;	/* default */


  if ((_res.options & RES_INIT) == 0 && res_init() == -1) {
    h_errno   = NETDB_INTERNAL;
    *h_errnop = NETDB_INTERNAL;
    return hostinfo;
  } 


#ifdef DEBUG
  /* enable debugging information */
  _res.options+=2;
#endif


  class = C_IN;
  type  = T_A;

  /* creating the query */
  n = res_mkquery(QUERY, name, C_IN, T_A, NULL, 0, NULL,
		  buffer, buflen);


  if (n <= 0) {
    if (_res.options & RES_DEBUG)
      printf(";; res_query: mkquery failed\n");
    h_errno   = NO_RECOVERY;
    *h_errnop = NO_RECOVERY;
    return hostinfo;
  }

  status = sendquery((const u_char *)buffer, n, (u_char **)&answer, &anslen,(int **) &length);





  /* Sending the query */
  if (status < 0) {
    if (_res.options & RES_DEBUG)
      printf(";; res_query: send error\n");
    h_errno   = TRY_AGAIN;
    *h_errnop = TRY_AGAIN;
    return hostinfo;
  }


  if (_res.options & RES_DEBUG) {
    printf(";; gethostbyname_r() output of sendrequest:\n");
    printf(";; length=%d anslen=%d\n", *length, anslen);
    printf(";; hp->rcode=%d\n", hp->rcode);
  }


  status = decode((u_char *)answer, MAXPACKET, *length, hostinfo);
  


  if ((_res.options & RES_DEBUG) && (hp->rcode == 0) ) {
    struct sockaddr_in sina;
    printf(";; gethostbyname_r() output of status:%d\n",status);
    printf(";; gethostbyname_r() output of hostinfo->h_length:%d\n",hostinfo->h_length);
    memcpy((void *)&sina.sin_addr.s_addr, hostinfo->h_addr_list[0], sizeof(u_int32_t));
    printf(";; gethostbyname_r() address #1 is %s\n",inet_ntoa(sina.sin_addr));
    printf(";; gethostbymname_r() Official Name %s\n",hostinfo->h_name);
    printf(";; --------------------------------------\n");
  }


  free(answer);
  *h_errnop = hp->rcode;
  if (hp->rcode == 0) {
    memcpy((void *)result,(void *)hostinfo,sizeof(struct hostent));
    return hostinfo;
  }
  else {
    return NULL;
  }

}

/*
 * sendquery sends the query to the DNS server 
 */
int sendquery(const u_char *buf, int buflen,u_char **answer, u_int *anslen, int **trueLenPtr)
{
  u_short id, len;
  u_char *cp;
  fd_set dsmask;
  struct timeval timeout;
  struct iovec iov[2];
  int terrno = ETIMEDOUT;
  char junk[512];
  int nscount;
  HEADER *anhp = (HEADER *) *answer;
  const HEADER *hp = (HEADER *) buf;
  int s = -1; /* socket descriptor */
  int n, try, v_circuit, resplen, salen;
  int gotsomewhere = 0, connected = 0;
  struct sockaddr_in sin;
  int connreset = 0;

  int counter;
  struct sockaddr sa;


  nscount=_res.nscount;
  
  if (_res.options & RES_DEBUG2) {
    printf("------------\nsendquery(), len %d\n", buflen);
  }
  sin.sin_family	= AF_INET;
  sin.sin_port	= htons(53);


  /*
   * Send request, RETRY times, in all name servers or until successful
   */
  for (counter=0; counter<=nscount-1; counter++) {

    sin.sin_addr = _res.nsaddr_list[counter].sin_addr;
  
    v_circuit = (_res.options & RES_USEVC) || buflen > PACKETSZ;
    id = hp->id;

    
    for (try = 0; try < _res.retry; try++) {
    usevc:
      if (v_circuit) {
	int truncated = 0;
	
	/*
	 * Use virtual circuit;
	 * at most one attempt per server.
	 */
	try = _res.retry;
	if (s < 0) {
	  s = socket(AF_INET, SOCK_STREAM, 0);
	  if (s < 0) {
	    terrno = errno;
	    if (_res.options & RES_DEBUG)
	      perror("socket (vc) failed");
	    continue;
	  }
	  if (connect(s, (struct sockaddr *)&sin,
		      sizeof(struct sockaddr)) < 0) {
	    terrno = errno;
	    if (_res.options & RES_DEBUG)
	      perror("connect failed");
	    (void) close(s);
	    s = -1;
	    continue;
	  }
	}
	/*
	 * Send length & message
	 */
	__putshort(buflen, (u_char *)&len);
	iov[0].iov_base = (caddr_t)&len;
	iov[0].iov_len = INT16SZ;
	iov[1].iov_base = (caddr_t)buf;
	iov[1].iov_len = buflen;
	if (writev(s, iov, 2) != INT16SZ + buflen) {
	  terrno = errno;
	  if (_res.options & RES_DEBUG)
	    perror("write failed");
	  (void) close(s);
	  s = -1;
	  continue;
	}
	/*
	 * Receive length & response
	 */
	cp = *answer;
	len = INT16SZ;
	while ((n = read(s, (char *)cp, (int)len)) > 0) {
	  cp += n;
	  if ((len -= n) <= 0)
	    break;
	}
	if (n <= 0) {
	  terrno = errno;
	  if (_res.options & RES_DEBUG)
	    perror("read failed");
	  (void) close(s);
	  s = -1;
	  /*
	   * A long running process might get its TCP
	   * connection reset if the remote server was
	   * restarted.  Requery the server instead of
	   * trying a new one.  When there is only one
	   * server, this means that a query might work
	   * instead of failing.  We only allow one reset
	   * per query to prevent looping.
	   */
	  if (terrno == ECONNRESET && !connreset) {
	    connreset = 1;
	  }
	  continue;
	}
	cp = *answer;
	if ((resplen = ns_get16((u_char*)cp)) > *anslen) {
	  if (_res.options & RES_DEBUG)
	    fprintf(stderr, "response truncated\n");
	  len = *anslen;
	  truncated = 1;
	} else
	  len = resplen;
	while (len != 0 &&
	       (n = read(s, (char *)cp, (int)len)) > 0) {
	  cp += n;
	  len -= n;
	}
	if (n <= 0) {
	  terrno = errno;
	  if (_res.options & RES_DEBUG)
	    perror("read failed");
	  (void) close(s);
	  s = -1;
	  continue;
	}
	if (truncated) {
	  /*
	   * Flush rest of answer
	   * so connection stays in synch.
	   */
	  anhp->tc = 1;
	  len = resplen - *anslen;
	  while (len != 0) {
	    n = (len > sizeof(junk) ?
		 sizeof(junk) : len);
	    if ((n = read(s, junk, n)) > 0)
	      len -= n;
	    else
	      break;
	  }
	}
      } else {
	/*
	 * Use datagrams.
	 */
	if (s < 0) {
	  s = socket(AF_INET, SOCK_DGRAM, 0);
	  if (s < 0) {
	    terrno = errno;
	    if (_res.options & RES_DEBUG)
	      perror("socket (dg) failed");
	    continue;
	  }
	}
#if BSD >= 43
	if (connected == 0) {
	  if (connect(s, (struct sockaddr *)&sin,
		      sizeof sin) < 0) {
	    if (_res.options & RES_DEBUG)
	      perror("connect");
	    continue;
	  }
	  connected = 1;
	}
	if (send(s, buf, buflen, 0) != buflen) {
	  if (_res.options & RES_DEBUG)
	    perror("send");
	  continue;
	}
#else /* BSD */
	if (sendto(s, (const char *)buf, buflen, 0,
		   (struct sockaddr *) &sin,
		   sizeof sin) != buflen) {
	  if (_res.options & RES_DEBUG)
	    perror("sendto");
	  continue;
	}
#endif
	/*
	 * Wait for reply
	 */
	timeout.tv_sec = (_res.retrans << try);
	if (timeout.tv_sec <= 0)
	  timeout.tv_sec = 1;
	timeout.tv_usec = 0;
      wait:
	FD_ZERO(&dsmask);
	FD_SET(s, &dsmask);
	n = select(s+1, &dsmask, (fd_set *)NULL,
		   (fd_set *)NULL, &timeout);
	if (n < 0) {
	  if (_res.options & RES_DEBUG)
	    perror("select");
	  continue;
	}
	if (n == 0) {
	  /*
	   * timeout
	   */
	  if (_res.options & RES_DEBUG)
	    printf("timeout\n");
#if BSD >= 43
	  gotsomewhere = 1;
#endif
	  continue;
	}
	
	salen = sizeof sa;
	resplen = recvfrom(s, (void *)*answer, (size_t) *anslen, 0,
			   (struct sockaddr *)&sa, &salen);
	
	if (resplen <= 0) {
	  if (_res.options & RES_DEBUG)
	    perror("recvfrom");
	  continue;
	}
	gotsomewhere = 1;
	if (id != anhp->id) {
	  /*
	   * response from old query, ignore it
	   */
	  if (_res.options & RES_DEBUG2) {
	    printf("------------\nOld answer:\n");
	  }
	  goto wait;
	}
	if (!(_res.options & RES_IGNTC) && anhp->tc) {
	  /*
	   * get rest of answer;
	   * use TCP with same server.
	   */
	  if (_res.options & RES_DEBUG)
	    printf("truncated answer\n");
	  (void) close(s);
	  s = -1;
	  v_circuit = 1;
	  goto usevc;
	}
      }
      if (_res.options & RES_DEBUG) {
	if (_res.options & RES_DEBUG2)
	  printf("------------\nGot answer (%d bytes):\n",
		 resplen);
	else
	  printf("------------\nGot answer:\n");
      }
      (void) close(s);
      s = -1;
      **trueLenPtr = resplen;
      return (SUCCESS);
    }
    if (s >= 0) {
      (void) close(s);
      s = -1;
    }
    if (v_circuit == 0)
      if (gotsomewhere == 0)
	return NO_RESPONSE;	/* no nameservers found */
      else
	return TIME_OUT;	/* no answer obtained */
    else
      if (errno == ECONNREFUSED)
	return NO_RESPONSE;
      else
	return ERROR;
  }
  /* Shouldn't happen I *think* - jl*/
  return ERROR;
}


/* 
 * decode decodes the answer from the DNS server
 */
int decode(u_char *answer, u_int anslen, int length, struct hostent *hostinfo)
{
  HEADER *hp = (HEADER *) answer;
  u_char *eom; /* eom = end of message */
  u_char *cp;  /* cp = current pointer in the message */
  u_char hostbuf[MAXDNAME]; /* MAXDNAME = 1025 */
  int qdcount, ancount, arcount, nscount, buflen;
  u_char *bp;
  int n;
  int i;
  int type, class;
  int dlen;
  char **aliasPtr;
  char **addrPtr;
  int numAliases = 0;
  int numAddresses = 0;
  int haveAnswer;
  int len;
  char *addr_list[MAXADDRS + 1];
  char *host_aliases[MAXALIASES];
  int host_aliases_len[MAXALIASES];
  int origClass = -1;
  int queryType = T_A;

  if (_res.options & RES_DEBUG) {
    printf(";; decode: hp->rcode=%d\n",hp->rcode);
  }


  /* Initialize unpacking staff */
  eom = (u_char *) answer + length;
  cp = (u_char *) answer + HFIXEDSZ; /* HFIXEDSZ bytes of fixed data in header = 12 */
  

  if (hp->rcode != NOERROR) {
    return (hp->rcode);
  }

  qdcount = ntohs(hp->qdcount);
  ancount = ntohs(hp->ancount);
  arcount = ntohs(hp->arcount);
  nscount = ntohs(hp->nscount);

  /*
   * If there are no answer, n.s. or additional records
   * then return with an error.
   */
  if (ancount == 0 && nscount == 0 && arcount == 0) {
    return (NO_INFO);
  }



  bp = hostbuf; /* pointer for expanded data */
  buflen = sizeof(hostbuf); /* size of the buffer of expanded */


  /* Skip over question section. */
  while (qdcount-- > 0) {
    n = dn_skipname(cp, eom);
    if (n < 0)
      return (ERROR);
    cp += n + QFIXEDSZ;
    if (cp > eom)
      return (ERROR);
  }



  if (_res.options & RES_DEBUG) {
    printf(";; decode: After skipping headers we have:\n");
    printf(";; eom=%d cp=%d start=%d\n",(int)eom,(int)cp,(int)answer);

  }

  aliasPtr = host_aliases;
  addrPtr = addr_list;
  haveAnswer = FALSE;


  /* We still have answers to examine */
  if (ancount != 0) {
    while (--ancount >= 0 && cp < eom) {
      n = dn_expand(answer, eom, cp, (char *)bp, buflen);
      if (n < 0)
	return(ERROR);
      cp += n;		
      if (cp + 3 * INT16SZ + INT32SZ > eom)
	return (ERROR);
      type = GetShort(cp);
      class = GetShort(cp);
      cp += INT32SZ; /* skip TTL */
      dlen = GetShort(cp);
      if (cp + dlen > eom)
	return (ERROR);
      if (type == T_CNAME) {
	/*
	 * Found an alias.
	 */
	cp += dlen;
	if (aliasPtr >= &host_aliases[MAXALIASES-1]) {
	  continue;
	}
      *aliasPtr++ = (char *)bp;
      n = strlen((char *)bp) + 1;
      host_aliases_len[numAliases] = n;
      numAliases++;
      bp += n;
      buflen -= n;
      continue;
      } else if (type == T_PTR) {
	/*
	 *  Found a "pointer" to the real name.
	 */
	n = dn_expand(answer, eom, cp, (char *)bp, buflen);
	if (n < 0) {
	  cp += n;
	  continue;
	}
	cp += n;
	len = strlen((char *)bp) + 1;
	hostinfo->h_name = calloc(1, len);
	memcpy(hostinfo->h_name, bp, len);
	haveAnswer = TRUE;
	break;
      } else if (type != T_A) {
	cp += dlen;
	continue;
      }

      /* 
       * Not so sure about this
       * if (dlen != INADDRSZ)
       * return (ERROR);
       *
       */

      if (haveAnswer) {
	/*
	 * If we've already got 1 address, we aren't interested
	 * in addresses with a different length or class.
	 */
	if (dlen != hostinfo->h_length) {
	  cp += dlen;
	  continue;
	}
	if (class != origClass) {
	  cp += dlen;
	  continue;
	}
      } else {
	/*
	 * First address: record its length and class so we
	 * only save additonal ones with the same attributes.
	 */
	hostinfo->h_length = dlen;
	origClass = class;
	hostinfo->h_addrtype = (class == C_IN) ? AF_INET : AF_UNSPEC;
	len = strlen((char *)bp) + 1;
	hostinfo->h_name = calloc(1, len);
	memcpy(hostinfo->h_name, bp, len);
      }
      bp += (((u_long)bp) % sizeof(char));
      if (bp + dlen >= &hostbuf[sizeof(hostbuf)]) {
	if (_res.options & RES_DEBUG) {
	  printf("Size (%d) too big\n", dlen);
	}
	break;
      }
      if (numAddresses >= MAXADDRS) {
	printf("MAXADDRS exceeded: skipping address\n");
	cp += dlen;
	continue;
      }
      memcpy(*addrPtr++ = (char *)bp, cp, dlen);
      bp += dlen;
      cp += dlen;
      numAddresses++;
      haveAnswer = TRUE;
    }
  }



  if ((queryType == T_A || queryType == T_PTR) && haveAnswer) {
    
    /*
     *  Go through the alias and address lists and return them
     *  in the hostPtr variable.
     */
    
    if (numAliases > 0) {
      hostinfo->h_aliases = (char **) calloc(1 + numAliases, sizeof(char *));
      for (i = 0; i < numAliases; i++) {
	hostinfo->h_aliases[i] = calloc(1, host_aliases_len[i]);
	memcpy(hostinfo->h_aliases[i], host_aliases[i],
	       host_aliases_len[i]);
      }
      hostinfo->h_aliases[i] = NULL;
    }
    if (numAddresses > 0) {
      hostinfo->h_addr_list =
	(char **)calloc(1+numAddresses, sizeof(char *));
      for (i = 0; i < numAddresses; i++) {
	hostinfo->h_addr_list[i] = calloc(1, hostinfo->h_length);
	memcpy(hostinfo->h_addr_list[i], addr_list[i], hostinfo->h_length);
      }
      hostinfo->h_addr_list[i] = NULL;
    }
    if ((_res.options & RES_DEBUG) && (hp->rcode == 0)) {
      struct sockaddr_in sina;
      int c;
      for (c=0; c<=numAddresses-1; c++) {
	memcpy((void *)&sina.sin_addr.s_addr, hostinfo->h_addr_list[c], sizeof(u_int32_t));
	printf(";; decode() address #%d is %s\n",c+1,inet_ntoa(sina.sin_addr));
      }
      for (c=0; c<=numAliases-1; c++) {
	printf(";; decode() alias #%d is %s\n",c+1,hostinfo->h_aliases[0]);
      }
      printf(";; decode() Official Name %s\n",hostinfo->h_name);
      printf(";; decode() number of IP address:%d\n",numAddresses);
      printf(";; --------------------------------------\n");
    }

#ifdef verbose
    if (hp->aa || nscount == 0) {
      return (SUCCESS);
    }
#else
    return (SUCCESS);
#endif
  }

  return 0;
}
