/* gsearch.c: simple glob matching. */
/* Extraced from /usr/src/bin/csh/glob.c in FreeBSD 3.5.
 * Reformatted, commented, ANSI-fied, constified, made re-entrant,
 * and fixed memory leaks.
 */

/* Copyright 2001 Columbia University.  All rights reserved.
 * Also see original copyright notice below.
 */

/*-
 * Copyright (c) 1980, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Gmatch()
 * Perform a glob match.  See if the string 'string' matches the pattern
 * 'pattern'.
 * 
 * The glob matcher recognizes the following glob pattern constructs:
 * ? - any single character
 * * - any string of characters
 * [xy] - a set of characters.  Also supports the range notation a-z.
 *        ^ beginning the set negates it.
 *        ] doesn't close the range if it's the first character of a set.
 *        - doesn't create a range if it's the first or last character of a
 *          set.
 * {foo,bar} - the choice of patterns foo or bar
 * ^ - as the first character of a pattern, invert the sense of the pattern
 * \x - a metacharacter verbatim
 * Any other character matches itself verbatim.
 *
 * This pattern syntax is the same as that of csh(1), from which the
 * pattern matching code was taken.
 *
 * The pattern syntax only works correctly on seven-bit-clean text.
 *
 * Return >0 if match successful; 0 if match unsuccessful;
 * <0 on error.  Values <0 can be passed to gmatch_strerror() for a
 * human-readable string.
 */

#ifndef lint
#if 0
static char sccsid[] = "@(#)glob.c	8.1 (Berkeley) 5/31/93";
#else
static const char rcsid[] =
  "$FreeBSD: src/bin/csh/glob.c,v 1.11.2.1 1999/08/29 14:11:46 peter Exp $";
#endif
#endif /* not lint */

#include <sys/param.h>
#include <limits.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <setjmp.h>
#include <stdio.h>

#include "sysdep.h"
#include "strdupn.h" /* free_valid */
#include "gmatch.h"

typedef void *ptr_t;


#define	GLOBSPACE	100	/* Alloc increment			*/

#define LBRC '{'
#define RBRC '}'
#define LBRK '['
#define RBRK ']'
#define EOS '\0'

#define ERR_MISSING_LBRC (-(int)LBRC)
#define ERR_MISSING_RBRC (-(int)RBRC)
#define ERR_MISSING_LBRK (-(int)LBRK)
#define ERR_MISSING_RBRK (-(int)RBRK)
#define ERR_NOMEM        (-256)

/*
 * The eighth bit of characters is used to prevent recognition,
 * and eventually stripped.
 */
# define        QUOTE   ((char) 0200)   /* Eighth char bit used for 'ing */
# define        TRIM            0177    /* Mask to strip quote bit */
#ifndef MAXPATHLEN
#define MAXPATHLEN  8192
#endif

/*
 * globbing is now done in two stages. In the first pass we expand
 * a subset of csh globbing idiom "{" and then we proceed doing the normal
 * globbing if needed "?*["
 */

static int	globbrace (char *, char *, char ***);
static int	expbrace (char ***, char ***, int);
static int	pmatch (const char *, const char *);
static int      collate_range_cmp (int, int);
static char *   quote(char *);

/*
 * strsave()
 * Equivalent to strdup, except that NULL gets dup'd as "".
 * Return NULL on malloc failure.
 */
char *strsave(const char *s)
{
  char *n, *p;
  const char *cp;

  if (s == NULL)
    s = "";
  for (cp = s; *cp != '\0'; cp++)
    continue;
  n = p = (char *) malloc((size_t) (cp - s + 1));
  if (p == NULL) return NULL;
  while ((*p++ = *s++) != '\0')
    continue;
  return (n);
} /* strsave */


/*
 * blklen()
 * Given a block (an array of strings, terminated by NULL) return
 * its total length.
 */
int blklen(char **av)
{
  int i = 0;

  while (*av++)
    i++;
  return (i);
} /* blklen */


/*
 * blkfree()
 * Free a block.
 */
void blkfree(char **av0)
{
  char **av = av0;

  if (!av0)
    return;
  for (; *av; av++)
    free_valid((ptr_t) * av);
  free_valid((ptr_t) av0);
}


/*
 * collate_range_cmp()
 * A utility function for handling glob ranges like [a-z].
 * This is a wrapper around strcoll(); it falls back to a literal comparison
 * if strcoll() considers two characters equivalent when they're not equal.
 * (This can happen in some locales.)
 */
static int collate_range_cmp (int c1, int c2)
{
  static char s1[2], s2[2];
  int ret;

  c1 &= UCHAR_MAX;
  c2 &= UCHAR_MAX;
  if (c1 == c2)
    return (0);
  s1[0] = c1;
  s2[0] = c2;
  if ((ret = strcoll(s1, s2)) != 0)
    return (ret);
  return (c1 - c2);
} /* collate_range_cmp */


/*
 * globbrace()
 * Given a pattern like {pat1,pat2,pat3}, expand it into
 * several patterns of the form pat1 pat2 pat3 (in the 'bl' structure).
 * Only expand one level: {a{a,b},c{c,d}} should expand to the patterns
 * a{a,b} and c{c,d}. expbrace() will call this function again for the
 * sub-patterns. 
 * Return the number of patterns expanded, or a (negative) error code.
 */
static int globbrace(char *s, char *p, char ***bl)
{
  int     i, len;
  char   *pm, *pe, *lm, *pl;
  char  **nv, **vl;
  char    gbuf[MAXPATHLEN];
  int     size = GLOBSPACE;

  nv = vl = (char **) malloc((size_t) sizeof(char *) * size);
  if (vl == NULL) return ERR_NOMEM;
  *vl = NULL;

  len = 0;
  /* copy part up to the brace */
  for (lm = gbuf, p = s; *p != LBRC; *lm++ = *p++)
    continue;

  /* check for balanced braces */
  for (i = 0, pe = ++p; *pe; pe++)
    if (*pe == LBRK) {
      /* Ignore everything between [] */
      for (++pe; *pe != RBRK && *pe != EOS; pe++)
        continue;
      if (*pe == EOS) {
        blkfree(nv);
        return (ERR_MISSING_RBRK);
      }
    }
    else if (*pe == LBRC)
      i++;
    else if (*pe == RBRC) {
      if (i == 0)
        break;
      i--;
    }

  if (i != 0 || *pe == '\0') {
    blkfree(nv);
    return (ERR_MISSING_RBRC);
  }

  for (i = 0, pl = pm = p; pm <= pe; pm++)
    switch (*pm) {
    case LBRK:
      for (++pm; *pm != RBRK && *pm != EOS; pm++)
        continue;
      if (*pm == EOS) {
        *vl = NULL;
        blkfree(nv);
        return (ERR_MISSING_RBRK);
      }
      break;
    case LBRC:
      i++;
      break;
    case RBRC:
      if (i) {
        i--;
        break;
      }
      /* FALLTHROUGH */
    case ',':
      if (i && *pm == ',')
        break;
      else {
        char    savec = *pm;

        *pm = EOS;
        (void) strcpy(lm, pl);
        (void) strcat(gbuf, pe + 1);
        *pm = savec;
        *vl = strsave(gbuf);
        if (*vl == NULL) {
          blkfree(nv);
          return ERR_NOMEM;
        }
        vl++;
        len++;
        pl = pm + 1;
        if (vl == &nv[size]) {
          char **nv_new;
          size += GLOBSPACE;
          nv_new = (char **) realloc((ptr_t) nv, (size_t)
                                     size * sizeof(char *));
          if (nv_new == NULL) {
            blkfree(nv);
            return ERR_NOMEM;
          }
          nv = nv_new;
          vl = &nv[size - GLOBSPACE];
        }
      }
      break;
    default:
      break;
    }
  *vl = NULL;
  *bl = nv;
  return (len);
} /* globbrace */


/*
 * expbrace()
 * Expand all the braces, calling globbrace() to expand each one.
 *
 * More specifically: we're passed a (pointer to) a block of patterns,
 * initially containing only one pattern, the original pattern.
 * Now, loop as follows:
 *   If the current pattern in the block contains no braces, advance to the
 *       the next pattern. 
 *   Else, globexpand it.
 *     If globexpand returned only a single brace, it was a (useless)
 *         comma-free brace pattern {foo}.  globexpand stripped off the
 *         braces on the pattern, so point to the stripped pattern.
 *     Else, 
 *         * move all the patterns after the current pattern down after the
 *           expansion of this pattern
 *         * point to the first newly-expanded pattern
 * until we reach the end of the block (a NULL pattern).
 *
 * The resulting number of patterns is exponential in the number of braced
 * expressions.  This is amazingly suboptimal, since a smarter algorithm
 * would be linear(!) using a DFA.  For quick use-once expressions where
 * {}'s are rare this is probably more efficient, though.
 *
 * I don't understand how the variable names were chosen.  They seem
 * awfully obscure.  UCB programming practice circa 1980, I guess.
 * Summary:
 *  - nvp: pointer to passed-in block
 *  - elp: pointer to end of passed-in block (I think?  Not used in this
 *         subset of the code.)
 *  - size: the size of the block -- size of nvp when passed in, after
 *         that size of nv
 *  - nv: the current "working" block.
 *  - vl: pointer to the current pattern (an offset from nv).
 *  - el: pointer to the last pattern of the block.
 *  - s:  the current pattern.
 *  - b:  the first brace in the current pattern (offset from s)
 *  - vp: "write" pointer in vl, for copying from bl to vl.
 *  - bp: "read" pointer in bl, for copying from bl to vl.
 *  - bl: the block of new patterns created by globexpand() expanding s.
 *  - len: the number of patterns created by globexpand() expanding s (length
 *        of bl).
 *
 * Return 0 on success, (negative) error code on failure.  On failure,
 * also free the passed-in block (holder and content).
 */
static int expbrace(char ***nvp, char ***elp, int size)
{
  char **vl, **el, **nv, *s;

  vl = nv = *nvp;
  if (elp != NULL)
    el = *elp;
  else
    for (el = vl; *el; el++)
      continue;

  for (s = *vl; s; s = *++vl) {
    char   *b;
    char  **vp, **bp;

    /* leave {} untouched for find */
    if (s[0] == '{' && (s[1] == '\0' || (s[1] == '}' && s[2] == '\0')))
      continue;
    if ((b = strchr(s, '{')) != NULL) {
      char  **bl;
      int     len;

      if ((len = globbrace(s, b, &bl)) < 0) {
        blkfree(nv);
        return(len);
      }
      /* Originally, 's' was freed here.  However, adding error checking
       * forced it to be postponed -- 'nv' points to it, so if the
       * realloc() below fails, we'll free it in the blkfree().  If
       * it's been freed already, this means a double-free.
       */
      if (len == 1) {
        *vl-- = *bl;
        free_valid((ptr_t) s);
        free_valid((ptr_t) bl);
        continue;
      }
      len = blklen(bl);
      if (&el[len] >= &nv[size]) {
        /* We need more space to hold all the new patterns we got from
         * globbrace().
         */
        int     l, e;
        char **nv_new;

        l = &el[len] - &nv[size];
        size += GLOBSPACE > l ? GLOBSPACE : l;
        l = vl - nv;
        e = el - nv;
        nv_new = (char **) realloc((ptr_t) nv, (size_t)
                                   size * sizeof(char *));
        if (nv_new == NULL) {
          blkfree(bl);
          blkfree(nv);
          return ERR_NOMEM;
        }
        nv = nv_new;
        vl = nv + l;
        el = nv + e;
      }
      free_valid((ptr_t) s);
      vp = vl--;
      *vp = *bl;
      len--;
      for (bp = el; bp != vp; bp--)
        bp[len] = *bp;
      el += len;
      vp++;
      for (bp = bl + 1; *bp; *vp++ = *bp++)
        continue;
      free_valid((ptr_t) bl);
    }

  }
  if (elp != NULL)
    *elp = el;
  *nvp = nv;
  return(0);
} /* expbrace */


/*
 * Gmatch1()
 * Core of the glob matcher.  Expand the pattern's braces, then check each
 * generated pattern to see if the string matches it.
 * Also check if the pattern began with a ^, and if so reverse the sense
 * of the test.
 * Return 1 if we matched some pattern (or no pattern if inverted);
 * error code < 0 on an error; otherwise 0.
 */
static int Gmatch1(const char *string, const char *pattern)
{
  char **blk;
  char * const *p;
  int	   gpol = 1, gres = 0;
  int ret;

  if (*pattern == '^') {
    gpol = 0;
    pattern++;
  }

  blk = (char **) malloc(GLOBSPACE * sizeof(char *));
  if (blk == NULL) return ERR_NOMEM;
  blk[0] = strsave(pattern);
  blk[1] = NULL;

  if (blk[0] == NULL) {
    blkfree(blk);
    return ERR_NOMEM;
  }

  if ((ret = expbrace(&blk, NULL, GLOBSPACE)) < 0) {
    return ret;
  }

  for (p = blk; *p; p++) {
    ret = pmatch(string, *p);
    if (ret < 0) {
      blkfree(blk);
      return ret;
    }
    gres |= ret;
  }

  blkfree(blk);
  return(gres == gpol);
} /* Gmatch1 */


/*
 * Gmatch()
 * Wrapper around Gmatch1.  Calls quote().
 */
int Gmatch(const char *string, const char *pattern)
{
  char *quoted_pattern;
  int ret;

  quoted_pattern = strdup(pattern);
  if (quoted_pattern == NULL) return ERR_NOMEM;
  quoted_pattern = quote(quoted_pattern);

  ret = Gmatch1(string, quoted_pattern);

  free_valid(quoted_pattern);
  return ret;
}

/*
 * pmatch()
 * Match an ordinary (non-braced) glob pattern.
 * For each metacharacter, check if it matches; return 1 when we hit the
 * end of the string and pattern simultaneously.  On error return error
 * code < 0.
 */
static int pmatch(const char *string, const char *pattern)
{
  char stringc, patternc;
  int     match, negate_range;
  char    rangec;

  for (;; ++string) {
    stringc = *string & TRIM;
    patternc = *pattern++;
    switch (patternc) {
    case 0:
      return (stringc == 0);
    case '?':
      if (stringc == 0)
        return (0);
      break;
    case '*':
      if (!*pattern)
        return (1);
      while (*string) {
        /* Skip N characters, then try to match the rest of the pattern. */
        /* Note: this is O(n^2). */
        int ret;
        ret = Gmatch1(string++, pattern);
        if (ret) 
          return ret; /* 1 on success; error code on error. */
      }
      return (0);
    case '[':
      match = 0;
      if ((negate_range = (*pattern == '^')) != 0)
        pattern++;
      while ((rangec = *pattern++) != '\0') {
        if (rangec == ']' && *(pattern-2) != '[')
          break;
        if (match)
          continue;
        if (rangec == '-' && *(pattern-2) != '[' && *pattern  != ']') {
          match = (   collate_range_cmp(stringc, *pattern & TRIM) <= 0
                      && collate_range_cmp(*(pattern-2) & TRIM, stringc) <= 0
                      );
          pattern++;
        }
        else
          match = (stringc == (rangec & TRIM));
      }
      if (rangec == 0)
        return ERR_MISSING_RBRK;
      if (match == negate_range)
        return (0);
      break;
    default:
      if ((patternc & TRIM) != stringc)
        return (0);
      break;

    }
  }
} /* pmatch */


/*
 * quote()
 * "quote" a string, in place.  Replace all occurences of \x in the string
 * with x with its high bit set.
 * Yes, this means that gmatch doesn't work correctly for eight-bit
 * text.  However, it should fail relatively sensibly, since eight-bit
 * characters will be interpreted as quoted characters, not smashed down
 * to seven-bit (possibly gramatically significant) characters.
 */
char *quote(char *orig)
{
  char *r, *w;

  for (r = w = orig; *r != '\0'; r++, w++) {
    if (*r == '\\' && *(r+1) != '\0') {
      *w = *++r | QUOTE;
    }
    else {
      *w = *r;
    }
  }
  *w = '\0';
  return orig;
} /* quote */


/*
 * gmatch_strerror()
 * Return a string corresponding to a Gmatch error code.
 */
const char *gmatch_strerror(int err)
{
  switch(err) {
  case ERR_MISSING_LBRC:
    return "Missing {";
  case ERR_MISSING_RBRC:
    return "Missing }";
  case ERR_MISSING_LBRK:
    return "Missing [";
  case ERR_MISSING_RBRK:
    return "Missing ]";
  case ERR_NOMEM:
    return "Out of memory";
  default:
    return "Invalid Error";
  }
} /* gmatch_strerror */
