/*
headers.h - add headers to outgoing requests and responses

Copyright 1998-1999 by Columbia University; all rights reserved
*/

#include "headers.h"
#include <stdio.h>   /* vsprintf() */
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>  /* strdup() */
#include "sysdep.h"
#include "strdupn.h" /* strassign() */
#include "shortform.h"


#define PRINTF_MAX 500

/*
* Add a response header. Create dynamic storage.
*/
void HeaderAppend(headers_t *hh, const char *name, const char *value)
{
  struct header_t *h;  

  if (name && value) {
    h = malloc(sizeof(struct header_t));
    if (hh->last) hh->last->next = h;
    else hh->headers = h;
    hh->last = h;
    h->next  = NULL;
    h->name  = strdup(name);
    h->value = strdup(value); 
  }
} /* HeaderAppend */


/*
 * Change a response header to the given value, or replace if
 * no such header currently exists.
 * Replaces first instance of new header; deletes all subsequent ones.
 */
void HeaderReplaceOrAppend(headers_t *hh, const char *name, const char *value)
{
  struct header_t *h;
  const char *otherform;

  otherform = get_otherform(name);

  for (h = hh->headers; h != NULL; h = h->next) {
    if (strcasecmp(h->name, name) == 0
        || (otherform != NULL && strcasecmp(h->name, otherform) == 0)) {
      strassign(&h->value, value);
      break;
    }
  }
  if (h == NULL) {
    /* We didn't find any matching headers */
    HeaderAppend(hh, name, value);
  }
  else {
    /* We found one; clear out remaining matching headers. */
    HeaderDeleteCore(hh, name, h, -1);
  }
} /* HeaderReplaceOrAppend */


/*
 * HeaderDeleteCore()
 * Delete all the response headers with the given name from the given set.
 * 'start' is the header *before* the one you want to start deleting from
 * (NULL if you want to search the whole thing).
 * Set occurrence if you want to delete only one header, counting from 1.
 * 0 means delete all.
 * This is the function which other HeaderDelete*() functions are a wrapper
 * for.
 */
void HeaderDeleteCore(headers_t *hh, const char *name,
                      struct header_t *start, int occurrence)
{
  struct header_t *h, *h_prev = NULL, *h_next;
  int count = 0;
  const char *otherform;

  otherform = get_otherform(name);

  if (start == NULL) {
    h = hh->headers;
  } else {
    h_prev = start;
    h = start->next;
  }

  for ( ; h != NULL; h = h_next) {
    h_next = h->next;
    if (strcasecmp(h->name, name) == 0
        || (otherform != NULL && strcasecmp(h->name, otherform) == 0)) {
      if (occurrence < 0 || count == occurrence) {
        /* We match, so delete this entry. */
        if (hh->headers == h) {
          hh->headers = h_next;
        } else {
          h_prev->next = h_next;
        }
        if (hh->last == h) {
          hh->last = h_prev;
        }
        free(h->name);
        free(h->value);
        free(h);
      }
      else {
        h_prev = h;
      }
      count++;
    }
    else {
      h_prev = h;
    }
  }
} /* HeaderDeleteCore() */


/*
 * Delete all the response headers with the given name from the given set.
 * SimpleWrapperAround HeaderDeleteCore().
 */
void HeaderDelete(headers_t *hh, const char *name)
{
  HeaderDeleteCore(hh, name, NULL, -1);
}


/*
 * HeaderDeleteFirst()
 * Delete the first occurrence of the header with the given name from the
 * given set.
 * Simple wrapper around HeaderDeleteCore().
 */
void HeaderDeleteFirst(headers_t *hh, const char *name)
{
  HeaderDeleteCore(hh, name, NULL, 0);
}


/*
 * HeaderDeleteSet()
 * Delete all the headers mentioned in 'del' from the set 'hh'.
 * This is O(n^2), but I don't imagine the deletion set will typically
 * be large.
 * Values of headers in 'del' are ignored; only their names matter.
 */
void HeaderDeleteSet(headers_t *hh, headers_t *del)
{
  struct header_t *h;
  for (h = del->headers; h; h = h->next) {
    HeaderDelete(hh, h->name);
  }
} /* HeaderDeleteSet */


/*
 * HeaderPrintf()
 * Create a header from a 'fmt' format string, as in sprintf(). All
 * strings are copied into internal storage. If 'replace' is true,
 * replace existing header.
 * On platforms with vsnprintf, fail fairly cleanly.  On other platforms,
 * we've probably already clobbered memory, so there's not much else we can
 * do -- just assert().
*/
void HeaderPrintf(headers_t *hh, int replace, const char *name,
                  const char *format, ...)
{
  int n;                  /* characters written */
  char value[PRINTF_MAX]; /* destination buffer */
  va_list ap;
  va_start(ap, format);

#if HAVE_VSNPRINTF
  n = vsnprintf(value, PRINTF_MAX, format, ap);
#else
  n = vsprintf(value, format, ap);
  assert(n < PRINTF_MAX);
#endif

  replace ? HeaderReplaceOrAppend(hh, name, value) : 
            HeaderAppend(hh, name, value);

  va_end(ap);
} /* HeaderPrintf */


/*
* Append a date header to the list of headers.
*/
void HeaderCftime(headers_t *hh, const char *name, const char *format,
                  time_t t, int localtime)
{
  DString d;
  struct header_t *h;

  DStringInit(&d);
  DStringCftime(&d, format, t, localtime);

  /* append to headers */
  h = malloc(sizeof(struct header_t));
  if (hh->last) hh->last->next = h;
  else hh->headers = h;
  hh->last = h;
  h->next  = NULL;
  h->name  = strdup(name);
  h->value = DStringValue(&d);
} /* HeaderCftime */


/*
* Add a header to beginning of response. Assume that 'name' is static
* and 'value' needs to be saved.
*/
void HeaderPrepend(headers_t *hh, const char *name, const char *value)
{
  struct header_t *h;

  h = malloc(sizeof(struct header_t));
  h->next  = hh->headers;
  h->name  = strdup(name);
  h->value = strdup(value); 
  hh->headers = h;
  if (!hh->last) hh->last = h;
} /* HeaderPrepend */


/*
 * HeaderMerge()
 * Merge the headers of src into dst.
 * If headers in dst already exist in src, replace the ones in src.  However,
 * if multiple headers of the same type are in 'src', copy them all.
 * Note: if this O(n^2) approch proves to be a bottleneck, an O(n log n)
 * algorithm is probably possible.
 */
void HeaderMerge(headers_t *dst, const headers_t *src)
{
  struct header_t *h, *i;
  const char *otherform;
  int first;

  for (h = src->headers; h != NULL; h = h->next) {
    /* Search src for an earlier header with this name */
    otherform = get_otherform(h->name);
    first = 1;
    for (i = src->headers; i != NULL && i != h; i = i->next) {
      if (strcasecmp(h->name, i->name) == 0
          || (otherform != NULL && strcasecmp(otherform, i->name) == 0)) {
        first = 0;
        break;
      }
    }
    if (first) 
      HeaderReplaceOrAppend(dst, h->name, h->value);
    else
      HeaderAppend(dst, h->name, h->value);
  }
} /* HeaderMerge */


/*
* Get response header value starting at location hp.
* Value returned is pointer into 'r'.
* occurrence is the occurrence of this specific header
* in the packet, starting from 1.....
*/
char *HeaderGet(headers_t *hh, const char *name, int occurrence, void **hp)
{
  char *value = "";
  struct header_t *h;
  int count = 0;
  const char *otherform;

  otherform = get_otherform(name);

  if (*hp == NULL) *hp = (void *)hh->headers;
  for (h = (struct header_t *)*hp; h != NULL; h = h->next) {
    if (strcasecmp(h->name, name) == 0
        || (otherform != NULL && strcasecmp(h->name, otherform) == 0)) {
      count++;
      if (occurrence == count) {
        value = h->value; 
        break;
      }
    }
  }
  *hp = h;
  return value;
} /* HeaderGet */


/* 
 * HeaderCopyAllBut()
 * Copy all headers, except those matching the given header, from 'src'
 * to 'dest'.
 */
void HeaderCopyAllBut(headers_t *dest, const headers_t *src, const char *name)
{
  struct header_t *h;
  const char *otherform;
  if (name != NULL) otherform = get_otherform(name);
  else otherform = NULL;

  for (h = src->headers; h; h = h->next) {
    if ((name == NULL || strcasecmp(h->name, name) != 0) 
        && (otherform == NULL || strcasecmp(h->name, otherform) != 0)) {
      HeaderAppend(dest, h->name, h->value);
    }
  }
} /* HeaderCopyAllBut */


/* HeaderCopyParsedHeaders()
 * Copy all the headers from an HTTP parse structure to an outgoing
 * headers structure.
 * Outgoing headers and parsed headers use different structures because
 * they're different; outgoing headers distinguish between line-separated
 * and comma-separated values, whereas parsed headers don't.  (Is this a
 * good thing?  I'm not sure.) This function turns everything into
 * line-separated values. 
 *
 * N.B.: the current parser puts the headers in reversed order from the
 * order they appeared in the original SIP message.  (Henning has promised
 * to fix this in the next version of the parser).  Thus, we have to reverse
 * the parsed headers before copying them.  Note that the *elements* of the
 * headers are in correct order, so we have to traverse the headers
 * backwards, but the elements forwards.  Aargh.
 */
void HeaderCopyParsedHeaders(headers_t *dest, parser_header_t *src)
{
  parser_header_t *h, *h_copy, *h_list = NULL;
  parser_element_t *e;

  /* Make a reversed copy of the headers of src.  This only needs to be a
     shallow copy. */

  for (h = src; h != NULL; h = h->next) {
    h_copy = (parser_header_t *) calloc(1, sizeof(parser_header_t));
    /* XXX check for memory failures */
    h_copy->name         = h->name;
    h_copy->elements     = h->elements;
    h_copy->last_element = h->last_element;

    h_copy->next = h_list;
    h_list = h_copy;
  }

  /* Copy elements from h_list */
  for (h = h_list; h != NULL; h = h->next) {
    for (e = h->elements; e != NULL && e->element != NULL; e = e->next) {
      HeaderAppend(dest, h->name, e->element);
    }
  }

  /* Free h_list */
  while (h_list != NULL) {
    h = h_list;
    h_list = h->next;
    free(h);
  }
} /* HeaderCopyParsedHeaders */


/* HeaderCopyOneParsedHeader()
 * Append the header with the name 'name' from an HTTP parse structure to an
 * outgoing headers structure, if it exists in the original.  If it does
 * not, do nothing.
 * If multiple headers with this name exist, they will all be appended, one
 * per line.  */
void HeaderCopyOneParsedHeader(headers_t *dest, parser_header_t *src,
                               const char *name)
{
  parser_header_t *h;
  parser_element_t *e;

  h = RFC822_Header(src, name);

  if (h == NULL)
    return;

  for (e = h->elements; e != NULL && e->element != NULL; e = e->next) {
    HeaderAppend(dest, h->name, e->element);
  }
} /* HeaderCopyOneParsedHeader */


/*
* Convert headers into message, appending to DString 'msg'.
*/
void HeaderMessage(DString *msg, const headers_t *hh)
{
  struct header_t *h;

  for (h = hh->headers; h; h = h->next) {
    /* Use consecutive DStringAppend's rather than DStringPrintf
     * here, or we can fail on very long headers. */
    DStringAppend(msg, h->name, -1);
    DStringAppend(msg, ": ", 2);
    DStringAppend(msg, h->value, -1);
    DStringAppend(msg, "\r\n", 2);
  }
} /* HeaderMessage */


/*
* Free all headers.
*/
void HeaderFree(headers_t *hh)
{
  struct header_t *h, *h_next;

  for (h = hh->headers; h; h = h_next) {
    h_next = h->next;
    free(h->name);
    free(h->value);
    free(h);
  }
  hh->headers = NULL;
  hh->last = NULL;
} /* HeaderFree */
