/*
* parser.c - Internet header parser.
*
* Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#include <stdio.h>      /* sscanf() */
#include <string.h>     /* strcpy(), strncpy(), strlen() */
#include <stdlib.h>     /* atoi(), atof(), calloc(), free() */
#include <ctype.h>      /* isspace() */
#include "rfc822.h"
#include "shortform.h"
#include "dstring.h"
#include "error.h"
#include "sysdep.h"
#include "strdupn.h"

/*
***
*** Utility functions. ***
***
*/

/*
* isws()
* Returns 1 if character is WS (linear white space, tab or space), 0
* otherwise.
*/
static int isws(unsigned char c)
{
  return ((c == '\t') || (c == ' ') || (c == '\r') || (c == '\n'));
} /* isws */


/*
* skipws()
* Return pointer to character after WS.
*/
static unsigned char *skipws(unsigned char *s)
{
  for (; *s && isws(*s); s++) ;
  return s;
} /* skipws */


/*
* iseol()
* Returns 1 if character marks end of line (LF (\n) or CR (\r))
*/
static int iseol(unsigned char c)
{
  return ((c == '\r') || (c == '\n'));
} /* iseol */
 

/*
 * istokenchar()
 * Returns 1 if character is a valid token character, 0 otherwise.
 * Definition taken from appendix C.1 of RFC2543bis.
 */
static int istokenchar(unsigned char c)
{
  return (isalnum(c) ||
          c == '-' || c == '.' || c == '!' || c == '%' || c == '*' ||
          c == '_' || c == '+' || c == '`' || c == '\'' || c == '~');
} /* istokenchar */


/*
* strdupn_flat()
* Create a buffer from 'n' characters starting at 'string'. Pad with zero and
* remove any embedded CR and LF. If 'n' < 0, create an empty string.
*/
static char *strdupn_flat(char *string, int n)
{
  char *buffer;
  char *s, *d;
  int i;

  if (n < 0) n = 0;
  buffer = malloc(n+1);
  for (s = string, d = buffer, i = 0; *s && i < n; s++, i++) {
    if (!iseol(*s)) *d++ = *s;
  }
  *d = '\0';

  return buffer;
} /* strdupn_flat */


/*
* Return type of header.
*/
static header_type_t header_type(const char *name)
{
  int i;
  struct {
    const char *name;
    header_type_t type;
  } list[] = {
     {"accept",              HT_standard},
     {"accept-disposition",  HT_standard},
     {"accept-encoding",     HT_standard},
     {"accept-language",     HT_standard},
     {"allow",               HT_standard},
     {"authorization",       HT_comma},
     {"call-id",             HT_standard},
     {"contact",             HT_nameaddr},
     {"content-disposition", HT_standard},
     {"content-encoding",    HT_standard},
     {"content-length",      HT_standard},
     {"content-type",        HT_unstructured},
     {"cseq",                HT_standard},
     {"date",                HT_date},
     {"encryption",          HT_comma},
     {"expires",             HT_date},
     {"from",                HT_nameaddr},
     {"hide",                HT_standard},
     {"max-forwards",        HT_standard},
     {"organization",        HT_unstructured},
     {"priority",            HT_standard},
     {"proxy-authenticate",  HT_comma},
     {"proxy-authorization", HT_comma},
     {"proxy-require",       HT_standard},
     {"require",             HT_standard},
     {"referer",             HT_unstructured},
     {"record-route",        HT_standard},
     {"response-key",        HT_standard},
     {"retry-after",         HT_date},
     {"route",               HT_standard},
     {"server",              HT_unstructured},
     {"subject",             HT_unstructured},
     {"to",                  HT_nameaddr},
     {"timestamp",           HT_date},
     {"unsupported",         HT_standard},
     {"user-agent",          HT_unstructured},
     {"via",                 HT_standard},
     {"warning",             HT_standard},
     {"www-authenticate",    HT_comma},
     {NULL,                  0}
  };

  /* Short-form header */
  if (name[0] != '\0' && name[1] == '\0') {
    const char* longname = get_longform(name);
    if (longname) name = longname;
  }

  for (i = 0; list[i].name; i++) {
    if (!strcasecmp(list[i].name, name)) return list[i].type; 
  }
  return HT_unstructured;
} /* header_type */


/*
***
*** Parser ***
***
*/


/*
* RFC822_ParseHeader()
* Parse a single element into top-level value (ignoring comments) and
* parameters, e.g., 'foo bar (comment) ;parameter1=x;parameter2="a
* string"' or 'value parameter=1,parameter2="string"'.  The type of
* header is determined by 'htype'.  Returns 0 if ok, -1 if syntax error. 
*/
int RFC822_ParseHeader(char *element,
  char **value, char **comment, header_type_t htype, parser_parameter_t **ppp) 
{
  unsigned char *s;
  unsigned char *start;
  int quoted_string = 0;  /* quoted string */
  int addr = 0;           /* <url> */
  int clevel = 0;         /* parentheses in comments */
  int backslash = 0;      /* saw backslash for escaping */
  int retval = 0;         /* return value */
  char parsep = ';';      /* parameter separator */
  enum {H_space, H_comment, H_attribute, H_eq_parsep, H_value,
    H_quoted_value, H_parsep } state = H_parsep, next = H_parsep;
  DString v, vv, c;
  parser_parameter_t *p_first = NULL, *p_last = NULL, *p;

  DStringInit(&v);
  DStringInit(&c);

  /* If parameters are separated by comma, set parsep accordingly. */
  if (htype == HT_comma) parsep = ',';

  /* global value, with quoted strings or <> or comment */
  s = skipws(element);
  for (; *s; s++) {
    if (quoted_string) {
      if (backslash) backslash = 0;
      else if (*s == '\\') backslash = 1;
      else if (*s == '"') quoted_string = 0;
    }
    else if (clevel > 0) {
      if (*s == ')' && *(s-1) != '\\') clevel--;
      else if (*s == '(' && *(s-1) != '\\') clevel++;
    }
    else if (addr > 0) {
      if (*s == '>') addr = 0;
    }
    else if (*s == '"') {
      quoted_string = 1;
    }
    else if (*s == '(') {
      clevel++;
    }
    else if (*s == '<') {
      addr = 1;
    }
    /* 
     * End of value string in HT_comma header, followed by parameter,
     * without leading comma.
     */
    else if (htype == HT_comma && isspace(*s)) {
      state = H_space;
      next  = H_attribute; 
      break;
    }
    /* end of value string and beginning of parameter */
    else if (*s == parsep) {
      break;
    }
    if (!backslash) {
      if (clevel > 0) DStringAppend(&c, s, 1);
      else DStringAppend(&v, s, 1);
    }
  }

  do {
    switch(state) {
    case H_space:
      if (*s == '(') {
        clevel++;
        state = H_comment;
      }
      else if (!isspace((int)*s) && *s) {
        state = next;
        start = s;
        s--;
      }
      break;

    case H_comment:
      if (*s == '(') {
        clevel++;
      }
      else if (*s == ')') {
        clevel--;
        if (clevel == 0) {
          state = next;
        }
      }
      else if (*s == '\\') {
        s++;
      } 
      if (state == H_comment) DStringAppend(&c, s, 1);
      break;

    /* look for parameter separator (comma or semicolon) */
    case H_parsep:
      if (*s == parsep) {
        state = H_space;
        next  = H_attribute;
        start = s;
      }
      else if (*s == '(') {
        state = H_comment;
        next  = H_attribute;
        clevel++;
      }
      /* anything but space or comment is an error */
      else if (!isspace(*s)) {
        retval = -1;
      }
      break;

    /* name of parameter, followed by equal sign or space or par. separ. */
    case H_attribute:
      if (!(istokenchar(*s))) {
        p = calloc(1, sizeof(parser_parameter_t));
        if (p_last)   p_last->next = p;
        if (!p_first) p_first = p;
        p_last = p;
        p->attribute = strdupn_flat(start, s - start);
        DStringInit(&vv);
        state = H_eq_parsep;
        s--;
      }
      break;

    /* looking for equal sign */
    case H_eq_parsep:
      if (*s == '=') {
        state = H_space;
        next  = H_value;
      }
      else if (*s == parsep) {
        state = H_space;
        next  = H_attribute;
      }
      break;

    /* unquoted value (token) */
    case H_value:
      if ((s == start) && (*s == '"')) {
        state = H_quoted_value;
      }
      else if (*s == parsep) {
        p->value = DStringValue(&vv);
        p->value_type = PT_token;
        state = H_space;
        next  = H_attribute;
        start = s;
      }
      else if ((*s == '\0') || isspace(*s)) {
        p->value = DStringValue(&vv);
        p->value_type = PT_token;
        state = H_space;
        next  = H_parsep;
      }
      else {
        DStringAppend(&vv, s, 1);
      }
      break;

    /* 
     * Value enclosed in quotation marks (quoted string); with \ as escape.
     */ 
    case H_quoted_value:
      if ((*s == '"')) {
        p->value = DStringValue(&vv);
        p->value_type = PT_quoted;
        state = H_parsep;
      }
      else if (*s == '\\') {
        s++;
        DStringAppend(&vv, s, 1);
      }
      else if (*s == '\0') {
        p->value = DStringValue(&vv);
        p->value_type = PT_quoted;
        retval = -1;
      }
      else {
        DStringAppend(&vv, s, 1);
      }
      break;
    }
  } while (*s++ != '\0' && retval == 0);

  /* Return first (global) value. */
  if (value) *value = DStringTrim(&v);
  else DStringFree(&v);

  if ( (value != NULL) && (*value == NULL)) {
    *value = strdup(""); /* set the value to "" if NULL */
  }

  *ppp = p_first;
  return retval;
} /* RFC822_ParseHeader */


/*
 * RFC822_Header
 * Find specific header in parsed message.
 */
parser_header_t *RFC822_Header(parser_header_t *h, const char *name)
{
  const char *otherform = get_otherform(name);

  for (; h; h = h->next) {
    if (!strcasecmp(h->name, name)
        || (otherform && !strcasecmp(h->name, otherform))) return h;
  }
  return NULL;
} /* RFC822_Header */


/*
 * RFC822_Element_Ordered
 * Find specific instance of header element in parsed message. 'order=1' is first.
 */
parser_element_t *RFC822_Element_Ordered(parser_header_t *h, const char *name,
                                        int order)
{
  const char *otherform = get_otherform(name);
  parser_element_t *e;

  for (; h != NULL; h = h->next) {
    if (!strcasecmp(h->name, name)
        || (otherform && !strcasecmp(h->name, otherform))) {
      for (e = h->elements; e != NULL; e = e->next) {
        order--;
        if (order == 0) return e;
      }
    }
  }
  return NULL;
} /* RFC822_Element_Ordered */


/*
* Append a single new element and parse into parameters if not
* unstructured.  'first' point to the first character, 'last' points to
* last character in element.  Return 0 if ok, -1 if error.
*/
static int element(parser_header_t *h, header_type_t htype, char *value)
{
  int status = 0;
  parser_element_t *e;

  e = calloc(1, sizeof(parser_element_t));
  if (!h->elements) h->elements = e;
  if (h->last_element) h->last_element->next = e;
  h->last_element = e;

  e->element = strdup(value);
  if (htype == HT_unstructured || htype == HT_date)
    e->value = e->element;
  else
    status = RFC822_ParseHeader(e->element, &e->value, &e->comment,
               htype, &e->parameters);
  return status;
} /* element */


/*
* RFC822_ParseLine()
* Extracts header from line and passes it to RFC822_???  for processing. 
* Create list of headers.  Handle case of comma-divided list as well as
* several headers with same name.  Set 'ph' to list of headers and sets
* 'length' to total length of headers.  Return 0 if ok, < 0 if syntax
* error.
*/
int RFC822_ParseLine(unsigned char *line, int linelength,
  parser_header_t **php)
{
  unsigned char *s;             /* string pointer */
  int backslash = 0;            /* saw backslash for escaping */
  parser_header_t  *h_list = *php, *h;
  header_type_t htype = HT_standard;

  /* Mark end of line; if \r\n, remove last \r. */
  if (linelength > 0 && line[linelength-1] == '\r') linelength--;
  line[linelength] = '\0';

  /* Get name. */
  for (s = line; *s; s++) {
    if (*s == ':' || isws(*s)) {
      char *name;

      /* Check if name exists already. */
      name = strdupn(line, s - line);
      h = RFC822_Header(h_list, name);
      /* If not, prepend to list. */
      if (!h) {
        h = calloc(1, sizeof(parser_header_t));
        h->name = name;
        h->next = h_list;
        h_list = h;
      }
      /* If name exists, append as element. */
      else {
        free(name);
      }
      htype = header_type(h->name);
      break;
    }
  }

  /* Position on first character after colon. */
  s = skipws(s);
  if (*s != ':') {
    return -1;
  }
  s++;

  /* Unstructured field, comma-separated parameters, or a date
   * (which contains a comma). */
  if (htype == HT_unstructured || htype == HT_comma || htype == HT_date) {
    s = skipws(s);
    element(h, htype, s);
  }
  /* Comma-separated elements. */
  else {
    int done;
    int quoted_string = 0;
    int comment = 0;
    unsigned char *start;

    do {
      start = s = skipws(s);
      for (; *s; s++) {
        if (quoted_string) {
          if (backslash) backslash = 0;
          else if (*s == '\\') backslash = 1;
          else if (*s == '"') quoted_string = 0;
        }
        else if (comment > 0) {
          if (*s == ')' && *(s-1) != '\\') comment--;
          else if (*s == '(' && *(s-1) != '\\') comment++;
        }
        else {
          if (*s == '"') quoted_string = 1; 
          else if (*s == '(') comment++;
          else if (*s == ',') break; 
        }
      }
      if (quoted_string || comment) return -2;
      done = (*s == 0);
      *s = '\0';
      element(h, htype, start);
      if (!done) s++;
    } while (!done);
  }

  *php = h_list;
  return 0;
} /* RFC822_ParseLine() */


/*
* RFC822_Parse()
* Break header into logical lines by folding continuation lines.
* Return 0 if ok, -1 if syntax error.
*/
int RFC822_Parse(char *header, int *headerlength, parser_header_t **php)
{
  unsigned char *s, *line = header;
  int linelength = 0;     /* length of line to be processed */
  int lf = 0;             /* number of LFs in a row */
  int retval = 0;

  /* Find length of each header line. */
  *php = NULL;
  for (s = header; *s && retval >= 0; s++) {
    if (*s == '\n') lf++;
    if (lf == 1) {
      /* continuation line */
      if (*(s+1) == ' ' || *(s+1) == '\t') {
        linelength++;
        lf = 0;
      }
      /* CR or LF */
      else if (*(s+1) == '\r' || *(s+1) == '\n') {
#if 0 /* XXX I think this line is wrong? */
        linelength++;
#endif
      }
      /* Not CR, LF or WS, thus a new header. */
      else {
        retval = RFC822_ParseLine(line, linelength, php);
        linelength = 0;
        line = s+1;
        lf = 0;
      }
    }
    /* After 2 LFs, we've reached the end of header. */
    else if (lf == 2) {
      if (headerlength) *headerlength = ((char *)s - header) + 1;
      retval = RFC822_ParseLine(line, linelength, php);
      break;
    }
    /* Just a regular character; count it. */
    else {
      linelength++;
    }
  }

  if (lf < 2) { /* lf should be at least 2 after parsing */
    error("RFC822_Parse", "didn't detect \"\\n\\n\"");
    /* 
     * But compute a reasonable header length anyway, otherwise, offset
     * in ReceiveUDP will only be size of request line, causing sipd to
     * think another request is in the same datagram.  we don't add 1,
     * because s is already at end of message.
     */
    if (headerlength) *headerlength = (char *)s - header;
  }
  return retval;
} /* RFC822_Parse */


/*
 * RFC822_ParameterDuplicate()
 * Duplicate a parameter list.  The new list is output in *dst.  Any
 * old lists pointed to by *dst are overwritten.
 */
void RFC822_ParameterDuplicate(parser_parameter_t **dst,
                               parser_parameter_t *src)
{
  parser_parameter_t *p, **p_new;

  p_new = dst;

  for (p = src; p != NULL; p = p->next) {
    *p_new = (parser_parameter_t *) calloc(1, sizeof(parser_parameter_t));
    if (*p_new == NULL) {
      error("RFC822_ParameterDuplicate",
            "Couldn't allocate duplicate parameter");
      goto fail;
    }
    
    strassign(&((*p_new)->attribute), p->attribute);
    strassign(&((*p_new)->value), p->value);
    (*p_new)->value_type = p->value_type;
    p_new = &((*p_new)->next);
  }

  *p_new = NULL;
  return;
    
fail:
  RFC822_ParameterFree(*dst);
  *dst = NULL;
  return;
} /* RFC822_ParameterDuplicate */


/*
* RFC822_ParameterFree()
* Free a parameter list.
*/
void RFC822_ParameterFree(parser_parameter_t *plist)
{
  parser_parameter_t *p, *p_next;

  for (p = plist; p; p = p_next) {
    p_next = p->next;
    if (p->attribute) free(p->attribute);
    if (p->value)     free(p->value);
    free(p);
  }
} /* RFC822_ParameterFree */


/*
 * RFC822_Duplicate()
 * Duplicate a list of headers.  The new list is output in *dst.  Any
 * old lists pointed to by *dst are overwritten.
 */
void RFC822_Duplicate(parser_header_t **dst, parser_header_t *src)
{
  parser_header_t   *h, **h_new;
  parser_element_t  *e, **e_new;

  if (src == NULL) {
    *dst = NULL;
    return;
  }
  
  h_new = dst;

  for (h = src; h != NULL; h = h->next) {
    *h_new = (parser_header_t *) calloc(1, sizeof(parser_header_t));
    if (*h_new == NULL) {
      error("RFC822_Duplicate", "Couldn't allocate duplicate header");
      goto fail;
    }

    strassign(&((*h_new)->name), h->name);

    if (h->elements == NULL) {
      (*h_new)->elements = NULL;
    }
    else {
      e_new = &((*h_new)->elements);
      for (e = h->elements; e != NULL; e = e->next) {
        *e_new = (parser_element_t *) calloc(1, sizeof(parser_element_t));
        if (*e_new == NULL) {
          error("RFC822_Duplicate", "Couldn't allocate duplicate element");
          *h_new = NULL;
          goto fail;
        }
        
        strassign(&((*e_new)->element), e->element);
        if (e->value == e->element)
          (*e_new)->value = (*e_new)->element;
        else
          strassign(&((*e_new)->value), e->value);
        /* ??? comment field not used? */
        RFC822_ParameterDuplicate(&((*e_new)->parameters), e->parameters);

        e_new = &((*e_new)->next);
      }

      *e_new = NULL;
    }

    h_new = &((*h_new)->next);
  }

  *h_new = NULL;

  return;

fail:
  RFC822_Free(*dst);
  *dst = NULL;
  return;
} /* RFC822_Duplicate */


/*
 * RFC822_DuplicateSingleHeader()
 * Duplicate single header.
 */
void RFC822_DuplicateSingleHeader(parser_header_t **dst, 
                                  parser_header_t *src)
{
  parser_header_t   *h, **h_new;
  parser_element_t  *e, **e_new;

  if (src == NULL) {
    RFC822_Free(*dst);
    *dst = NULL;
    return;
  }
  
  h_new = dst;
  RFC822_Free(*dst);
  *dst = NULL; /* Is this correct */

  h = src;

  *h_new = (parser_header_t *) calloc(1, sizeof(parser_header_t));
  if (*h_new == NULL) {
    error("RFC822_Duplicate", "Couldn't allocate duplicate header");
    goto fail;
  }
  
  strassign(&((*h_new)->name), h->name);
  
  if (h->elements == NULL) {
    (*h_new)->elements = NULL;
  }
  else {
    e_new = &((*h_new)->elements);
    for (e = h->elements; e != NULL; e = e->next) {
      *e_new = (parser_element_t *) calloc(1, sizeof(parser_element_t));
      if (*e_new == NULL) {
        error("RFC822_Duplicate", "Couldn't allocate duplicate element");
        *h_new = NULL;
        goto fail;
      }
      
      strassign(&((*e_new)->element), e->element);
      if (e->value == e->element)
        (*e_new)->value = (*e_new)->element;
      else
        strassign(&((*e_new)->value), e->value);
      /* ??? comment field not used? */
      RFC822_ParameterDuplicate(&((*e_new)->parameters), e->parameters);
      
      e_new = &((*e_new)->next);
    }
    
    *e_new = NULL;
  }

  h_new = &((*h_new)->next);

  *h_new = NULL;

  return;

fail:
  RFC822_Free(*dst);
  *dst = NULL;
  return;
} /* RFC822_Duplicate */


/*
 * RFC822_Reverse()
 * Reverse the order of the header elements.
 */
void RFC822_Reverse(parser_header_t *dst)
{
  parser_element_t *e, *prev;
  parser_element_t *first = dst->elements;
  parser_element_t *last = dst->last_element;

  /* Swap the first and last pointer */
  dst->elements = last;
  dst->last_element = first;

  /* Reverse the singly linked list */
  for ( prev=NULL, e = first; e; ) {
    parser_element_t *next = e->next;
    parser_element_t *this = e;
    
    e->next = prev;
    prev = this;
    e = next;
  }

  return;
} /* RFC822_Reverse */

/*
* RFC822_Free()
* Free a list of headers.
*/
void RFC822_Free(parser_header_t *hlist)
{
  parser_header_t    *h, *h_next;
  parser_element_t   *e, *e_next;

  for (h = hlist; h; h = h_next) {
    h_next = h->next;

    if (h->name) {
      free(h->name);
      h->name = 0;
    }

    for (e = h->elements; e; e = e_next) {
      e_next = e->next;
      if (e->value && e->value != e->element) {
        free(e->value);
      }
      if (e->element) free(e->element);
      RFC822_ParameterFree(e->parameters);
      free(e);
    }
    free(h);
  }
} /* RFC822_Free */


/*
* RFC822_Attribute1()
* Fetch an attribute value from a parameter for a given parameter list. 
* If 'name' is 0, retrieve the value.  If attribute has no value, return
* "".  If the attribute doesn't exist, return 0.
*/
char *RFC822_Attribute1(parser_parameter_t *plist, char *attribute)
{
  parser_parameter_t *p;

  for (p = plist; p; p = p->next) {
    if (!strcasecmp(p->attribute, attribute))
      return p->value ? (char *)p->value : "";
  }
  return NULL;
} /* RFC822_Attribute1 */


/*
 * RFC822_Attribute()
 * Fetch an attribute from a parameter list for the 'n'th element
 * (n=0,...).  If 'attribute' is NULL, retrieve the value.  If attribute
 * has no value, return "".
 */
char *RFC822_Attribute(parser_header_t *h, int index, char *attribute)
{
  parser_element_t   *e;
  int n = 0;

  for (e = h->elements; e && n < index; e = e->next, n++) {
  } 

  if (e) {
    if (attribute) {
      return RFC822_Attribute1(e->parameters, attribute);
    }
    else {
      return e->value;
    }
  }

  return NULL;
} /* RFC822_Attribute */

/*
 * RFC822_AddAttribute()
 * Adds a new attribute to the parameter list.
 * This is useful for adding tag attributes to the To:
 * or From: headers.
 * Default type value is PT_token. This method will
 * typically be called by the user.cpp user policy thread.
 */
void RFC822_AddAttribute(parser_header_t *h, char *attribute, char *value,
                         parameter_value_type_t type)
{
  parser_parameter_t *p;
  DString to;

  char *v;
  if ( !h ) 
    return;
  v = RFC822_Attribute1(h->elements->parameters, attribute);
  if ( v || !attribute || !value ) {
    /* Attribute exists, do not overwrite that.
     * Do not do anything if either attribute or value
     * are null.
     */
  }
  else {
    /* insert the parameter in the begining */
    if ( type == PT_unknown ) {
      type = PT_token;
    }

    p = (parser_parameter_t *) calloc (1, sizeof(parser_parameter_t));
    p->attribute  = strdup(attribute);
    p->next       = h->elements->parameters;
    p->value      = strdup(value);
    p->value_type = type;
    h->elements->parameters = p;

    /* The raw string should also be updated */
    DStringInit(&to);

    /* Note that we are removing the comments here.
     * Does it matter for an user agent ? */
    DStringPrintf(&to, "%s; %s=%s", 
                  h->elements->element, p->attribute, p->value);
    strassign(&h->elements->element, DStringValue(&to));
    DStringFree(&to);
  }
  return;
}
