/*
 * pattern.c - match (possibly negated) regex or glob patterns.
 */

/* Copyright 2001 Columbia University.  All rights reserved. */

#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <regex.h>

#include "glob2regex.h"
#include "dstring.h"
#include "strdupn.h"

#include "pattern.h"


/*
 * seterrorstr()
 * Set an error string inside an string of n characters.
 * The "set zeroth character to null and strcat" trick is used because
 * strncpy doesn't do what you expect.
 */
static char* seterrorstr(char *dest, const char *src, size_t n)
{
  *dest = '\0';
  strncat(dest, src, n);

  return dest;
} /* seterrorstr */


/*
 * categorize_pattern()
 * Figure out the type of pattern this is.
 * There are four types of patterns: globs, negated globs, regexes, and
 * negated regexes.
 * Regexes look like '/.../'.  Negated regexes look like '!/.../'.
 * Globs look like '???'.  Negated globs look like '^???'.
 * On error, return false and set 'error' if possible
 */
static int categorize_pattern(char **pattern,
                              pattern_type_t *pat_type,
                              int *pat_negate,
                              char *errorstr,
                              size_t errorstr_len)
{
  int pat_len;
 
  pat_len = strlen(*pattern);

  if (pat_len == 0) {
    seterrorstr(errorstr, "Zero-length pattern", errorstr_len);
    return 0;
  }

  if ((*pattern)[0] == '!') {
    /* Probably a negated regex */
    if ((*pattern)[1] != '/') {
      seterrorstr(errorstr, "\"!\" not before regex pattern",
                  errorstr_len);
      return 0;
    }
    if (pat_len < 3 || (*pattern)[pat_len-1] != '/') {
      seterrorstr(errorstr, "Malformed regex pattern",
                  errorstr_len);
      return 0;
    }
    /* Definitely a negated regex */
    *pat_type = pattern_regex;
    *pat_negate = 1;
    (*pattern)[pat_len-1] = '\0';
    *pattern += 2;
  }
  else if ((*pattern)[0] == '/') {
    /* Probably a forward regex */
    if (pat_len < 2 || (*pattern)[pat_len-1] != '/') {
      seterrorstr(errorstr, "Malformed regex pattern",
                  errorstr_len);
      return 0;
    }
    *pat_type = pattern_regex;
    *pat_negate = 0;
    (*pattern)[pat_len-1] = '\0';
    *pattern += 1;
  }
  else if ((*pattern)[0] == '^') {
    /* A negated glob */
    *pat_type = pattern_glob;
    *pat_negate = 1;
    *pattern += 1;
  }
  else {
    /* A forward glob */
    *pat_type = pattern_glob;
    *pat_negate = 0;
  }
  return 1;
} /* categorize_dialplan_pattern */

      
/*
 * pattern_compile()
 * Compile a pattern, filling in the fields of 'pattern'.
 * On error, fill in "errorstr" (up to "errorstr_len") and return false.
 */
int pattern_compile(pattern_t *comp, const char *orig_pattern,
                    char *errorstr, size_t errorstr_len)
{
  pattern_type_t pat_type;
  int pat_negate;
  int ret;
  char *pattern, *pattern_cpy, *re_pat = NULL;
  int suffix_offset;

  memset(comp, 0, sizeof(pattern_t));

  pattern_cpy = strdup(orig_pattern);
  
  pattern = pattern_cpy;

  /* Determine the type of the pattern */
  if (!categorize_pattern(&pattern, &pat_type, &pat_negate,
                          errorstr, errorstr_len)) {
    ret = 0;
    goto done;
  }

  comp->negate = pat_negate;

  if (pat_type == pattern_glob) {
    if ((ret = glob2regex(pattern, &re_pat, 1, &suffix_offset)) != 0) {
      glob2regex_regerror(ret, NULL, errorstr, errorstr_len);
      ret = 0;
      goto done;
    }

    if (pat_negate && suffix_offset != 0) {
      seterrorstr(errorstr, 
                  "Can't use parenthesized prefix with negated glob",
                  errorstr_len);
      free_valid(re_pat);
      ret = 0;
      goto done;
    }

    comp->type = pat_type;

    if ((ret = regcomp(&comp->re, re_pat, REG_EXTENDED)) != 0) {
      glob2regex_regerror(ret, &comp->re, errorstr, errorstr_len);
      free_valid(re_pat);
      ret = 0;
      goto done;
    }

    comp->suffix_offset = suffix_offset;
    comp->match_count = suffix_offset + 1;

    comp->re_str = re_pat;

    ret = 1;
  }
  else if (pat_type == pattern_regex) {
    comp->type = pat_type;

    if ((ret = regcomp(&comp->re, pattern, REG_EXTENDED)) != 0) {
      regerror(ret, &comp->re, errorstr, errorstr_len);
      ret = 0;
      goto done;
    }

    /* We support reffing back-references 0 (the whole string) and
       1 through 9 */
    comp->match_count = comp->re.re_nsub + 1 < 10 ? comp->re.re_nsub + 1 : 10;
    ret = 1;
  }      
  else {
    seterrorstr(errorstr, "Unknown pattern type", errorstr_len);
    ret = 0;
  }
  
done:
  free_valid(pattern_cpy);
  return ret;
} /* pattern_compile */


/*
 * pattern_match_core()
 * Match the string 'string' against the compiled pattern 'comp'.
 * If it matches, return > 0; if no match, return 0; on error, return < 0
 * and fill in 'error'.
 */
static int pattern_match_core(const char *string, const pattern_t *comp,
                              size_t nmatch, regmatch_t pmatch[],
                              char *errorstr, size_t errorstr_len)
{
  int ret;

  ret = regexec(&comp->re, string, nmatch, pmatch, 0);
  if (ret != 0 && ret != REG_NOMATCH) {
    /* Error of some kind */
    if (comp->type == pattern_glob) {
      glob2regex_regerror(ret, &comp->re, errorstr, errorstr_len);
    }
    else {
      regerror(ret, &comp->re, errorstr, errorstr_len);
    }
    return -1;
  }
  else if ((ret == REG_NOMATCH) != (comp->negate != 0)) {
    return 0;
  }
  else {
    return 1;
  }
} /* pattern_match_core */


/*
 * dup_and_interpolate_matches()
 * Duplicate interpolate_str.  If any $n appears in it, interpolate
 * the substring of src_str described by pmatch[n], if it exists, or 
 * the empty string.
 * XXX ???: should we instead give error messages for non-existant refs?
 * The caller must free the resulting string.
 */
static char *dup_and_interpolate_matches(const char *interpolate_str,
                                         const char *match_str,
                                         size_t nmatch,
                                         const regmatch_t *pmatch)
{
  const char *src;
  DString dst;
  int n;

  DStringInit(&dst);

  for (src = interpolate_str; *src != 0; src++) {
    if (*src == '$' && isdigit((int)(*(src+1)))) {
      src++;
      n = *src - '0';
      if (n >= nmatch ||
          pmatch[n].rm_so < 0 || pmatch[n].rm_eo < 0) {
        /* "Append" the empty string, i.e. do nothing */
      }
      else {
        DStringAppend(&dst, match_str + pmatch[n].rm_so,
                      pmatch[n].rm_eo - pmatch[n].rm_so);
      }
    }
    else {
      DStringAppend(&dst, src, 1);
    }
  }
  return DStringValue(&dst);
} /* dup_and_interpolate_matches */


/*
 * dup_and_interpolate_dollar()
 * Duplicate interpolate_str.  If any $ appears in it, interpolate
 * the substring of match_str described by pmatch[suffix_offset].
 * The caller must free the resulting string.
 */
static char *dup_and_interpolate_dollar(const char *interpolate_str,
                                        const char *match_str,
                                        int suffix_offset,
                                        size_t nmatch,
                                        const regmatch_t *pmatch)
{
  const char *src;
  DString dst;

  DStringInit(&dst);

  for (src = interpolate_str; *src != 0; src++) {
    if (*src == '$') {
      DStringAppend(&dst, match_str + pmatch[suffix_offset].rm_so,
                    pmatch[suffix_offset].rm_eo - pmatch[suffix_offset].rm_so);
    }
    else {
      DStringAppend(&dst, src, 1);
    }
  }
  return DStringValue(&dst);
} /* dup_and_interpolate_dollar */


/*
 * pattern_match()
 * Match the string 'string' against the compiled pattern 'comp'.
 * If it matches, return > 0; if no match, return 0; on error, return < 0
 * and fill in 'error'.
 * Public wrapper around pattern_match_core that doesn't try to do
 * interpolation.
 */
int pattern_match(const char *string, const pattern_t *comp,
                  char *errorstr, size_t errorstr_len)
{
  return pattern_match_core(string, comp, 0, NULL, errorstr, errorstr_len);
} /* pattern_match */


/*
 * pattern_match_and_interpolate()
 * Match the string 'string' against the compiled pattern 'comp'.  If it
 * matches, fill in 'deststr' based on 'interpolatestr'.  Otherwise
 * behave as with pattern_match().  deststr must be freed by the caller. 
 */
int pattern_interpolate(const char *string, const pattern_t *comp,
                        const char *interpolate_str, char **deststr,
                        char *errorstr, size_t errorstr_len)
{
  regmatch_t *pmatch;
  int ret;

  pmatch = (regmatch_t *) calloc(comp->match_count, sizeof(regmatch_t));
  if (pmatch == NULL) {
    seterrorstr(errorstr, strerror(ENOMEM), errorstr_len);
    return -1;
  }

  ret = pattern_match_core(string, comp, comp->match_count, pmatch,
                           errorstr, errorstr_len);
  if (ret <= 0) {
    free_valid(pmatch);
    return ret;
  }

  /* Successful match! */
  if (comp->negate) {
    /* Need to mock up a fake pmatch for interpolation. */
    int i;
    pmatch[0].rm_so = 0;
    pmatch[0].rm_eo = strlen(string);
    for (i = 1; i < comp->match_count; i++) {
      pmatch[i].rm_so = -1;
      pmatch[i].rm_eo = -1;
    }
  }        
  if (comp->type == pattern_glob) {
    *deststr = dup_and_interpolate_dollar(interpolate_str, string,
                                          comp->suffix_offset,
                                          comp->match_count, pmatch);
  }
  else {
    *deststr = dup_and_interpolate_matches(interpolate_str, string,
                                           comp->match_count, pmatch);
  }      
  free_valid(pmatch);
  return 1;
} /* pattern_interpolate */


/*
 * pattern_free()
 * Free a pattern_t.
 */
extern void pattern_free(pattern_t *comp)
{
  if (comp->type != pattern_unknown) {
    regfree(&comp->re);
  }
  free_valid(comp->pat_str);
  free_valid(comp->re_str);
} /* pattern_free */
