#ifndef RFC822_H
#define RFC822_H

#include <time.h>

#define MAX_MIME_TYPE 32

typedef enum {
  HT_standard,      /* value [;parameter] (comment) */
  HT_comma,         /* value [,parameter=value] [...] */
  HT_parameters,    /* parameter=value [;parameter=value] [...] */
  HT_date,          /* date or seconds */
  HT_nameaddr,      /* addr or Name <addr> [;parameter] (comment) */
  HT_unstructured   /* value, uninterpreted */
} header_type_t;


typedef enum {
  PT_unknown = 0,   /* Unknown parameter value format */
  PT_token,         /* parameter=token */
  PT_quoted         /* parameter="quoted string" */
} parameter_value_type_t;

/*
 * Each header consist of a name and a list of comma-separated elements. 
 * Each element has zero or more unnamed values and a set of parameters,
 * each again consisting of a name and a value. Parameter values may be
 * enclosed in quotation marks.
 */
typedef struct parser_parameter {     /* header parameters */
  char *attribute; 
  char *value;
  parameter_value_type_t value_type;
  struct parser_parameter *next;
} parser_parameter_t;

typedef struct parser_element {
  char *element;                      /* The whole element */
  char *value;                        /* top-level value */  
  char *comment;                      /* comment (in parentheses) */
  parser_parameter_t *parameters;     /* list of parameters */
  struct parser_element *next;
} parser_element_t;

typedef struct parser_header {        /* header */
  char *name;
  parser_element_t *elements;         /* first element in list */
  parser_element_t *last_element;     /* last element in list */
  struct parser_header *next;
} parser_header_t;


/* RFC822 header extraction and parsing */
extern int                 RFC822_Parse(char *header, int *length, parser_header_t **ph);
extern parser_header_t    *RFC822_Header(parser_header_t *, const char *name);
extern parser_element_t   *RFC822_Element_Ordered(parser_header_t *, const char *name, int order);
extern char               *RFC822_Attribute(parser_header_t *h, int index,
                             char *attribute);
extern char               *RFC822_Attribute1(parser_parameter_t *plist,
                              char *attribute);
extern int                 RFC822_ParseHeader(char *element,
                              char **value, char **comment,
                              header_type_t htype, parser_parameter_t **ppp);
extern void                RFC822_ParameterDuplicate(parser_parameter_t **dst,
                               parser_parameter_t *src);
extern void                RFC822_ParameterFree(parser_parameter_t *plist);
extern void                RFC822_Duplicate(parser_header_t **dst,
                              parser_header_t *src);
extern void                RFC822_DuplicateSingleHeader(
                              parser_header_t **dst,
                              parser_header_t *src);
extern void                RFC822_Reverse(parser_header_t *dest);
extern void                RFC822_Free(parser_header_t *hlist);
extern void                RFC822_AddAttribute(parser_header_t *h, 
                               char *attribute, char *value, 
                               parameter_value_type_t type);

#endif /* RFC822_H */
