/*
tcp.c - handle incoming TCP connections and requests

Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#include <stdio.h>     /* FILE */
#include <string.h>    /* strerror */
#include <pthread.h>   /* pthread_ */
#include <stdlib.h>    /* malloc/free */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>    /* close */

#include "sysdep.h"
#include "request-common.h"
#include "error.h"
#include "pthread-util.h"
#include "thread_pool.h"
#include "tcp.h"

static int tcpsock4 = -1;   /* for in TCP connection requests. IPv4 */
static int tcpsock6 = -1;   /* TCP/IPv6 */

/*
 * ReceiveTCPRequests()
 * Process incoming TCP requests and send out responses.
 * Messages without Content-Length are assumed to have no body. (This
 * differs from HTTP/1.0 conventions.)
 */
void *ReceiveTCPRequests(void *arg)
{
  struct socket_info *si = (struct socket_info *)arg;
  enum {Header, CR, CRNL, CRNLCR, NL, Body} state = Header;
  int content_length;      /* Content-Length field */
  int cl = 0;              /* bytes read in content (body) */
  int c;                   /* current character */
  char ch;                 /* same, as char */
  DString header, body;    /* header and body */
  FILE_SOCKET *f;          /* socket, as FILE */
  request_common_t *r;     /* request parameters */

  fd_set read_fd;
  struct timeval timeout;
  int retval;

  if (si == NULL) {
    error("ReceiveTCPRequests", "param is NULL\n");
    return NULL;
  }

  f = fdopen_socket(si->socket, "r");
  if (f == NULL) {
    error ("ReceiveTCPRequests", "Couldn't fdopen socket %d for reading",
           si->socket);
    free(si);
    return NULL;
  }

  c = EOF; /* To avoid warning */  
  /* Once around the loop for each message. */
  do {
    DStringInit(&header);
    DStringInit(&body);
    state          = Header;
    content_length = INFINITE_CONTENT_LENGTH;
    cl             = 0;
    r              = NULL;
    while ((cl < content_length) && ((c = getc_socket(f)) != EOF)) {
      ch = c;
      DStringAppend(state == Body ? &body : &header, &ch, 1);
      switch (state) {
      case Header:
        if (c == '\r') state = CR;
        else if (c == '\n') state = NL;
        break;
      case CR:
        if (c == '\n') state = CRNL;
        else state = Header;
        break;
      case CRNL:
        if (c == '\r') state = CRNLCR;
        else state = Header;
        break;
      case NL:
      case CRNLCR:
        if (c == '\n') {
          if (header.length > 4) {
            debug2("ReceiveTCPRequests", (struct sockaddr*)&si->sin,
                   "Incoming TCP message on socket %d:\n%s\n",
                   si->socket,
                   DStringValue(&header));
            r = (request_common_t *)
              RequestCreate(SOCK_STREAM, si->socket, &si->sin);
            r->server = si->server; /* To know original server */

            if (HTTP_Parse(DStringValue(&header), &r->p) < 0) {
             /* If this was a request whose headers didn't parse, we want
              * to try to send a 400 response to it.  On the other hand,
              * there's not much we can do with responses that didn't parse,
              * so just mark them as the unknown message type so that the
              * server ignores them and frees their memory.
             */
             if (r->p.type == http_request) {
               ResponseSetStatus((request_t *)r, 400, NULL);
             } 
             else {
               r->p.type = http_unknown_type;
             }
            }
            content_length = HTTP_ContentLength(&r->p);
            if (missing_content_length_means_zero) { /* RTSP */
              /* INFINITE_CONTENT_LENGTH is the "default" value returned by
               * the parser, implying content length was not specified 
               */
              if (content_length == INFINITE_CONTENT_LENGTH) {
                content_length = 0;
              }
            }
          }
          else 
            content_length = -1;
          
          DStringFree(&header);
          state = Body;
        }
        else 
          state = Header;
        break; 
      case Body:
        cl++;
        break;
      }
    }
    /* We now have a complete message, with header and content. */
    if (r) {
      r->request_body.content = DStringValue(&body);
      r->request_body.content_length = content_length;
      RequestProcess((request_t *)r);

      FD_ZERO(&read_fd);
      FD_SET(si->socket, &read_fd);
      timeout.tv_sec = 120;
      timeout.tv_usec = 0;
      
      retval = select(si->socket+1, &read_fd, NULL, NULL, &timeout);
        
      if (retval < 0) {
        error("ReceiveTCPRequests", "select returns an error\n");
        break;
      }
        
      if (!retval && !FD_ISSET(si->socket, &read_fd)) {
        debug("ReceiveTCPRequests", 
              "No activity on socket %d. Freeing resources\n", si->socket);
        break;
      }
    } /* if (r) */
  } while (c != EOF);

  fclose_socket(f);
  
  debug2("ReceiveTCPRequests", (struct sockaddr*)&si->sin, 
         "Connection closed (socket %d)\n", si->socket);
  
  free(si);
  
  return NULL;
} /* ReceiveTCPRequests */


/*
 * TCP request receiver. 
*/
void *ReceiveTCP(void *arg)
{
  struct tcp_thread_arg *ptr = (struct tcp_thread_arg *)arg; 
  void *server               = ptr->server;
  char *primaryIP4           = ptr->primaryIP;
  char *primaryIP6           = ptr->primaryIP6;
  unsigned short port        = ptr->port;  
  struct socket_info *si;                 /* accept socket information */
  struct sockaddr_storage sin4;           /* bind socket address/ ipv4 */
  struct sockaddr_storage sin6;           
  fd_set fds;                             /* for select */
  int sinlen;                             /* length of address */
  int thr_err;                            /* thread error condition */
  int retval;
  int max;

  /* Create TCP/IPv4 listening socket. */
  memset(&sin4, 0, sizeof(sin4));
  sin4.ss_family = AF_INET;

  if (primaryIP4) {
    retval = AFIGetAddrByName(&sin4, primaryIP4);
    if (!retval) {
      error("ReceiveTCP", 
            "Bad address %s. Defaulting to (any)\n", primaryIP4);
    }
  }
  
  AFISetPort(&sin4, port);
  if ((tcpsock4 = AFIGetTCPServerSocket(&sin4)) < 0) {
    error("ReceiveTCP", "socket\n");
    return NULL;
  }

  /* if port is 0, the OS would have assigned a port number.
   * find it out, so that we can use the same for IPv6
   */
  if (port == 0) {
    struct sockaddr_in sin;
    int sinlen = sizeof(sin);
    if (getsockname(tcpsock4, (struct sockaddr *)&sin, &sinlen) == 0) {
      port = ntohs(sin.sin_port);
      ptr->port = port;
      AFISetPort(&sin4, port);
      debug("ReceiveTCP", "bind to port %hu\n", port);
    }
    else {
      error("ReceiveTCP", "getsockname\n");
    }
  }

  /* Create TCP/IPv6 listening socket. */
  if (AFIIPv6Active()) {
    memset(&sin6, 0, sizeof(sin6));
    sin6.ss_family = AF_INET6;
  
    if (primaryIP6) {
      retval = AFIGetAddrByName(&sin6, primaryIP6);
      if (!retval) {
        error("ReceiveTCP", 
              "Bad address %s. Defaulting to (any)\n", primaryIP6);
      }
    }
    
    AFISetPort(&sin6, port);
 
    if ((tcpsock6 = AFIGetTCPServerSocket(&sin6)) < 0) {
      error("ReceiveTCP", "socket. IPv6 TCP disabled...\n");
      return NULL;
    }
  }

  retval = 0;
  while (retval != -1) {
  
    /* since we have two, tcpsock4 and tcpsock6, do a select on
     * all applicable sockets. when data is available, dispatch
     * it after doing a accept.
    */
    FD_ZERO(&fds);
    FD_SET(tcpsock4, &fds);
    max = tcpsock4 + 1;   
    if (tcpsock6 > -1) {
      FD_SET(tcpsock6, &fds);
      max = tcpsock6 + 1;
    }
    
    retval = select(max, &fds, NULL, NULL, NULL);
 
    if (retval >= 1) {
      if (FD_ISSET(tcpsock4, &fds)) {
        si = (struct socket_info *) malloc(sizeof(struct socket_info));
        if (si == NULL) {
          error("ReceiveTCP", "malloc\n");
          return NULL;
        }
    
        si->server  = server;
      
        sinlen = sizeof(struct sockaddr_in);
        if ((si->socket = accept(tcpsock4, 
                                (struct sockaddr *)&si->sin, &sinlen)) < 0) {
          error("ReceiveTCP", "accept\n");
          free(si);
          return NULL;
        }
    
        debug2("ReceiveTCP", (struct sockaddr*)&si->sin, 
               "Got connection (socket %d)\n", si->socket);

        if ((thr_err = pool_schedule(ReceiveTCPRequests, (void *)si))) {
          error("ReceiveTCP", "unable to get thread from pool\n");
          closesocket(si->socket);
          free(si);
        }
      } /* FD_ISSET(tcpsock4, ...) */
        
      if (tcpsock6>0 && FD_ISSET(tcpsock6, &fds)) {
        si = (struct socket_info *) malloc(sizeof(struct socket_info));
        if (si == NULL) {
          error("ReceiveTCP", "malloc\n");
          return NULL;
        }
    
        si->server  = server;
      
        sinlen = sizeof(struct sockaddr_in6);
        if ((si->socket = accept(tcpsock6, 
                                (struct sockaddr *)&si->sin, &sinlen)) < 0) {
          error("ReceiveTCP", "accept\n");
          free(si);
          return NULL;
        }
    
        debug2("ReceiveTCP", (struct sockaddr*)&si->sin, 
               "Got connection (socket %d)\n", si->socket);

        if ((thr_err = pool_schedule(ReceiveTCPRequests, (void *)si))) {
          error("ReceiveTCP", "unable to get thread from pool\n");
          closesocket(si->socket);
          free(si);
        }
      } /* FD_ISSET(tcpsock6, ...) */
    } /* retval >= 1 */
  } /* while */

  /*
   * XXX ptr must not be free'd because otherwise libsip++ code may access
   * it without knowing that the thread terminated abnormally and
   * ptr is freed or not. Since this is a one time memory leak, it
   * is tolerable.
   */
  /* free(ptr); */

  return NULL;
} /* ReceiveTCP */


/*
 * CloseTCP
 *  -Close the tcp socket.
 *  -Todo: Kill all sleeping threads in ReceiveTCPRequests.
*/
void CloseTCP() {
 closesocket(tcpsock4);
 closesocket(tcpsock6);
}


