/* timegm.c - define timegm() if system doesn't have it. */
/* Copyright 2000 Columbia University */

#include <time.h>
#include "timegm.h"

/* timegm()
 * like mktime(), but in UTC rather than local timezone.
 *
 * HACK.  Convert the tm, as local time, to a time_t, with mktime().  Then
 * convert it back to a tm with gmtime(), and then convert it to a time_t
 * *again* with mktime().  The difference between these two values is the
 * GMT offset, so subtract this difference to the original time.
 *
 * Example: in US/Eastern:
 * tm1 = {100, 0, 1, 0, 0, 0} (Jan 1, 2000, 00:00:00).
 * time1 = mktime(tm1) = 946702800
 * tm2 = gmtime(time1) = {100, 0, 1, 5, 0, 0} (Jan 1, 2000, 05:00:00).
 * time2 = mktime(tm2) = 946720800
 * offset = time2 - time1 = 18000
 * answer = time1 - offset = 946684800
 *
 * NOTE: this will fail (pick the wrong offset) for times right before a
 * daylight-savings-time change: for example, for
 * tm1 = {100, 3, 2, 1, 0, 0} (Apr 2, 2000, 01:00:00)
 * it picks the offset 14400, when the correct offset is 18000.
 */
time_t timegm(struct tm *tm1p)
{
  time_t time1, time2, offset, answer;
  struct tm tm2, *tm2p;
  
  time1 = mktime(tm1p);
  tm2p = gmtime_r(&time1, &tm2);
  time2 = mktime(tm2p);
  offset = time2 - time1;
  answer = time1 - offset;

  /* If mktime fails, it'll return -1.  Do the same thing here. */
  if (time1 == -1 || time2 == -1)
    return -1;

  return answer;
} /* timegm */
