/*
 * udp.c - handle incoming UDP requests
 * Copyright 1998-2001 by Columbia University; all rights reserved.
 *   Rewritten October 2001 for IPv6 support.
*/

#include <stdio.h>                  /* perror() */
#include <stdlib.h>                 /* free() */
#include <string.h>                 /* strstr, ... */
#include <memory.h>                 /* memset() */
#include <unistd.h>                 /* close() */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include "afi.h"                    /* AFI socket functions */
#include "thread_pool.h"            /* thread pool */
#include "request-common.h"
#include "error.h"
#include "sysdep.h"
#include "udp.h"

#define MAX_REQUEST 8192            /* large enough for UDP packet */

typedef struct udp_worker_data {
  int sock_response;               /* the socket to use for responses */
  struct sockaddr_storage sin;     /* the address of remote */
  void* server;                    
  char *msg;                       /* the incoming datagram */
  int msglen;                      /* actual length of datagram */
} UDPWorkerData;


/* forward declarations */
static void* ProcessUDPPacket(void *arg);
static void join_sip_mcast_group(int sockid); 
static int UDPDispatch(int address_family, int sockid);
static int getMessageType(const char *buf);
void CloseUDP();

static int sock4;                   /* socket for in requests, ip4 */
static int sock6 = -1;              /* socket for in requests, ip6 */
static int sock_response4;          /* socket for responses, ip4 */
static int sock_response6 = -1;     /* socket for responses, ip6 */
static void *server;              

/*
 * ReceiveUDP()
 * UDP SIP receiver:  get message from socket and pass to
 * ProcessRequest(). One message may contain several requests.
 */
void *ReceiveUDP(void *arg)
{
  struct udp_thread_arg *ptr = (struct udp_thread_arg *)arg;
  char *primaryIP4 = ptr->primaryIP;  /* not null => overrides INADDR_ANY */
  char *primaryIP6 = ptr->primaryIP6; 
  unsigned short port = ptr->port;    /* UDP port */
  
  struct sockaddr_storage sin4;       /* ipv4 address for binding */
  struct sockaddr_storage sin6;       /* ipv6 address for binding */
  fd_set fds;                         /* select */
  int retval = 0;          
  int max; 
  
  server = ptr->server;

  /* Algorithm:
   * 1) Initialize ipv4 and ipv6 sockets for requests and responses.
   * 2) Do a select on all applicable sockets.
   * 3) When select indicates data, call UDPDispatch function with 
   *    that socket, and the appropriate address family.
   * 4) repeat this indefinitely until an error is encountered.
  */
  
  /* ipv4 init */
  memset(&sin4, 0, sizeof(sin4));
  sin4.ss_family = AF_INET;
  if (primaryIP4) {
    retval = AFIGetAddrByName(&sin4, primaryIP4);
    if (!retval) {
      error("ReceiveUDP", 
            "Bad address %s. Defaulting to (any)\n", primaryIP4);
    }
  }
  AFISetPort(&sin4, port);
  if ((sock4 = AFIGetUDPServerSocket(&sin4)) < 0) {
    error("ReceiveUDP", "Unable to get a udp/ipv4 socket.\n");
    return NULL;
  }
  
  /* Note that if the port were zero, the operating system
   * would have picked one port for UDP/IP4. Use the same
   * for IPv6 also. so, we should get the correct port
   * before binding ipv6.
  */
  if (port == 0) {
    int sinlen = sizeof(struct sockaddr_in);
    if (getsockname(sock4, (struct sockaddr *)&sin4, &sinlen) == 0) {
      port = AFIGetPort((struct sockaddr*)&sin4);
      ptr->port =  port;
      AFISetPort(&sin4, port);
      debug("ReceiveUDP", "bind to port %hu\n", port);
    }
    else {
      error("ReceiveUDP", "getsockname\n");
    }    
  }

  /* Create UDP send sockets for ipv4. */
  if ((sock_response4 = AFIGetUDPSocket(AF_INET)) < 0) {
    perror("UDP socket for responses");
  }

  /* ipv6 init */
  if (AFIIPv6Active()) { /* dont bother if ipv6 is not usable */
    memset(&sin6, 0, sizeof(sin6));
    sin6.ss_family = AF_INET6;
    
    if (primaryIP6) {
      retval = AFIGetAddrByName(&sin6, primaryIP6);
      if (!retval) {
        error("ReceiveUDP", 
              "Bad address %s. Defaulting to (any)\n", primaryIP6);
      }
    }
    
    AFISetPort(&sin6, port);
    
    if ((sock6 = AFIGetUDPServerSocket(&sin6)) < 0) {
      error("ReceiveUDP", "Unable to bind ipv6 socket. IPv6 disabled\n");
    }
    else {
      debug("ReceiveUDP", "Accepting requests on IPv6 port %d\n", port);
    }

  
    /* Create UDP send socket for ipv6 only if sock6 >0. */
    if (sock6 > 0 && (sock_response6 = AFIGetUDPSocket(AF_INET6)) < 0) {
      perror("UDP/IPv6 socket for responses");
    }
  } /* AFIIPv6Active */

  /* Accept packets on multicast address sip.mcast.net as well. */
  join_sip_mcast_group(sock4);
 
  while (retval != -1) {
  
    /* since we have two sockets sock4 and sock6, do a select on
     * all applicable sockets. when data is available, dispatch
     * it after doing a recvfrom.
    */
    FD_ZERO(&fds);
    FD_SET(sock4, &fds);
    max = sock4 + 1;
    if (sock6 > -1) {
      FD_SET(sock6, &fds);
      max = sock6 + 1;
    }
    
    retval = select(max, &fds, NULL, NULL, NULL);
 
    if (retval >= 1) {
      if (FD_ISSET(sock4, &fds)) {
        retval = UDPDispatch(AF_INET, sock4);
      }
        
      if (sock6>0 && FD_ISSET(sock6, &fds)) {
        retval = UDPDispatch(AF_INET6, sock6);
      }
    }
  } /* for */

  CloseUDP();

  /*
   * XXX ptr must not be free'd because libsip++ code may access
   * it without knowing that the thread terminated abnormally and
   * ptr is freed or not. Since this is a one time memory leak, it
   * is tolerable.
   */
  /* free(ptr); */
  
  return NULL;
} /* ReceiveUDP */

/* UDPDispatch
 *  Read an incoming datagram from sockid, and dispatch it to 
 *  ProcessUDPRequests in a pool thread.
*/
int UDPDispatch(int address_family, int sockid) {
  
  int sinlen = sizeof(struct sockaddr_in);
  int request_type;             
  int maxqueuesize, curqueuesize; /* whether to parse or drop */
  UDPWorkerData *workerData;

  if (address_family == AF_INET6) {
    sinlen = sizeof(struct sockaddr_in6);
  }

  /* workerData is the arg to pass to ProcessUDPRequests. malloc */
  workerData = (UDPWorkerData*) calloc(1, sizeof(UDPWorkerData));
  workerData->msg = (char*) malloc(MAX_REQUEST * sizeof(char));
  workerData->server = server;

  if (address_family == AF_INET) {
    workerData->sock_response = sock_response4;
  }
  else
    workerData->sock_response = sock_response6;

  workerData->msglen = recvfrom(sockid, workerData->msg, 
                                MAX_REQUEST - 1, 0, 
                                (struct sockaddr*)&workerData->sin, 
                                &sinlen);
    
  if (workerData->msglen < 0) {
    error("ReceiveUDP", "recvfrom\n");
    return -1;
  }
  
  workerData->msg[workerData->msglen] = '\0';
 
  debug2("ReceiveUDP", (struct sockaddr*)&workerData->sin, 
         "Got packet (%d) bytes\n%s\n",
         workerData->msglen, workerData->msg);
  
 /*
  * Check if this is a request or response. We check for responses
  * by seeing if this message starts with SIP or RTSP.
  *
  * If this is a response we always schedule it in a new thread,
  * which does parsing, and response processing.
  *
  * If this is a request, we schedule it if the current_queue_length
  * is less than max_queue_length/2 (the reasoning is: max_queue_length
  * is set to 2*total_threads. even if all threads are active, this will
  * ensure that a thread picks up this job, once it has completed what
  * it is doing currently - which typically will be less than 10ms).
  */
  pool_get_status(NULL, &maxqueuesize, &curqueuesize);
    
  if ((request_type = getMessageType(workerData->msg)) != http_response 
       && (curqueuesize >= (maxqueuesize >> 1))) {
    error("udp","dropping request\n");
    free(workerData->msg);
    free(workerData);
    return 0;
  }
    
 /*
  * what happens if pool_schedule returns < 0 (i.e., EBUSY) ???
  * in this case we drop requests. for responses, we serve it
  * iteratively (i.e., instead of calling ProcessUDPRequest in 
  * a new thread, we call it in this thread itself.
  */
  if (pool_schedule(ProcessUDPPacket, (void*) workerData) != 0) {
    if (request_type == http_request)  {
      error("udp","dropping packet2\n");
      return 0;
    }
    else
     ProcessUDPPacket((void*)workerData);
  }

  return 0;
} /* UDPDispatch */


/* ProcessUDPPacket
 *  Parse the incoming datagram, and dispatch it to RequestCreate if it
 *  is well-formed.
*/
void* ProcessUDPPacket(void *arg) 
{

  UDPWorkerData* workerData = (UDPWorkerData*) arg;

  request_common_t *r;     /* incoming request */
  int offset = 0;          /* offset into message */  
  int content_length = 0;  /* content_length */
  DString body;            /* space for message body */
  char* msg;

  /* 
   * Extract all requests from message.  For invalid messages, the
   * header length is zero; we must exit in that case to avoid looping
   * forever.
   */
  msg = workerData->msg;

  do {
  
    r = (request_common_t *) RequestCreate(SOCK_DGRAM, 
                                           workerData->sock_response,
                                           &workerData->sin);

    r->server = workerData->server; /* server/port at which recvd */

    if (HTTP_Parse(&msg[offset], &r->p) < 0) {
     /* If this was a request whose headers didn't parse, we want
      * to try to send a 400 response to it.  On the other hand,
      * there's not much we can do with responses that didn't parse,
      * so just mark them as the unknown message type so that the
      * server ignores them and frees their memory.
      */
      if (r->p.type == http_request) {
        ResponseSetStatus((request_t *)r, 400, NULL);
      } 
      else {
        r->p.type = http_unknown_type;
      }
    }
    offset += r->p.header_length;
    content_length = HTTP_ContentLength(&r->p);
    if (content_length > 0) {
      if (offset + content_length > workerData->msglen) 
        content_length = workerData->msglen - offset;
      DStringInit(&body);
      DStringAppend(&body, &msg[offset], content_length);
      r->request_body.content = DStringValue(&body);
      r->request_body.content_length = content_length;
   }
   offset += content_length;
      
   RequestProcess((request_t *)r);
      
  } while ((offset+1 < workerData->msglen) && (r->p.header_length > 0));

  debug("ProcessUDPPacket", "Finished parse and schedule...\n");
  
  free(workerData->msg);
  workerData->msg = NULL;
  free(workerData);
  
  return NULL;
} /* ProcessUDPRequests */


/*
 * CloseUDP
 *  Should be called during cleanup. Close all sockets.
*/
void CloseUDP() {
  closesocket(sock4);
  closesocket(sock6);
  closesocket(sock_response4);
  closesocket(sock_response6);
}


/*
 * getMessageType
 *
 * Tentatively decide whether the incoming message is a sip/rtsp
 * request or response.
 */
static int getMessageType(const char *buf) 
{
  char line[1024];
  char *p;
  int tocopy = sizeof(line);
  
  /* copy the first line */
  if (!(p = strchr(buf,'\n'))) {
    return http_response;
  }
  
  if ((p - buf) < tocopy) {
    tocopy = p-buf;
  }
  
  strncpy(line, buf, tocopy);
  
  if ((p = strstr(line, "SIP"))) {
    return (p == line) ? http_response : http_request;
  }
  
  if ((p = strstr(line,"RTSP"))) {
    return (p == line) ? http_response : http_request;
  }
  
  /*
   * better to err on the wrong side.
   */
  return http_response;
  
} /* getMessageType */


/* join_sip_mcast_group
 *  For the socket 'sockid', add ability to listen to sip.mcast.net.
 *  XXX: Todo IPv6 equivalent.
*/
static void join_sip_mcast_group(int sockid) {

#ifdef IP_ADD_MEMBERSHIP     
  
  struct ip_mreq mreq;               /* multicast group */
  
  mreq.imr_multiaddr.s_addr = inet_addr("224.0.1.75");
  mreq.imr_interface.s_addr = htonl(INADDR_ANY);
  
  if (setsockopt(sockid, IPPROTO_IP, IP_ADD_MEMBERSHIP, 
                (char *)&mreq, sizeof(mreq)) < 0) {
    error("ReceiveUDP", "IP_ADD_MEMBERSHIP\n");
  }

#endif /* IP_ADD_MEMBERSHIP */

  return;
} /* join_sip_mcast_group */
   
