/*
 * cinema_fastsql_table
 *
 * -Subclass of cinema_table that manages a in-memory database of 
 * the put, aliases, and contacts (currently) tables and does cache
 * management.
 *
 * This is the class that needs to be used for database access.
 *
 * (c) 2001 Columbia University. All rights reserved.
 */
#ifndef CINEMA_FASTSQL_TABLE_H__
#define CINEMA_FASTSQL_TABLE_H__	

#include <string>

#include "cinema_table.h"

#include "cinema_fastsql_cache.h"

class cinema_fastsql_table : public cinema_table  
{
  
 private:
  
  cinema_fastsql_cache *cacheRef; /* the cache manager */
  
 public:
  cinema_fastsql_table(std::string uri);
  cinema_fastsql_table(uri_t& uri);
  
  virtual ~cinema_fastsql_table();

  virtual char* get_key(const uri_t *uri) const;
  virtual uri_t get_uri(const char *key) const;
  
  virtual alias_table_entry *get_aliases_by_alias(const char* alias);
  
  virtual put_table_entry *get_put_by_user(const char* user);
  
  virtual contact_table_entry** get_contacts_by_user(const char *user,
                                                     int *count = NULL);
  
  virtual void insert_contact(const contact_table_entry *entry);
  
  virtual void delete_contacts_by_user(const char *user);
  
  virtual void delete_contact_by_contact(const char *user,
                                         const char *contact);
  
  virtual void delete_expired_contacts_by_user(const char *user, 
                                               time_t expire_time);

  virtual sub_table_entry *get_sub_by_fromto(const char* from, const char* to);
  
  virtual int num_subscriptions(const char *to);
  
  virtual sub_table_entry **get_all_subscriptions(const char* to, int *count);
  
  virtual void insert_sub(const sub_table_entry *entry);  

  virtual void delete_sub_by_fromto(const char* from, const char *to);
  
  TABLE_SCHEME("fastsql");
  
}; /* cinema_fastsql_table */


/*
 * Since all the functions of this class delegate to cinema_fastsql_cache,
 * its worth inlining them. so, all method definitions are kept here,
 * instead of a cpp file.
 */

#define CHECK_VALID(xptr) \
  if(!xptr) \
  { \
      throw(new cinema_error(std::string("Cache not running...\n"))); \
  } 

/*
 * cinema_fastsql_table::cinema_fastsql_table(string uri)
 */
inline cinema_fastsql_table::cinema_fastsql_table(std::string uri)
{
  
  cacheRef = (cinema_fastsql_cache*)cinema_fastsql_cache::get_fastsql_ref(uri);
  
} /* cinema_fastsql_table::cinema_fastsql_table(string) */


/*
 * cinema_fastsql_table::cinema_fastsql_table(uri_t&)
 */
inline cinema_fastsql_table::cinema_fastsql_table(uri_t& uri) 
{
  cacheRef = (cinema_fastsql_cache*)cinema_fastsql_cache::get_fastsql_ref(uri);
  
} /* cinema_fastsql_table::cinema_fastsql_table(uri_t) */


/*
 * cinema_fastsql_table::~cinema_fastsql_table
 */
inline cinema_fastsql_table::~cinema_fastsql_table()
{
  cinema_fastsql_cache::free_fastsql_ref(cacheRef);
}

/*
 * get_key from cinema_table.
 */
inline char* cinema_fastsql_table::get_key(const uri_t *uri) const 
{
  
  CHECK_VALID(cacheRef); 
  
  return cacheRef->get_key(uri);

} /* cinema_fastsql_table::get_key(const uri_t *uri) const */


/*
 * get_uri from cinema_table.
 */
inline uri_t cinema_fastsql_table::get_uri(const char *key) const 
{
  
  CHECK_VALID(cacheRef); 
  
  return cacheRef->get_uri(key);
  
} /*  cinema_fastsql_table::get_uri(const char *key) const  */


/*
 * get_aliases_by_alias from cinema_table.
 */
inline alias_table_entry*
cinema_fastsql_table::get_aliases_by_alias(const char* alias)
{
  
  CHECK_VALID(cacheRef); 
  
  return cacheRef->get_aliases_by_alias(alias);

}/* cinema_fastsql_table::get_aliases_by_alias(const char* alias) */


/*
 * get_put_by_user from cinema_table.
 */
inline put_table_entry*
cinema_fastsql_table::get_put_by_user(const char* user)
{
  
  CHECK_VALID(cacheRef); 
  
  return cacheRef->get_put_by_user(user);

} /* cinema_fastsql_table::get_put_by_user(const char* user) */


/*
 * get_contacts_by_user from cinema_table.
 */
inline contact_table_entry** 
cinema_fastsql_table::get_contacts_by_user(const char *user,
					   int *count)
{

  CHECK_VALID(cacheRef); 

  return cacheRef->get_contacts_by_user(user, count);

} /* cinema_fastsql_table::get_contacts_by_user(const char *user,
     int *count) */


/*
 * insert_contact from cinema_table.
 */
inline void 
cinema_fastsql_table::insert_contact(const contact_table_entry *entry)
{
  
  CHECK_VALID(cacheRef); 
  
  cacheRef->insert_contact(entry);

} /* cinema_fastsql_table::insert_contact(const contact_table_entry *entry) */


/*
 * delete_contacts_by_user from cinema_table.
 */
inline void 
cinema_fastsql_table::delete_contacts_by_user(const char *user)
{
  
  CHECK_VALID(cacheRef); 
  
  cacheRef->delete_contacts_by_user(user);

} /* cinema_fastsql_table::delete_contacts_by_user(const char *user) */


/*
 * delete_contact_by_contact from cinema_table.
 */
inline void 
cinema_fastsql_table::delete_contact_by_contact(const char *user,
						const char *contact)
{
  
  CHECK_VALID(cacheRef); 
  
  cacheRef->delete_contact_by_contact(user,contact);

} /* cinema_fastsql_table::delete_contact_by_contact(const char *user,
     const char *contact) */


/*
 * delete_expired_contacts_by_user from cinema_table.
 */
inline void 
cinema_fastsql_table::delete_expired_contacts_by_user(const char *user, 
						      time_t expire_time)
{
  
  CHECK_VALID(cacheRef); 
  
  cacheRef->delete_expired_contacts_by_user(user,expire_time);

} /* cinema_fastsql_table::delete_expired_contacts_by_user(const char *user, 
     time_t expire_time) */

inline sub_table_entry*
cinema_fastsql_table::get_sub_by_fromto(const char* from, const char* to)
{
  
  CHECK_VALID(cacheRef); 
  
  return cacheRef->get_sub_by_fromto(from, to);
  
} /* cinema_fastsql_table::get_sub_by_fromto */


inline void 
cinema_fastsql_table::insert_sub(const sub_table_entry *entry)
{
  CHECK_VALID(cacheRef);

  cacheRef->insert_sub(entry);

} /* cinema_fastsql_table::insert_sub */


inline int 
cinema_fastsql_table::num_subscriptions(const char *to)
{
  
  CHECK_VALID(cacheRef);
  
  return cacheRef->num_subscriptions(to);
  
} /* cinema_fastsql_table::num_subscriptions */

inline sub_table_entry **
cinema_fastsql_table::get_all_subscriptions(const char* to, int *count) 
{
  
  CHECK_VALID(cacheRef);
  
  return cacheRef->get_all_subscriptions(to, count);
  
} /* cinema_fastsql_table::get_all_subscriptions */


inline void 
cinema_fastsql_table::delete_sub_by_fromto(const char* from, const char *to)
{
  CHECK_VALID(cacheRef);

  cacheRef->delete_sub_by_fromto(from, to);
} /* cinema_fastsql_table::delete_sub_by_fromto */


#endif   /* cinema_fastsql_table.h/cpp */
