/* -*- C++ -*- */
/* cinema_mysql_table.h -- A mysql database implementation, intended as a
 * sample from which real database implementations can be written. */
   
/* Copyright 2001 Columbia University.  All rights reserved. */
#ifdef __linux__
#define _GNU_SOURCE /* Needed for strptime declaration from <time.h> */
#endif

#include <strstream>
#include <string>
#include <vector>

#include <cstdlib>
#include <ctime>

extern "C" {
#include <pthread.h>

#ifndef HAVE_TIMEGM
#include "timegm.h"
#endif

#include "dstring.h"
#include "uri.h"
#include "strdupn.h"
}

#include "stack_lock.h"

#include "cinema_mysql_table.h"


using namespace std;

/* 
 * SQL_DATETIME_BUFSIZE is the strlen of the string
 * "YYYY-MM-DD HH:MM:SSn" 
 */
#define SQL_DATETIME_BUFSIZE 32

/*
 * cinema_mysql_table::cinema_mysql_table(string)
 * Construct a cinema mysql table, based on the given URI given as a string.
 */
cinema_mysql_table::cinema_mysql_table(string uri)
{
  table_uri = URI_Parse(uri.c_str());
  if (table_uri.scheme != table_scheme()) {
    URI_Free(&table_uri);
    throw cinema_invalid_arg("Invalid table scheme for mysql table");
  }
  
  try {
    init_connection();
  }
  catch(...) {
    URI_Free(&table_uri);
    throw;
  }
} /* cinema_mysql_table::cinema_mysql_table(string) */


/*
 * cinema_mysql_table::cinema_mysql_table(uri_t&)
 * Construct a cinema mysql table, based on the given URI given as a uri_t.
 */
cinema_mysql_table::cinema_mysql_table(uri_t& uri)
{
  if (uri.scheme != table_scheme()) {
    throw cinema_invalid_arg("Invalid table scheme for mysql table");
  }

  table_uri = URI_Copy(uri);

  init_connection();
} /* cinema_mysql_table::cinema_mysql_table(uri_t) */


/*
 * cinema_mysql_table::~cinema_mysql_table()
 * Destruct a cinema mysql table.
 */
cinema_mysql_table::~cinema_mysql_table()
{
  delete conn;
  URI_Free(&table_uri);
} /* cinema_mysql_table::~cinema_mysql_table() */


/*
 * cinema_mysql_table::init_connection()
 * Initialize conn, based on table_uri.  Throw an exception
 * on error.
 */
void
cinema_mysql_table::init_connection()
{
  if (table_uri.password != NULL &&
      strcmp(table_uri.password, "NULL") == 0) {
    /* Backward misfeature-compatibility -- treat the literal password
     * "NULL" as the null password.
     * This was a workaround for an earlier, broken URI parser.
     */
    free(table_uri.password);
    table_uri.password = NULL;
  }

  /* if no database is specified, default to sip */ 
  if (!table_uri.path)
	strassign(&table_uri.path, "sip");
 
  try {
    conn = 
      new MysqlConnection(table_uri.user, table_uri.password,
                          table_uri.path, table_uri.host, table_uri.port);

  }
  catch (exception& e) {
    URI_Free(&table_uri);
    throw cinema_internal_err(string("Couldn't create MySQL table: ") +
                              e.what());
  }
} /* cinema_mysql_table::init_connection() */


/*
 * cinema_mysql_table::get_key(const uri_t *)
 * Get a database key suitable for the mysql table, based on a URI.
 * For SIP, this is currently just user@host; other URI schemes use
 * the whole URI.  XXX TODO?: make SIP the whole URI too.  Will
 * require updating the database.
 * Use DString rather than std::string so we can use URI_SPrint.
 */
char*
cinema_mysql_table::get_key(const uri_t *uri) const
{
  DString userid;

  DStringInit(&userid);

  if (strcasecmp(uri->scheme, "sip") != 0) {
    URI_SPrint(uri, &userid);
  }
  else {
    if (uri->user == NULL && uri->host == NULL) {
      /* Shouldn't happen */
      DStringFree(&userid);
      throw cinema_invalid_arg("URI argument has neither user nor host");
    }
    if (uri->user != NULL) {
      DStringAppend(&userid, uri->user, -1);
      DStringAppend(&userid, "@", -1);
    }
    if (uri->host != NULL) {
      DStringAppend(&userid, uri->host, -1);
    }
  }
  
  return DStringValue(&userid);
} /* cinema_mysql_table::cinema_mysql_table(const uri_t *) */


/*
 * cinema_mysql_table::get_uri(const char*)
 * Get a canonical URI which represents the given database key.
 */
uri_t
cinema_mysql_table::get_uri(const char *key_cp) const
{
  string key(key_cp);

  if (key.find(':') == string::npos) {
    /* Keys without a scheme are user@host, and indicate sip */
    key = "sip:" + key;
  }

  uri_t uri = URI_Parse(key.c_str());

  return uri;
} /* cinema_mysql_table::get_uri(const char*) */


/*
 * cinema_mysql_table::sanity_check_result(int, bool, unsigned int,
 *                                         const char*, const char*,
 *                                         const char*)
 * After a call to queryTable, we want to make sure that the result is
 * reasonable.  Check it.
 * Parameters:
 *   succ: the return code of cinema_mysql_table::query_table().
 *   primary: whether the query was for a primary key (i.e. we expect exactly
 *            one row in the result)
 *   columns: the number of columns we expect in the result
 *   keyname: the name of the key with which the query was made (used for
 *            strings in exceptions)
 *   keyval: the value of the key with which the query was made (likewise)
 *   table: the name of the table against which the query was made (likewise)
 * Throws a cinema_error if any of the checks fail.
 */
void
cinema_mysql_table::sanity_check_result(int succ, bool primary,
                                        unsigned int columns,
                                        const char* keyname,
                                        const char* keyval,
                                        const char* table)
{
  ostrstream s;

  if (succ != 0) {
    s << "Error reading from MySQL " << table << " database: "
      << conn->getLastError() << ends;
    throw cinema_internal_err(s);
  }
      
  if (conn->getNumRowsReturned() == 0) {
    s << table << ": ";
    s << keyname << " " << keyval;
    s << ": not found" << ends;
    throw cinema_notfound(s);
  }

  if (primary && conn->getNumRowsReturned() > 1) {
    s << "MySQL Sanity check failed! ";
    s << conn->getNumRowsReturned() << " rows returned in table ";
    s << table << "for primary key \'" << keyname << "\' = " << keyval << ends;

    throw cinema_internal_err(s);
  }

  if (conn->getColumnCount() != columns) {
    s << "MySQL Sanity check failed!  Asked " << columns << " questions, got ";
    s << conn->getColumnCount() << " answers!" << ends;

    throw cinema_internal_err(s);
  }
} /* cinema_mysql_table::sanity_check_result(int, bool, unsigned int,
                                             const char*, const char*,
                                             const char*) */


/*
 *Max time_t vaule for infinite time
 */
#define MAX_TIME_T  ((time_t) 0x7fffffff)

/*
 * cinema_mysql_table::sql_datetime_to_time(const char *)
 * Convert string representation of MySQL date-time to time_t.
 * db string: YYYY-MM-DD HH:MM:SS 
 * Infinite time (MAX_TIME_T) corresponds to
 * 0000-00-00 00:00:00
 * All others are converted as per GMT 
 */
time_t
cinema_mysql_table::sql_datetime_to_time(const char *str)
{
  tm tm;
  time_t ret;
  
  if (str == NULL) {
    return MAX_TIME_T;
  }
  if (!strcmp(str, "0000-00-00 00:00:00")) {
    /* If the string is unspecified; i.e. all zeros then 
     * registration is always active */
    return MAX_TIME_T; /* Max Possible */
  }
  if (!strcmp(str, "9999-12-31 23:59:59")) {
    /* Similarly, the "max possible" value. */
    return MAX_TIME_T; /* Max Possible */
  }

  /* Some strptime() implementations (glibc, FreeBSD) don't initialize
     all the fields of struct tm, so set it to zero first. */
  memset(&tm, '\0', sizeof(tm));
  strptime(str, "%Y-%m-%d %H:%M:%S", &tm);
  ret = timegm(&tm); /* convert GMT struct tm to time_t */
  if (ret == -1) {
    /* timegm was unhappy.  Interpret as 0. */
    /* (XXX: Or throw exception?) */
    return 0;
  }
  return ret;
} /* cinema_mysql_table::sql_datetime_to_time(const char *) */


/*
 * cinema_mysql_table::time_to_sql_datetime(time_t, char *)
 * Convert a time_t into a string representation of a MySQL date-time,
 * (YYYY-MM-DD HH:MM:SS) as per GMT.
 * Buf must be caller-allocated and contain at least SQL_DATETIME_BUFSIZE
 * characters.
 */
void
cinema_mysql_table::time_to_sql_datetime(time_t time, char *buf)
{
  struct tm tm;

  if (time == MAX_TIME_T) {
    strcpy(buf, "0000-00-00 00:00:00");
    return;
  }

  gmtime_r(&time, &tm);

  strftime(buf, SQL_DATETIME_BUFSIZE, "%Y-%m-%d %H:%M:%S", &tm);
} /* cinema_mysql_table::time_to_sql_datetime(time_t, char *) */


/*
 * cinema_mysql_table::sql_timestamp14_to_time(const char *)
 * Convert string representation of MySQL timestamp(14) to time_t.
 * db string: YYYYMMDDHHMMSS 
 * Converted as per GMT 
 */
time_t
cinema_mysql_table::sql_timestamp14_to_time(const char *str)
{
  tm tm;
  
  if (str == NULL) {
    return MAX_TIME_T;
  }

  /* Some strptime() implementations (glibc, FreeBSD) don't initialize
     all the fields of struct tm, so set it to zero first. */
  memset(&tm, '\0', sizeof(tm));
  strptime(str, "%Y%m%d%H%M%S", &tm);
  return timegm(&tm); /* convert GMT struct tm to time_t */
} /* cinema_mysql_table::sql_timestamp14_to_time(const char *) */


/*
 * cinema_mysql_table::deserialize_next_alias_row()
 * Get the next MYSQL_ROW from the connection, and convert it to an
 * alias_table_entry*.
 * The previous query must have been for the alias table.
 */
alias_table_entry*
cinema_mysql_table::deserialize_next_alias_row()
{
  alias_table_entry* result = (alias_table_entry *)
    calloc(1, sizeof(alias_table_entry));
  if (result == NULL) throw bad_alloc();

  MYSQL_ROW row = conn->getNextRow();

  strassign(&result->alias, row[0]);
  strassign(&result->primary_user, row[1]);
  result->last_modified = sql_timestamp14_to_time(row[2]);
  
  return result;
} /* cinema_mysql_table::deserialize_next_alias_row */


/*
 * cinema_mysql_table::get_aliases_by_alias(const char *)
 * Get an alias entry for the given database key.
 */
alias_table_entry*
cinema_mysql_table::get_aliases_by_alias(const char *key)
{
  // Lock database for the duration of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  try {
    int succ = conn->queryTable("aliases", "alias", key,
                                "alias,primary_user,last_modified");
    
    sanity_check_result(succ, true, 3, "alias", key, "aliases");
    
    alias_table_entry *result = deserialize_next_alias_row();

    conn->freeResultSet();

    return result;
  }
  catch (...) {
    conn->freeResultSet();
    throw;
  }
} /* cinema_mysql_alias_table::get_aliases_by_alias(const char *) */


/*
 * cinema_mysql_table::parse_sip_groups(const char *)
 * Parse a sip_groups string into a bitwise 'or' of put_sip_group values.
 */
int
cinema_mysql_table::parse_sip_groups(const char* orig_groups)
{
  int ret = 0;

  vector<string> group_list = split_string(orig_groups, " ");
  vector<string>::iterator this_group;

  for (this_group = group_list.begin();
       this_group != group_list.end();
       this_group++) {
    if (strcasecmp(this_group->c_str(), "admin") == 0) {
      ret |= put_sip_group_admin;
    }
    else if (strcasecmp(this_group->c_str(), "cgi") == 0) {
      ret |= put_sip_group_cgi;
    }
    else if (strcasecmp(this_group->c_str(), "voicemail") == 0) {
      ret |= put_sip_group_voicemail;
    }
  }

  return ret;
} /* cinema_mysql_table::parse_sip_groups(const char *) */


/*
 * cinema_mysql_table::parse_sip_methods(const char *)
 * Parse a sip_methods string into a bitwise 'or' of put_sip_method values.
 */
int
cinema_mysql_table::parse_sip_methods(const char* orig_methods)
{
  int ret = 0;

  vector<string> group_list = split_string(orig_methods, " ");
  vector<string>::iterator this_group;

  for (this_group = group_list.begin();
       this_group != group_list.end();
       this_group++) {
    if (strcasecmp(this_group->c_str(), "INVITE") == 0) {
      ret |= put_sip_method_INVITE;
    }
    else if (strcasecmp(this_group->c_str(), "REGISTER") == 0) {
      ret |= put_sip_method_REGISTER;
    }
    else if (strcasecmp(this_group->c_str(), "any") == 0) {
      ret |= put_sip_method_any;
    }
  }

  return ret;
} /* cinema_mysql_table::parse_sip_methods(const char *) */

/*
 * cinema_mysql_table::parse_auth(const char *)
 * Parse an auth string into a put_auth enum.
 */
put_auth
cinema_mysql_table::parse_auth(const char* auth)
{
  if (auth == NULL) {
    /* Fail closed */
    return put_auth_required;
  }
  else if (strcasecmp(auth, "required") == 0) {
    return put_auth_required;
  }
  else if (strcasecmp(auth, "request") == 0) {
    return put_auth_request;
  }
  else if (strcasecmp(auth, "never") == 0) {
    return put_auth_never;
  }
  else {
    /* Fail closed. */
    return put_auth_required;
  }
} /* cinema_mysql_table::parse_auth(const char *) */

/*
 * cinema_mysql_table::deserialize_next_put_row()
 * Get the next MYSQL_ROW from the connection, and convert it to an
 * put_table_entry*.
 * The previous query must have been for the PUT table.
 */
put_table_entry*
cinema_mysql_table::deserialize_next_put_row()
{
  put_table_entry* result = (put_table_entry *)
    calloc(1, sizeof(put_table_entry));
  if (result == NULL) throw bad_alloc();

  MYSQL_ROW row = conn->getNextRow();

  strassign(&result->user, row[0]);

  strassign(&result->hash_value, row[1]);
  strassign(&result->realm, row[2]);

  result->sip_groups = parse_sip_groups(row[3]);

  result->auth = parse_auth(row[4]);

  strassign(&result->algorithm, row[5]);

  result->sip_methods = parse_sip_methods(row[6]);

  result->remote_user = strvec_to_c_strs(split_string(row[7], " "));

  strassign(&result->busy, row[8]);
  strassign(&result->noresponse, row[9]);
  strassign(&result->message_template, row[10]);

  if (row[11] != NULL) {
    istrstream r11(row[11]);
    r11 >> result->um_timeout;
  }
  else {
    result->um_timeout = 0; /* XXX ??? */
  }

  istrstream r12(row[12]);
  r12 >> result->max_msgsize_kb;

  result->last_modified = sql_timestamp14_to_time(row[13]);

  strassign(&result->gwclass, row[14]);
  
  return result;
} /* cinema_mysql_table::deserialize_next_put_row */


/*
 * cinema_mysql_table::get_put_by_user(const char *)
 * Get a PUT entry for the given database key.
 */
put_table_entry*
cinema_mysql_table::get_put_by_user(const char *key)
{
  // Lock database for the duration of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  try {
    int succ = conn->queryTable("put", "user", key,
                                "user,hash_value,realm,sip_groups,auth,"
                                "algorithm,sip_methods,remote_user,busy,"
                                "noresponse,message_template,um_timeout,"
                                "max_msgsize_kb,last_modified,gwclass");
    
    sanity_check_result(succ, true, 15, "user", key, "put");
    
    put_table_entry *result = deserialize_next_put_row();

    conn->freeResultSet();
    
    return result;
  }
  catch (...) {
    conn->freeResultSet();
    throw;
  }
} /* cinema_mysql_table::get_put_by_user(const char *) */


/*
 * cinema_mysql_table::parse_action(const char*)
 * Parse an action string into a put_sip_action enum.
 */
contact_sip_action
cinema_mysql_table::parse_action(const char* action)
{
  if (action == NULL) {
    return contact_sip_action_NULL;
  }
  else if (strcasecmp(action, "Proxy") == 0) {
    return contact_sip_action_Proxy;
  }
  else if (strcasecmp(action, "Redirect") == 0) {
    return contact_sip_action_Redirect;
  }
  else {
    return contact_sip_action_NULL;
  }
} /* cinema_mysql_table::parse_action(const char*) */


/*
 * cinema_mysql_table::deserialize_next_contact_row()
 * Get the next MYSQL_ROW from the connection, and convert it to an
 * contact_table_entry*.
 * The previous query must have been for the contacts table.
 */
contact_table_entry*
cinema_mysql_table::deserialize_next_contact_row()
{
  contact_table_entry* result = (contact_table_entry *)
    calloc(1, sizeof(contact_table_entry));
  if (result == NULL) throw bad_alloc();

  MYSQL_ROW row = conn->getNextRow();

  strassign(&result->user, row[0]);

  strassign(&result->contact, row[1]);

  if (row[2] != NULL) {
    result->expires = sql_datetime_to_time(row[2]);
  }
  else {
    result->expires = MAX_TIME_T;
  }

  if (row[3] != NULL) {
    istrstream r3(row[3]);
    r3 >> result->q;
  }
  else {
    result->q = 1.0;
  }

  result->action = parse_action(row[4]);

  result->last_modified = sql_timestamp14_to_time(row[5]);

  strassign(&result->display_name, row[6]);

  strassign(&result->sip_methods, row[7]);

  return result;
} /* cinema_mysql_table::deserialize_next_contact_row() */


/*
 * cinema_mysql_table::get_contacts_by_user(const char *)
 * Get a list of contact entries for the given database key.
 */
contact_table_entry** 
cinema_mysql_table::get_contacts_by_user(const char *user, int *count)
{
  // Lock database for the duration of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  try {
    int succ = conn->queryTable("contacts", "user", user,
                                "user,contact,expires,q,action,"
                                "last_modified,display_name,sip_methods",
                                "deleted", "false");
    
    sanity_check_result(succ, false, 8, "user", user, "contacts");
    
    contact_table_entry ** result = (contact_table_entry **)
      calloc(sizeof(contact_table_entry *), conn->getNumRowsReturned() + 1);
    if (result == NULL) throw bad_alloc();
    
    for (int i = 0; i < conn->getNumRowsReturned(); i++) {
      result[i] = deserialize_next_contact_row();
    }
    
    if (count != NULL) {
      *count = conn->getNumRowsReturned();
    }

    conn->freeResultSet();
    
    return result;
  }
  catch (...) {
    conn->freeResultSet();
    throw;
  }
} /* cinema_mysql_table::get_contacts_by_user(const char *) */



/*
 * cinema_mysql_table::insert_contact(const char *)
 * Insert the given contact entry into the database.  If a contact already
 * exists with the given user/contact pair, replace it.
 */
void
cinema_mysql_table::insert_contact(const contact_table_entry *entry)
{
  /* First, serialize the entry. */
  /* This isn't split out to a separate function because it's a lot cleaner
   * to keep the "names" and "values" arrays on the stack. */
  const char* names[7];
  const char* values[7];
  int num_columns;

  char expires_buf[SQL_DATETIME_BUFSIZE];
  ostrstream q_buf;

  names[0] = "user";
  values[0] = entry->user;

  names[1] = "contact";
  values[1] = entry->contact;

  num_columns = 2;

  if (entry->expires != 0) {
    time_to_sql_datetime(entry->expires, expires_buf);
    names[num_columns] = "expires";
    values[num_columns] = expires_buf;
    num_columns++;
  }

  if (entry->q >= 0.) {
    q_buf.precision(2);
    q_buf << entry->q << ends;
    names[num_columns] = "!q";
    values[num_columns] = q_buf.str();
    num_columns++;
  }

  if (entry->action != contact_sip_action_NULL) {
    names[num_columns] = "action";
    switch (entry->action) {
    case contact_sip_action_Proxy:
      values[num_columns] = "Proxy";
      break;

    case contact_sip_action_Redirect:
      values[num_columns] = "Redirect";
      break;

    default:
      values[num_columns] = "NULL";
      break;
    }
    num_columns++;
  }
      
  /* last_modified is updated automatically. */

  if (entry->display_name != NULL) {
    names[num_columns] = "display_name";
    values[num_columns] = entry->display_name;
    num_columns++;
  }

  if (entry->sip_methods != NULL) {
    names[num_columns] = "sip_methods";
    values[num_columns] = entry->sip_methods;
    num_columns++;
  }

  // Lock database for the remainder of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  int succ = conn->replaceRow("contacts", names, values, num_columns);

  q_buf.rdbuf()->freeze(0); /* Done with the buffer */

  if (succ != 0) {
    ostrstream s;

    s << "Error writing to MySQL contacts database: "
      << conn->getLastError() << ends;
    throw cinema_internal_err(s);
  }
} /* cinema_mysql_table::insert_contact(const char *) */


/*
 * cinema_mysql_table::delete_contacts_by_user(const char *)
 * Delete all contacts for the given user from the database.
 * If no such contacts exist, this is a no-op.
 */
void
cinema_mysql_table::delete_contacts_by_user(const char *user)
{
  // Lock database for the duration of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  int succ = conn->deleteRow("contacts", "user", user);

  if (succ != 0) {
    ostrstream s;

    s << "Error writing to MySQL contacts database: "
      << conn->getLastError() << ends;
    throw cinema_internal_err(s);
  }
} /* cinema_mysql_table::delete_contacts_by_user(const char *) */


/*
 * cinema_mysql_table::delete_contact_by_contact(const char *, const char *)
 * Delete the given user/contact from the database table.
 * If no such contact exists, this is a no op.
 */
void
cinema_mysql_table::delete_contact_by_contact(const char *user,
                                              const char *contact)
{
  // Lock database for the duration of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  const char *names[] = {"user", "contact"};
  const char *values[] = {user, contact};

  int succ = conn->deleteRow("contacts", names, values, 2);

  if (succ != 0) {
    ostrstream s;

    s << "Error writing to MySQL contacts database: "
      << conn->getLastError() << ends;
    throw cinema_internal_err(s);
  }
} /* cinema_mysql_table::delete_contact_by_contact(const char *,
   *                                               const char *) */


/*
 * cinema_mysql_table::delete_expired_contacts_by_user(const char *,
 *                                                     time_t)
 * Delete all expired contacts (those whose expires value is earlier than
 * 'expire_time') for the given user from the database.  If no such contacts
 * exist, this is a no-op.
 */
void
cinema_mysql_table::delete_expired_contacts_by_user(const char *user,
                                                    time_t expire_time)
{
  ostrstream query;

  char expires_buf[SQL_DATETIME_BUFSIZE];
  time_to_sql_datetime(expire_time, expires_buf);

  query << "delete from ";
  query << "contacts";
  query << " where ";
  query << "user=\'" << user << "\'";
  query << " and ";
  query << "expires != 0"; /* Equiv. to 0000-00-00 00:00:00 */
  query << " and ";
  query << "expires < \'" << expires_buf << "\'";
  query << ends;

  // Lock database for the remainder of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  int succ = conn->rawQueryTable(query.str());

  query.rdbuf()->freeze(0); /* Done with the buffer */
  conn->freeResultSet(); /* Done with the query result */

  if (succ != 0) {
    ostrstream s;

    s << "Error writing to MySQL contacts database: "
      << conn->getLastError() << ends;
    throw cinema_internal_err(s);
  }
} /* cinema_mysql_table::delete_expired_contacts_by_user(const char *,
   *                                                     time_t) */


/*
 * cinema_mysql_table::deserialize_next_sub_row()
 * Get the next MYSQL_ROW from the connection, and convert it to an
 * sub_table_entry*.
 * The previous query must have been for the subscription table.
 */
sub_table_entry*
cinema_mysql_table::deserialize_next_sub_row()
{
  sub_table_entry* result = (sub_table_entry *)
    calloc(1, sizeof(sub_table_entry));
  if (result == NULL) throw bad_alloc();

  MYSQL_ROW row = conn->getNextRow();

  strassign(&result->sub_to, row[0]);
  strassign(&result->sub_from, row[1]);

  if (row[2] != NULL) {
    result->expires = sql_datetime_to_time(row[2]);
  }
  else {
    result->expires = MAX_TIME_T;
  }
  
  strassign(&result->contact, row[3]);
  strassign(&result->call_id, row[4]);
  strassign(&result->record_route, row[5]);

  result->last_modified = sql_timestamp14_to_time(row[6]);
  
  return result;
} /* cinema_mysql_table::deserialize_next_sub_row */

/*
 * cinema_mysql_table::get_sub_by_fromto(const char *, const char*)
 * Get a subscription entry for the given database key.
 */
sub_table_entry* 
cinema_mysql_table::get_sub_by_fromto(const char* from, const char* to)
{

  // Lock database for the duration of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  try {
    int succ = conn->queryTable("subscription", "sub_to", to,
                                "sub_to, sub_from, expires, contact, call_id,"
				"record_route, last_modified",
				"sub_from", from);
    
    sanity_check_result(succ, true, 7, "sub_to", to, "subscription");
    
    sub_table_entry *result = deserialize_next_sub_row();

    conn->freeResultSet();
    
    return result;
  }
  catch (...) {
    conn->freeResultSet();
    throw;
  }
} /* cinema_mysql_table::get_sub_by_fromto(const char *, const char*) */

/*
 * cinema_mysql_table::num_subscriptions
 * Return the number of subscriptions for this user.
 */
int cinema_mysql_table::num_subscriptions(const char *to)
{

  // Lock database for the duration of this function
  pthread_mutex_stack_obj the_lock(conn_lock);
  char buf[512];
  MYSQL_ROW row;
  int retval = 0;

  sprintf(buf,"select count(*) from subscription where sub_to='%s'",to);
  try {
    int succ =  conn->rawQueryTable(buf);
   
    if (!succ && conn->getNumRowsReturned() == 1) {
     row = conn->getNextRow(); 
     retval = atoi(row[0]);
    }
    
    conn->freeResultSet();
    
    return retval;
  }
  catch (...) {
    conn->freeResultSet();
    throw;
  }
} /* cinema_mysql_table::num_subscriptions */


sub_table_entry **cinema_mysql_table::get_all_subscriptions(const char* to, int *count)
{
  // Lock database for the duration of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  int succ;

  try {
  
  succ = conn->queryTable("subscription", "sub_to", to,
                           "sub_to, sub_from, expires, contact, call_id,"
				                   "record_route, last_modified");

  sanity_check_result(succ, true, 7, "sub_to", to, "subscription");
    
  sub_table_entry **result = (sub_table_entry **)
    calloc(sizeof(sub_table_entry *), conn->getNumRowsReturned() + 1);
    
  if (result == NULL) throw bad_alloc();
    
  for (int i = 0; i < conn->getNumRowsReturned(); i++) {
      result[i] = deserialize_next_sub_row();
  }
    
  if (count != NULL) 
      *count = conn->getNumRowsReturned();

  conn->freeResultSet();
    
   return result;
  }
  catch (...) {
    conn->freeResultSet();
    throw;
  }
} /* cinema_mysql_table::get_all_subscriptions */


/*
 * cinema_mysql_table::insert_sub(const sub_table_entry*)
 * Insert the given subscription entry into the database.  If a subscription
 * already exists with the given to/from pair, replace it.
 */
void cinema_mysql_table::insert_sub(const sub_table_entry *entry) 
{
  /* First, serialize the entry. */
  /* This isn't split out to a separate function because it's a lot cleaner
   * to keep the "names" and "values" arrays on the stack. */
  const char* names[6];
  const char* values[6];
  int num_columns;

  char expires_buf[SQL_DATETIME_BUFSIZE];
  ostrstream q_buf;

  names[0] = "sub_to";
  values[0] = entry->sub_to;

  names[1] = "sub_from";
  values[1] = entry->sub_from;

  num_columns = 2;

  if (entry->expires != 0) {
    time_to_sql_datetime(entry->expires, expires_buf);
    names[num_columns] = "expires";
    values[num_columns] = expires_buf;
    num_columns++;
  }
  
  /* last_modified is updated automatically. */
  
  if (entry->contact != NULL) {
    names[num_columns] = "contact";
    values[num_columns] = entry->contact;
    num_columns++;
  }

  if (entry->record_route != NULL) {
    names[num_columns] = "record_route";
    values[num_columns] = entry->record_route;
    num_columns++;
  }

  if (entry->call_id != NULL) {
    names[num_columns] = "call_id";
    values[num_columns] = entry->call_id;
    num_columns++;
  }

  // Lock database for the remainder of this function
  pthread_mutex_stack_obj the_lock(conn_lock);

  int succ = conn->replaceRow("subscription", names, values, num_columns);

  if (succ != 0) {
    ostrstream s;

    s << "Error writing to MySQL subscription database: "
      << conn->getLastError() << ends;
    throw cinema_internal_err(s);
  }
} /* cinema_mysql_table::insert_sub */


void cinema_mysql_table::delete_sub_by_fromto(const char* from, const char *to)
{
  
  pthread_mutex_stack_obj the_lock(conn_lock);
  
  const char *names[] = {"sub_to", "sub_from"};
  const char* values[2];
  
  values[0] = to;
  values[1] = from;
  
  int succ = conn->deleteRow("subscription", names, values, 2);
  
  if (succ != 0) {
    ostrstream s;
    s << "Error deleting from MySql subscription database: "
      << conn->getLastError() << ends;
    throw cinema_internal_err(s);
  }

  return;
  
} /* cinema_mysql_table::delete_sub_by_fromto */
