/* -*- C++ -*- */
/* cinema_table.cpp -- Virtual base class representing cinema database
 * tables. */
   
/* Copyright 2001 Columbia University.  All rights reserved. */

#include "cinema_table.h"
#include "cinema_dummy_table.h"
#include "cinema_file_table.h"

#ifdef USE_MYSQL
#include "cinema_mysql_table.h"
#include "cinema_fastsql_table.h"
#endif

#include "uri.h"

#include <string>

using namespace std;

/*
 * cinema_table::create()
 * Create a new cinema table, based on URI.  If no table mapping is
 * found based on this URI, throw cinema_notfound.
 */
cinema_table* cinema_table::create(string table_uri_str)
{
  uri_t table_uri;

  cinema_table *table = NULL;

  table_uri = URI_Parse(table_uri_str.c_str());

  try {
#ifdef USE_MYSQL
    if (table_uri.scheme == cinema_mysql_table::table_scheme()) {
      table = new cinema_mysql_table(table_uri);
    }
    else
	if (table_uri.scheme == cinema_fastsql_table::table_scheme()) {
	  table = new cinema_fastsql_table(table_uri);
        }
	else
#endif /* USE_MYSQL */
    if (table_uri.scheme == cinema_file_table::table_scheme()) {
      table = new cinema_file_table(table_uri);
    }
    else if (table_uri.scheme == cinema_dummy_table::table_scheme()) {
      table = new cinema_dummy_table(table_uri);
    }
    else {
      throw cinema_notfound("Unknown table scheme");
    }
  }
  catch (...) {
    URI_Free(&table_uri);
    throw;
  }

  URI_Free(&table_uri);

  return table;
}




/*
 * split_string(const char *, const char *)
 * Split a string into a vector of substrings.  Analogous to strtok()
 * in C.
 */
vector<string>
split_string(const char* c_string, const char *separators)
{
  if (c_string == NULL) {
    c_string = "";
  }

  string orig_string(c_string);
  vector<string> ret;

  string::size_type start, end;

  end = 0; /* Start searching at beginning */

  for (;;) {
    start = orig_string.find_first_not_of(separators, end);
    if (start == string::npos) {
      /* No more tokens */
      break;
    }
    end = orig_string.find_first_of(separators, start);
    if (end == string::npos) {
      /* Reached end of string */
      end = orig_string.length();
    }

    ret.push_back(orig_string.substr(start, end - start));
  }

  return ret;
} /* split_string(const char *, const char *) */


/*
 * free_cstrs(char **)
 * Free a NULL-terminated array of pointer to char, as generated by the below.
 */
void
free_cstrs(char **array)
{
  int i;
  for (i = 0; array[i] != NULL; i++) {
    free(array[i]);
  }
  
  free(array);
} /* free_cstrs */


/*
 * strvec_to_c_strs(const vector<string>)
 * Copy a vector<string> to a NULL-terminated array of pointer to char.
 */
char **
strvec_to_c_strs(const vector<string> vec)
{
  char **ret;
  vector<string>::size_type i;
  
  ret = (char **) calloc(vec.size() + 1, sizeof(char *));
  if (ret == NULL) {
    throw bad_alloc();
  }
  for (i = 0; i < vec.size(); i++) {
    ret[i] = strdup(vec[i].c_str());
    if (ret[i] == NULL) {
      free_cstrs(ret);
      throw bad_alloc();
    }
  }
  
  ret[vec.size()] = NULL;

  return ret;
} /* strvec_to_c_strs(const vector<string>) */
