/* dbaccess.cpp -- C interface to libdb++ cinema_table functions. */

/* Copyright 2001 Columbia University.  All rights reserved. */

#include "dbaccess.h"
#include "cinema_table.h"

extern "C" {
#include <stdlib.h>
#include "strdupn.h"
}

#include <new> // bad_alloc

#define CATCH_ERRS_RETURN_NULL \
  catch (cinema_error& e) { \
    fill_in_err_and_reason(e, err, reason); \
    return NULL; \
  } \
  catch (std::exception& e) { \
    fill_in_err_and_reason(DB_UNKNOWN, e.what(), \
                           err, reason); \
    return NULL; \
  } \
  catch (...) { \
    fill_in_err_and_reason(DB_UNKNOWN, "Caught unknown exception", \
                           err, reason); \
    return NULL; \
  }

#define CATCH_ERRS_RETURN_ZERO \
  catch (cinema_error& e) { \
    fill_in_err_and_reason(e, err, reason); \
    return 0; \
  } \
  catch (std::exception& e) { \
    fill_in_err_and_reason(DB_UNKNOWN, e.what(), \
                           err, reason); \
    return 0; \
  } \
  catch (...) { \
    fill_in_err_and_reason(DB_UNKNOWN, "Caught unknown exception", \
                           err, reason); \
    return 0; \
  }

#define CATCH_ERRS_RETURN_DB_ERR \
  catch (cinema_error& e) { \
    return fill_in_err_and_reason(e, err, reason); \
  } \
  catch (std::exception& e) { \
    return fill_in_err_and_reason(DB_UNKNOWN, e.what(), \
                                  err, reason); \
  } \
  catch (...) { \
    return fill_in_err_and_reason(DB_UNKNOWN, "Caught unknown exception", \
                                  err, reason); \
  }

/*
 * fill_in_err_and_reason(cinema_error&, db_err *, char **)
 * Fill in 'err' and 'reason', if they are non-NULL, based on 'e'.
 * Used for C interface to C++ exceptions.
 */
static db_err
fill_in_err_and_reason(cinema_error& e, db_err *err, char **reason)
{
  if (err != NULL) {
    *err = e.get_errcode();
  }

  if (reason != NULL) {
    *reason = strdup(e.what());
  }
  return e.get_errcode();
} /* fill_in_err_and_reason(cinema_error&, db_err *, char **) */



/*
 * fill_in_err_and_reason(db_err, const char *, db_err *, char **)
 * Fill in 'err' and 'reason', if they are non-NULL, based on the given
 * strings.  Used if we don't have a cinema_error to base this on -- we
 * caught an unknown exception, or locally detected an error.
 */
static db_err
fill_in_err_and_reason(db_err my_err, const char *my_reason,
                       db_err *err, char **reason)
{
  if (err != NULL) {
    *err = my_err;
  }

  if (reason != NULL) {
    *reason = strdup(my_reason);
  }
  return my_err;
} /* fill_in_err_and_reason(db_err, const char *, db_err *, char **) */


/*
 * get_table_key_from_URI()
 * Given a URI, return the table key corresponding to that URI, for
 * the underlying table type.
 * On error, return NULL, and fill in err and reason if they are non-NULL.
 */
extern "C"
char* get_table_key_from_URI(cinema_table *table, const uri_t *uri,
                             db_err *err, char **reason)
{
  try {
    return table->get_key(uri);
  }
  CATCH_ERRS_RETURN_NULL
} /* get_table_key_from_URI */


/*
 * get_URI_from_table_key()
 * Given a table key, return a canonical URI corresponding to that table key,
 * for the underlying table type.
 * On error, return NULL, and fill in err and reason if they are non-NULL.
 */
extern "C"
uri_t* get_URI_from_table_key(cinema_table *table, const char *key,
                              db_err *err, char **reason)
{
  try {
    uri_t *ret = (uri_t *) calloc(1, sizeof(uri_t));
    if (ret == NULL) {
      throw std::bad_alloc();
    }
    *ret = table->get_uri(key);
    return ret;
  }
  CATCH_ERRS_RETURN_NULL
} /* get_URI_from_table_key */


/*
 * free_table()
 * Free a cinema_table
 */
extern "C"
void free_table(cinema_table *table)
{
  delete table;
} /* free_table */


/*
 * new_cinema_table()
 * Create a new cinema table, based on table_uri.
 * On error, return NULL, and fill in err and reason if they are non-NULL.
 */
extern "C"
cinema_table* new_cinema_table(const char *table_uri,
                               db_err *err, char **reason)
{
  try {
    return cinema_table::create(table_uri);
  }
  CATCH_ERRS_RETURN_NULL
} /* new_cinema_table */


/*
 * alias_get_by_alias()
 * Get an alias_table_entry from table, based on 'alias'.
 * On error, return NULL, and fill in err and reason if they are non-NULL.
 */
extern "C"
alias_table_entry* alias_get_by_alias(cinema_table *table,
                                      const char *alias,
                                      db_err *err, char **reason)
{
  try {
    return table->get_aliases_by_alias(alias);
  }
  CATCH_ERRS_RETURN_NULL
} /* alias_get_by_alias */



/*
 * put_get_by_user()
 * Get an put_table_entry from table, based on 'user'.
 * On error, return NULL, and fill in err and reason if they are non-NULL.
 */
extern "C"
put_table_entry* put_get_by_user(cinema_table *table,
                                 const char *user,
                                 db_err *err, char **reason)
{
  try {
    return table->get_put_by_user(user);
  }
  CATCH_ERRS_RETURN_NULL
} /* put_get_by_user */



/*
 * contacts_get_by_user()
 * Get a NULL-terminated array of contact_table_entrys from table, based on
 * 'user'.
 * On error, return NULL, and fill in err and reason if they are non-NULL.
 */
extern "C"
contact_table_entry** contacts_get_by_user(cinema_table *table,
                                           const char *user,
                                           int *count,
                                           db_err *err, char **reason)
{
  try {
    return table->get_contacts_by_user(user, count);
  }
  CATCH_ERRS_RETURN_NULL
} /* contacts_get_by_user */


/*
 * contacts_insert_contact()
 * Insert the contact table entry 'entry' into the contact table.
 * Subsequent queries can find/alter it based on its 'user' and 'contact'
 * parameters. 
 * If a contact already exists with this user/contact pair, replace it.
 * On error, fill in err and reason if they are non-NULL, and return 
 * the value that (was/would be) filled in to err.
 */
extern "C"
db_err contact_insert_contact(cinema_table *table,
                              const contact_table_entry *entry,
                              db_err *err, char **reason)
{
  try {
    table->insert_contact(entry);
    return DB_OK;
  }
  CATCH_ERRS_RETURN_DB_ERR
} /* contacts_insert_contact */


/*
 * contact_delete_contacts()
 * Delete all contacts for the user 'user' from the contact table.
 * On error, fill in err and reason if they are non-NULL, and return 
 * the value that (was/would be) filled in to err.
 * If no contacts exist for the user, it is not an error (just a no-op).
 */
extern "C"
db_err contact_delete_contacts(cinema_table *table, const char *user,
                               db_err *err, char **reason)
{
  try {
    table->delete_contacts_by_user(user);
    return DB_OK;
  }
  CATCH_ERRS_RETURN_DB_ERR
} /* contact_delete_contacts */


/*
 * contact_delete_contact()
 * Delete the contact 'contact' for the user 'user' from the contact table.
 * On error, fill in err and reason if they are non-NULL, and return 
 * the value that (was/would be) filled in to err.
 * If the specified contact doesn't exist, it is not an error (just a no-op).
 */
extern "C"
db_err contact_delete_contact(cinema_table *table, const char *user,
                              const char *contact, db_err *err, char **reason)
{
  try {
    table->delete_contact_by_contact(user, contact);
    return DB_OK;
  }
  CATCH_ERRS_RETURN_DB_ERR
} /* contact_delete_contact */


/*
 * contact_delete_expired_contacts()
 * Delete all those contacts for 'user' which have an expiration 
 * earlier than 'time'.
 * On error, fill in err and reason if they are non-NULL, and return 
 * the value that (was/would be) filled in to err.
 * If no expired contacts exist for the user, it is not an error
 * (just a no-op).
 */
db_err contact_delete_expired_contacts(cinema_table *table, const char *user,
                                       time_t expire_time,
                                       db_err *err, char **reason)
{
  try {
    table->delete_expired_contacts_by_user(user, expire_time);
    return DB_OK;
  }
  CATCH_ERRS_RETURN_DB_ERR
} /* contact_delete_expired_contacts */



/*
 * free_alias_table_entry()
 * Free an alias table entry.  Since alias_table_entry is a 
 * plain-old-data type, so it can be used from C code,
 * free it using C mechanisms.
 */
extern "C"
void free_alias_table_entry(alias_table_entry *entry)
{
  free_valid(entry->alias);
  free_valid(entry->primary_user);
  free_valid(entry);
} /* free_alias_table_entry */


/*
 * free_put_table_entry()
 * Free a primary user table entry.  Since put_table_entry is a 
 * plain-old-data type, so it can be used from C code,
 * free it using C mechanisms.
 */
extern "C"
void free_put_table_entry(put_table_entry *entry)
{
  free_valid(entry->user);

  free_valid(entry->hash_value);
  free_valid(entry->realm);
  free_valid(entry->algorithm);

  if (entry->remote_user != NULL) {
    for (int i = 0; entry->remote_user[i] != NULL; i++) {
      free_valid(entry->remote_user[i]);
    }
  }
  free_valid(entry->remote_user);

  free_valid(entry->busy);
  free_valid(entry->noresponse);
  free_valid(entry->message_template);

  free_valid(entry->gwclass);

  free_valid(entry);
} /* free_put_table_entry */


/*
 * free_contact_table_entry()
 * Free a contact table entry.  Since contact_table_entry is a 
 * plain-old-data type, so it can be used from C code,
 * free it using C mechanisms.
 */
extern "C"
void free_contact_table_entry(contact_table_entry *entry)
{
  free_valid(entry->user);
  free_valid(entry->contact);
  free_valid(entry->display_name);
  free_valid(entry->sip_methods);

  free_valid(entry);
} /* free_contact_table_entry */


/*
 * free_contact_table_entries()
 * Free a NULL-terminated array of contact table entries, as returned by 
 * contacts_get_by_user().
 */
extern "C"
void free_contact_table_entries(contact_table_entry **entries)
{
  for (contact_table_entry **entry_p = entries; *entry_p != NULL; entry_p++) {
    free_contact_table_entry(*entry_p);
  }

  free_valid(entries);
} /* free_contact_table_entry */


/*
 * sub_get_by_fromto
 * Get a sub_table_entry from table, based on 'from' and 'to'.
 * On error, return NULL, and fill in err and reason if they are non-NULL.
 */
extern "C"
sub_table_entry* sub_get_by_fromto(cinema_table *table, 
				   const char* from, 
				   const char* to,
				   db_err *err, char **reason) 
{
  try {
    return table->get_sub_by_fromto(from, to);
  }
  CATCH_ERRS_RETURN_NULL
} /* sub_get_by_fromto */


/*
 * delete_sub_by_fromto
 * Delete the record whose sub_to = "to" and sub_from="from" from
 * subscription table.
 */
extern "C"
db_err delete_sub_by_fromto(cinema_table *table, const char* from,
			    const char* to, db_err *err, char **reason)
{
  try {
    table->delete_sub_by_fromto(from, to);
    return DB_OK; 
  }
  CATCH_ERRS_RETURN_DB_ERR
} /* delete_sub_by_fromto */

/*
 * any_subscriptipns
 * Count the number of subscriptions.
 */
extern "C"
int num_subscriptions(cinema_table *table,
                      const char* to, db_err *err, char **reason)
{
  try {
    return table->num_subscriptions(to);
  }CATCH_ERRS_RETURN_ZERO
} /* num_subscriptions */

/*
 * get_all_subscriptions
 * Return a vector of all the subscriptions to this presentity.
 */
sub_table_entry** get_all_subscriptions(cinema_table *table,
                      const char* to, int *count, db_err *err, char **reason)
{
  try {
    return table->get_all_subscriptions(to, count);
  }CATCH_ERRS_RETURN_NULL
} /* get_all_subscriptions */

/*
 * sub_insert
 * Insert the subscription table entry 'entry' into the subscription table.
 * Subsequent queries can find/alter it based on its 'from' and 'to'
 * parameters. 
 * If a subscription already exists with this from/to pair, replace it.
 * On error, fill in err and reason if they are non-NULL, and return 
 * the value that (was/would be) filled in to err.
 */
extern "C"                            
db_err sub_insert(cinema_table *table,
			  const sub_table_entry *entry,
			  db_err *err, char **reason)
{
  try {
    table->insert_sub(entry);
    return DB_OK;
  }
  CATCH_ERRS_RETURN_DB_ERR
} /* sub_insert */

/*
 * free_table_entry()
 * Free a subscription table entry.  Since sub_table_entry is a 
 * plain-old-data type, so it can be used from C code,
 * free it using C mechanisms.
 */
extern "C"
void free_sub_table_entry(sub_table_entry *entry) 
{
  free_valid(entry->sub_to);
  free_valid(entry->sub_from);
  free_valid(entry->contact);
  free_valid(entry->call_id);
  free_valid(entry->record_route);

  free_valid(entry);
} /* free_sub_table_entry */

/*
 * free_sub_table_entries()
 * Free a NULL-terminated array of subscription table entries, as returned by 
 * get_all_subscriptions().
 */
extern "C"
void free_sub_table_entries(sub_table_entry **entries)
{
  for (sub_table_entry **entry_p = entries; *entry_p != NULL; entry_p++) {
    free_sub_table_entry(*entry_p);
  }
  
  free_valid(entries);
} /* free_sub_table_entries */


