/* dbaccess.h -- abstract interfaces to access the various tables of the
 * cinema database. */

/* Note: update scripts/createsip.sql in parallel with this file! */

/* Copyright 2001 Columbia University.  All rights reserved. */

#ifndef DBACCESS_H_
#define DBACCESS_H_

#ifdef  __cplusplus
extern "C" {
#endif

#include <time.h>

#include "uri.h"

/* Error codes returned from db query functions. */
typedef enum {
  /* Normal responses */ 
  DB_NO = 0,
  DB_OK = 1,
  /* Normal failures */
  DB_NOTFOUND = -1,
  /* Abnormal failures */
  DB_ERR = -1001,
  DB_INVALIDARG = -1002,
  DB_UNKNOWN = -2000
} db_err;

/* Representation of a table.  Opaque type to C. */
#ifdef __cplusplus
class cinema_table;
#else
typedef struct cinema_table cinema_table;
#endif

/* Generic transformations a table can perform */
char *get_table_key_from_URI(cinema_table *table, const uri_t *uri,
                             db_err *err, char **reason);
uri_t *get_URI_from_table_key(cinema_table *table, const char *key,
                              db_err *err, char **reason);


/* Free a table */
void free_table(cinema_table *table);


/* Functions creating tables */
cinema_table* new_cinema_table(const char *table_uri,
                               db_err *err, char **reason);


/* Read-only tables */

/*
 * alias varchar(255) primary key,
 * primary_user varchar(255) not null,
 * last_modified timestamp(14)
 */

typedef struct {
  char *alias;
  char *primary_user;
  time_t last_modified;
} alias_table_entry;


/*
 * user varchar(255) primary key,
 * hash_value varchar(255) not null,
 * realm varchar(150) not null,
 * sip_groups varchar(100),
 * auth enum('required','request','never'),
 * algorithm varchar(20) not null,
 * sip_methods varchar(100),
 * remote_user text,
 * busy varchar(255),
 * noresponse varchar(255),
 * message_template text,
 * um_timeout smallint(5) unsigned,
 * max_msgsize_kb int(10) unsigned not null default 200,
 * last_modified timestamp(14),
 * gwclass varchar(100)
 */
typedef enum {
  put_sip_group_admin = 1 << 0,
  put_sip_group_cgi = 1 << 1,
  put_sip_group_voicemail = 1 << 2
} put_sip_group;

typedef enum {
  put_auth_required,
  put_auth_request,
  put_auth_never
} put_auth;

typedef enum {
  put_sip_method_any = 1 << 0,
  put_sip_method_INVITE = 1 << 1,
  put_sip_method_REGISTER = 1 << 2
} put_sip_method;

typedef struct {
  char *user;
  char *hash_value;
  char *realm;
  int sip_groups; /* bitwise or of put_sip_group */
  put_auth auth;
  char *algorithm;
  int sip_methods; /* bitwise or of put_sip_method */
  char **remote_user;
  char *busy;
  char *noresponse;
  char *message_template;
  int um_timeout;
  int max_msgsize_kb;
  time_t last_modified;
  char *gwclass; /* TODO: correlate with gwclass table? */
} put_table_entry;


/* Semi-XXX: largely duplicates sip_action_t in libsip/sip.h */
typedef enum {
  contact_sip_action_NULL = 0,
  contact_sip_action_Proxy,
  contact_sip_action_Redirect
} contact_sip_action;

/*
 * user varchar(255) not null,
 * contact varchar(255) not null,
 * expires datetime,
 * q float(10,2),
 * action enum('Proxy','Redirect'),
 * last_modified timestamp(14),
 * display_name varchar(100),
 * sip_methods varchar(100) default 'any',
 * key(user)
 */
typedef struct {
  char *user;
  char *contact;
  time_t expires;
  float q;
  contact_sip_action action;
  time_t last_modified;
  char *display_name;
  char *sip_methods;
} contact_table_entry;



/* sub_to varchar(255) not null
 * sub_from varchar(255) not null
 * expires datetime
 * contact varchar(255) not null
 * call_id varchar(255) not null
 * record_route text
 * last_modified timestamp(14)
 */
typedef struct {
  char *sub_to;
  char *sub_from;
  time_t expires;
  char *contact;
  char *call_id;
  char *record_route;
  time_t last_modified;
} sub_table_entry;


/* Free table entries */
void free_alias_table_entry(alias_table_entry *entry);

void free_put_table_entry(put_table_entry *entry);

void free_contact_table_entry(contact_table_entry *entry);
void free_contact_table_entries(contact_table_entry **entries);

void free_sub_table_entry(sub_table_entry *entry);
void free_sub_table_entries(sub_table_entry **entries);

/* Functions accessing the tables */
alias_table_entry* alias_get_by_alias(cinema_table *table,
                                      const char *alias,
                                      db_err *err, char **reason);

put_table_entry* put_get_by_user(cinema_table *table,
                                 const char *user,
                                 db_err *err, char **reason);


contact_table_entry** contacts_get_by_user(cinema_table *table,
                                           const char *user,
                                           int *count,
                                           db_err *err, char **reason);

db_err contact_insert_contact(cinema_table *table,
                              const contact_table_entry *entry,
                              db_err *err, char **reason);

db_err contact_delete_contacts(cinema_table *table, const char *user,
                               db_err *err, char **reason);

db_err contact_delete_contact(cinema_table *table, const char *user,
                              const char *contact, db_err *err, char **reason);

db_err contact_delete_expired_contacts(cinema_table *table, const char *user,
                                       time_t expire_time,
                                       db_err *err, char **reason);
                                       
sub_table_entry* sub_get_by_fromto(cinema_table *table, 
				   const char* from, 
				   const char* to,
				   db_err *err, char **reason);
  
int num_subscriptions(cinema_table *table,
		      const char* to, 
		      db_err *err, char **reason);
  
sub_table_entry** get_all_subscriptions(cinema_table *table,
					const char* to, int *count, 
					db_err *err, char **reason);
                                      
db_err sub_insert(cinema_table *table,
		  const sub_table_entry *entry,
		  db_err *err, char **reason);

db_err delete_sub_by_fromto(cinema_table *table, const char* from,
			    const char* to, db_err *err, char **reason);

#ifdef  __cplusplus
} /* extern "C" */
#endif

#endif /* DBACCESS_H_ */
