#ifndef _DB_API_H_
#define _DB_API_H_

/*
 * Put all the mysql definition inside #ifdef USE_MYSQL so that
 * compiler does not give errors for modules which do not
 * use MySQL but still have #include dbapi.h
 */
#ifdef USE_MYSQL

#include <stdio.h>
#include <mysql.h>

#ifdef  __cplusplus
extern "C" {
#endif

typedef MYSQL_ROW* ResultTable;

/* 
 * function which initializes connection to Mysql Database.  The user,
 * passwd, database, and host of mysql server have to be passed in.  If no
 * port is desired, the port should be 0.
 */
int initializeConnection(const char * user, const char * passwd, 
                         const char * database, const char * sqlServer, 
                         int port);

/*
 * closes connection to Mysql Database.
 */
void closeConnection();

/*
 * function which executes a 'SELECT' on a 'table' with search key
 * equal to 'value'.  It only selects the columns specified in 'columns'.
 * A two-dimension table for each row (and columns within the row) are
 * returned.  Also, the number of rows retrieved are stored in rowsReturned.
 * If no rows were retrieved from the query, then NULL is returned.
 */
ResultTable query(const char * table, const char * key, const char * value, 
		  const char * columns, int * rowsReturned);

/*
 * function which executes a 'DELETE' on a 'table' with deletion key equal
 * to 'value'.  Returns 0 if successful, -1 if not successful.  This function
 * should be used to remove all rows with specified key.
 */
int removeAllData(const char * table, const char * key, const char * value);

/*
 * function which executes a 'DELETE' on a 'table' with multiple conditions.
 * Returns 0 if successful, -1 if not successful.  This function should be 
 * used to remove all rows where one column = "xxxx" AND second columns = 
 * "yyyy" AND third column = "zzzz", etc ....  Values which are not strings
 * should have a column name starting with '#' for int values and '!' for 
 * float values.  (See below on example).
 */
int removeData(const char * table, const char ** columns,
               const char ** values, int num_columns);

/*
 * function which executes an 'INSERT' on a 'table'.  Each column value to be
 * inserted into the table has to be explicitly passed (in columns).  The value
 * types for each column is assumed to be 'string' type unless the column name
 * (stored in columns**) starts with a '#' which means store as int value or
 * '!' which means store as float value.  For example, a column which is named
 * "gadgets" that needs an int value, would be passed into the columns** as 
 * "#gadgets", rather than just "gadgets".  Returns 0 if successful, -1 if not
 * successful.
 */
int insert(const char * table, const char ** columns, const char ** values, 
	   int num_columns);

/*
 * function which executes an 'UPDATE' on a 'table'.  Each column value to be
 * inserted into the table has to be explicitly passed (in columns).  The value
 * types for each column is assumed to be 'string' type unless the column name
 * (stored in columns**) starts with a '#' which means store as int value or
 * '!' which means store as float value.  For example, a column which is named
 * "gadgets" that needs an int value, would be passed into the columns** as 
 * "#gadgets", rather than just "gadgets".  Returns 0 if successful, -1 if not
 * successful.
 */
int updateOnly(const char * table, const char * key, const char * value, 
	   const char ** columns, const char ** values, int num_columns);


/*
 * UpdateOnly2 is similar to updateOnly() except that only
 * the entries which match both key=value and
 * secondkey=secondvalue are updated. 
 * This is typically used
 * when the table has MUL (multiple) keys with same value.
 * A secondary key can identify the row uniquely. 
 * It is the responsibility of the application to make sure
 * that this is maintained.
 */
int updateOnly2(const char * table, const char * key, const char * value, 
                const char *secondkey, const char *secondvalue,
                const char ** columns, const char ** values,
                int num_columns, int start_update_from);


/*
 * UpdateMatching is similar to update() except that only
 * the entries which match both key=value and
 * secondkey=secondvalue are updated. Also if the
 * entry does not exist then it is created (inserted).
 * This is typically used
 * when the table has MUL (multiple) keys with same value.
 * A secondary key can identify the row uniquely. 
 * It is the responsibility of the application to make sure
 * that this is maintained.
 */
int updateOrInsert(const char * table, const char * key, const char * value, 
                   const char *secondkey, const char *secondvalue,
                   const char ** columns, const char ** values,
                   int num_columns, int start_update_from);


/*
 * returns last mysql error 
 */
char * getMysqlError();

/*
 * checks to make sure connection to MYSQL Database exists.  Returns 1
 * if there is currently a connection, 0 if htere is no connection.
 */
int isConnectionEstablished();

/*
 * frees the memory of the ResultTable 'table' (returned from the last 
 * query function).
 */
void freeResultTable(ResultTable table, int row_count, int column_count);


/*
 * function which executes a well-formed sql query on the database.
 * A two-dimension table for each row (and columns within the row) are
 * returned.  Also, the number of rows retrieved are stored in rowsReturned.
 * If no rows were retrieved from the query, then NULL is returned.
 */
ResultTable rawQuery(const char * sqlQuery,int * rowsReturned);

#ifdef  __cplusplus
}
#endif

#endif /* USE_MYSQL */

#endif /* _DB_API_H_ */


