/* ads.h - Common data structures and templates for list, arrays etc.
 * (c) 2001-2001 Columbia University. All rights reserved.
*/

#ifndef _ADS_H_
#define _ADS_H_

#define NO_TEMPLATES 1

#include <iostream.h>  /* ostream */
/**@name Abstract data structures. */
//@{

#ifdef NO_TEMPLATES

/**
 * This class is used for defining a list of any datatype.
 * Individual class lists should be derived from this.
 * Application should use DEFINE_AS_LIST macro to
 * define a list of a particular data type. Currently,
 * it is implemented as a linked list of objects.
 * Please note that you must have a copy constructor and
 * an = operator defined for any class for which you
 * want to create a list.
 *
 * @see DEFINE_AS_LIST
 */
class GenericList
{
 public:
  /**@name Constructors and destructors*/
  //@{
  /**
   * Construct an empty list. The members of the list are
   * automatically destroyed when the list is deleted.
   */
  GenericList();

  /**
   * Construct an empty list. 
   * @param allow_delete determines whether to delete the
   * members of the list automatically when the list is
   * deleted.
   */
  GenericList(bool allow_delete);

  /** Copy constructor */
  GenericList(const GenericList & list);

  /** 
   * Default destructor. It deletes all the linked list elements.
   * If the allow_delete member is set to true then it
   * calls the destructors on all the list members.
   */
  virtual ~GenericList();
  //@}

  /**@name List methods */
  //@{
  /**
   * Retrive number of list elements.
   *
   * @return the number of elements in the list.
   */
  virtual int GetSize() const {return size;} ;

  /**
   * Retrive the pointer to a element. This
   * allows for array like access to the actual
   * object (an not a copy) stored in the list.
   *
   * @param index of the object to retrieve.
   *
   * @return pointer to the object at location "index".
   */
  virtual void * GetAt(int index) const;

  /**
   * Append an element at the end of the list.
   * This allows for list like append operation.
   *
   * @param pointer to the object of given type.
   * make sure that your object types are consistent and
   * as specified in the definition using DEFINE_AS_LIST.
   *
   * @return the index of the newly added object.
   */
  virtual int  Append(void *); /* returns index */

  /**
   * Inserts an element in the begining of the list.
   * This allows for list like insert operation.
   *
   * @param pointer to the object of given type. make sure
   * that your object types are consistent and as specified in the
   * definition using DEFINE_AS_LIST.
   */
  virtual void Insert(void *); /* insert in begining */

  /**
   * Removes an element at a particular index.
   * This removes the element at index from the list.
   * It also calls the destructor on the element if
   * allow_delete is set to true.
   *
   * note that after this method, the index of other elements
   * (which are after this element) will be changed
   * (decremented by one).
   *
   * @param index of the element to be removed.
   */
  virtual void RemoveAt(int index);

  /**
   * Removes all the elements from the list.
   */
  virtual void RemoveAll();   /* Remove all elements */

  /**
   * Copy operator. This is used to copy one list to another.
   * All the members are copied appropriately.
   *
   * @param list The list from where to copy.
   *
   * @return the current (this) list.
   *
   * Make sure to define the copy constructor and the
   * copy operator for the classes for which you want a list.
   */
  virtual GenericList & operator=(const GenericList & list);
  //@}

  /**@name Misc */
  //@{
  /**
   * Print the list to output stream.
   * This is used for debugging purpose. Application need not
   * use this.
   *
   * e.g., cout << "List is : " << my_list << endl; 
   */
  friend ostream & operator<<(ostream & o, const GenericList & list);
  //@}

  //protected:
  /**@name Protected members, used internally. */
  //@{
  /**
   * Delete an element.
   * This is supposed to be a pure virtual method. 
   * DEFINE_AS_LIST defines the method for appropriate types.
   */
  virtual void DeleteElement(void *data)
    { cout << "Dummy DeleteElement() called" << endl; }

  /**
   * Allocate and copy an element.
   * This is supposed to be a pure virtual method.
   * DEFINE_AS_LIST defines the method for appropriate types.
   */
  virtual void *AllocateElement(void *p) 
    { /*cout << "Dummy AllocateElement() called" << endl;*/ return NULL;}

  /**
   * Element class. This stores a pointer to an element.
   * This is a linked list element of the internal linked
   * list implementation.
   */
  class Element {
  public:
    Element(void * d) { data = d; next = NULL; };
    ~Element() { next = NULL; data = NULL; }
    Element * next;
    void * data;
  };

  /**
   * Whether to allow delete on members or not.
   * This is used by applications to specify whether to
   * call the destructors of the members automatically
   * when they are removed or the list is deleted.
   * default value is true.
   */
  bool allow_delete;

  /**
   * Implementation type. This specifies if the implementation
   * is of linked list (false) or as an array (true). This is
   * internal member and should not be used by application.
   * Currently only linkedlist is implemented.
   */
  bool is_array; /* whether it is a linked list or an array */

  /**
   * Linked list. This is a pointer to linked list of
   * elements in the list. 
   */
  Element *head; /* Head of the list */

  /**
   * Number of elements. This stores the number of elements
   * present in the list 
   */
  int     size;  /* From head till last element */
  //@}
};

/**
 * This macro is used for defining a new list of a given type.
 * This macro is useful for defining a list of any type.
 * This derives a new class from #GenericList# and
 * fills out the missing pieces to create a list class.
\begin{verbatim}
 DEFINE_AS_LIST(IptelAddressList, IptelAddress); //define
 IptelAddressList mylist;   //instantiate
 mylist = anotherlist;  // Copy the complete list
 IptelAddress a = mylist[12];   // access the 13th member
                                // Counting as 0,1,..12
 mylist.Append("128.59.19.200"); //Append a new member.
 mylist.Insert("sip:kns10@home");//Insert a member.
 mylist.GetSize();  // returns the number of elements in list.
 mylist.RemoveAt(10); //remove the 11th element. 
\end{verbatim}
* Note that the element for which you want to
* define a list must be a C++ object, with a
* copy constructor and a copy operator (=).
*
* @see GenericList
*/

#define DEFINE_AS_LIST(cls,T)                                          \
  class cls : public GenericList {                                     \
  public:                                                              \
  cls() : GenericList(true) { /*cout << #cls << "()" << endl;*/ }          \
  cls(const cls & val) : GenericList(val) { /*cout << #cls <<"()" << endl;*/ } \
  virtual ~cls() { /*cout << "~" << #cls << endl; */ RemoveAll(); }    \
  virtual cls & operator=(const cls & val) { GenericList::operator=(val); return *this; } \
  T & operator[](int index) const                                      \
      { return  (T &) (*(T *)GetAt(index)); }                          \
  int Append(const T & e) {                                              \
     T *n = new T(e); return GenericList::Append((void *)n); }         \
  void Insert(const T & e) {                                             \
     T *n = new T(e); GenericList::Insert((void *)n); }                \
  int Append(T* e) {                                                   \
     return GenericList::Append((void *)e); }                          \
  void Insert(T* e) {                                                  \
     GenericList::Insert((void *)e); }                                 \
  /* protected: */                                                     \
  void DeleteElement(void *data)                                       \
    { /*cout << #cls << " Delete Element" << endl;*/ if ( data!=NULL) delete (T*)data;}  \
  void *AllocateElement(void *v)                                       \
    { /*cout << #cls << " Allocate Element" << endl;*/ T* n = new T; if ( v!=NULL) *n =  *((T*)v); return n;} \
};

#else

/* Use templates here */

template <class T> class GenericListTemplate : protected list
{
 public:
  GenericListTemplate():list();
  GenericListTemplate(const GenericListTemplate &);
  virtual ~GenericListTemplate();
  T & operator[](int index) const;
  GenericListTemplate & operator=(const GenericListTemplate &);
  GenericListTemplate * Clone() const;
  int Append(const T e);
  void Insert(const T e);
  int Append(T * e);
  void Insert(T * e);
}

#define DEFINE_AS_ARRAY(cls,T) typedef GenericListTemplate<T> cls


#endif

/** 
 * GenericHeaderValue is a pair of header name and value field.
 * e.g.,
\begin{verbatim}

 GenericHeaderValue hv("To","sip:alice@home.com");
 (const char *)hv;  //this gives sip:alice@home.com
 hv = "sip:bob@office.com"; //This should update the hv value.
 hv = GenericHeaderValue("From","sip:myself"); // this updates both
                                               //header and value.

\end{verbatim}
 */
class GenericHeaderValue
{
 public:
  /**Construct an empty header value. */
  GenericHeaderValue();

  /**
   * Construct an object with assigned header and value.
   * e.g., GenericHeaderValue("To","sip:kns10@columbia.edu"); 
   */
  GenericHeaderValue(const char *attrib, const char *value);

  /**
   * Construct an object by using the given separator.
   * e.g., GenericHeaderValue("s=Dummy Subject", '=');
   * will create a object with header = "s" and
   * value = "Dummy Subject".
   */
  GenericHeaderValue(const char *attrib, char separator);

  /** Copy Constructor */
  GenericHeaderValue(const GenericHeaderValue & val);

  /**Destroy the object. Both header and value are freed. */

  ~GenericHeaderValue();

  /**
   * Typecast to string. 
   * @return the value of the header-value pair.
   * e.g., (const char * ) GenericHeaderValue("To","kns10")
   * gives "kns10".
   */
  operator const char *() const;

  /**Set the header. Overwrites previous.*/
  void SetType(const char *header);

  /**Set the value. Overwrites previous. */
  void SetValue(const char *value);

  /** Retrieve the header name.  */
  const char *GetType() const;

  /**Retrieve the value. */
  const char *GetValue() const;

  /**
   * Assign the value from a string.
   * e.g., hv = "Kns10" will set the value of hv to "Kns10".
   * header remains unaffected.
   */
  GenericHeaderValue & operator=(const char *); /* assign the value */

  /**Copy operator. */
  GenericHeaderValue & operator=(const GenericHeaderValue & val);

  /**Compare the two header-values. 
   * @return true if same, else false. */
  bool operator==(const GenericHeaderValue hv) const;

 private:
  char *header;
  char *value;
};

/** 
 * GenericHeadersList is a list for GenericHeaderValue.
 * 
 * @see GenericHeaderValue, GenericList, DEFINE_AS_LIST
 */
DEFINE_AS_LIST(GenericHeadersList,GenericHeaderValue);

/**
 * GenericHeaders: is a generic header-value hash table.
 * Since most of the protocols (SIP,SDP) have header-value
 * format for messages, this is very important.
 * Moreover this can be used to store any information for
 * other protocols also.
 */
class GenericHeaders
{
 public:
  /**Construtor. The separator is used to
     separate header from value. e.g., if separator is
     ':' then the printed string becomes, "header:value".
     SDP uses '='.
  */
  GenericHeaders(const char seperator= '=');

  /** Copy constructor */
  GenericHeaders(const GenericHeaders & h);

  /**Destroy the object*/
  ~GenericHeaders();

  /** Copy operator */
  GenericHeaders & operator=(const GenericHeaders & h);

  /**Retrive the number of header-values */
  int GetSize() const;

  /**Return reference to ith header value.
     Use index as -1 to insert something and
     index as size of this to append something.
   e.g., a[0] = GenericHeaderValue("To","alice@home.com") will 
   set the 0th element to "To:alice@home.com" while
   a[-1] = GenericHeaderValue("From","alice@home.com") will
   insert a new header-value "From:alice@home.com".
  */
  GenericHeaderValue & operator[](int index);

  /**Return the ith headervalue.
   e.g., char *value = (const char *) a[10] */
  const GenericHeaderValue operator[](int index) const;

  /**Return header value referencecorresponding to this attribute. 
   e.g., a["To"] = "sip:alice@sip.com" */
  GenericHeaderValue & operator[](const char *attrib);

  /**Return the header value corresponding to this attribute.
   e.g., char *ptr = a["From"] */
  const GenericHeaderValue operator[](const char *attrib) const;
  
  /**Add a new header value pair. Duplicates are not
   * checked. 
   */
  GenericHeaders & operator+= (const GenericHeaderValue g);

  /**Add a new header value pair from the string.
   * Use the separator to separate out the header and value.
   */
  GenericHeaders & operator+= (const char *headervalue);

  /**Remove the particular header field. All occcurances are 
   * removed. 
   */
  GenericHeaders & operator-= (const char *header);

  /**Remove the specific header value */
  GenericHeaders & operator-= (const GenericHeaderValue g);

  /**Test if the header is present or not
   * @return true if the header element is present. 
   */
  bool operator & (const char *header) const;

  /**Test if the header-value is present or not.
   * @return true if the header-value is present.
   */
  bool operator & (const GenericHeaderValue g) const;

 protected:
  GenericHeadersList  headers;
  char separator;
};

//@}

#endif /* _ADS_H_ */
