/* iptelapi.h - Common Header for SIP/H.323 C++ library.
 *  (c) 1999-2001 Columbia University. All rights reserved.
*/
#ifndef _IPTEL_API_H_
#define _IPTEL_API_H_

#include <iostream.h>  /* ostream */
#include "ads.h"       /* GenericHeaders and GenericList */

/**@name Common Header for SIP/H.323 C++ library */
//@{
#ifdef __cplusplus
extern "C" { /* For including C headers */
#endif

/*
 * We use a stub for pthread related things on windows.
*/
#ifdef WIN32 /* For windows */
#ifndef _WINSOCKAPI_
  #include <winsock2.h>
#endif
#endif

#include <time.h>  /* time_t */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>
#include "afi.h"
#include "sysdep.h"
#ifdef __cplusplus
}
#endif

/**
 * SESSIONID datatype: defined for easy portablity.
 * This is an unique identifier for the session.
 */
typedef unsigned long SESSIONID;

class IptelIPAddress; // forward declaration.

/**
 * IptelAddress: is a string used for storing addresses, for example, self
 * address, remote address, etc. This is a class and not typedefined
 * as 'char *' so that other protocols may override this value
 * to their requirements. 
 */
class IptelAddress
{
  public:
    /**@name Constructors
     * Constructors: either empty address or from const char * 
     * or String. When constructing from char * one
     * of the following formats is assumed and parsed in that
     * order.
     * \begin{verbatim}
       a.b.c.d:port    - IP Address and port
       a.b.c.d         - IP Address
       tel:number      - telephone number.
       h323:anything   - Any H323 address (resolved by gatekeeper).
       sip:user@host   - SIP address. 
       \end{verbatim}
     * Both escaped and unescaped URLs are supported.
     * "h323:Kundan Singh" is same as "h323:Kundan%20Singh".
     */
    //@{
    /** default constructor */
    IptelAddress();

    /** Construct from a string */
    IptelAddress(const char *address);

    /** Copy constructor */
    IptelAddress(const IptelAddress & address);
    //@}

    /**@name Destructors */
    //@{
    /** 
     * Destructor: destroys the object
     */
    virtual ~IptelAddress();
    //@}

    /**@name Operators */
    //@{
    /**
     * Copy operators. IptelAddress can be copied from
     * string.
     */
    virtual IptelAddress& operator=(const char * address);

    /** copy from another IptelAddress address */
    virtual IptelAddress& operator=(const IptelAddress & address);

    /** Copy from IP address */
    virtual IptelAddress& operator=(const IptelIPAddress & ipaddress);

    /** Comparing two IptelAddress'es.
     * @return true if same, false otherwise
     */
    virtual bool operator==(const IptelAddress & address) const;

    /** Comparing two IptelAddress'es 
     * @return true if same.
     */
    virtual bool operator==(const char * address) const;

    /**Typecast to string */
    operator const char *() const { return (const char *)Address; };

    /**Typecast to IptelIPAddress */
    operator IptelIPAddress () const;
    //@}

    /**@name URI related methods */
    //@{
    /**
     * GetScheme(): returns the string representation of 
     * URI scheme. NULL if none or error. Caller should
     * free the memory for returned string.
     */
    virtual char *GetScheme() const;

    /**
     * SetScheme(): Sets the URI scheme. Removes if NULL.
     */
    virtual void SetScheme(const char *s);

    /**
     * GetUser(): returns the string representation of 
     * Username. NULL if none or error. Caller should
     * free the memory for returned string.
     */
    virtual char *GetUser() const;

    /**
     * SetUser(): Sets the URI username. Removes if NULL.
     */
    virtual void SetUser(const char *s);

    /**
     * GetDisplayName(): Gets the display name for the address.
     */
    virtual char *GetDisplayName() const;

    /**
     * SetDisplayName(): Sets the display name 
     */
    virtual void SetDisplayName(const char *s);

    /**
     * GetURI(): Get the URI portion of the address.
     */
    virtual char *GetURI() const;

    /**
     * GetPassword(): returns the string representation of 
     * URI password. NULL if none or error. Caller should
     * free the memory for returned string.
     */
    virtual char *GetPassword() const;

    /**
     * SetPassword(): Sets the URI password. Removes if NULL.
     */
    virtual void SetPassword(const char *s);

    /**
     * GetHost(): returns the string representation of 
     * URI hostname. NULL if none or error. Caller should
     * free the memory for returned string.
     */
    virtual char *GetHost() const;

    /**
     * SetHost(): Sets the URI host. Removes if NULL.
     */
    virtual void SetHost(const char *s);

    /**
     * GetPort(): returns the
     * URI port number. 0 if none or error.
     */
    virtual int   GetPort() const;

    /**
     * SetPort(): Sets the URI port. Removes if 0.
     */
    virtual void SetPort(int s);

    /**
     * Get the Path for RTSP/HTTP like URLs.
     */
    virtual char *GetPath() const;

    /**
     * Set the Path for RTSP/HTTP like URLs 
     */
    virtual void SetPath(const char *s);

    /**
     * IsParameter(): returns true if the URI parameter is
     * presesnt. false otherwise.
     */
    virtual bool   IsParameter(const char *pname) const;

    /**
     * GetParameter(): returns the string representation of
     * the URI parameter, NULL if not present or on error.
     * Caller should free the memory of returned string.
     */
    virtual char * GetParameter(const char *pname) const;

    /**
     * SetParameter(): Sets the URI parameter. Removes if NULL.
     */
    virtual void SetParameter(const char *pname,
                              const char *value);

    //@}

    /**@name Other methods */
    //@{
    /**
     * GetAddressString() : returns the string representation
     *  of the Address. Return value points to an internal
     *  buffer and should not be freed. It is recommended that
     *  the application immediately copies the String (not the
     *  pointer) to another buffer because the content may change
     *  later e.g. after a call to operator=
     */
    virtual const char *GetAddressString() const;

    /**
     * IsEmpty():
     * This returns true if the address is empty and
     * false otherwise.
     */
    virtual bool IsEmpty() const;

    /**
     * IsTelephoneNumber():
     * This returns true if the address is a telephone number
     * without any scheme and host part.
     * e.g., +1-80-5556668.
     * Characters, "+", "-", "(", ")", "#", "'" , "*", 0-9 
     * are recognized.
     */
    virtual bool IsTelephoneNumber() const;

    /**
     * IsIpAddress():
     * Returns true if the address is an IP address.
     * With valid IP address as host portion.
     */
    virtual bool IsIpAddress() const;

    /**
     * IsIp6Address():
     * Returns true if the address is an IPv6 address.
     * With valid IPv6 address as host portion.
     */
    virtual bool IsIp6Address() const;

    /** operator << for printing onto output stream */
    friend ostream & operator<<(ostream & o, IptelAddress & addr);
    //@}

  protected:
    char *Address;
};

/**
 * IptelAddressList: is a list of IptelAddresses.
 */
DEFINE_AS_LIST(IptelAddressList, IptelAddress);

/**
 * IptelIPAddress:
 * Used to store IP v4 or v6 Address + port number.
 */
class IptelIPAddress
{
  public:
    /**@name Constructors */
    //@{
    
    /**
     * Default constructor - Initialize to AF_INET/INADDR_ANY/0.
     */
    IptelIPAddress(); 
    
    /** Try to resolve 'address' (host name, ip, or dotted addresses)
     * into usable IP v4 or v6.
    */
    IptelIPAddress(const char *address);

    /** Use the contents of sin.
    */
    IptelIPAddress(const struct sockaddr_storage& sin);

    /** Use in.
    */
    IptelIPAddress(const struct in_addr& in);

    /** Use in6.
    */
    IptelIPAddress(const struct in6_addr& in6);

    /**
     * IptelIPAddress can be constructed from IptelAddress. 
     * If IptelAddress does not have a valid IP Address 
     * then IptelIPAddress stores invalid value.
     */
    IptelIPAddress(const IptelAddress& address);

    /** From another IptelIPAddress */
    IptelIPAddress(const IptelIPAddress& address);
    //@}

    /**@name Destructors */
    //@{
    /** Destructor */
    virtual ~IptelIPAddress();
    //@}

    /**@name Copy operators.
     * IptelIPAddress can be copied from
     * IptelIPAddress, 
     * or IptelAddress object.
     */
    //@{
    /** Copy from another IP address */
    virtual IptelIPAddress& operator=(const IptelIPAddress& address);

    /** Parse and copy from a generic address */
    virtual IptelIPAddress& operator=(const IptelAddress & address);

    /** parse and copy from a string */
    virtual IptelIPAddress& operator=(const char* address);

    /** Copy.
    */
    virtual IptelIPAddress& operator=(const struct sockaddr_storage &sin);

    /** Copy from in_addr.
    */
    virtual IptelIPAddress& operator=(const struct in_addr& in);

    /** Copy from in_addr.
    */
    virtual IptelIPAddress& operator=(const struct in6_addr& in6);

    /** Compare with another IptelIPAddress */
    virtual bool operator==(const IptelIPAddress & address) const;

    /** Typecast to an IptelAddress */
    virtual operator IptelAddress() const;

    /** Typecast to a struct sockaddr_storage */
    virtual operator sockaddr_storage() const;

    /** Typecast to a struct in_addr */
    virtual operator in_addr() const;

    /** Typecast to a struct in6_addr */
    virtual operator in6_addr() const;
    /** Typecast to char* representation. Caller should free memory */
    virtual operator char*() const;

    //@}

    /**@name Accessing IP address and Port number */
    //@{
    /**
     * Accessing Port number in HOST BYTE ORDER.
     * Internally sockaddr_in stores in network byte order.
     * DO NOT call ntohs(GetPort()) or SetPort(htons(port));
     * instead use GetPort() and SetPort(port) directly.
     */
    unsigned short GetPort() const;
    /**
     * Set the port in host order.
     */
    void SetPort(unsigned short);

    /**
     * Accessing IP address. If successful returns 0 else
     * negative value. For GetIpAddress if successful 
     * the corresponding param passed is updated. 
     * Address and Port will be in NBO.
     */
    int  GetIpAddress(struct sockaddr_storage& saddr_in) const;

    /** Set using sockaddr_storage */
    void SetIpAddress(const struct sockaddr_storage& saddr_in);

    //@}

    /**@name Other methods */
    //@{
    /**
     * Retrun true if the IP address is valid.
     * 0.x.y.z is considered an invalid IP address.
     * port number is not examined.
     */
    bool IsValid() const;

    /**
     * Return true if the IP addres is a multicast address.
     * f0.0.0.0 to ff.0.0.0 in hex for IPv4. */
    bool IsMulticast() const;


    /**
     * Return true if the IP addres is a IPv4 address.
    */
    bool IsIP4() const;

    /**
     * Return true if the IP addres is a IPv6 address.
    */
    bool IsIP6() const;

    /** operator << for printing onto output stream */
    friend ostream & operator<<(ostream & o, IptelIPAddress & addr);
    //@}

  private :
    /*
     * sockaddr_in from socket library is used to store the 
     * IP v4 address internally
     */
    struct sockaddr_storage    Address;
};

/**
 * IptelIPAddressList is a list of IptelIPAddreses
 */
DEFINE_AS_LIST(IptelIPAddressList, IptelIPAddress);

/**
 * FormatInfo:
 * This defines the payload type and definitions of the payload
 * types.
 */
class FormatInfo
{
 public:
  /** Construct without ant predefined parameter */
  FormatInfo();

  /**Construct a format from predefined payload types */
  FormatInfo(int payloadtype);

  /** Construct from a mime subtype,e.g., audio/basic */
  FormatInfo(const char *mimetype);

  /**Construct a format from a predefined definition.
   */
  FormatInfo(const char *definition,
             int clockrate, /*default takes 8000 for audio */
             int channelcount = 1);

  /** Construct from a mime subtype and assign
   * appropriate payloadtype 
   */
  FormatInfo(int payloadtype,
             const char *mimetype);

  /**Construct a format from a payloadtype and a definition
   and clockrate and channel count */
  FormatInfo(int payloadtype, const char *definition,
             int clockrate, /* default takes 8000 for audio */
             int channelcount = 1);

  /**Copy constructor */
  FormatInfo(const FormatInfo & format);

  /** Destructor destroys the object */
  virtual ~FormatInfo();

  /** Get the payload type */
  virtual int  GetPayloadType() const {return payloadtype; };
  
  /** get name of algorithm e.g., pcmu */
  virtual const char *GetFormatName() const {return description; };
  
  /** get the clock rate */
  virtual int  GetClockRate() const {return clockrate; } ;
  
  /** get the channel count */
  virtual int  GetChannelCount() const {return channelcount; };

  /** Set the payload type */
  virtual void SetPayloadType(int pt) {payloadtype = pt;} ;
  
  /** Set the description */
  virtual void SetFormatName(const char *description);
  
  /** Set channel cound */
  virtual void SetChannelCount(int cound) {channelcount = cound;};

  /** Set the clock rate. */
  virtual void SetClockRate(int rate) {clockrate = rate; } ;

  /** Assignment operator */
  virtual FormatInfo & operator=(const FormatInfo & format);

  /** Comparision operator to compare with a payload type.
   * If payloadtype is -1 then it compares everything.
   * If description if -1 then it compares everything.
   * If description, clockrate are present then
   * comparision is done based on description, clockrate and
   * channelcount and not on payloadtype.*/
  virtual bool operator==(const FormatInfo format) const;

  /** Typecast to a int as a payload type */
  virtual operator int() const {return payloadtype;} ;
  
  /** Typecast to a string as the description of format,
   * e.g., pcmu.
   */
  virtual operator const char *() const {return description; } ;

  /** operator << for printing onto output stream */
  friend ostream & operator<<(ostream & o, FormatInfo & addr);

 private:
  int payloadtype;
  char *description;
  int clockrate;
  int channelcount;
};

/**
 * FormatInfoList is a list of FormatInfo 
 */
DEFINE_AS_LIST(FormatInfoList, FormatInfo);

/**
 * MediaInfo:
 * Used to store information about a session. H245 or SDP
 * specific session information can be stored.
 * A session refers to 
 * audio,video,data,application or control media type and may be
 * unidirectional or bidirectional. 
 * (Following description is copied from SDP rfc2327):
 * The difference between "application" and "data" is
 * that the former is a media flow such as whiteboard information, and
 * the latter is bulk-data transfer such as multicasting of program
 * executables which will not typically be displayed to the user.
 * "control" is used to specify an additional conference control
 * channel for the session. Typically formats will correspond to
 * one Media (DataType). If multiple media is multiplexed in 
 * single format e.g. audio and video then corresponding MIME type
 * should be defined and is used for specifying DataType and
 * Format.
 *
 * Example of Corelation to MIME:
 *   a mime type of application/wb (For white board) corresponds
 *   to DataType = "Application" and Format = "wb".
 */
class MediaInfo
{
  public:
    /**
     * Direction of the session which respect to this endpoint.
     * Transmit and Receive are aplicable for unidirectional
     * sessions and "TransmitAndReceive" is applicable to bidirectional
     * sessions.
     * \begin{verbatim}
      UnknownDirection
      Transmit,
      Receive,
      TransmitAndReceive
      \end{verbatim}
     */
    enum Direction
    {
      UnknownDirection   = 0,
      Transmit           = 1, /* 0x01 */
      Receive            = 2, /* 0x02 */
      TransmitAndReceive = 3  /* 0x01 | 0x02 == 0x03 */
    } direction;

    /**
     * Datatype: Datatype (audio/video) for the mediainfo.
     * Values are Audio, Video, Application, Data, or Control 
     * used for the session.
     * Single session may not have multiple media types.
     */
    enum DataType
    {
      UnknownData = 0,
      Audio,                  /* pcmu, gsm */
      Video,                  /* h.261, mpeg */
      Application,            /* wb, chat */
      Data,                   /* ? */
      Control,                /* h.323 */
      /*
       * Add additional DataType here.
       */
      DataTypeCount
    } datatype;

    /**
     * TransportType: RTP/AVP, UDP or TCP.
     * Values are RTP_AVP_Transport, UDP_Transport,
     * or TCP_Transport.
     */
    enum TransportType
    {
      UnknownTransport = 0,
      RTP_AVP_Transport,      /* RTP/AVP */
      UDP_Transport,          /* udp */
      TCP_Transport,          /* tcp */
      H323_Transport          /* H.323 */
    } transport;

    /**
     * NetworkType: IN=Internet
     * values are NetworkIN
     */
    enum NetworkType
    {
      UnknownNetwork = 0,
      NetworkIN
    } network;

    /**
     * AddressType: IP4=IP version 4 address/ IP6 is IP version 6 address
     */
    enum AddressType
    {
      UnknownAddress = 0,
      AddressIP4,
      AddressIP6
    } addresstype;

    /**
     * Unknown data types and other unknown values 
     * can be stored in this;
     * Should these be private XXX.
     * Possible indices are datatype, transport, network, address.
     */
    GenericHeaders parameters; 

    /**
     * Constructor: An empty object is allocated
     * while constructing.
     */
    MediaInfo();

    /**
     * Constructor to allocate a MediaInfo of given type.
     */
    MediaInfo(Direction dir,
              DataType type = Audio,
              TransportType transport = RTP_AVP_Transport,
              NetworkType net = NetworkIN,
              AddressType addr = AddressIP4);

    /** Copy constructor */
    MediaInfo(const MediaInfo & media);

    /** Destructor for media info */
    virtual ~MediaInfo() { };

    /**@name Get or Set for format (payload) types..
     * Get and Set methods for different parameters of the session.
     * Some parameters may not be applicable for a particular
     * protocol say H245 and should not be used.
     */
    //@{
    /** Set the data type: audio,video... */
    virtual void   SetDataType(MediaInfo::DataType dt) {datatype =dt;} 

    /** get the datatype: audio,video ... */
    virtual MediaInfo::DataType  GetDataType() const {return datatype;}

    /** set the direction: transmit, receive */
    virtual void   SetDirection(MediaInfo::Direction dir) {direction=dir;}

    /** get the direction: transmit, receive */
    virtual MediaInfo::Direction  GetDirection() const {return direction;}

    /** Set the transport type */
    virtual void SetTransportType(MediaInfo::TransportType t) {transport=t;}

    /** retrieve the transport type */
    virtual MediaInfo::TransportType GetTransportType() 
                                     const {return transport;}

    /** Set the transport type */
    virtual void SetNetworkType(MediaInfo::NetworkType t) {network=t;}

    /** retrieve the transport type */
    virtual MediaInfo::NetworkType GetNetworkType() const {return network;}

    /** Set the transport type */
    virtual void SetAddressType(MediaInfo::AddressType t) {addresstype=t;}

    /** retrieve the transport type */
    virtual MediaInfo::AddressType GetAddressType() const 
                                   {return addresstype;}

    /** Adding Audio/Video predefined payload types */
    virtual int  AddFormat(const FormatInfo & format)
      { formats.Append(format); return 0; }
 
    /** 
     * Removing Format from the Format List of Media.
     * payload == -1 removes all entry.
     */
    virtual int  RemoveFormat(int payload) {
      FormatInfo f(payload,NULL); return RemoveFormat(f);}

    /** remove format by name of algorithm, e.g., pcmu */
    virtual int  RemoveFormat(const char *name) { FormatInfo f;
    f.SetFormatName(name); return RemoveFormat(f);}

    /** remove format by comparing with format specified */
    virtual int RemoveFormat(const FormatInfo & format);

    virtual int RemoveAllFormats();

    /** Querring if the format is present */
    virtual bool IsFormatPresent(int payload) const {
      FormatInfo f(payload); return IsFormatPresent(f);}

    /** Querring if the format is present by algorithm */
    virtual bool IsFormatPresent(const char *name) const {
      FormatInfo f; f.SetFormatName(name); return IsFormatPresent(f);}

    /** Querring if the format is present by FormatInfo */
    virtual bool IsFormatPresent(const FormatInfo format) const 
      { return (GetFormatIndex(format) >=0); }

    /** Getting the index of the formatinfo. Returns -1 if not present */
    virtual int GetFormatIndex(const FormatInfo & format) const;

    /** Extract the first format info */
    virtual FormatInfo & GetFormatInfo() const { return formats[0];}

    /** Extract the format info with given payload type */
    virtual FormatInfo & GetFormatInfo(int payloadtype) const
      {int i=GetFormatIndex(FormatInfo(payloadtype)); return formats[i]; }

    /** Extract the format info with given name */
    virtual FormatInfo & GetFormatInfo(const char *name) const
      {int i=GetFormatIndex(FormatInfo(name,0)); return formats[i];}

    /** Return the number of formats */
    virtual int GetFormatCount() const {return formats.GetSize(); }

    /** Operator to extract the ith format */
    virtual FormatInfo & operator[](int index) const {return formats[index];}

    /** Operator to add a format */
    virtual MediaInfo & operator+=(const FormatInfo & format)
    {AddFormat(format); return *this;} ;

    /** Operator to remove a format */
    virtual MediaInfo & operator-=(const FormatInfo & format)
    {RemoveFormat(format); return *this;};

    /** Operator to check if a format is present */
    virtual bool operator & (const FormatInfo & format) const 
      {return IsFormatPresent(format); }

    /** Copy operator */
    virtual MediaInfo & operator=(const MediaInfo & media);

    /** Operator to compare the media info.
     * Only direction, transport and datatype are compared.
     * Unknown values in second argument of binary operator matches
     * anything. */
    virtual bool operator==(const MediaInfo & media) const;

    //@}

    /**@name Media Transport */
    //@{

    /** Set the transmit port. Not used in SIP. Use in host order. */
    virtual void           SetTxPort(unsigned short port) {tx.SetPort(port);}

    /** Retrieve the trsnamit port. Returns in host order. */
    virtual unsigned short GetTxPort() const {return tx.GetPort();}

    /** Set the receive port at which RTP can be received, inhost order */
    virtual void           SetRxPort(unsigned short port) {rx.SetPort(port);}

    /** Get the receive port at which RTP can be received, in host order */
    virtual unsigned short GetRxPort() const {return rx.GetPort();}

    /**
     * [Set,Get][Self,Remote]Address():
     * for self and remote. Only IPv4/v6 Address (not port number)
     * can be accessed using these. If Get*Address returns invalid
     * IptelIPAddress then IP Address of the SessionDescription Class
     * which contains list of this (MediaInfo) class objects should
     * be used.
     */
    /** Set the transmit address. Not used in SIP */
    virtual void           SetTxAddress(const IptelIPAddress &
                                        address) {tx = address;}

    /** Get the transmit address */
    virtual IptelIPAddress GetTxAddress() const {return tx;}

    /** Set the receive address at which the RTP can be received */
    virtual void           SetRxAddress(const IptelIPAddress &
                                        address) {rx = address;}

    /** Get the receive address at which the RTP can be received */
    virtual IptelIPAddress GetRxAddress() const {return rx;}

    /**Retrive the TTL value in Rx address if it is multicast address
     */
    virtual int  GetRxTtl() const {return rx_ttl;}

    /** Set the TTL value in Tx address if it is multicast address */
    virtual void SetRxTtl(int ttl) {rx_ttl = ttl;}

    //@}

    /**@name misc methods */
    //@{
    /**
     * [Set,Get][Tx,Rx]Bandwidth(): in bps.
     * for tx_bandwidth and rx_bandwidth. These methods
     * have no effect if the bandwidth is limited by
     * total call bandwidth or the payload type.
     * e.g. if payload type is G711 U law 64K (PCMU)
     * then bandwidth will always be 64K.
     */
    virtual void          SetTxBandwidth(unsigned long bandwidth)
      {tx_bandwidth= bandwidth;};
    ///
    virtual unsigned long GetTxBandwidth() const { return tx_bandwidth;} ;
    ///
    virtual void          SetRxBandwidth(unsigned long bandwidth)
      {rx_bandwidth=bandwidth;};
    ///
    virtual unsigned long GetRxBandwidth() const {return rx_bandwidth;};

    /** 
     * To access Media specific attributes.
     * SetAttribute() : 
     * GetAttribute() : returns string containing value
     *    of the attribute or NULL if not present. 
     *    Application should not modify the pointer returned
     *    and should copy the content to another local
     *    buffer immediately.
     * IsAttributePresent(): return true if present.
     */
    virtual void  SetAttribute(const char *attr,
                               const char *value)
    { headers[attr] = value; }
    ///
    virtual const char *GetAttribute(const char *attr) const
    { return headers[attr]; }
    ///
    virtual bool  IsAttributePresent(const char *attr) const
    { return headers & attr; }

    //@}
    /**@name Other methods */

    //@{
    /** operator << for printing onto output stream */
    friend ostream & operator<<(ostream & o, MediaInfo & addr);
    //@}

  protected:
    /* 
     * private and protected members are not 
     * part of the interface 
     */
    FormatInfoList formats;  /* Dynamic array for formats */

    IptelIPAddress tx;      /* ipaddress:portnumber for session */
    IptelIPAddress rx;      /* ipaddress:portnumber for session */
    int rx_ttl;             /* TTL for rx if it is multicast address */
    /* Remote address may not be used by an application which uses SDP */
   
    unsigned long  tx_bandwidth; /* Max Transmit bandwidth */
    unsigned long  rx_bandwidth; /* Max Receive bandwidth */
  
  public:
    /* SDP attributes */
    GenericHeaders headers;

  friend class SessionDescription;
};

/** MediaInfoList is list of MediaInfo */
DEFINE_AS_LIST(MediaInfoList, MediaInfo);

/**
 * SessionDescription: contains list of session information. 
 * Each session is represented by a MediaInfo 
 * object in the list.
 */
class SessionDescription
{
  public:
    /** 
     * Constructors and Destructors
     */
    SessionDescription();
    /** Copy constructor*/
    SessionDescription(const SessionDescription & session);

    ///
    virtual ~SessionDescription() { };

    /**
     * Copy operator 
     */
    virtual SessionDescription & operator=(const SessionDescription
                                           &s);
    
    /**
     * AddSession():
     * to add a MediaInfo to the list.
     */
    virtual void AddSession(const MediaInfo & session) {
      sessions.Append(session); }

    /**
     * DeleteSession():
     * to delete a MediaInfo from the list.
     * If all the parameters are unknown it matches to any 
     * value. 
     * So 
     * DeleteSession(MediaInfo::Transmit,MediaInfo::UnknownData);
     * deletes all transmit sessions from the sessionDescription.
     */
    virtual void DeleteSession(const MediaInfo & session);
    ///
    virtual void DeleteSession(
      MediaInfo::Direction dir= MediaInfo::UnknownDirection,
      MediaInfo::DataType  dat= MediaInfo::UnknownData) { 
      MediaInfo m(dir,dat); DeleteSession(m);}

    /** Find the session index. session index>=0 */
    virtual int FindSessionIndex(const MediaInfo & session) const;

    /**Operator to add a media info */
    virtual SessionDescription& operator+=(const MediaInfo & media)
    {AddSession(media); return *this;}

    /**Operator to remove a media info */
    virtual SessionDescription& operator-=(const MediaInfo & media)
    {DeleteSession(media); return *this;}

    /**Method to return the number of media infos */
    virtual int GetMediaCount() const {return sessions.GetSize();}

    /** operator to get the ith MediaInfo */
    virtual MediaInfo & operator[](int index) const {return sessions[index];}

    /** 
     * FindSession():
     * to find the specific session from the list.
     * If DataType is Unknown then all datatypes are 
     * examined. If Direction is Unknown then
     * any direction is acceptable.
     * If transport is Unknown any is acceptable.
     * returns true if found, false otherwise.
     */
    virtual bool FindSession(MediaInfo & mediainfo,
      MediaInfo::Direction dir  = MediaInfo::UnknownDirection,
      MediaInfo::DataType  data = MediaInfo::UnknownData,
      MediaInfo::TransportType  transport = MediaInfo::RTP_AVP_Transport,
      MediaInfo::AddressType address = MediaInfo::UnknownAddress
                           ) const;

    /** Set the session ID */
    void       SetSessionId(SESSIONID sid) {sessionId = sid;};
    /** Get the session ID. Generate a new if none exists */
    SESSIONID  GetSessionId();

    /** 
     * Returns a photocopy of itself.
     * All the internal members are cloned.
     */
    virtual SessionDescription *Clone();

    /**
     * GetAttributes: is used to retrieve the
     * attribute value corresponding to given header (o=,i=) etc.
     * Defined attributes are
     * v, o, s, i, u, e, p, c, b, etc (refer RFC2327).
     * Only session description attributes are defined/used here.
     * Media description attributes are part of MediaInfo.
     */
    virtual const char *GetAttribute(const char *header) const 
      { return headers[header]; }

    /**
     * SetAttribute: is used to set the attributes of the
     * corresponding header (o=, i=, etc.). A
     * value of NULL resets and removes the attrib.
     */
    virtual void SetAttribute(const char *attr, const char *value)
      { headers[attr] = value; }

    /** operator << for printing onto output stream */
    friend ostream & operator<<(ostream & o, SessionDescription & session);

  protected:
    /*
     * Protected members are not part of the interface.
     * Some of the elements like bandwidth,encryption,title
     * can be included in the derived classes of this class.
     */
    MediaInfoList sessions;
    IptelIPAddress tx; /* Transmit media transport  */
    IptelIPAddress rx; /* Receive media transport */

    /* SDP related fields */
    SESSIONID           sessionId;

  public:
    GenericHeaders      headers;
};


class IptelCall;    // forward declaration for IptelCall.

/**
 * IptelEndpoint: represents a protocol server or client
 * A server will be listening for incoming calls. while
 * a client won't. Note however that to place an outgoing call
 * you need IptelCall and not IptelEndpoint. IptelEndpoint
 * is an abstraction of a terminal or IP part of a gateway
 * or any other entity which has a fixed set of Capabilities
 * and local addresses for all the calls.
 */
class IptelEndpoint
{
  public:
    /**
     * IptelEndpoint::IptelEndpoint() : 
     * It initialises the protocol endpoint client/server.
     * When called with server option
     * This function will typically create a server thread
     * and allocate sockets for listening. Implementation for
     * individual protocols should write their own constructor 
     * to achieve this functionality. If IptelAddress is passed
     * then the server listens for incoming calls
     * at that address if possible (e.g. in H323).
     * A client endpoint will create an object to hold 
     * client properties. Not that the IptelIPAddress argument 
     * is not the local name for the server but the prefered 
     * address at which the server should listen. The address is 
     * typically of types Ip address 0.0.0.0:portno or
     * 127.0.0.1:portno i.e. only the
     * port number is used for most of the protocols.
     *
     * IsRegistrar = true indicates that this endpoint object will
     * be used as a registrar e.g. H323 Gatekeeper or 
     * SIP registrar. So the endpoint should listen at the 
     * appropriate ports for listening to Register requests.
     * If registration_address is specified then the server listens
     * for incoming registration requests at that address. Although
     * IsRegistrar = true and IsServer = False does not make sense for
     * SIP server but it is possible for H323 gatekeepers which want
     * to do just address translation and not accept calls.
     *
     * Dynamic change in server address is not allowed since that
     * may cause lot of implementation complexity.
     */
    IptelEndpoint(bool IsServer    = false, 
                  bool IsRegistrar = false);
    
    ///
    IptelEndpoint(bool IsServer, 
                  const IptelAddress& registrationAddress);

     ///
    IptelEndpoint(const IptelIPAddress& listeningAddress, 
                  bool IsRegistrar);

    ///v4 and registrar address.
    IptelEndpoint(const IptelIPAddress& listeningAddress4, 
                  const IptelIPAddress& registrationAddress,
                  bool  IsRegistrar);

    ///v4, v6 and registrar
    IptelEndpoint(const IptelIPAddress& listeningAddress4, 
                  bool  IsRegistrar,
                  const IptelIPAddress& listeningAddress6);

    ///this constructor will activate v4, v6 + registration.
    ///note the order of arguments.
    IptelEndpoint(const IptelIPAddress& listeningAddress4, 
                  const IptelIPAddress& registrationAddress,
                  const IptelIPAddress& listeningAddress6,
                  bool  IsRegistrar);
 
    /**
     * IptelEndpoint::~IptelEndpoint() : Deletes the server instance.
     */
    virtual ~IptelEndpoint();
    
    /**
     * IptelEndpoint::IsServer():
     * returns true if this is a server i.e. listening for 
     * incoming calls.
     */
    bool IsServer() const {return isserver;}

    /**
     * IptelEndpoint::IsRegistrar():
     * returns true if this is a registrar i.e. H323 Gatekeeper.
     */
    bool IsRegistrar() const {return isregistrar;} 

    /**
     * IptelEndpoint::AllocateNewCallObject():
     * This allocates a new object of the appropriate
     * Call. Default behaviour should be to allocate an
     * object of type IptelCall or any inherited class.
     * @param callid represents the call. If NULL then create a new call.
     */

    virtual IptelCall *AllocateNewCallObject(const char *callid) = 0;

    /**
     * IptelEndpoint::GetServerAddress():
     * Returns the address at which the server is listening 
     * for incoming calls. e.g. 0.0.0.0:1720 or 127.0.0.1:1720 for H323.
     * This default behaviour returns
     * Invalid IptelIPAddress as well as port = 0
     * for client only endpoints (non-servers).
     */
    virtual const IptelIPAddress GetServerAddress() const {return
                                                       ListeningAddress;}

    /**
     * IptelEndpoint::GetServerAddress6():
     * Returns the IPv6 address at which the server is listening 
     * for incoming calls. The default behaviour returns
     * Invalid IptelIPAddress as well as port = 0
     * for client only endpoints (non-servers).
     */
    virtual const IptelIPAddress GetServerAddress6() const {return
                                                       ListeningAddress6;}
    /**
     * IptelAddress::[Get,Set]LocalName():
     * To access the local Endpoint Name. This name
     * is used by default as self address for outgoing or
     * incoming calls. Default behaviour does not do
     * anything.
     */
    virtual IptelAddress GetLocalName() const {return LocalName;}
    ///
    virtual void SetLocalName(const IptelAddress & name) {LocalName =
                                                            name;}

    /**
     * IptelAddress::[Get,Set]Registrar():
     * To access the Registrars address. This name
     * is used by default when registering or unregistering.
     */
    virtual IptelAddress GetRegistrar() const {return Registrar;}
    ///
    virtual void SetRegistrar(const IptelAddress & name) {Registrar =
                                                            name;}

    /**
     * IptelAddress::[Get,Set]ContactAddress():
     * To access the Contact Address. This name
     * is used by default as contact address while
     * registering or unregistering. 
     */
    virtual IptelAddress GetContactAddress() const {return
                                                        contactAddress;}
    ///
    virtual void SetContactAddress(const IptelAddress & name)
      {contactAddress = name;}

    /**
     * RegistrationInfo: 
     * Information about registration.
     * @see #SIPRegistrationInfo#
     */
    class RegistrationInfo 
    {
    public:
      /** Address of the registrar. If absent, then derived from localName. */
      IptelAddress registrar;
      /** Address to be registered. This must be present */
      IptelAddress localName; 
      /** Contact address. If absent then local host name is used */
      IptelAddress contactAddress;
    };

    /** RegistrationInfoList is a list of RegistrationInfo */
    DEFINE_AS_LIST(RegistrationInfoList, RegistrationInfo);

    /**
     * Register/Unregister():
     * To do user registration.
     * If Registrar is not set then default registrar is used.
     * If localAddress is not set then Localendpoint name is used.
     * If contactAddress is not set then local ListeningAddress or
     * localName is used.
     * Unregister with no parameter unregisters all the names.
     * These return 0 on success and <0 on failure. Return value
     * indicate that the process of registration was initiated or
     * not. Actual registration happened or not is conveyed by
     * a callback method.
     */
    virtual int Register();
    ///
    virtual int Unregister();
    ///
    virtual int Register(
      const IptelEndpoint::RegistrationInfo & reginfo);
    ///
    virtual int Unregister(
      const IptelEndpoint::RegistrationInfo & reginfo);

    /**
     * If userName is supplied then it is used instead of
     * localName. Similarly when contactAddress is supplied then
     * it is used instead of protected member. 
     * This could be useful in gateways when single endpoint has to
     * register with different users and different contacts.
     *
     */
    virtual int Register(const IptelAddress & user);
    ///
    virtual int Register(const IptelAddress & user,
                         const IptelAddress & contact);
    ///
    virtual int Unregister(const IptelAddress & user);
    ///
    virtual int Unregister(const IptelAddress & user,
                           const IptelAddress & contact);

    /**
     * GetRegistration(): and GetRegistrationCount();
     * Returns the registrationInfo for this endpoint.
     * Array reginfo[] should be of size "size".
     * GetRegistrationCount returns number of registrations for
     * this endpoint and GetRegistration() returns numbers of
     * registrations filled in reginfo[] and -1 in case of error.
     * If localName is specified then Registrations for only
     * that name is returned.
     */
    virtual int GetRegistrationCount() const;
    ///
    virtual int GetRegistration(
                    IptelEndpoint::RegistrationInfo reginfo[],
                    int size) const;
    ///
    virtual int GetRegistrationCount(const IptelAddress localName) const;
    ///
    virtual int GetRegistration(const IptelAddress localName,
                    IptelEndpoint::RegistrationInfo reginfo[],
                    int size) const;
    /**Operator to return the ith registration */
    virtual RegistrationInfo & operator[](int index) const 
                                     { return reg_list[index]; }

    /**
     * OnIncomingRegistration
     * Invoked when some other endpoint registers with this endpoint.
     * This is possible when this endpoint is used in a registrar.
     * e.g. H323 gatekeeper or SIP registrar.
     *
     * reginfo contains the information about the incoming registration
     * as follows: \begin{verbatim}
       IptelAddress registrar = this endpoints Address;
       IptelAddress localName = remotes address to be registered;
       IptelAddress contactAddress = remotes contact info;
       time_t       expiry    = registration expiration time; 
       float        q_value   = q value as specified by SIP;
       \end{verbatim}
     *
     * The overloaded method should return >= 0 if it wants to accept the
     * registration or -1 if it wants to reject the registration.
     *
     * Default behaviour returns -1 since registrar is not implemented.
     */

    virtual int OnIncomingRegistration(const 
                     IptelEndpoint::RegistrationInfo reginfo);


    /**
     * OnIncomingUnregistration.
     * Invoked when a registered endpoint unregisters with
     * this endpoint. This is possible when this is used in a
     * registrar, e.g., H.323 gatekeeper or SIP registrar.
     * Default behaviour does not do anything.
     */
    virtual void OnIncomingUnregistration(const 
                     IptelEndpoint::RegistrationInfo reginfo);

    /** operator << for printing onto output stream */
    friend ostream & operator<<(ostream & o, IptelEndpoint & ep);

  protected:
    bool          isserver;
    bool          isregistrar;

    IptelIPAddress  ListeningAddress;
    IptelIPAddress  ListeningAddress6;
    IptelIPAddress  RegistrationAddress;
    IptelAddress    LocalName;
    IptelAddress    Registrar; /* SIP registrar or H323 Gatekeeper */
    IptelAddress    contactAddress;

    RegistrationInfoList reg_list;
};


/**
 * IptelCall: represents a call. Various methods are provided to use 
 * the services of a Call like initiating a call, alerting remote
 * party, accepting an incoming call etc. Various virtual methods
 * are also provided to indicate events related to the call like
 * alerting received, call closed by remote, error indication etc.
 * All the methods are made virtual so that the inherited class can do
 * some account/status keeping.
 */
class IptelCall 
{
  public:
    /**
     * IptelCall::IptelCall(): Initializer for the call.
     */
    IptelCall();

    /**
     * IptelCall::~IptelCall() destructor for the call.
     */
    virtual ~IptelCall();

    /**
     * IptelCall::IsActive(): 
     * Returns H323Success if Active or H323Failure if not 
     * active.
     */
    virtual bool   IsActive() const {return isactive;}

    /**
     * IptelCall::Initiate(): Method to initiate a call. "to" is the
     * address of the callee and "from" is the address
     * of the caller. from is optional and may be empty.
     */
    virtual int   Initiate(const IptelAddress & to,
                           const IptelAddress & from) = 0;

    /**
     * IptelCall::Hangup() : Method to terminate a call.
     */
    virtual int   Hangup(
          int reason = 0, /* 0 for normal termination */
          const char *strReason = "Normal Termination") = 0;

    /**
     * IptelCall::Alert(): Method to inform the remote party that the 
     * user has been alerted i.e. the phone is ringing.
     */
    virtual int   Alert() = 0;


    /**
     * IptelCall::Accept(): Method to accept an incoming call.
     */
    virtual int   Accept() = 0;
    
    /**
     * IptelCall::Reject(): Method to reject an incoming call.
     */
    virtual int   Reject() = 0;

    /**
     * IptelCall::OnNewIncomingCall() : 
     * This is called by the server thread whenever a new Incoming
     * call is received. API users must derive another
     * class from IptelCall and give an implementation for
     * this method. Typical implementation would inform the
     * user of the incoming call and either accept or reject the
     * call.
     */
    virtual void OnNewIncomingCall(IptelEndpoint & ep) = 0;

    /**
     * IptelCall::OnAlerting(): Invoked when the remote user is
     * informed about the call. 
     */
    virtual void   OnAlerting() = 0;

    /**
     * IptelCall::WaitForResponse(): Allows the caller thread to
     * wait till the system receives the Call established or
     * call rejected response from remote. If this responses is
     * already received then returns immediately.
     * After the call returns the caller should check using 
     * IptelCall::IsActive() to findout if the call was 
     * established or failed. If the client application does
     * not use this method then it has to write its own 
     * mechanism for waiting on the response using
     * OnCallEstablished() and OnCallRejected(). 'timeout' 
     * parameter specifies the timeout for wait in milliseconds.
     * However the timer precision is system dependent.
     * Return value of Success indicates either call is successfully
     * accepted or rejected. Return value of 1 indicates 
     * that the request timed out. Return of 0 indicates
     * system has received response from remote. Return value of -1
     * indicates some failure in executing the command.
     * Timeout of -1 indicates wait
     * indefinitely.
     */
    virtual int   WaitForResponse(int timeout = -1) = 0;
    
    /**
     * IptelCall::OnCallEstablished(): Invoked when an 
     * outgoing/incoming call is successfully established. 
     * Default behaviour does nothing.
     */
     
    virtual void   OnCallEstablished() = 0;

    /**
     * IptelCall::OnCallRejected(): Invoked when an outgoing
     * call can not be established. System removes the call 
     * information and then calls this method.
     * Default behaviour does nothing. Application can get
     * the errno and find out the reason for failure.
     *
     * @return Application should return 0 if you want the object
     * to be deleted automatically by the library.
     * Otherwise return non-zero.
     */
    virtual int   OnCallRejected() = 0;

    /**
     * IptelCall::OnHangup(): Invoked when the remote user closes
     * the call. The call context is removed by the system and
     * the application is informed by this method.
     * Default behaviour does nothing.
     *
     * @return Application should return 0 if you want the object to
     * be deleted automatically by the library. Otherwise return
     * non-zero.
     */
    virtual int   OnHangup() = 0;

    /**
     * IptelCall::WaitForTermination(): Allows the caller thread to
     * wait till the remote terminates the call or the 
     * timeout happens. If the timeout happens then the call
     * remains active and you have to explicitly call
     * hangup().
     */
    virtual int   WaitForTermination(int timeout = -1) = 0;
    
    /**
     * IptelCall::[Set,Get]xyz():
     * The access methods for the parameters of the call. 
     * The memory for arguments (if they are pointers) shall not
     * be freed by Setxyz(). Getxyz() returns the internal
     * pointers (if the return value is a pointer) and
     * should not be freed by the caller.
     */

    /**
     * IptelCall::[Set,Get]RemoteAddress():
     * To set or get the remote address of the call. For out
     * going call this is destination address and for incoming call
     * this is usually the source address.
     */
    virtual int  SetRemoteAddress(const  IptelAddress & r) 
                                  { remote = r; return 0;}
    /// 
    virtual IptelAddress     GetRemoteAddress() const {return remote;}

    /**
     * IptelCall::[Set,Get]RemoteAddress():
     * To set or get the self address of the call. For
     * outgoing call this is usually the source address and for
     * incoming call the destination address.
     */
    virtual int  SetSelfAddress(const IptelAddress & s) 
                               {self = s; return 0;}
    /// 
    virtual IptelAddress     GetSelfAddress() const {return self;}

    /**
     * IptelCall::[Set,Get]ProxyAddress():
     * To set or get the address of the intermediate proxy
     * or gateway if any. This maps to the H323 requestURI.
     */
    virtual int  SetProxyAddress(const IptelAddress & proxy) = 0;
    /// 
    virtual IptelAddress     GetProxyAddress() = 0;

    /**
     * IptelCall::[Set,Get]SessionDescription():
     * To set or get the session information for the call.
     */
    virtual int SetSelfSessionDescription(
                          SessionDescription * session);
    ///
    virtual int SetRemoteSessionDescription(
                          SessionDescription * session);
    ///
    virtual SessionDescription *GetSelfSessionDescription();
    ///
    virtual SessionDescription *GetRemoteSessionDescription();
    
    /**
     * IptelCall::[Set,Get]CallBandwidth():
     * to set or get the maximum call bandwidth for the call.
     * This includes bandwidth for all sessions in each 
     * direction. Get() returns 0 in case of error
     * or value not known.
     */
    unsigned long  SetCallBandwidth(unsigned long b) { return bandwidth = b;}
    ///
    unsigned long  GetCallBandwidth() {return bandwidth;}

    /** 
     * IptelCall::ErrorType: Error values for the protocol 
     * or the implementation. 
     * Most of the error values are imported from H323 
     * specification.
     */
    enum ErrorType
    {
      IPTELERR_NO_ERROR = 0,
      IPTELERR_NOT_SUPPORTED,
      IPTELERR_MEMORY_FAILURE,
      IPTELERR_NETWORK_ERROR,
      IPTELERR_PROTOCOL_ERROR,
      IPTELERR_INVALID_ADDRESS,
      IPTELERR_OPENH323_INTF,
      IPTELERR_OPENH323_CORE,
      /*
       * Add new values here.
       */
      IPTELERR_UNDEFINED,
      IPTELERR_ERROR_COUNT
    };

    /**
     * IptelCall::OnError(int errno): This is invoked when the 
     * system encounters an error. Default behaviour depends on
     * the error message. It is recommended that the application
     * calls the default behaviour in the overridden method.
     */
    virtual void   OnError(int errorno) = 0;

    /** Retrieve the last error, also clears it */
    virtual int  GetLastError(); /* clears the error */

    /** Set the last error variable */
    virtual void SetLastError(int errorno);

    /** Explicitly clears the last error */
    virtual void ClearError();

    /**
     * IptelCall::strerror(): This method returns a string represent-
     * ation of the error message stored in errno. This method
     * can be used typically to print the error message to
     * output stream.
     */
    static const char *strerror(int e);

    /** operator << for printing onto output stream */
    friend ostream & operator<<(ostream & o, IptelCall & addr);

  protected:
    IptelAddress        remote;
    IptelAddress        self;
    bool                isactive;
    SessionDescription *selfsession;
    SessionDescription *remotesession;
    unsigned long       bandwidth;

    /**
     * IptelCall::errno: Whenever a method returns FAILURE errno stores 
     * the reason for failure. This value is not updated
     * when the method returns SUCCESS.
     */
    int errorno;

};

//@}

#endif /* _IPTEL_API_H_ */

