/*sipapi.cpp - SIP API Implementation.
 * (c) 1999-2001. Columbia University. All rights reserved.
*/

#include <time.h>       /* time_t, ... */

#include "iptelapi.h"   /* IptelAddress,Session etc */
#include "sipapi.h"     /* SIPCall, Endpoint etc */
#include "msgflow.h"    /* For msg flow trace */
#include "sipcall.h"

extern "C" {
#include <string.h>    /* strdup */
#include <stdlib.h>    /* malloc */
#include <sys/time.h>  /* struct timeval */
#include <sys/types.h> /* uint,... */ 
#include <sys/stat.h>  /* stat,... */
#include <sys/socket.h>/* socket, ... */
#include <netinet/in.h>/* AF_INET, ... */
#include <arpa/inet.h> /* inet_addr, inet_ntoa */
#include <unistd.h>    /* gethostname */
#include <stdarg.h>    /* vsprintf, ... */
#include <pthread.h>   /* pthread_create, ... */

#include "thread_pool.h" /* InitThreadPool and DestroyThreadPool */
#include "udp.h"         /* udp_thread_arg, ... */  
#include "tcp.h"         /* tcp_thread_arg, ... */   
#include "request-sip.h" /* request_t */
#include "canonicalizer_interface.h" /* Dummy canonicalizer implementation */
#include "strdupn.h"     /* strdupn */
#include "uri.h"
#include "policy.h"
#include "pthread-util.h"
#include "config.h"      /* config_t */
#include "authenticateSIP.h"
}

/* Default listening port for SIP */
#define DEFAULT_SIP_PORTNO  5060

/* Memory management */
#define AllocateObject(type,count)    \
    (((count) == 1) ? (new type) : (new type[count]));
#define DeallocateObject(obj)  if (obj) delete (obj);

/* Indicates whether different "C" componenets of the
 * library are initialized or not 
*/
static bool toLibraryInitialized = false;

extern "C" {
  void InitLibSip();
  unsigned long random32();
  struct in_addr host2ip(char *host);
  void DebugInit(char *);
  void ErrorInit(const char *filename);
  int MessageFlowFile(FILE *fp); 
  char *base64_encode_string(char *);
  void CloseTCP();
  void CloseUDP();
}

/* Globals needed by the sip library */
extern "C" int debug_flag;
extern "C" const char *server_name;
extern  struct config_t config;
extern "C" transaction_policy_t userapi_policy;
extern "C" int global_udp_port;
extern "C" int global_tcp_port;

const char *server_name;

extern IptelAddress SipUriToIptelAddress(uri_t uri);
extern sip_addr_t   URI_ParseNew(const char *);
request_t * FormatOptions(const char *method, int type, int af, 
                          const char *callid,
                          int cseq, const IptelAddress & to,
                          const IptelAddress & from, body_t *body, 
                          const SIPEndpoint & endpoint);

void  RemoveEndpointFromAllReg(SIPEndpoint *);

int   SIPLibrary::GeneralTimeout = 30;
int   SIPLibrary::CgiTimeout     = 15;
int   SIPLibrary::TraceLevel     = 3;
FILE* SIPLibrary::TraceFile      = NULL;
int   SIPLibrary::EndpointCount  = 0;
int   SIPLibrary::CallCount      = 0;
bool  SIPLibrary::oUseNumericAddress = false;

/* addresses and names*/
char *SIPLibrary::HostName4      = NULL;
char *SIPLibrary::HostName6      = NULL;
struct in_addr  SIPLibrary::HostInAddr4;
struct in6_addr SIPLibrary::HostInAddr6;

char *SIPLibrary::UAID           = NULL;   /* dynamically generated */
SessionDescription *SIPEndpoint::default_session = NULL;


/* Should the SIPLibrary::globals be freed on cleanup ?
 * Not necessary since, they should be available till the
 * application is running. No point in freeing
 * when application terminates, since OS takes care of it.
 */
static pthread_mutex_t library_mutex; /* to lock globals */

/* SIPLibrary Initializer */
int SIPLibrary::Initialize()
{
  struct sockaddr_storage sin;

  if (toLibraryInitialized)
    return 0;

  toLibraryInitialized = true;
  /* 
   * This part should be moved to dllmain() for Windows.
   * Otherwise this should be present in SIPCall
   * constructor also.
  */
  InitLibSip();    /* Initialize libsip */
  startupSocket(); 

  /* Register the MsgFlow classes */
  MessageFlowFile(NULL); /* switch off Message Flow */
  RegisterMsgClass("SIPAPI", 3, "API", "This", "Remote"); 

  SIPLibrary::GeneralTimeout = 30;
  SIPLibrary::CgiTimeout     = 15;

  SIPLibrary::TraceLevel     = 3;
  SIPLibrary::TraceFile      = stderr;

  SIPLibrary::EndpointCount  = 0;
  SIPLibrary::CallCount      = 0;

  config.NumericVia = 
    (SIPLibrary::oUseNumericAddress) ? 1 : 0;

  if (!SIPLibrary::oUseNumericAddress) 
    SIPLibrary::HostName4 = AFIGetOurName(AF_INET);
  else
    SIPLibrary::HostName4 = AFIGetOurDottedAddress(AF_INET);
  
  if (!SIPLibrary::HostName4) {
    SIPLibrary::HostName4 = strdup("localhost");
  }

  if (AFIIPv6Active()) {
    if (!SIPLibrary::HostName6) { /* user did not give a hostname */
      if (!SIPLibrary::oUseNumericAddress) 
        SIPLibrary::HostName6 = AFIGetOurName(AF_INET6);
      else 
        SIPLibrary::HostName6 = AFIGetOurDottedAddress(AF_INET6);
      
      if (!SIPLibrary::HostName6) {
        debug("SIPLibrary::Initialize", 
              "It seems your IPv4 and v6 hostnames are different."
             " Or you may not have usable v6. Turning off v6\n");
        AFIUseIPv6(0);
      }
    } /* !HostName6 */
    else {
      /* the user specified a IP6 host name. check if it is numeric, 
       * else try to translate.
      */
      if (SIPLibrary::oUseNumericAddress) {
        int i = AFITranslateNumeric(&sin, SIPLibrary::HostName6);
        if (i != 0) { /* this is not a numeric address, translate */
          sin.ss_family = AF_INET6;
          if (AFIGetAddrByName(&sin, SIPLibrary::HostName6) == 0) {
            free(SIPLibrary::HostName6);
            SIPLibrary::HostName6 = 
                      AFIGetDottedAddress((struct sockaddr*)&sin);
          }
          else {
            debug("SIPLibrary::Initialize", "You specified a hostname %s. It"
                " is unresolvable. Turning off IPv6.\n", 
                SIPLibrary::HostName6);
            AFIUseIPv6(0);
          }
        }
      }
    } /* else !HostName6 */
  } /* AFIIPv6Active */
        
  if (!SIPLibrary::HostName6) 
    SIPLibrary::HostName6 = strdup(SIPLibrary::HostName4);

  memset(&sin, 0, sizeof(sin));
  sin.ss_family = AF_INET;
  if (AFIGetOurAddr(&sin) == 0) {
    SIPLibrary::HostInAddr4 = ((struct sockaddr_in*)&sin)->sin_addr;
  }
  else {
    SIPLibrary::HostInAddr4.s_addr = inet_addr("127.0.0.1");
  }
  
  if (AFIIPv6Active()) {
    sin.ss_family = AF_INET6;
    if (AFIGetAddrByName(&sin, SIPLibrary::HostName6) == 0) 
      SIPLibrary::HostInAddr6 = ((struct sockaddr_in6*)&sin)->sin6_addr;
  }
     
  DString uaid;
  DStringInit(&uaid);
  DStringPrintf(&uaid, "%lu", SIPLibrary::Random32());
  SIPLibrary::UAID = strdup(DStringValue(&uaid));
  DStringFree(&uaid);

  /* local C variable config */
  config.timeout      = SIPLibrary::GeneralTimeout;
  config.ProxyName    = strdup(SIPLibrary::HostName4);
  if (SIPLibrary::oUseNumericAddress) {
    char buf[512];
    sprintf(buf, "[%s]", SIPLibrary::HostName6);
    config.ProxyName6 = strdup(buf);
  }
  else 
    config.ProxyName6 = strdup(SIPLibrary::HostName6);

  config.HostName6    = strdup(SIPLibrary::HostName6);
  config.ClockSkew    = 300; /* +- seconds */
  config.NonceLifeTime = 60; /* seconds */
  config.AuthMethod   = auth_none;
  config.expires      = 3600;
  config.cgi_timeout  = SIPLibrary::CgiTimeout;
  config.port         = DEFAULT_SIP_PORTNO;  
  config.PrimaryIP    = NULL; 
  config.PrimaryIP6   = NULL; 
  config.ThreadPoolSize = 10; 
  config.UseThreadPool = 1;   
  InitThreadPool(config.ThreadPoolSize);

  config.ServerName = strdup(SIPLibrary::HostName4);

  if (strstr(SIPLibrary::HostName4, "cs.columbia.edu")) {
    config.ServerName   = strdup("cs.columbia.edu");
  }
  pthread_mutex_init(&library_mutex, NULL); /* lock initialized */
  
  return 0;
}

/* Error reporting and debugging functions */
void SIPLibrary::DebugOn(const char *filename)
{
  debug_flag = 1;
  DebugInit((char *)filename);
}

void SIPLibrary::DebugOn()
{
  debug_flag = 1;
}

void SIPLibrary::DebugOff()
{
  debug_flag = 0;
}

void SIPLibrary::ErrorOn(const char *filename)
{
  ErrorInit((char *)filename);
}

void SIPLibrary::MessageFlowOn(FILE *fp)
{
  MessageFlowFile(fp);
}

void SIPLibrary::Debug(char *functionname, char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  debug (functionname, format, ap);
  va_end(ap);
}

void SIPLibrary::Error(char *functionname, char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  error (functionname, format, ap);
  va_end(ap);
}

void SIPLibrary::Fatal(char *functionname, char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  fatal (functionname, format, ap);
  va_end(ap);
}

void SIPLibrary::SetName(const char *name)
{
  server_name = name;
}

unsigned long SIPLibrary::Random32()
{
  return (unsigned long) random32();
}

void SIPEndpoint::Startup() {

  int e;
  struct tcp_thread_arg *t_arg = NULL;
  struct udp_thread_arg *u_arg = NULL;

  /* Do one time initialization */
  SIPLibrary::Initialize();

#ifdef MEDIACONTROL_SIPCOMM
  initMediaControl(NULL);
#endif

  userid = NULL;
  passwd = NULL;
  
  pthread_mutex_init(&mutex, NULL);

  unsigned short port = DEFAULT_SIP_PORTNO; /* Default SIP listening port */
   
  struct in_addr  primary_addr = (in_addr)ListeningAddress;
  struct in6_addr primary_addr6 = (in6_addr)ListeningAddress6;

  if (primary_addr.s_addr != INADDR_ANY &&
      primary_addr.s_addr != (unsigned) -1) {
    char* primaryIP = inet_ntoa(primary_addr);
    config.ProxyName = strdup(primaryIP);  /* PrimaryIP overrides proxy name */
    config.PrimaryIP = strdup(primaryIP);
    debug("SIPEndPoint", "Using %s as the default address\n", primaryIP);
  }
  else { /* INADDR_ANY is for binding. set the ip to our interface ip */
    ListeningAddress = SIPLibrary::HostInAddr4;
  }
  
  if (!IN6_IS_ADDR_UNSPECIFIED(&primary_addr6)) {
    char buf[256];

    if (inet_ntop(AF_INET6, &primary_addr6, buf, sizeof(buf)) != NULL) {
      char proxyName[256];
      
      config.PrimaryIP6 = strdup(buf);

      sprintf(proxyName, "[%s]", buf); /* ProxyName for v6 be in [] */
      
      config.ProxyName6 = strdup(proxyName);   
      
     debug("SIPEndPoint", "Using %s as the default v6 address\n",
           config.PrimaryIP6);
    }
  }
  else {
    ListeningAddress6 = SIPLibrary::HostInAddr6;
  }

  if (ListeningAddress.GetPort() > 1) {
    port = ListeningAddress.GetPort();
    debug("SipEndPoint","Listening on %d\n",port);
  }

  t_arg = (struct tcp_thread_arg*) calloc(1, sizeof(tcp_thread_arg));
  u_arg = (struct udp_thread_arg*) calloc(1, sizeof(udp_thread_arg));
  
  if (t_arg && u_arg) {
    if (config.PrimaryIP) {
      strassign(&t_arg->primaryIP, config.PrimaryIP);
      strassign(&u_arg->primaryIP, config.PrimaryIP);
    }

    if (config.PrimaryIP6) {
      strassign(&t_arg->primaryIP6, config.PrimaryIP6);
      strassign(&u_arg->primaryIP6, config.PrimaryIP6);
    }

    u_arg->port = t_arg->port = ListeningAddress.GetPort();
    if (t_arg->port == 0) {
      t_arg->port = u_arg->port = port;
    }
    else if (t_arg->port == 1) {
      t_arg->port = 0; /* Dynamic port */
    }
    else {
      port = t_arg->port;
    }

    t_arg->server = this;
    u_arg->server = this;
    
    if ((e = pthread_create_detached(&tid[1], NULL, ReceiveTCP, 
                                    (void *)t_arg)) != 0) {
      error("main", "Cannot create TCP thread\n");
      free_valid(t_arg);
      t_arg = NULL;
    }
  
    if ((e = pthread_create_detached(&tid[0], NULL, 
                                     ReceiveUDP, (void *)u_arg)) != 0) {
      error("main", "Cannot create UDP thread");
      free_valid(u_arg);
      u_arg = NULL;
    }
  } /* t_arg && u_arg */
  else {
    error("main", "memory allocation fails\n");
    u_arg = NULL;
    t_arg = NULL;
  }

  if (ListeningAddress.GetPort() <= 1) { /* get the ports */
    int retry = 1;

    /* Sleep for some time to wait for the
     * TCP/UDP thread to allocate the port 
    */
    while (u_arg != NULL && u_arg->port == 0 && retry <= 100) {
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 100000; /* 100 millisec */
      debug("main", "waiting for udp port (%hu)\n", u_arg->port);
      select(0, NULL, NULL, NULL, &tv);
      retry++;
    }
    debug("main", "port is set to %hu\n", u_arg == NULL ? 0 :
          u_arg->port);
    port = u_arg->port;
    ListeningAddress.SetPort(port);
  }

  config.port = port; /* What about the TCP port XXX */
  lastcseq = 0;

  /* Update statistics */
  pthread_mutex_lock(&library_mutex);
  SIPLibrary::EndpointCount ++;
  pthread_mutex_unlock(&library_mutex);

  return;
}

/* SIP Endpoint Constructor */
SIPEndpoint::SIPEndpoint(const IptelIPAddress& listeningAddress, 
                         bool IsRegistrar)
                        : IptelEndpoint(listeningAddress, IsRegistrar)
{
  Startup();  
}

SIPEndpoint::SIPEndpoint(const IptelIPAddress& listeningAddress4, 
                         const IptelIPAddress& listeningAddress6,
                         bool  IsRegistrar)
                         : IptelEndpoint(listeningAddress4, 
                                         IsRegistrar,
                                         listeningAddress6)
{
  Startup();
}
  
    
SIPEndpoint::SIPEndpoint(bool IsServer, bool IsRegistrar)
                        :IptelEndpoint(IsServer,IsRegistrar)
{
  Startup();
}

SIPEndpoint::~SIPEndpoint()
{
  /* Wait for thread completion. */

  /* XXX Send a signal to terminate these threads.
   * This is a dirty hack to not wait for the threads to
   * terminate. */
  //WaitForMultipleObjects(2, tid, TRUE, INFINITE);

  pthread_cancel(tid[0]);
  pthread_cancel(tid[1]);

  CloseUDP();
  CloseTCP();

  DestroyThreadPool();  
  KillLibSipThreads();

  /* Remove the registration ifnormation from this endpoint */
  RemoveEndpointFromAllReg(this);

  free_valid(userid);
  free_valid(passwd);


  /* Update statistics */
  pthread_mutex_lock(&library_mutex);
  SIPLibrary::EndpointCount --;
  pthread_mutex_unlock(&library_mutex);

  pthread_mutex_destroy(&mutex);

#ifdef MEDIACONTROL_SIPCOMM
  shutdownMediaControl();
#endif
}

void SIPEndpoint::Lock()
{
  debug("SIPEndpoint::Lock","entering\n");
  pthread_mutex_lock(&mutex);
  debug("SIPEndpoint::Lock","exiting\n");
}

void SIPEndpoint::Unlock()
{
  debug("SIPEndpoint::Unlock","entering\n");
  pthread_mutex_unlock(&mutex);
  debug("SIPEndpoint::Unlock","exiting\n");
}

/* Allocate a new SIPCall object - default behaviour */
IptelCall *SIPEndpoint::AllocateNewCallObject(const char *callid)
{
  return new SIPCall(*this, callid);
}

/* When the server receives a REGISTER from remote */
int SIPEndpoint::OnIncomingRegistration(
          const SIPRegistrationInfo & reginfo)
{
  /* Do nothing just accept */
  /*
   * XXX add the code to append the new registration
   * information in the local structure if maintained.
   * currently we do not maintain any local list of
   * incoming registrations.
   */
  return 0;
}

SessionDescription *SIPEndpoint::GetRemoteSessionDescription(
             const IptelAddress & to,
             const IptelAddress & from)
{
  /* 
   * XXX provide another parameter for blocking/non-blocking mode.
   */
  Lock();

  int af;
  if (to.IsIp6Address()) 
    af = AF_INET6;
  else
    af = AF_INET;

  request_t *r = NULL;

  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","Options");
  });

  /* Build a default body, or from the params */
  body_t body;
  memset(&body, 0, sizeof(body));

  SessionDescription *session = 
    SIPEndpoint::GetDefaultSessionDescription();
  char *self_sdp = NULL;
  if (session) {
    self_sdp = SIPCall::FormatSDP(*session);
  }

  /* Check for Accept header if present and send accordingly */
  if (self_sdp) {
    body.content = self_sdp;
    body.content_type = "application/sdp";
    body.content_length = strlen(self_sdp);
  }

  /* Free mem for sdp */
  if (session != NULL) {
    delete session;
  }
  /* Do not free self_sdp here */

  DString callid;
  DStringInit(&callid);
  DStringPrintf(&callid, "%lu@%s", 
                SIPLibrary::Random32(), SIPLibrary::HostName4); 

  r = FormatOptions("OPTIONS", 
    SOCK_DGRAM, /* XXX should allow TCP based */
    af,
    DStringValue(&callid), ++lastcseq, 
    to, from,
    body.content ? &body : NULL,
    *this);

  DStringFree(&callid);

  if (r != NULL) {
    /* Create the request policy thread */
    uri_t proxy;
    memset(&proxy, 0, sizeof(proxy));
    /* Set the URI parameter to appropriate TCP/TCP */
    if (CallCreatePolicyThread(r, r->p.URI, proxy) != 0) {
      pthread_mutex_unlock(&r->mutex);
      RequestFree(r);
      r = NULL;
    }
    URI_Free(&proxy);
  }

  free_valid(body.content);
  
  Unlock();

  /* 
  * Should wait here for the respnse and then return the SDP
  */
  return NULL;
}

/***********************************************************************
******************* SIPCall Object Definition **************************
*/

/* SIPCall Constructors */
SIPCall::SIPCall(const char *mycallid) 
       :IptelCall()
#ifdef MEDIACONTROL_SIPCOMM
, m_pMSInfo (NULL)
#endif
{
  isoriginator = false;

  endpoint = NULL;
  strcpy(ErrorString,"");
  LastStatus = 0;
  response_initialized = 0;
  response_received = 0;  /* 0 means NO and 1 means yes */
  mark_done = 0;
  userid = NULL;
  passwd = NULL;

  pthread_mutex_lock(&library_mutex);
  SIPLibrary::CallCount ++;
  pthread_mutex_unlock(&library_mutex);

  if (mycallid == NULL) {
    /* Create the call thread and call object */
    sipcall_t *call = CallCreate(NULL /* this will allocate a call id */);
    if (call == NULL) {
       return;
    }
    else {
      /* Store the call id */
      //pthread_mutex_lock(&call->mutex);
      callid = strdup(call->callid);
      call->callobj = this;
      pthread_mutex_unlock(&call->mutex);
    }
  }
  else {
    /* Callid is already present, means the
     * call thread is active. Just create the
     * callid element and copy it */

    callid = strdup(mycallid);

    /*
    * !!! do not call CallSearch(callid),
    * because that will result in a deadlock.
    * Let some other entity set the call->callobj 
    * as needed. 
    */
  }
}

SIPCall::SIPCall(SIPEndpoint & ep, const char *mycallid)
#ifdef MEDIA_CONTROL_SIPCOMM
  : m_pMSInfo(NULL)
#endif
{
  /* This part is copied from SIPCall(const char *); */
  isoriginator = false;

  strcpy(ErrorString,"");
  LastStatus = 0;
  response_initialized = 0;
  response_received = 0;  /* 0 means NO and 1 means yes */
  mark_done = 0;
  userid = NULL;
  passwd = NULL;

  endpoint = &ep;

  pthread_mutex_lock(&library_mutex);
  SIPLibrary::CallCount ++;
  pthread_mutex_unlock(&library_mutex);

  if (mycallid == NULL) {
    /* Create the call thread and call object */
    sipcall_t *call = CallCreate(NULL /* this will allocate a call id */);
    if (call == NULL) {
       return;
    }
    else {
      /* Store the call id */
      //pthread_mutex_lock(&call->mutex);
      callid = strdup(call->callid);
      call->callobj = this;
      pthread_mutex_unlock(&call->mutex);
    }
  }
  else {
    /* Callid is already present, means the
     * call thread is active. Just create the
     * callid element and copy it */

    callid = strdup(mycallid);

    /*
    * !!! do not call CallSearch(callid),
    * because that will result in a deadlock.
    * Let some other entity set the call->callobj 
    * as needed. 
    */
  }
}

/* Destructor */
SIPCall::~SIPCall()
{
  debug("~SIPCall", "destructor called\n");
  sipcall_t *call;
  if (mark_done == 0) {
    /* Inform the call thread that we are dying */
    if ((call = CallSearch(callid)) == NULL) {
      error("SIPCall::~SIPCall", "No call with callid=%s\n", callid ?
            callid : "NULL");
      mark_done = 1;
    }
    else {
      //CallDone(call);
      call->callobj = NULL;
      pthread_cond_broadcast(&call->cond); /* signal the condition */
      pthread_mutex_unlock(&call->mutex); /* release the call mutex */
    }
  }

#if 0
  while (!mark_done) {
    /* There should be some other way to wait XXX ? */
    struct timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 50000; /* wait for 50 millisec */
    debug("~SIPCall", "Waiting for mark_done\n");
    select(0,NULL,NULL,NULL, &tv);
  }
#endif

#ifdef MEDIACONTROL_SIPCOMM
  m_pMSInfo = NULL;
#endif

  free_valid(userid);
  free_valid(passwd);

  pthread_mutex_lock(&library_mutex);
  SIPLibrary::CallCount --;
  pthread_mutex_unlock(&library_mutex);

  if (callid != NULL) {
    free(callid);
  }

  /*
  if (selfsession != NULL) {
    delete selfsession;
  }

  if (remotesession != NULL) {
    delete remotesession;
  }
  */

  debug("~SIPCall", "destructor exiting\n");
}

/* 
 * Initiate an outgoing call. A dummy SIP INVITE 
 * request is constructed using the parameters
 * available. The HTTP_Parser is invoked to
 * parse it and update all the relevant fields
 * of request structure. This technique helps in
 * future upgradation of HTTP_,Request related
 * modules without affecting this. 
 */
int SIPCall::Initiate(const IptelAddress & to, const IptelAddress & from)
{
  SetSelfAddress(from);
  SetRemoteAddress(to);
  
  return Initiate(); /* call Initiate to initate thecall */
}

int SIPCall::Initiate()
{
  int oRet = 0;

  isoriginator = true; /* outgoing request */

  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","Initiate");
  });

  if (selfsession == NULL) {
    debug("Initiate", "No SDP, using default\n");

    /* Set the session description to default if absent */
#ifdef MEDIACONTROL_SIPCOMM
    SetUpSelfSessionDescription();
#else
    SetSelfSessionDescription(SIPEndpoint::GetDefaultSessionDescription());
#endif
  }

  int af = AF_INET;
  if (this->GetRemoteAddress().IsIp6Address() ||  
      this->GetSelfAddress().IsIp6Address()) 
    af = AF_INET6;

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::Initiate", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    oRet = -1;
  }
  else {

    call->af = af;

    MessageEnqueueApiRequest(&call->mqueue, "INVITE", nobody_handled);
    pthread_cond_broadcast(&call->cond);
    pthread_mutex_unlock(&call->mutex);
  }
  return oRet;
}

int   SIPCall::Hangup(int reason,const char *strReason)
{
  /* reason and strReason are ignored, XXX ? */

  int oRet = 0;

  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","Hangup");
  });
   
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::Hangup", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    oRet = -1;
  }
  else {
    /* Even CANCEL is handled depending on thecall state by the
    * call thread. We need to send just BYE */
    MessageEnqueueApiRequest(&call->mqueue, "BYE", nobody_handled);
    pthread_cond_broadcast(&call->cond);
    pthread_mutex_unlock(&call->mutex);
  }
#ifdef MEDIACONTROL_SIPCOMM
  CloseMediaStream();
  if (m_pMSInfo != NULL) {
    terminateMediaSession(m_pMSInfo->id);
    m_pMSInfo = NULL;
  }
#endif
  return oRet;
}

int    SIPCall::Alert()
{

  int oRet = 0;

  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","Alert");
  });
   
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::Alert", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    oRet = -1;
  }
  else {
    MessageEnqueueApiResponse(&call->mqueue, 180, "Alerting", nobody_handled);
    pthread_cond_broadcast(&call->cond);
    pthread_mutex_unlock(&call->mutex);
  }
  return oRet;
}

int    SIPCall::Reinvite()
{

  int oRet = 0;

  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","reinvite");
  });

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::Reinvite", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    oRet = -1;
  }
  else {
    MessageEnqueueApiRequest(&call->mqueue, "INVITE", nobody_handled);
    pthread_cond_broadcast(&call->cond);
    pthread_mutex_unlock(&call->mutex);
  }
  return oRet;
}

int    SIPCall::Accept()
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","Accept");
  });

  int oRet = 0;

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::Accept", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    oRet = -1;
  }
  else {
    MessageEnqueueApiResponse(&call->mqueue, 200, NULL, nobody_handled);
    pthread_cond_broadcast(&call->cond);
    pthread_mutex_unlock(&call->mutex);
  }

#ifdef MEDIACONTROL_SIPCOMM
  OpenMediaStream();
#endif
  return oRet;
}

int   SIPCall::Reject()
{
  return Reject(0, NULL);
}

int    SIPCall::Reject(int status, const char *reason)
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","Accept");
  });

  int oRet = 0;

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::Reject", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    oRet = -1;
  }
  else {
    /* If status is 0 then alredy set status is used  XXX */
    if (status <= 0) {
      status = 603; /* User decline */
    }
    
    MessageEnqueueApiResponse(&call->mqueue, status, reason, nobody_handled);
    pthread_cond_broadcast(&call->cond);
    pthread_mutex_unlock(&call->mutex);
  }
  return oRet;
}

int SIPCall::Redirect(const IptelAddress newaddress,
                      const char *reason)
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","Redirect");
  });

  int oRet = 0;

  /* Set the Contact address */
  SetOutgoingResponseHeader("Contact", (const char *) newaddress);

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::Redirect", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    oRet = -1;
  }
  else {
    /* Set the Contact address for the next response */
    MessageEnqueueApiResponse(&call->mqueue, 
                              302 /* moved temporarily */, 
                              reason,
                              nobody_handled);
    pthread_cond_broadcast(&call->cond);
    pthread_mutex_unlock(&call->mutex);
  }
  return oRet;
}

/* Authenticate SIP call INVITE,REGISTER etc... */

bool  SIPCall::Authenticate(const char *method, /* basic digest,...*/
                            const char *userfile)
{
  bool oRet = true;

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::Redirect", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    oRet = false;
  }
  else {
    oRet = (CallAuthenticate(call, method, userfile) == 0);
    pthread_mutex_unlock(&call->mutex);
  }

  return oRet;
}

void   SIPCall::OnNewIncomingCall(IptelEndpoint & ep)
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","This","API","OnNewIncominCall");
  });

#ifdef MEDIACONTROL_SIPCOMM 
  SetUpSelfSessionDescription();
#endif
  isoriginator = false;
  return;
}

void   SIPCall::OnAlerting()
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","This", "API","OnAlerting");
  });

  return;
}

int    SIPCall::WaitForResponse(int timeout)
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","This", "API","WaitForResponse");
  });
  
  if (response_initialized == 0)
  {
    pthread_cond_init(&response_cond,NULL);
    pthread_mutex_init(&response_mutex,NULL);
    response_initialized = 1;
  }

  pthread_mutex_lock(&response_mutex);
  while (response_received == 0) {
    pthread_cond_wait(&response_cond,&response_mutex);
  }
  pthread_mutex_unlock(&response_mutex);

  return 0;
}

void   SIPCall::OnCallEstablished()
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","This", "API","OnCallEstablished");
  });

  isactive = true;
  if (response_initialized)
  {
    pthread_mutex_lock(&response_mutex);
    response_received = 1;
    pthread_cond_signal(&response_cond);
    pthread_mutex_unlock(&response_mutex);
  }

#ifdef MEDIACONTROL_SIPCOMM
  OpenMediaStream();
#endif
  return;
}

int   SIPCall::OnCallRejected()
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","This", "API","OnCallRejected");
  });

  isactive = false;
  if (response_initialized)
  {
    pthread_mutex_lock(&response_mutex);
    response_received = 1;
    pthread_cond_signal(&response_cond);
    pthread_mutex_unlock(&response_mutex);
  }
  /* return 0, so that this object is deleted */
  return 0;
}

int   SIPCall::OnHangup()
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","This", "API","OnHangup");
  });

  isactive = false;
  if (response_initialized)
  {
    pthread_mutex_lock(&response_mutex);
    response_received = 1;
    pthread_cond_signal(&response_cond);
    pthread_mutex_unlock(&response_mutex);
  }

/* return 0 so that this object is deleted. */
#ifdef MEDIACONTROL_SIPCOMM
  CloseMediaStream();
#endif
  return 0;
} 

int    SIPCall::WaitForTermination(int timeout)
{
  DRAW_MESSAGE_FLOW({
    DrawMessageFlow("SIPAPI","API","This","WaitForTermination");
  });
  
  /*
   * XXX use logic similar to WaitForResponse() to implement this.
   */
  return 0;
}

int SIPCall::SetProxyAddress(const IptelAddress & proxy)
{
  /* Set the outbound proxy address, IP or domain name XXX */
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetProxyAddress", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    return -1;
  }
  else {
    char *uri = proxy.GetURI();
    if (uri != NULL && *uri) {
      sip_addr_t addr = URI_ParseNew(uri);
      call->proxy = addr.uri;
      free_valid(addr.display_name);
      strassign(&call->proxy.scheme, "sip"); /* only sip: is supported */
    }
    else {
      URI_Free(&call->proxy);
      memset(&call->proxy, 0, sizeof(call->proxy));
    }
    free_valid(uri);
    pthread_mutex_unlock(&call->mutex);
    return 0;
  }
}

IptelAddress SIPCall::GetProxyAddress()
{
  sipcall_t *call;
  IptelAddress ipdummy;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetProxyAddress", 
          "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
      if (call->proxy.host) {
        IptelAddress ip(call->proxy.host);
        ip.SetScheme(call->proxy.scheme);
        ip.SetPort(call->proxy.port);
        pthread_mutex_unlock(&call->mutex);
        return ip;
      }
      pthread_mutex_unlock(&call->mutex);
  }
  return ipdummy;
}

void   SIPCall::OnError(int errorno)
{
  /* default implementation does nothing */
  return;
}


/*
* Generic Setter and getters
*/
int SIPCall::SetRemoteAddress(const IptelAddress & r)
{
  IptelCall::SetRemoteAddress(r);

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetRemoteAddress", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    return -1;
  }
  else {
    strassign(&call->remote, (const char *) r);
    pthread_mutex_unlock(&call->mutex);
    return 0;
  }
}

IptelAddress SIPCall::GetRemoteAddress()
{
  IptelAddress address;
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetRemoteAddress", 
          "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    address = call->remote;
    pthread_mutex_unlock(&call->mutex);
  }
  remote = address;
  return address;
}

int SIPCall::SetSelfAddress(const IptelAddress & r)
{
  IptelCall::SetSelfAddress(r);

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetSelfAddress", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    return -1;
  }
  else {
    strassign(&call->self, (const char *) r);
    pthread_mutex_unlock(&call->mutex);
    return 0;
  }
}

IptelAddress SIPCall::GetSelfAddress()
{
  IptelAddress address;
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetSelfAddress", 
          "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    address = call->self;
    pthread_mutex_unlock(&call->mutex);
  }

  self = address;
  return address;
}

int SIPCall::SetSelfSessionDescription(SessionDescription *session)
{
  IptelCall::SetSelfSessionDescription(session);

  char *sdp = NULL;

  if (session == NULL) {
    debug("SetSelfSessionDescription", "session is NULL, removing older\n");
    return -1;
  }
  else {
    sdp = SIPCall::FormatSDP(*session, NULL);
    if (sdp == NULL) {
      debug("SetSelfSessionDescription", "FormatSDP() failed\n");
      return -1;
    }
  }

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetSelfSessionDescription", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    free_valid(sdp);
    return -1;
  }
  else {
    /*
     * Should we store the self_sdp if this is a re-INVITE.
     * Because is re-INVITE gets rejected then we should be able to use
     * the original SDP.
     * I assume that the application keeps track of original SDP
     * So the library need not worry about it.
     */
    free_valid(call->self_sdp);
    call->self_sdp = sdp;

    /* Should we send re-INVITE in any state other than
     * IDLE or TERMINATING XXX. I think that will complicate
     * the logic and state transition. */
    if (call->state == SIPCALL_ACTIVE) {
      /* We should trigger a re-invite */
      /* There is potential problem here.
       * The application may call this method
       * in response to OnRemoteSessionChanged() callback,
       * which in turn happened due to the incoming re-INVITE.
       * We would trigger another re-INVITE. To avoid this,
       * we must check if an incoming re-INVITE is pending,
       * then do not send message.
       */
      if (!call->incoming_reinvite) {
        MessageEnqueueApiRequest(&call->mqueue, "INVITE",
                                 nobody_handled);
        pthread_cond_broadcast(&call->cond);
      }
    }

    pthread_mutex_unlock(&call->mutex);
    return 0;
  }
}

SessionDescription *SIPCall::GetSelfSessionDescription()
{
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetSelfSessionDescription", 
            "No call with callid=%s\n", callid ? callid : "NULL");
    return NULL;
  }
  else {
    SessionDescription * session = SIPCall::DeformatSDP(call->self_sdp, -1);
    IptelCall::SetSelfSessionDescription(session);
    pthread_mutex_unlock(&call->mutex);
    return selfsession;
  }
}

int SIPCall::SetRemoteSessionDescription(SessionDescription *session)
{
  debug("SetRemoteSessionDescription", "This should not be called\n");

  IptelCall::SetRemoteSessionDescription(session);

  char *sdp = NULL;

  if (session == NULL) {
    debug("SetRemoteSessionDescription", "session is NULL, removing older\n");
  }
  else {
    sdp = SIPCall::FormatSDP(*session, NULL);
    if (sdp == NULL) {
      debug("SetRemoteSessionDescription", "FormatSDP() failed\n");
      return -1;
    }
  }

  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetRemoteSessionDescription", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    free_valid(sdp);
    return -1;
  }
  else {
    free_valid(call->remote_sdp);
    call->remote_sdp = sdp;
    pthread_mutex_unlock(&call->mutex);
    return 0;
  }
}

SessionDescription *SIPCall::GetRemoteSessionDescription()
{
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetRemoteSessionDescription", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    return NULL;
  }
  else {
    SessionDescription *session = SIPCall::DeformatSDP(call->remote_sdp, -1);
    IptelCall::SetRemoteSessionDescription(session);
    pthread_mutex_unlock(&call->mutex);
    return remotesession;
  }
}

void SIPCall::SetStatus(int status, const char *reason)
{
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetStatus", 
            "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    call->p.status = status;
    strassign(&call->p.reason, reason);
    pthread_mutex_unlock(&call->mutex);
  }

  if (reason != NULL) {
    strncpy(ErrorString,reason,sizeof(ErrorString)-1);
  }
  LastStatus = status;
}

int SIPCall::GetStatus()
{
  /*
   * Define XXX other methods for acesssing the received or
   * sent status.
   */
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetStatus", 
          "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    /* XXX use p.status if p_irs.status is invalid XXX */
    LastStatus = call->p_irs.status;
    if (call->p.reason != NULL) {
      strncpy(ErrorString,call->p_irs.reason,sizeof(ErrorString)-1);
    }
    else {
      strncpy(ErrorString, GetStandardReason(call->p_irs.status),
        sizeof(ErrorString)-1);
    }
    pthread_mutex_unlock(&call->mutex);
  }

  return LastStatus;
}

const char *SIPCall::GetStatusString()
{
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetStatusString", 
          "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    /* XXX use p.status if p_irs.status is invalid XXX */
    LastStatus = call->p_irs.status;
    if (call->p.reason != NULL) {
      strncpy(ErrorString,call->p_irs.reason,sizeof(ErrorString)-1);
    }
    else {
      strncpy(ErrorString, GetStandardReason(call->p_irs.status),
        sizeof(ErrorString)-1);
    }
    pthread_mutex_unlock(&call->mutex);
  }

  return ErrorString;
}

const char *SIPCall::GetHeader(const char *hdr)
{
  char *v = (char *) GetIncomingRequestHeader(hdr);
  if (v == NULL) {
    v = (char *) GetOutgoingRequestHeader(hdr);
  }
  return v;
}

const char * SIPCall::GetParameter(const char * headerName, 
                                 const char * paramName)
{
  /* XXX implement this */
  debug("GetParamter()", "is not implemented\n");
  return NULL;
}

void SIPCall::SetContentInfo(const char * contentType,
                             const char * contentBody)
{
  /* XXX implement this */
  debug("SetContentInfo()", "is not implemented\n");
  /* Nothing is done now. */
}

int  SIPCall::SetRequestURI(const IptelAddress & address)
{
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetRequestURI", 
          "No call with callid=%s\n", callid ? callid : "NULL");
    return -1;
  }
  else {
    char *str = strdup((const char *) address);
    sip_addr_t uri;
    uri = URI_ParseNew(str);
    URI_Free(&call->p.URI);
    call->p.URI = uri.uri;
    /* Do not free uri */
    free_valid(uri.display_name);
    free_valid(str);
    pthread_mutex_unlock(&call->mutex);
  }
  return 0;
}

IptelAddress     SIPCall::GetRequestURI()
{
  IptelAddress address;
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetRequestURI", 
          "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    address = SipUriToIptelAddress(call->p.URI);
    pthread_mutex_unlock(&call->mutex);
  }
  return address;
}

void SIPCall::SetTransportType(int type)
{
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetTransportType", 
          "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    call->type = type;
    pthread_mutex_unlock(&call->mutex);
  }
}

void SIPCall::SetAddressFamily(int af)
{
  sipcall_t *call;
  if (af != AF_INET && af != AF_INET6) {
    error("SIPCall::SetAddressFamily", "Invalid address family\n");
  }

  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetAddressFamily", 
          "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    call->af = af;
    pthread_mutex_unlock(&call->mutex);
  }
}

char * SIPEndpoint::BuildAuthorizationHeader(const char *user,
                                             const char *passwd)
{
  char *ret = NULL;
  DString h, v;
  DStringInit(&h);
  DStringInit(&v);

  DStringPrintf(&h, "%s:%s", user, passwd);
  char *base64_user_pass = base64_encode_string(DStringValue(&h));
  if (base64_user_pass != NULL) {
    DStringPrintf(&v, "Basic %s", base64_user_pass);
    free_valid(base64_user_pass);
    ret = strdup(DStringValue(&v)); 
  }

  DStringFree(&h);
  DStringFree(&v);
  return ret;
}

void SIPCall::SetAuthorization(const char *user, const char *pass)
{
  strassign(&userid, user);
  strassign(&passwd, pass);
}

/* Endpoint authorization method is also here 
* for easy reference. Infact the definition is
* almost the same for the two methods.
* We should combine them in future. XXX. */
void SIPEndpoint::SetAuthorization(const char *user, const char *pass)
{
  strassign(&userid, user);
  strassign(&passwd, pass);
}

void SIPEndpoint::SetOutgoingRequestHeader(
              const char *h, const char *v, bool append)
{  
  if (!h) {
    return;
  }

  Lock();

  if (!append) {
    h_orq -= (const char *) h;
  }

  if (v != NULL) {
    h_orq += GenericHeaderValue(h, v);
  }
  /* If v is null, just remove that */

  Unlock();
}

void SIPCall::SetOutgoingRequestHeader(const char *h, const char *v, 
                                       bool append)
{  
  sipcall_t *call;

  if (!h) {
    return;
  }

  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetOutgoingRequestHeader", 
      "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    if (v == NULL) {
      /* remove the headers */
      HeaderDelete(&call->h_orq, (char *) h);
    }
    else if (append == false) {
      /* replace the header if already exists */
      HeaderReplaceOrAppend(&call->h_orq, (char *) h, (char *) v); 
    }
    else {
      /* just append the header */
      HeaderAppend(&call->h_orq, (char *) h, (char *) v);
    }
    pthread_mutex_unlock(&call->mutex);
  }
}

void SIPCall::SetOutgoingResponseHeader(const char *h, const char *v, 
                                        bool append)
{  
  sipcall_t *call;

  if (!h) {
    return;
  }

  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetOutgoingResponseHeader", 
      "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    if (v == NULL) {
      /* remove the headers */
      HeaderDelete(&call->h_ors, (char *) h);
    }
    else if (append == false) {
      /* replace the header if already exists */
      HeaderReplaceOrAppend(&call->h_ors, (char *) h, (char *) v); 
    }
    else {
      /* just append the header */
      HeaderAppend(&call->h_ors, (char *) h, (char *) v);
    }
    pthread_mutex_unlock(&call->mutex);
  }
}

const char *SIPCall::GetOutgoingRequestHeader(const char *h)
{
  sipcall_t *call;
  if (!h) {
    return NULL;
  }
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetOutgoingRequestHeader", 
      "No call with callid=%s\n", callid ? callid : "NULL");
    return NULL;
  }
  else {
    /* just append the header */
    void *hp = NULL;
    char *v = (char *) HeaderGet(&call->h_orq, (char *) h, 1, &hp);
    /*
     * XXX potential problem here, because the returned pointer
     * may be used when it is already deallocated.
     * The problem is there for other Get**Header() methods also.
     */
    pthread_mutex_unlock(&call->mutex);
    return v;
  }  
}

const char *SIPCall::GetOutgoingResponseHeader(const char *h)
{
  sipcall_t *call;
  if (!h) {
    return NULL;
  }
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetOutgoingResponseHeader", 
      "No call with callid=%s\n", callid ? callid : "NULL");
    return NULL;
  }
  else {
    /* just append the header */
    void *hp = NULL;
    char *v = (char *) HeaderGet(&call->h_ors, (char *) h, 1, &hp);
    pthread_mutex_unlock(&call->mutex);
    return v;
  }  
}

const char *SIPCall::GetIncomingRequestHeader(const char *h)
{
  sipcall_t *call;

  if (h==NULL) {
    return NULL;
  }

  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetIncomingRequestHeader", 
      "No call with callid=%s\n", callid ? callid : "NULL");
    return NULL;
  }
  else {
    /* just append the header */
    char *v = NULL;
    parser_header_t *H;  /* request header */
    if ((H = RFC822_Header(call->p_irq.headers, (char *) h)) && H->elements) {
      v = H->elements->element;
    }
    else {
      debug("GetIncomingRequestHeader", "Header %p does not exist\n", h);
    }
    pthread_mutex_unlock(&call->mutex);
    return v;
  }  
}

const char *SIPCall::GetIncomingResponseHeader(const char *h)
{
  sipcall_t *call;

  if (h==NULL) {
    return NULL;
  }

  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::GetIncomingResponseHeader", 
      "No call with callid=%s\n", callid ? callid : "NULL");
    return NULL;
  }
  else {
    /* just append the header */
    char *v = NULL;
    parser_header_t *H;  /* response header */
    if ((H = RFC822_Header(call->p_irs.headers, (char *) h)) && H->elements) {
      v = H->elements->element;
    }
    else {
      debug("GetIncomingResponseHeader", "Header %p does not exist\n", h);
    }
    pthread_mutex_unlock(&call->mutex);
    return v;
  }  
}

/************************ DEFINITION OF SIPMessage *************************
*/

#if 0

SIPMessage::SIPMessage(const SIPMessage & msg)  
{ 
  request = msg.request; 
  call = msg.call;
  o_outgoing=msg.o_outgoing; 
  remoteIpAddress = msg.remoteIpAddress;
  request_uri = msg.request_uri; 
  strncpy(method, msg.method, 20);
  strncpy(reason, msg.reason, 128); 
  status = msg.status;
}

SIPMessage & SIPMessage::operator= (const SIPMessage & msg)  
{ 
  request = msg.request; 
  call = msg.call;
  o_outgoing=msg.o_outgoing; 
  remoteIpAddress = msg.remoteIpAddress;
  request_uri = msg.request_uri; 
  strncpy(method, msg.method, 20);
  strncpy(reason, msg.reason, 128); 
  status = msg.status;
  return *this;
}
#endif

/************************** SIP Message Body ******************************
*/
SIPMessageBody::SIPMessageBody()
{
  content = NULL;
  content_length = 0;
  content_type = strdup("text/plain");
  content_disposition = strdup("render");
  content_handling = strdup("optional");
}

SIPMessageBody::SIPMessageBody(const char *body,
               const char *type,
               int   length,
               const char *disposition,
               const char *handling)
{
  content = (body != NULL) ? strdup(body) : NULL;
  content_length = length;
  content_type = strdup(type);
  content_disposition = strdup(disposition);
  content_handling = strdup(handling);
}

SIPMessageBody::SIPMessageBody(const SIPMessageBody & body)
{
  content = (body.content != NULL) ? strdup(body.content) : NULL;

  content_length = body.content_length;

  content_type = (body.content_type != NULL) ? 
                   strdup(body.content_type) : NULL;

  content_disposition = (body.content_disposition != NULL) ? 
                   strdup(body.content_disposition) : NULL;

  content_handling = (body.content_handling != NULL) ? 
                    strdup(body.content_handling) : NULL;
}

SIPMessageBody::~SIPMessageBody()
{
  free_valid(content);
  free_valid(content_type);
  free_valid(content_disposition);
  free_valid(content_handling);
}

SIPMessageBody & SIPMessageBody::operator=(const SIPMessageBody & body)
{
  strassign(&content, body.content);
  content_length = body.content_length;
  strassign(&content_type, body.content_type);
  strassign(&content_disposition, body.content_disposition);
  strassign(&content_handling, body.content_handling);
  return *this;
}


void SIPMessageBody::SetContent(const char *body)
{
  strassign(&content, body);
}

void SIPMessageBody::SetContentType(const char *type)
{
  strassign(&content_type, type);
}

void SIPMessageBody::SetContentDisposition(const char *disposition)
{
  strassign(&content_disposition, disposition);
}

void SIPMessageBody::SetContentHandling(const char *handling)
{
  strassign(&content_handling, handling);
}

void SIPCall::SetMessageBody(const SIPMessageBody & body)
{
  sipcall_t *call;
  if ((call = CallSearch(callid)) == NULL) {
    error("SIPCall::SetMessageBody", 
          "No call with callid=%s\n", callid ? callid : "NULL");
  }
  else {
    DString disp;

    strassign(&call->body.content, body.GetContent());
    strassign(&call->body.content_type, body.GetContentType());
    call->body.content_length = body.GetContentLength();
    if (call->body.content_length == 0) {
      call->body.content_length = strlen(call->body.content);
    }

    DStringInit(&disp);
    DStringPrintf(&disp, "%s; handling=%s", body.GetContentDisposition(),
                                            body.GetContentHandling());
    /*
     * Currently we only add in the response header.
     * SHould this be added for request headers also.
     * Also you should make sure to use 
     * SDP body is present for 2xx responses.
     */
    HeaderReplaceOrAppend(&call->h_ors, "Content-Disposition", 
                          DStringValue(&disp));
    DStringFree(&disp);
    pthread_mutex_unlock(&call->mutex);
  }
  return;
}

void SIPCall::OnNewMessageBody(const SIPMessageBody & body)
{
  /* do nothing with the message body.
   * It is assumed the the application will 
   * override this methos if it wants to
   * process non-SDP bodies. */
}

int SIPCall::OnRemoteSessionChanged(const SessionDescription & sdp)
{
  /*
   * we return -1 here and the application
   * can override the return value to 0 
   * if it wants to support re-INVITE
   * XXX ideally we should compare this with the existing
   * sessions and return appropriately.
   */
  return -1;
}

#ifdef MEDIACONTROL_SIPCOMM

void SIPCall::OpenMediaStream()
{
  m_pMSInfo = openMediaSession();
  if (m_pMSInfo) {
      /* Get remote IP and port */
      SessionDescription* rem_session = GetRemoteSessionDescription();
    
      MediaInfo rem_media;
      rem_session->FindSession(rem_media, 
                               MediaInfo::Receive, MediaInfo::Audio);
      
      IptelIPAddress rem_addr = rem_media.GetRxAddress();
      
      struct sockaddr_in addr;
      rem_addr.GetIpAddress(addr);
      m_pMSInfo->peerAddr.nPort = rem_media.GetRxPort();
      m_pMSInfo->peerAddr.ulIPAddr = (unsigned long)(rem_addr);
 
      SessionDescription* host_session = GetSelfSessionDescription();
     
      MediaInfo host_media;
      host_session->FindSession(host_media, 
                                MediaInfo::Receive, MediaInfo::Audio);
      
      IptelIPAddress host_addr = host_media.GetRxAddress();
      
      struct sockaddr_in rx_addr;
      host_addr.GetIpAddress(rx_addr);
      
      m_pMSInfo->selfAddr.ulIPAddr = (unsigned long)(host_addr);
      m_pMSInfo->selfAddr.nPort =  host_media.GetRxPort();
      
      
      establishMediaSession(m_pMSInfo->id);
      
  }
}

void SIPCall::CloseMediaStream()
{
  if (m_pMSInfo != NULL) {
    terminateMediaSession(m_pMSInfo->id);
    m_pMSInfo = NULL;
  }
}

int SIPCall::SwitchMediaComponentMode(int MediaSession_Mode, char* WaveFileName)
{
  return SetMediaComponentMode(m_pMSInfo->id,MediaSession_Mode, WaveFileName);
} 

time_t SIPCall::GetStartTimestamp ()
{
  return GetTimeStamp(m_pMSInfo->id, 10);
}

time_t SIPCall::GetEndTimestamp ()
{
  return GetTimeStamp(m_pMSInfo->id, 20);
}


void SIPCall::SetUpSelfSessionDescription()
{
  SessionDescription *session;
  MediaInfo media;

  /* Allocate self session info */
  session = new SessionDescription;
  media.SetDirection(MediaInfo::Receive);
  media.SetDataType(MediaInfo::Audio);
  IptelIPAddress rxaddr(SIPLibrary::HostInAddr);
  media.SetRxAddress(rxaddr);
  srand(time(NULL));
  media.SetRxPort(4000 + (rand() % 6000)*2);
  FormatInfo f(0);
  media.AddFormat(f); /* PCMU */
  session->AddSession(media);
  SetSelfSessionDescription(session);
}

HWAVEOUT SIPCall::GetWaveOutHandle()
{
  return getWaveOutHandle(m_pMSInfo->id);
}

bool SIPCall::SetMuteState(BOOL mute)
{
  return setMute(m_pMSInfo->id,mute);
}

void SIPCall::PutCallOnHold()
{
  SessionDescription* host_session = GetSelfSessionDescription();
  
  MediaInfo host_media, host_media2;
  host_session->FindSession(host_media2, 
                            MediaInfo::Receive, MediaInfo::Audio);
  host_media = host_media2;
  
  host_session->DeleteSession(MediaInfo::Receive, MediaInfo::Audio);
  host_media.SetDirection(MediaInfo::Receive);
  host_media.SetDataType(MediaInfo::Audio);
  IptelIPAddress rxaddr("0.0.0.0");
  host_media.SetRxAddress(rxaddr);

  host_session->AddSession(host_media);
  SetSelfSessionDescription(host_session);
}

#endif

/* InitCanonicalize.
 * Dummy implementation.
 */
int InitCanonicalize(int argc, char *argv[]) {
 return 0;
} 

/*
 * InitCanonicalize2
 * Dummy implementation.
 */
int InitCanonicalize2(int argc, char *argv[], 
                      cinema_table *optional_database) {
 return 0;
}

/*
 * canonicalize_uri()
 * Dummy implementation. 
 */
int canonicalize_uri(uri_t* input_uri, canon_result* out_result) {
 return -1;
}

/*
 * CloseCanonicalize.
 *  Dummy implementation.
 */
int CloseCanonicalize() {
 return 0;
}

