/* sipcall.cpp - SIP Call handling.
 * 1999-2001 Columbia University. All rights reserved.
 */
extern "C" {
#include <stdio.h>                   /* basic i/o */
#include <string.h>                  /* strlen, ... */
#include <stdlib.h>                  /* malloc, ... */
#include <sys/errno.h>               /* errno, ... */
#include <pthread.h>                 /* pthread_create, ... */
#include <time.h>                    /* time, ... */
#include <sys/time.h>                /* gettimeofday, ... */
}

#include "iptelapi.h"
#include "sipapi.h"
#include "sipcall.h"

extern "C" {
#include "sysdep.h"
#include "strdupn.h"                 /* free_valid */
#include "request-sip.h"
#include "policy.h"
#include "timespec.h"
#include "pthread-util.h"
}


#include "msgflow.h"

static sipcall_t *C = NULL;    /* list of calls */
static pthread_mutex_t C_lock = PTHREAD_MUTEX_INITIALIZER;

/* XXX should we use SIPLibrary::Random32 in all c++ files ? */
extern "C" unsigned long random32();

/* Local prototypes */
void * CallThread(void *);

extern IptelAddress SipAddressToIptelAddress(sip_addr_t addr);

/*
* CallCreate()
* Create a new call object and the call thread.
* It also adds the call in the list.
* This is called from the constructor when the callid
* is not known (e.g., for a new call), typically
* for an outgoing call. For an incoming call
* the library itself calls this and then allocates
* a SIPCall object with a defined callid.
*/
sipcall_t *CallCreate(const char *id)
{
  int e; /* error */
  sipcall_t *c = (sipcall_t *) calloc(1, sizeof(sipcall_t));
  char *callid = NULL;

  c->state = SIPCALL_IDLE;
  if (id == NULL) {
    /* create a new callid */
    const char *host = SIPLibrary::HostName4;
    DString NewCallId;
    DStringInit(&NewCallId);
    DStringPrintf(&NewCallId,"%u@%s", (unsigned int) random32(), 
                  (host != NULL) ? 
                  host : "unknown");
    /* We use strdup() instead of directly taking the value
    * of DString() to save space. Note that DString()
    * may allocate some additional bytes 
    */
    c->callid = strdup(DStringValue(&NewCallId));
    callid = strdup(c->callid);
    DStringFree(&NewCallId);  
  }
  else {
    c->callid = strdup(id);
    callid = strdup(c->callid);
  }

  c->self      = NULL;
  c->remote    = NULL;
  c->self_tag  = NULL;
  c->remote_tag= NULL;

  c->lastcseq_rx  = 0;
  c->lastcseq_tx  = 0;
  c->mqueue    = NULL;
  //c->tid       = NULL; /* does not work for unix */
  c->userid    = NULL;
  c->password  = NULL;
  c->unauth_count = 0;

  c->route = NULL;

  c->next      = NULL;

  pthread_mutex_init(&c->mutex, NULL);
  pthread_cond_init(&c->cond, NULL);
  c->done  = 0;

  c->callobj = NULL;
  c->self_sdp = NULL;
  c->remote_sdp = NULL;
  c->type  = SOCK_DGRAM; /* default is SIP over UDP */  
  c->af    = AF_INET;    /* SIP/UDP/IPv4 */
  c->incoming_reinvite = 0;

  pthread_mutex_lock(&c->mutex);

  CallAdd(c);

  /* Create the thread */
  if ((e = pthread_create_detached(&c->tid, NULL, 
                          CallThread, (void *) strdup(callid))) != 0) {
    error("CallCreate","pthread_create_detached failed %d\n", e);
  }

  if (e != 0) {
    CallFree(c);
    c = NULL;
  }

  /*
  if (c) {
    pthread_mutex_unlock(&c->mutex);
  }
  */

  free(callid);

  return c;
} /* CallCreate */

/* 
* Add the call in the calls list,
* it is assumed that the call mutex is locked
*/
void CallAdd(sipcall_t *c)
{
  /* Add the new call in the begining */
  pthread_mutex_lock(&C_lock);
  c->next = C;
  C = c;
  pthread_mutex_unlock(&C_lock);
} /* CallAdd */

/*
* CallFree()
* If present, remove request from linked list of requests.
* It is assumed that the mutex is locked.
* If the call thread is active, this should only
* be called from that thread, to avoid any
* race condition !
*/
void CallFree(sipcall_t *c)
{
  sipcall_t *cp;
  message_t *m, *m_next;

  if (!c) return;

  debug("CallFree", "Freeing %p\n", c);

  /* set this to terminating no new call is tried. */
  /* This is against the normal state transition if the state
  * is already IDLE, since you are transitioning from 
  * IDLE to TERMINATING. But that's OK since we are terminating
  * the call thread (FSM) anyway.
  */
  c->state = SIPCALL_TERMINATING; 

  pthread_mutex_unlock(&c->mutex);
  pthread_mutex_lock(&C_lock);
  pthread_mutex_lock(&c->mutex);

  /* Remove from list. */
  if (c == C) {
    C = C->next;
  }
  else {
    for (cp = C; cp && cp->next != c; cp = cp->next) ;
    if (cp) cp->next = c->next;
  }
  pthread_mutex_unlock(&C_lock);

  /* Free the message queue */
  for (m = c->mqueue; m != NULL; m = m_next) {
    m_next = m->next;
    MessageDelete(m);
  }
  /* We set all the pointers to NULL,
  * after freeing to avoid any misuse of freed
  * memory !!! */

  c->mqueue = NULL;
  /* XXX note that c->callobj->mark_done must have been set to 1,
   * otherwise we have memory leak, because the
   * destructor ~SIPCall (for c->callobj) will
   * wait in infinite loop. We should not access
   * c->callobj here because it might have been deleted.
   * There should be some way to fix this problem !!!
   */
  c->callobj = NULL;
  free_valid(c->callid);     c->callid     = NULL;
  free_valid(c->self);       c->self       = NULL;
  free_valid(c->remote);     c->remote     = NULL;
  free_valid(c->self_tag);   c->self_tag   = NULL;
  free_valid(c->remote_tag); c->remote_tag = NULL;
  free_valid(c->self_sdp);   c->self_sdp   = NULL;
  free_valid(c->remote_sdp); c->remote_sdp = NULL;
  URI_Free(&c->remote_uri);
  URI_Free(&c->proxy);
  /* free the body */
  free_valid(c->body.content); c->body.content = NULL;
  free_valid(c->body.content_type); c->body.content_type = NULL;
  /* Free any request or response parsed or unparsed headers */
  HTTP_Free(&c->p_irq);
  HTTP_Free(&c->p_irs);
  HTTP_Free(&c->p);
  HeaderFree(&c->h_orq);
  HeaderFree(&c->h_ors);

  free_valid(c->userid);
  free_valid(c->password);

  if (c->route != NULL) {
    RFC822_Free(c->route);
    free_valid(c->route);
  }

  pthread_mutex_unlock(&c->mutex); /* Release lock before deleting */
  pthread_mutex_destroy(&c->mutex);
  pthread_cond_destroy(&c->cond);
  free(c);

  debug("CallFree", "Exiting\n");
  return;
} /* CallFree */

/*
* CallPrint()
* This prints the existing calls on the output stream FILE *fp
* If fp is NULL then prints to stdout.
*/
void CallPrint(FILE *fp)
{
  sipcall_t *cp;

  if (!fp) {
    fp = stdout;
  }

  pthread_mutex_lock(&C_lock);
  for (cp = C; cp; cp = cp->next) {
    /*  Print per call info here XXX */
  }
  pthread_mutex_unlock(&C_lock);
 
  return;
} /* callPrint */

 
/*
 * CallSearch()
 * Determine if the request 'c' with callid 'callid' exists in the list of
 * calls. 
 * N.B.: The returned call has its mutex locked!  You must unlock it
 * when you're done with it!  
 */
sipcall_t *CallSearch(const char *callid)
{
  sipcall_t *cp, *result = NULL;

  if (!callid)
    return NULL;

  pthread_mutex_lock(&C_lock);
  for (cp = C; cp; cp = cp->next) {
    if (cp->callid != NULL &&
        !strcmp(cp->callid, callid)) {
      result = cp;
      pthread_mutex_lock(&cp->mutex);
      pthread_mutex_unlock(&C_lock); /* Release R_lock to avoid deadlock */
      return result;
    }
  }
  pthread_mutex_unlock(&C_lock);

  return NULL; /* not found */
} /* CallSearch */


/*
 * CallDone();
 * This sets the flag for call thread to terminate
 * and signals the condition.
 * It is assumed that the mutex is locked.
 */
void CallDone(sipcall_t *c)
{
  c->done = 1;
  pthread_cond_broadcast(&c->cond);
}

/*
 * CallComplaint()
 */
void CallComplaint(sipcall_t *c)
{
  error("CallComplaint", "Somthing wrong happened\n");
}

/*
 * CallCleanup() is empty XXX 
 */
int CallCleanup(sipcall_t *c)
{
  return 0;
}

bool CallCompareContentType(const char *type,
                            parser_t *p)
{
  bool oMatching = false; /* whether the Accept header matches or not */
  if (type != NULL) {
    oMatching = true; /* If Accept header is
                       * not present then we return that it matches */
    parser_header_t *h;
    if ((h = RFC822_Header(p->headers, "Accept")) &&
      h->elements->element) {
      int i;
      oMatching = false; /* If Accept is present then must match it */
      for (i=0; ; i++) {
        char *type2;
        type2 = RFC822_Attribute(h, i, NULL /* get the value */);
        if (type2 == NULL || *type2 == 0) {
          /* Last element, exit from loop */
          break;
        }
        if (!strcmp(type, type2)) {
          oMatching = true;
          break;
        }
      }
    }
  }
  else {
    /* No need to match,
     * since there is no content-type
    * to be sent. */
    oMatching = true;
  }

  return oMatching;
}

/*
 * This function enqueues a new response by building
 * on the existing request/response by taking the
 * headers (additional) from h and body (if present)
 * from body. This is needed because there is no
 * method to enqueue a defined response to the
 * client/request threads.
 */
void MessageEnqueueResponseNew(message_t **list, 
              branch_t *b, handled_by_t handled, 
              int status, const char *reason,
              request_t *r, headers_t *new_headers, body_t *body)
{
  message_t *new1 = (message_t *) calloc(1, sizeof(message_t));
  if (new1 == NULL) {
    error("MessageCreate", "Couldn't allocate new memory space");
    return;
  }

  /* We build the default response and then add thebody and
   * headers to the r->headers, is this correct ? XXX */

  /* Let the existing methods fill the default headers */
  ResponseSetStatus(r, status, (char *) reason);
  ResponseBuildDefault(r);
  /*
   * Append the new headers and body.
   */
  if (body) {
    ResponseSetBody(r, body->content, body->content_type, body->content_length);
  }

  
  if (new_headers) {
    HeaderMerge(&r->headers, new_headers);
  }

  new1->handled = handled;
  new1->branch = b;
  new1->next = NULL;
  new1->type = message_real_message;
  new1->p.type = http_response;
  new1->p.status = status;
  new1->p.reason = reason ? strdup(reason) : NULL;
  
  MessageReallyEnqueue(list, new1);

  return;
}

/*
 * MessageEnqueueRequestNew is called to enqueue a request of
 * type ACK. Since the proxy code (in client.c) does not
 * automatically generate a ACK request this needs to be
 * generated by this module for user agents.
 * It copies the parsed haeders from r and 
 * appends a body if present.
 */
void MessageEnqueueRequestNew(message_t **list, 
              branch_t *b, handled_by_t handled, 
              const char *method,
              request_t *r, body_t *body, parser_header_t *route)
{
  headers_t hdrs; /* Followup request headers */ 
  parser_t  p;
  DString msg;
  char *version  = HTTP_Version(&r->p);
  char *protocol = HTTP_Protocol(&r->p);
  sip_cseq_t cseq = SIP_CSeq(&r->p);
  DString uri_str;

  DStringInit(&uri_str);

  hdrs.headers = NULL;
  hdrs.last = NULL;

  /* 
   * Create request message by copying all headers and body, 
   * adding the Via with the proxy in it 
   */
  DStringInit(&msg);
  if (route != NULL && route->elements != NULL &&
       route->elements->value) {
    sip_addr_t firstroute;

    memset(&firstroute, 0, sizeof(firstroute));

    firstroute = sip_addr(route->elements->value);
    if (firstroute.uri.scheme != NULL) {
      URI_SPrint(&firstroute.uri, &uri_str);
    }
    else {
      URI_SPrint(&r->p.URI, &uri_str);          
    }
    free_valid(firstroute.display_name);
    URI_Free(&firstroute.uri);
  }
  else {
    URI_SPrint(&r->p.URI, &uri_str);
  }
  DStringPrintf(&msg, "%s %s %s/%s\r\n", method,
                DStringValue(&uri_str), protocol, version);

  /*
   * Do not add via header . It is added automatically
   * by the client.c thread.
   */
  HeaderCopyOneParsedHeader(&hdrs, r->p.headers, "To");
  HeaderCopyOneParsedHeader(&hdrs, r->p.headers, "From");
  HeaderCopyOneParsedHeader(&hdrs, r->p.headers, "Call-ID");
  HeaderPrintf(&hdrs, 1, "Cseq", "%lu %s", (long)cseq.seq, method);

  if (route) {
    parser_element_t *e;

    /* Add the Route header */

    /* Add a dummy route header which will be ignored any way.
       XXX Not done. */
 
    for (e = route->elements; e; e=e->next) {
      HeaderPrintf(&hdrs, 0 /* Don't replace */,
                   "Route", "%s", e->element);
    }

  }

  /* XXX Do we want any others -- Date, User-Agent, etc? */

  /* Add the content-type,length */
  if (body && body->content_type && body->content) {
    HeaderPrintf(&hdrs, 1, "Content-Type", "%s", body->content_type);
    HeaderPrintf(&hdrs, 1, "Content-Length", "%u", 
      (body->content_length > 0) ? 
      body->content_length : strlen(body->content));
  }
  else {
    HeaderPrintf(&hdrs, 1, "Content-Length", "%u", 0);
  }

  HeaderMessage(&msg, &hdrs);
  DStringAppend(&msg, "\r\n", 2); 

  memset(&p, 0, sizeof(p));
  if (HTTP_Parse(DStringValue(&msg), &p) < 0) {
    error("MessageEnqueueRequestNew", "parse error\n");
  }
  else {
    struct body_t body_new;
    struct sockaddr_storage sin;
    memset(&sin, 0, sizeof(sin));
    sin.ss_family = AF_INET;
    if (body) {
      body_new = *body;
    }
    else {
      memset(&body_new, 0, sizeof(body_new));
    }
    MessageEnqueue(list, message_real_message,
      &p, &body_new, &sin, NULL, nobody_handled);
  }

  free(version);
  free(protocol);
  DStringFree(&uri_str);
  HeaderFree(&hdrs);
  HTTP_Free(&p);
  DStringFree(&msg);
}

/*
 * Set the call status 
 */
void CallSetStatus(sipcall_t *c, int status, const char *reason)
{
  c->p.status = status;
  strassign(&c->p.reason, reason);
}

/*
 * CallThread().
 * This implements the call thread 
 */
void *CallThread(void *cid)
{
  char *callid = (char *)cid;
  sipcall_t *c = NULL;
  struct message_t *m, *m_next, **prev_next;
  int thr_err;
  int first_time = 1;

  if (cid == NULL) {
    error("CallThread", "cid is NULL\n");
    return NULL;
  }

  debug("CallThread", "called with %s\n", callid);

  if ((c = CallSearch(callid)) == NULL) {
    debug("CallThread", "Call is not in the list\n");
    free_valid(callid);
    return NULL;
  }

  free_valid(callid);

  /* c->mutex is now locked. */

  while (!c->done) {
    /* Do not wait on the cond for the first time,
       to avoid lost wakeup problem. The
       API thread may signal cond before it is started.
    */
    if (!first_time) {
      if (c->callobj == NULL) {
        /* No associated call obj wait for 40 seconds */
        struct timeval now;

        gettimeofday(&now, NULL);
        struct timespec timeout;
        timeout = timespec_delay(now, 40.0);

        thr_err = pthread_cond_timedwait(&c->cond, &c->mutex, &timeout);
      }
      else {
        thr_err = pthread_cond_wait(&c->cond, &c->mutex);
      }

      if (thr_err) {
        if (thr_err != ETIMEDOUT) {
          error("CallThread", "Error waiting for proxy events: %s (%d)",
                strerror(thr_err), thr_err);
        }
        else {
          /* The timeout expired. */
          /* Don't trigger another timeout unless one is set */
          goto cleanup;
        }
        continue;
      }
    }
    first_time = 0;

    /* Walk the message queue, and send requests not already handled */
    /* If the handler for any message calls CallDone(), stop
       immediately. */
    for (m = c->mqueue; m != NULL && !c->done; m = m->next) {
      if (m->handled >= policy_handled)
        continue;

      if (m->type != message_real_message && m->type != message_pseudo_message &&
          m->type != message_api_message) {
        error("CallThread",
              "Surprising message type %d on mqueue", m->type);
        m->handled = finished_handling;
        continue;
      }

      if (m->p.type == http_request) {
        if (CallHandleRequest(c, m) != 0) {
          CallComplaint(c);
          goto cleanup;
        }
      }
      else if (m->p.type == http_response) {
        if (CallHandleResponse(c, m) != 0) {
          CallComplaint(c);
          goto cleanup;
        }
      }
      else {
        error("CallThread",
              "Surprising message parse type %d on mqueue", m->p.type);
        m->handled = finished_handling;
      }
    }

    /* Remove all nodes marked finished_handling from the queue. */
    prev_next = &c->mqueue;
    for (m = c->mqueue; m != NULL; m = m_next) {
      m_next = m->next;
      if (m->handled == finished_handling) {
        *prev_next = m->next;
        MessageDelete(m);
      } else {
        prev_next = &m->next;
      }
    }
  }

cleanup:  /* Jump here when any policy function returns non-zero */

  CallCleanup(c); /* should this be after cleanup: label XXX.
                   * what about in proxy.c and user.cpp */

  /* XXX Alert the requests to shut down */

  /* XXX Alert the API call objects to terminate */

  debug("CallThread", "c=%p finished\n", c);

  if (c->callobj) {
    c->callobj->mark_done = 1;
  }
  c->callobj = NULL;

  CallFree(c);

  return NULL;
}

int CallHandleRequest(sipcall_t *c, message_t *m)
{
  int from_api = 0;
  request_t *rp = NULL;

  debug("CallHandleRequest", "Entered\n");

  if (m->type == message_api_message) {
    from_api = 1;
  }

  if (m->type == message_real_message) {
    pthread_mutex_unlock(&c->mutex);
    rp = RequestFind(&m->p, 0 /* hash is not known */);
    pthread_mutex_lock(&c->mutex);

    debug("CallHandleRequest", "RequestFind returned %p\n", rp);

    /* Update the headers, parsers */
    if (c->state == SIPCALL_OUTGOINGCALL && !from_api &&
         !strcmp(m->p.method, "INVITE")) {
      /* Do not overwrite c->p if the same INVITE is looped back to
       * us when we are sending outgoing INVITE. */
    }
    else {
      /* Otherwise copt m->p to c->p */
      HTTP_Duplicate(&c->p, &m->p);
    }

    HTTP_Duplicate(&c->p_irq, &m->p);
  }
  else if (m->type == message_api_message) {
    if (strcasecmp(m->p.method, "CANCEL") == 0 ||
         (c->state == SIPCALL_OUTGOINGCALL && 
         strcasecmp(m->p.method, "BYE") == 0)) {
      /* Find the request for CANCEL or BYE */
      pthread_mutex_unlock(&c->mutex);
      rp = RequestFind(&c->p, 0 /* hash is not known */);
      pthread_mutex_lock(&c->mutex);
    }
  }
  else {
    debug("CallHandleRequest", "No original request\n");
  }


  if (!from_api) {
    /* 
     * If a UAS finds a Record-Route headers ina request, it copies
     * the Record-Route maddr parameters as well as other Record-Route
     * parameters and any port value, maintaining their ordering to
     * the c->route. The name-addr is taken from Contact (if present)
     * or From header. All URI parameters are also copied from
     * Contact or From as appropriate.
     */
    parser_header_t *h;
    if ((h = RFC822_Header(m->p.headers, "Record-Route")) &&
        h->elements) {
      parser_element_t *e;
      sip_contact_t *contact;
      sip_addr_t from;

      debug("RecordRoute", "Encountered Record-Route header\n");

      /* Duplicate the route header */
      RFC822_DuplicateSingleHeader(&c->route, h);
      if (c->route == NULL) {
        debug("RecordRoute", "Could not duplicate record route.\n");
      }
      else {
        debug("RecordRoute", "Duplicated the header\n");
        memset(&from, 0, sizeof(from));

        /* Get the contact or from address */
        if ((contact = SIP_Contact(&m->p)) == NULL ||
            contact[0].uri.scheme == NULL) {
          debug("RecordRoute", "No contact use from\n");
          from = SIP_From(&m->p);
          /* maddr is absent anyway */
        }
        else {
          debug("RecordRoute", "Contact is present\n");
          from.uri = URI_Copy(contact[0].uri);        
        }
        if (contact) {
          SIP_FreeContact(contact);
        }
        
        /* Parameters are kept as they are */
        /* Comment is also unchanged */
        for (e = c->route->elements; e; e = e->next) {
          sip_addr_t rh;
          sip_addr_t actual;
          char *maddrparam;

          debug("RecordRoute", "Header is %s\n", e->value);

          memset(&actual, 0, sizeof(actual));
          if (from.display_name) {
            actual.display_name = strdup(from.display_name);
          }
          actual.uri = URI_Copy(from.uri);
          
          rh = sip_addr(e->value); /* Parsed the SIP address */
          
          if (rh.uri.scheme == NULL) {
            debug("RecordRoute", "Could not parse the address\n");
            continue; /* To next element */
          }
          actual.uri.port = rh.uri.port; 

          maddrparam = (char*)URI_Parameter(rh.uri, "maddr");
          /* Do not free maddrparam */

          if (maddrparam == NULL) {
            debug("RecordRoute", "maddr is NULL\n");
          }
          else {
            char *nameaddr;
            
            debug("RecordRoute", "Maddr is present\n");

            /* Change the maddr param */
            URI_SetParameter(&actual.uri, "maddr", maddrparam);
            
            /* Build the header */
            nameaddr = SIP_BuildNameAddr(actual.display_name,
                                         &actual.uri);

            if (nameaddr == NULL || *nameaddr == '\0') {
              debug("RecordRoute", "Could not build name addr\n");
            }
            else {
              strassign(&e->value, nameaddr);             

              /* XXX this is wrong. We should put the complete header
               * here including parameters, comments.*/
              strassign(&e->element, nameaddr);

              debug("RecordRoute", "Added %s\n", nameaddr);
            }
            free_valid(nameaddr);
            
            /* Now set this in route header */
            
          } /* maddrparam is not null */

          /* Free the memory */
          free_valid(actual.display_name);
          URI_Free(&actual.uri);

          free_valid(rh.display_name);
          URI_Free(&rh.uri);

        } /* Loop through all Record-route headers */

        if (c->route && c->route->elements &&
             c->route->elements->value) {
          c->remote_uri = URI_Parse(c->route->elements->value);
        }

      } /* Could duplicate the headers */
    } /* Record-route is present */
  } /* Record-route handling */

  if (c->route && c->route->elements) {
    parser_element_t *e;
    for (e = c->route->elements; e; e = e->next) {
      debug("RecordRoute", "Element: %s\n", e->value ? e->value : "NULL");
    }
  }

  /* State machine processing */
  switch (c->state) {
    case SIPCALL_IDLE: /* No active call */
      debug("CallHandleRequest", "received %s in IDLE state\n", m->p.method);

      /*
      * Possible messages are from_api : INVITE
      *                      !from_api : INVITE
      */
      if (!from_api && strcmp(m->p.method, "INVITE") == 0) {
        /*
         * Incoming call,
         * inform the API about an incoming call.
         */
        if (c->callobj) {
          /* What is rp is NULL */
          if (rp != NULL) {
            c->state = SIPCALL_INCOMINGCALL; /* set state before calling callback */

            if (c->route && c->route->elements &&
                 c->route->elements->value) {
              /* It is already set */
            }
            else {

             sip_contact_t *contact = SIP_Contact(&rp->p);
             if (contact != NULL && contact[0].uri.scheme != NULL) {
              URI_Free(&c->remote_uri); /* free previous value if any
                                         */
                c->remote_uri = URI_Copy(contact[0].uri);
             }
             else {
              URI_Free(&c->remote_uri); /* free previous value if any */
              sip_addr_t from = SIP_From(&rp->p);
              c->remote_uri = from.uri;
              free_valid(from.display_name);
              /* do not free from.uri */
              strassign(&c->remote_uri.scheme, "sip"); /* to avoid errors */

              /* XXX is this correct to use the IP address, directly */
              //strassign(&c->remote_uri.host, inet_ntoa(rp->source_sin)); 
             }
             SIP_FreeContact(contact);
           }

            /*
             * Set the body here for the message 
             */
            if (m->body.content_type && m->body.content) {
              if (m->body.content_length <= 0) {
                m->body.content_length = strlen(m->body.content);
              }

              /* Is mime type case sensitive XXX */
              if (!strcasecmp(m->body.content_type, "application/sdp")) {
                /* do not use strassign for body, because length may
                 * be smaller than the actual string.
                 */
                free_valid(c->remote_sdp);
                if (m->body.content_length >= 100000) {
                  m->body.content_length = strlen(m->body.content);
                }
                c->remote_sdp = strdupn(m->body.content, m->body.content_length);
              }
              else {
                debug("CallHandleRequest", "content-type %s received in INVITE\n",
                   m->body.content_type);
                /* XXX respond with error code "415 Unsupported Media Type"
                 * and add the Accept header */
              }
            }

            /* Set the other headers, like cseq,
             * self and remote names etc... */
            sip_cseq_t cseq = SIP_CSeq(&m->p);
            c->lastcseq_rx = cseq.seq;
            IptelAddress addr;
            /* To becomes self */
            addr = SipAddressToIptelAddress(rp->to);
            strassign(&c->self, (const char *) addr);
            debug("CallProcessRequest", "Self is %s\n", (const char *) addr);
            /* From becomes remote */
            addr = SipAddressToIptelAddress(rp->from);
            strassign(&c->remote, (const char *) addr);

            /* set the socket type */
            c->type = rp->type;

            if (rp->server != NULL) {

              SIPEndpoint *server = (SIPEndpoint *)rp->server;
              /* unlock the original transaction and reset it */
              pthread_mutex_unlock(&rp->mutex);
              rp = NULL;

              pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
              c->callobj->OnNewIncomingCall( *server);
              pthread_mutex_lock(&c->mutex);
            }
            else {
              /* send back 500 response */
              MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 500);
              pthread_cond_broadcast(&rp->cond);
            }
          }
          else {
            /* XXX send back appropriate response */
            debug("CallHandleRequest", "No rp for incoming call\n");
          }
        }
        else {
          if (rp != NULL) {
            /* Send back internel error response */
            debug("CallHandleRequest", "No call present\n");
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 500);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      else if (from_api && strcmp(m->p.method, "INVITE") == 0) {

        if (c->callobj) {
          /*
           * outgoing call.
           * Invoke the request thread/policy for new INVITE.
           */
          c->state = SIPCALL_OUTGOINGCALL;

          request_t *r;
          body_t body;
          body.content_length = 0;
          if (c->self_sdp == NULL) {
            debug("CallhandleRequest", "No SDP specified\n");
          }
          else {
            /* We are sending the self session always */
            body.content_type = "application/sdp";
            body.content = c->self_sdp;
          }

          r = FormatRequest("INVITE", c->type, c->af,  
            c, ++c->lastcseq_tx, c->self_sdp ? &body : NULL);

          if (c->route && c->route->elements &&
               c->route->elements->value) {
            /* No change as already set */
          }
          else {
            URI_Free(&c->remote_uri);
            c->remote_uri = URI_Copy(r->p.URI);
          }

          /* Should store the outgoing request here */
          HTTP_Duplicate(&c->p, &r->p);

          /* Set the request body, SDP  XXX */
          /*
          if (c->proxy.scheme) {
            URI_SetParameter(&c->proxy, 
              "transport", c->type == SOCK_STREAM ? "tcp" : "udp");
          }
          else {
            URI_SetParameter(&c->remote_uri, 
              "transport", c->type == SOCK_STREAM ?
              "tcp" : "udp");
          }
          */
          CallCreatePolicyThread(r, c->remote_uri, c->proxy);
        }
        else {
          /* How do we send back the response to API */
          debug("CallHandleRequest", "No call object, ignoring\n");
        }
      }
      else {
        /* If this is a remote request then send back appropriate response. */
        debug("CallHandleRequest", "Message in unexpected state\n");
        if (!from_api && rp) {
          if (!strcmp(m->p.method, "BYE") || !strcmp(m->p.method, "CANCEL")) {
            /* Call leg / transaction does not exitst */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 481);
            pthread_cond_broadcast(&rp->cond);
          }
          else if (strcmp(m->p.method, "ACK")) {
            /* XXX should add a reason-phrase */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 400);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      break;

    case SIPCALL_OUTGOINGCALL: /* waiting for final response from remote */
      debug("CallHandleRequest", "received %s in OUTGOING state\n", m->p.method);
      if (from_api && 
        (strcmp(m->p.method, "CANCEL") == 0 ||
         strcmp(m->p.method, "BYE") == 0)) {
          request_t *r;

          c->state = SIPCALL_TERMINATING;

          /* XXX rp may be incorrect if the request is looped. */

          if (rp == NULL) {
            /* create a new request */
            /* should not increment cseq */
            r = FormatRequest("CANCEL", c->type, c->af, 
              c, c->lastcseq_tx, NULL);

            /*
            if (c->proxy.scheme) {
              URI_SetParameter(&c->proxy, 
                "transport", c->type == SOCK_STREAM ? "tcp" : "udp");
            }
            else {
              URI_SetParameter(&c->remote_uri, 
                "transport", c->type == SOCK_STREAM ?
                "tcp" : "udp");
            }
            */
            CallCreatePolicyThread(r, c->remote_uri, c->proxy);
          }
          else {
            /* use existing request */
            debug("CallHandleRequest", "Sending CANCEL to policy\n");
            MessageEnqueuePseudoRequest(&rp->mqueue, "CANCEL", 
              nobody_handled, NULL);
            pthread_cond_broadcast(&rp->cond);
          }
      }
      else {
        /* Should check for re-INVITE here before the call is establish...
         * use lastcseq_rx/tx.
         * is it allowed ? XXX */

        debug("CallHandleRequest", "Message in unexpected state\n");
        if (!from_api && rp) {
          if (!strcmp(m->p.method, "BYE") || !strcmp(m->p.method, "CANCEL")) {
            /* Call leg / transaction does not exitst */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 481);
            pthread_cond_broadcast(&rp->cond);
          }
          else if (!strcmp(m->p.method, "INVITE")) {
            /* Loop detected */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 482);
            pthread_cond_broadcast(&rp->cond);
            
          }
          else if (strcmp(m->p.method, "ACK")) {
            /* XXX should add a reason-phrase */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 400);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      break;

    case SIPCALL_INCOMINGCALL: /* waiting for API message */
      debug("CallHandleRequest", "received %s in INCOMING state\n", m->p.method);
      if (!from_api && strcmp(m->p.method, "CANCEL") == 0) {
        /*
        * inform api about call cancelled, XXX should
        * we have another callback for OnCallCancelled.
        */
        c->state = SIPCALL_IDLE;

        if (c->callobj != NULL) {
          pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
          if (c->callobj->OnHangup() == 0) {
            c->callobj->mark_done = 1;
            delete c->callobj;
            c->callobj = NULL;
          }
          pthread_mutex_lock(&c->mutex);
          if (c->callobj == NULL) {
            CallDone(c);
          }
        }
        if (rp) {
          /* No need to inform policy since already
           * response has been sent. */
        }
      }
      else {
        debug("CallHandleRequest", "Message in unexpected state\n");

        if (!from_api && rp) {
          if (!strcmp(m->p.method, "BYE") || !strcmp(m->p.method, "CANCEL")) {
            /* Call leg / transaction does not exitst */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 481);
            pthread_cond_broadcast(&rp->cond);
          }
          else if (strcmp(m->p.method, "ACK")) {
            /* XXX should add a reason-phrase */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 400);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      break;

    case SIPCALL_INCOMINGCALL_ACK: /* waiting for ACK from remote */
      debug("CallHandleRequest", "received %s in INCOMING_ACK state\n", m->p.method);
      if (!from_api && strcmp(m->p.method, "ACK") == 0) {
        /* 
         * inform the API that call is established, if not already done.
         */
        c->state = SIPCALL_ACTIVE;

        /*
         * Set the sdp if it is not already set and
         * this body contains it.
         */
        if (c->remote_sdp == NULL && m->body.content_type != NULL &&
             m->body.content != NULL) {
          
          if (m->body.content_length <= 0) {
            m->body.content_length = strlen(m->body.content);
          }
          
          /* Is mime type case sensitive XXX */
          if (!strcasecmp(m->body.content_type, "application/sdp")) {
            /* do not use strassign for body, because length may
             * be smaller than the actual string.
             */
            free_valid(c->remote_sdp);
            if (m->body.content_length >= 100000) {
              m->body.content_length = strlen(m->body.content);
            }
            c->remote_sdp = strdupn(m->body.content, m->body.content_length);
          }
          else {
            debug("CallHandleRequest", "content-type %s received in INVITE\n",
                  m->body.content_type);
            /*
             * Should inform the API about the new body.
             * should we reset the SDP content-type.
             * I think yes XXX 
             */
          }
          
          SIPMessageBody msgbody(m->body.content, 
                                 m->body.content_type,
                                 m->body.content_length);
          if (c->callobj) {
            
            pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
            if (!strcmp(msgbody.GetContentType(), "application/sdp")) {
              SessionDescription *sdp = SIPCall::DeformatSDP(msgbody.GetContent(), 0);
              if (sdp != NULL) {
                if (c->callobj->OnRemoteSessionChanged(*sdp) != 0) {
                  /* What should be the error code for rejecting the 
                   * change in SDP */
                  debug("OnRemoteSessionChanged", "rejected\n");
                }
                delete sdp;
              }
              else {
                debug("DeformatSDP", "SDP not deformated\n");
              }
            }
            else {
              c->callobj->OnNewMessageBody(msgbody);
            }
            pthread_mutex_lock(&c->mutex);
          }
          
        }
      }      
      else if (!from_api && strcmp(m->p.method, "CANCEL") == 0) {
      /*
        * inform api about call cancelled, XXX should
        * we have another callback for OnCallCancelled.
        */
        c->state = SIPCALL_IDLE;

        if (c->callobj != NULL) {
          pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
          if (c->callobj->OnHangup() == 0) {
            c->callobj->mark_done = 1;
            delete c->callobj;
            c->callobj = NULL;
          }
          pthread_mutex_lock(&c->mutex);
          if (c->callobj == NULL) {
            CallDone(c);
          }
        }

        if (rp) {
          /* No need to inform policy since appropriate
           * response has been already sent ! */
        }
      }
      else {
        debug("CallHandleRequest", "Message in unexpected state\n");
        if (!from_api && rp) {
          if (!strcmp(m->p.method, "BYE") || !strcmp(m->p.method, "CANCEL")) {
            /* Call leg / transaction does not exitst */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 481);
            pthread_cond_broadcast(&rp->cond);
          }
          else if (strcmp(m->p.method, "ACK")) {
            /* XXX should add a reason-phrase */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 400);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      break;

    case SIPCALL_ACTIVE:
      debug("CallHandleRequest", "received %s in ACTIVE state\n", m->p.method);
      if (!from_api && strcmp(m->p.method, "BYE") == 0) {
        /* Inform the API that the call is closed */
        c->state = SIPCALL_IDLE;

        if (c->callobj != NULL) {
          if (rp) {
            /* BYE is always successful */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 200);
            pthread_cond_broadcast(&rp->cond);
          }
          else {
            debug("CallHandleRequest", "No request for BYE, strange !!\n");
          }

          pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
          if (c->callobj->OnHangup() == 0) {
            c->callobj->mark_done = 1;
            delete c->callobj;
            c->callobj = NULL;
          }
          pthread_mutex_lock(&c->mutex);
          if (c->callobj == NULL) {
            CallDone(c);
          }
        }
        else {
          error("CallHandleRequest", "No call object found, ignored\n");
          if (rp) {
            /* Call leg / transaction does not exist */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 481);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      else if (from_api && strcmp(m->p.method, "BYE") == 0) {
        /* send BYE to remote */
        c->state = SIPCALL_TERMINATING;

        /* Send a new BYE request */

        request_t *r;
        r = FormatRequest("BYE", c->type, c->af,  
          c, ++c->lastcseq_tx, NULL);

        /* Set the request body, SDP  XXX */
        /*
        if (c->proxy.scheme) {
          URI_SetParameter(&c->proxy, 
            "transport", c->type == SOCK_STREAM ? "tcp" : "udp");
        }
        else {
          URI_SetParameter(&c->remote_uri, 
            "transport", c->type == SOCK_STREAM ?
            "tcp" : "udp");
        }
        */
        CallCreatePolicyThread(r, c->remote_uri, c->proxy);
      }
      else if (from_api && strcmp(m->p.method, "INVITE") == 0) {
        /* API wants a re-INVITE to be sent */
        if (c->callobj) {
          request_t *r;
          body_t body;
          body.content_length = 0;
          if (c->self_sdp == NULL) {
            debug("CallhandleRequest", "No SDP specified\n");
          }
          else {
            /* We are sending the self session always */
            body.content_type = "application/sdp";
            body.content = c->self_sdp;
          }

          r = FormatRequest("INVITE", c->type, c->af, 
            c, ++c->lastcseq_tx, c->self_sdp ? &body : NULL);

          /*
          URI_Free(&c->remote_uri);
          c->remote_uri = URI_Copy(r->p.URI);
          */

          /* Should store the outgoing request here */
          HTTP_Duplicate(&c->p, &r->p);

          /* Set the request body, SDP  XXX */
          /*
          if (c->proxy.scheme) {
            URI_SetParameter(&c->proxy, 
              "transport", c->type == SOCK_STREAM ? "tcp" : "udp");
          }
          else {
            URI_SetParameter(&c->remote_uri, 
              "transport", c->type == SOCK_STREAM ?
              "tcp" : "udp");
          }
          */
          CallCreatePolicyThread(r, c->remote_uri, c->proxy);
        }
        else {
          /* How do we send back the response to API */
          debug("CallHandleRequest", "No call object, ignoring\n");
        }
      }
      else if (!from_api && strcmp(m->p.method, "INVITE") == 0) {
        /* Received a re-INVITE */
        if (c->callobj) {
          int status = 200;

          /* Set the flag so that application
           * does not invoke another re-INVITE */
          c->incoming_reinvite = 1;

          /* What if rp is NULL.
           * This part is copied from IDLE state processing.
           * should have a common function instead.
           */
          if (rp != NULL) {
            sip_cseq_t cseq = SIP_CSeq(&m->p);
            sip_addr_t to = SIP_To(&m->p);
            sip_addr_t from = SIP_From(&m->p);
            IptelAddress selfname = c->self;
            IptelAddress remotename = c->remote;
            char *selfstr = selfname.GetURI();
            char *remotestr = remotename.GetURI();
            uri_t selfuri;
            uri_t remoteuri;
            selfuri = URI_Parse(selfstr);
            remoteuri = URI_Parse(remotestr);
            free_valid(selfstr);
            free_valid(remotestr);

            if (! URI_Equal_With_Defaults(to.uri, selfuri, sip_uri_defaults) ||
              ! URI_Equal_With_Defaults(from.uri, remoteuri, sip_uri_defaults)) {
              /* This is a different call leg, respond with 481 */
              /* XXX we must compare the Tag values also, if any XXX */
              debug("CallHandleRequest", "INVITE for a different call leg\n");
              MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 481);
              pthread_cond_broadcast(&rp->cond);
            }
            else if (cseq.seq <= c->lastcseq_rx) {
              /* This is probably a retransmission, it should not 
               * come here. We should respond with 400. */
              debug("CallHandleRequest", "Probably a retransmission of lost INVITE\n");
              MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 400);
              pthread_cond_broadcast(&rp->cond);
            }
            else {
              c->lastcseq_rx = cseq.seq;

              if (c->route && c->route->elements &&
                   c->route->elements->value) {
                /* No change as already set */
              }
              else {
                sip_contact_t *contact = SIP_Contact(&rp->p);
                if (contact != NULL && contact[0].uri.scheme != NULL) {
                  URI_Free(&c->remote_uri); /* free previous value if any */
                  c->remote_uri = URI_Copy(contact[0].uri);
                }
                /* If no contact specified, ignore */
                SIP_FreeContact(contact);
              }

              /*
               * Set the body here for the message 
               */
              if (m->body.content_type && m->body.content) {
                if (m->body.content_length <= 0) {
                  m->body.content_length = strlen(m->body.content);
                }

                /* Is mime type case sensitive XXX */
                if (!strcasecmp(m->body.content_type, "application/sdp")) {
                  /* do not use strassign for body, because length may
                   * be smaller than the actual string.
                   */
                  free_valid(c->remote_sdp);
                  if (m->body.content_length >= 100000) {
                    m->body.content_length = strlen(m->body.content);
                  }
                  c->remote_sdp = strdupn(m->body.content, m->body.content_length);
                }
                else {
                  debug("CallHandleRequest", "content-type %s received in INVITE\n",
                     m->body.content_type);
                  /*
                   * Should inform the API about the new body.
                   * should we reset the SDP content-type.
                   * I think yes XXX 
                   */
                }

                SIPMessageBody msgbody(m->body.content, 
                                       m->body.content_type,
                                       m->body.content_length);
                if (c->callobj) {

                  pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
                  if (!strcmp(msgbody.GetContentType(), "application/sdp")) {
                    SessionDescription *sdp = SIPCall::DeformatSDP(msgbody.GetContent(), 0);
                    if (sdp != NULL) {
                      if (c->callobj->OnRemoteSessionChanged(*sdp) != 0) {
                        /* What should be the error code for rejecting the 
                         * change in SDP */
                        status = 400;
                      }
                      delete sdp;
                    }
                    else {
                      status = 500; /* What should be the correct status here.
                                     * The SDP could not deformatted */
                    }
                  }
                  else {
                    c->callobj->OnNewMessageBody(msgbody);
                  }
                  pthread_mutex_lock(&c->mutex);
                }
              }

              /*
               * Note that the Application may set the new
               * session description here.
               * So we should re-read the SDP and
               * send it to remote 
               */

              /* Respond back */
              /* Send the succssful response to remote */
              if (status == 200) {
                body_t body;
                int sdp_mismatch = 0;
                char *sdp_mismatch_str = NULL;

                memset(&body, 0, sizeof(body));

                if (c->self_sdp) {
                  body.content_type = "application/sdp";
                  char * sdp = 
                    CallMatchSessionWithSource(c->remote_sdp, c->self_sdp, 
                                                       &sdp_mismatch_str);
                  /* Note that we do not update the local.
                   * SDP here */
                  body.content = sdp;
                  body.content_length = sdp != NULL ? strlen(sdp) : 0;
                  /* Do not free sdp free_valid(sdp); */
                  if (body.content == NULL) {
                    /* sdp_mismatch indicates that the
                     * common subset can not be found.
                     * so we should respond with 4xx
                     * indicating the error.
                     */
                    sdp_mismatch = 1;
                  }
                }
                else if (c->body.content) {
                  /* Set to user defined content */
                  debug("Using user define content %s\n",
                    (c->body.content_type ? c->body.content_type : 
                     (char *) "<no-mime-type>"));

                  body = c->body;
                }
                else {
                  body.content_type = NULL;
                  body.content = NULL;
                  body.content_length = 0;
                }

                /* Check for Accept header if present and send accordingly */
                if (!sdp_mismatch) { 
                  if (CallCompareContentType(body.content_type, &rp->p) == true) { 
                    MessageEnqueueResponseNew(&rp->mqueue, 
                      NULL, nobody_handled, m->p.status, m->p.reason,
                      rp, &c->h_ors, body.content_type ? &body : NULL);
                  }
                  else {
                    /* Should we always accept application/sdp ? */
                    HeaderAppend(&c->h_ors, "Accept", body.content_type ?
                      body.content_type : (char *) "application/sdp" /* this is default */);
                    DString str;
                    DStringInit(&str);
                    DStringPrintf(&str, "You do not accept %s !", body.content_type ?
                      body.content_type : "application/sdp");

                    /* Should it be 406 or 415 !!! XXX. Should we add
                     * Other headers like Accept-Encoding and Accept-Language.
                     */
                    MessageEnqueueResponseNew(&rp->mqueue,
                      NULL, nobody_handled, 415, DStringValue(&str), rp, &c->h_ors, NULL);
                    DStringFree(&str);
                    
                    /* should we inform the API about the accept problem */
                  }
                }
                else {
                    /* Should we always accept application/sdp ? */
                    HeaderAppend(&c->h_ors, "Accept", "application/sdp");
                    if (sdp_mismatch_str) {
                      /* If the warning are given add them. */
                      DString str;
                      DStringInit(&str);
                      DStringPrintf(&str, "\"%s\"", sdp_mismatch_str);
                      HeaderAppend(&c->h_ors, "Warning", DStringValue(&str));
                      DStringFree(&str);
                    }

                    /* Should it be 488 or 606!!! 
                     * I think, this is local to this UA hence 488.
                     */
                    MessageEnqueueResponseNew(&rp->mqueue,
                      NULL, nobody_handled, 488, NULL, rp, &c->h_ors, NULL);

                    /* Should we inform the API about the mismatch */
                }

                free_valid(sdp_mismatch_str);
                free_valid(body.content);

                pthread_cond_broadcast(&rp->cond);
              }
              else {
                debug("CallHandleRequest", "User rejected re-INVITE\n");
                /* Should the status code be 6xx.
                 * Since this is a re-INVITE I think
                 * any failure response is OK */
                MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, status);
                pthread_cond_broadcast(&rp->cond);
              }
            }
            
            URI_Free(&to.uri);
            URI_Free(&from.uri);
            free_valid(to.display_name);
            free_valid(from.display_name);
            URI_Free(&selfuri);
            URI_Free(&remoteuri);
          }
          else {
            /* XXX send back appropriate response */
            debug("CallHandleRequest", "No rp for incoming call\n");
          }

          /* We have already responded to the INVITE */

          c->incoming_reinvite = 0;
        }
        else {
          if (rp != NULL) {
            /* Send back internel error response */
            debug("CallHandleRequest", "No call present\n");
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 500);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      else {
        debug("CallHandleRequest", "message in unexpected state\n");

        if (!from_api && rp) {
          if (!strcmp(m->p.method, "CANCEL")) {
            /* Call leg / transaction does not exitst */
            /*
             * What should be the appropriate reason for
             * responding to the CANCEL for a re-INVITE
             */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 400);
            pthread_cond_broadcast(&rp->cond);
          }
          else if (strcmp(m->p.method, "ACK") != 0) {
            /* XXX should add a reason-phrase for responding to noACK */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 400);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      break;

    case SIPCALL_TERMINATING: /* call is being terminated */
      debug("CallHandleRequest", "received %s in TERMINATING state\n", m->p.method);
      /* Remove the request */
      {
        if (!from_api && rp) {
          if (!strcmp(m->p.method, "BYE") || !strcmp(m->p.method, "CANCEL")) {
            /* Call leg / transaction does not exitst */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 481);
            pthread_cond_broadcast(&rp->cond);
          }
          else if (strcmp(m->p.method, "ACK")) {
            /* XXX should add a reason-phrase */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 400);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      break;

    default:
      error("CallHandleRequest", "surprising state %d\n", c->state);
      {
        if (!from_api && rp) {
          if (!strcmp(m->p.method, "BYE") || !strcmp(m->p.method, "CANCEL")) {
            /* Call leg / transaction does not exitst */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 481);
            pthread_cond_broadcast(&rp->cond);
          }
          else if (strcmp(m->p.method, "ACK")) {
            /* XXX should add a reason-phrase */
            MessageEnqueuePseudoResponse(&rp->mqueue, NULL, nobody_handled, 400);
            pthread_cond_broadcast(&rp->cond);
          }
        }
      }
      break;
  }

  if ( rp != NULL) {
    pthread_mutex_unlock(&rp->mutex);
  }

  m->handled = finished_handling;

  return 0;
}

int CallHandleResponse(sipcall_t *c, message_t *m)
{
  int from_api = 0;
  request_t *rp = NULL;

  if (m->type == message_api_message) {
    from_api = 1;
  }

  if (!from_api && m->type == message_real_message) {
    pthread_mutex_unlock(&c->mutex);
    rp = RequestFind(&m->p, 0 /* hash is not known */);
    pthread_mutex_lock(&c->mutex);

    /* Update the incoming response */
    HTTP_Duplicate(&c->p_irs, &m->p);
  }
  else if (from_api) {
    pthread_mutex_unlock(&c->mutex);
    rp = RequestFind(&c->p, 0 /*hash not known */);
    pthread_mutex_lock(&c->mutex);
  }
  else if (!from_api && m->type == message_pseudo_message) {
    pthread_mutex_unlock(&c->mutex);
    rp = RequestFind(&c->p, 0);
    pthread_mutex_lock(&c->mutex);
  }
  else {
    debug("CallHandleResponse", "No original request\n");
  }


  if (!from_api && m->p.status >= 200) {
    /* 
     * The UAC finds a Record-Route headers in a final response,
     * it copies it, including all the parameters, into
     * c->route, reversing the order of fields.
     * If the response has contact header then it is inserted as
     * the last Route-header.
     */
    parser_header_t *h;
    if ((h = RFC822_Header(m->p.headers, "Record-Route")) &&
        h->elements) {
      sip_contact_t *contact;

      RFC822_DuplicateSingleHeader(&c->route, h);
      RFC822_Reverse(c->route);

      contact = SIP_Contact(&m->p);
      if (c->route && contact && contact[0].uri.scheme) {
        parser_element_t *e;
        e = (parser_element_t *)calloc(1, sizeof(parser_element_t));
        e->element = strdup(contact[0].nameaddr);
        e->value = strdup(contact[0].nameaddr);
        e->next = NULL;
        if (c->route->last_element) {
          c->route->last_element->next = e;
          c->route->last_element = e;
        }
        else {
          c->route->last_element = e;
        }
        if (c->route->elements == NULL) {
          c->route->elements = e;
        }
      }
      SIP_FreeContact(contact);

      if (c->route && c->route->elements &&
           c->route->elements->value) {
        c->remote_uri = URI_Parse(c->route->elements->value);
      }
      
    } /* Record-route is present */
  } /* Record-route handling */


  /* State machine processing */
  switch (c->state) {
    case SIPCALL_IDLE: /* No active call */
      debug("CallHandleResponse", "received %d in IDLE state\n", m->p.status);
      /*
      * Possible messages : 1xx-6xx for an outgoing re-INVITE.
      */
      if (!from_api && m->p.status >= 100 && m->p.status < 700) {
        if (m->p.status >= 200) {
          /*
           * Set the contact address if present
           */
          if (c->route && c->route->elements &&
               c->route->elements->value) {
            /* No change as already set */
          }
          else {
            sip_contact_t *contact = SIP_Contact(&m->p);
            if (contact != NULL && contact[0].uri.scheme != NULL) {
              c->remote_uri = URI_Copy(contact[0].uri);
            }
            /* else no change in contact */

            SIP_FreeContact(contact);        
          }

          /* You should send the ACK directly to
           * the contact address specified. XXX ! */
          if (m->p.status < 300) {
            /* Set the unauth count to 0 */
            c->unauth_count = 0;

            /* Send ACK only for 2xx responses,
             * client.c takes care of rest */
            if (rp != NULL) {
              char *method = HTTP_Method(&rp->p);
              if (method && strcasecmp(method, "INVITE")==0) {
                URI_Free(&rp->p.URI); 
                rp->p.URI = URI_Copy(c->remote_uri);

                debug("CallHandleResponse", "Enqueuing the ACK to request thread\n");
                MessageEnqueueRequestNew(&rp->mqueue, NULL,
                   nobody_handled, "ACK", rp, /* no body_t */ NULL, c->route);
                pthread_cond_broadcast(&rp->cond);
              }
              else {
                /* Non-INVITE request. Do not send ACK */
              }
            }
            else {
              /* We do not need to terminate the call */
              debug("CallHandleResponse", "Can not send ACK, no request\n");
            }
          }
        }

        if (m->p.status < 200) {
          /* provisional response, We ignore even 180 Alerting. */
          /* XXX should we have c->callobj->OnProvisionalResponse() */
          debug("CallhandleResponse", "ignoring non 100 provisional response\n");
        }
        else if (m->p.status < 300) {
          /* Successful response */
          if (m->body.content_type != NULL &&
               m->body.content != NULL) {

            if (!strcasecmp(m->body.content_type, "application/sdp")) {

              if (m->body.content_length <= 0) {
                m->body.content_length = strlen(m->body.content);
              }
              /* Should call a callback for change in SDP */
              int o_sdp_changed = 0;
              if (c->remote_sdp != NULL) {
                if (strcmp(m->body.content, c->remote_sdp) != 0) {
                  o_sdp_changed = 1;
                }
              }
              free_valid(c->remote_sdp);
              c->remote_sdp = strdupn(m->body.content, m->body.content_length);

              if (o_sdp_changed) {
                /* Inform the API that the sessiion is changed */
                if (c->callobj) {
                  SessionDescription *sdp = SIPCall::DeformatSDP(m->body.content, 0);
                  if (sdp != NULL) {
                    c->incoming_reinvite = 1; /* This is to avoid 
                                               * self initiation of
                                               * reINVITE. Otherwise
                                               * the two endpoint may go
                                               * in an infinite loop
                                               * changing this session description.
                                               * This must be fixed. XXX
                                               */
                    pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
                    c->callobj->OnRemoteSessionChanged(*sdp);
                    pthread_mutex_lock(&c->mutex);
                    c->incoming_reinvite = 0;
                  }
                  else {
                    debug("CallHandleResponse", "Could not deformat SDP\n");
                  }
                }
                else {
                  debug("CallHandleResponse", "No call object\n");
                }
              }
            }
            else {
              debug("CallHandleResponse", "media type in response is not supported\n");
              /* Should we close the call here,
               * or inform API of new media type and content 
               */
              /* Inform about any provisional body to the higher layer */
              parser_header_t *h = RFC822_Header(m->p.headers, "Content-Disposition");
              char *disposition = NULL;
              char *handling = NULL;
              if (h && h->elements->element) {
                disposition = h->elements->value;
                if (disposition && !*disposition) {
                  /* disposition is "" */
                  disposition = NULL;
                }

                handling = RFC822_Attribute1(h->elements->parameters, "handling");
                if (handling && !*handling) {
                  /* Handling is "" */
                  handling = NULL;
                }
              }
              SIPMessageBody msgbody(m->body.content,
                                     m->body.content_type, 
                                     m->body.content_length,
                                     disposition, 
                                     handling
                                    );

              if (c->callobj) {
                pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
                c->callobj->OnNewMessageBody(msgbody);
                pthread_mutex_lock(&c->mutex);
              }
              else {
                debug("CallHandleResponse", "No call object\n");
              }
            }
          }
          else {
            debug("CallHandleResponse", "No body in 2xx response\n");
          }

          /* Should we inform the API that the re-INVITE was
           * successful */
        }
        else if (m->p.status < 700) {
          /* Store the new contact address for later Initiates XXX */
          if (m->p.status < 400) {
            /* Note that we should not update the remote_uri with
             * the new contact here in case of
             * re_INVITE. */
          }

          if (m->body.content_type != NULL &&
               m->body.content != NULL) {
            /* Inform about any provisional body to the higher layer */
            parser_header_t *h = RFC822_Header(m->p.headers, "Content-Disposition");
            char *disposition = NULL;
            char *handling = NULL;
            if (h && h->elements->element) {
              disposition = h->elements->value;
              if (disposition && !*disposition) {
                /* disposition is "" */
                disposition = NULL;
              }

              handling = RFC822_Attribute1(h->elements->parameters, "handling");
              if (handling && !*handling) {
                /* Handling is "" */
                handling = NULL;
              }
            }
            SIPMessageBody msgbody(m->body.content,
                                   m->body.content_type, 
                                   m->body.content_length,
                                   disposition, 
                                   handling
                                  );

            if (c->callobj) {
              pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
              c->callobj->OnNewMessageBody(msgbody);
              pthread_mutex_lock(&c->mutex);
            }
            else {
              debug("CallHandleResponse", "No call object\n");
            }
          }

          /* Do we need to inform the API that the re-INVITE was failed.
           */
        }
      }
      else {
        debug("CallHandleResponse", "Message in unexpected state\n");
      }
      break;

    case SIPCALL_OUTGOINGCALL: /* waiting for final response from remote */
      debug("CallHandleResponse", "received %d in OUTGOING state\n", m->p.status);
      /*
       * possible messages: from_api - none
       *                    !from_api- 1xx-6xx.
       */
      if (!from_api && m->p.status >= 100 && m->p.status < 700) {
        if (m->p.status >= 200) {
          /*
           * Set the contact address if present
           */
          if (c->route && c->route->elements &&
               c->route->elements->value) {
            /* No change as already set */
          }
          else {
            sip_contact_t *contact = SIP_Contact(&m->p);
            if (contact != NULL && contact[0].uri.scheme != NULL) {
              c->remote_uri = URI_Copy(contact[0].uri);
            }
            else {
              /* else no change in contact */
            }
            SIP_FreeContact(contact);        
          }

          /* You should send the ACK directly to
           * the contact address specified. XXX ! */
          if (m->p.status < 300) {
            /* Set the unauth count to 0 */
            c->unauth_count = 0;

            /* Send ACK only for 2xx responses,
             * client.c takes care of rest */
            if (rp != NULL) {
              URI_Free(&rp->p.URI); 
              rp->p.URI = URI_Copy(c->remote_uri);

	      debug("CallHandleResponse", "Enqueueing the ACK To the transaction\n");
              MessageEnqueueRequestNew(&rp->mqueue, NULL,
                nobody_handled, "ACK", rp, /* no body_t */ NULL, c->route);
              pthread_cond_broadcast(&rp->cond);
            }
            else {
              /* Should we create a new request here XXX, call OnError
               * instead of OnHangup() */
              debug("CallHandleResponse", "Can not send ACK, no request\n");

              if (c->callobj) {
                pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
                if (c->callobj->OnHangup() == 0) {
                  c->callobj->mark_done = 1;
                  delete c->callobj;
                  c->callobj = NULL;
                }
                pthread_mutex_lock(&c->mutex);
                if (c->callobj == NULL) {
                  CallDone(c);
                }
              }
            }
          }
        }

        if (m->p.status < 200) {
          /* provisional response */
          if (m->p.status == 180) {
            if (c->callobj) {

              pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
              c->callobj->OnAlerting();
              pthread_mutex_lock(&c->mutex);
            }
            else {
              debug("CallHandleResponse", "No call object\n");
            }
          }
          else {
            /* XXX should we have c->callobj->OnProvisionalResponse() */
            debug("CallhandleResponse", "ignoring non 180 provisional response\n");
          }
        }
        else if (m->p.status < 300) {
          /* Successful response */
          c->state = SIPCALL_ACTIVE;

          if (m->body.content_type != NULL &&
               m->body.content != NULL) {

            if (!strcasecmp(m->body.content_type, "application/sdp")) {

              if (m->body.content_length <= 0) {
                m->body.content_length = strlen(m->body.content);
              }
              /* Should call a callback for change in SDP */
              int o_sdp_changed = 0;
              if (c->remote_sdp != NULL) {
                o_sdp_changed = 1;
              }
              free_valid(c->remote_sdp);
              c->remote_sdp = strdupn(m->body.content, m->body.content_length);

              if (o_sdp_changed) {
                /* XXX c->callobj->OnRemoteSessionChange(); 
                 * be sure to unlock and lock the c->mutex here. */
              }
            }
            else {
              debug("CallHandleResponse", "media type in response is not supported\n");
              /* Should we close the call here,
               * or inform API of new media type and content 
               */
            }
          }
          else {
            debug("CallHandleResponse", "No body in 2xx response\n");
          }

          if (c->callobj) {
            /* XXX update the SDP */
            pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
            c->callobj->OnCallEstablished();
            pthread_mutex_lock(&c->mutex);
	    debug("CallhandleResponse", "Got the lock\n");
          }
          else {
            debug("CallHandleResponse", "No call object\n");
          }
        }
        else if (m->p.status < 700) {
          int e = -1; /* Error */

          /* Store the new contact address for later Initiates XXX */
          if (m->p.status < 400) {
            /* store the contact as new remote_uri, what if
             * multiple URIs are present in contact XXX */
            if (c->route && c->route->elements &&
                c->route->elements->value) {
              /* No change as already set */
            }
            else {
              sip_contact_t *contact = SIP_Contact(&m->p);
              if (contact && contact[0].uri.scheme) {
                URI_Free(&c->remote_uri); /* should this be before if() XXX */
                c->remote_uri = URI_Copy(contact[0].uri);
              }
              SIP_FreeContact(contact);
            }
          }

          c->state = SIPCALL_IDLE;


          if (c->callobj) {
            http_authenticate_t auth;
            char *userid = NULL;
            char *password = NULL;
            char *auth_method;

            memset(&auth, 0, sizeof(auth));

            if ((m->p.status == 401 || m->p.status == 407)
                 && (c->unauth_count < 1)) {
              /* Authorization needed */
              if (m->p.status == 401) {
                auth = HTTP_Authenticate(&m->p);
              }
              else {
                auth = HTTP_ProxyAuthenticate(&m->p);
              }
              strassign(&auth.http_method, "INVITE");
              
              c->unauth_count ++;

              switch(auth.method) {
              case auth_basic:
                auth_method = strdup("basic");
                break;
              case auth_digest:
                auth_method = strdup("digest");
                break;
              case auth_none:
                auth_method = strdup("none");
                break;
              default:
                auth_method = strdup("none");
                break;
              }

              pthread_mutex_unlock(&c->mutex);
              if (m->p.status == 401) {
                e = c->callobj->OnAuthorizationNeeded(auth_method,
                                                      auth.realm,
                                                      userid,
                                                      password);
              }
              else {
                e = c->callobj->OnProxyAuthorizationNeeded(auth_method,
                                                      auth.realm,
                                                      userid,
                                                      password);
              }
              pthread_mutex_lock(&c->mutex);

              free_valid(auth_method);

              strassign(&c->userid, userid);
              strassign(&c->password, password);
            }

            if (e == 0) {
              char * authorization;
              char *uristr;
              
              uristr = URI_AsString(c->remote_uri);

              authorization = HTTP_BuildAuthorizationHeader(
                                                            &auth,
                                                            userid,
                                                            password,
                                                            uristr);

              /*
              debug("CallHandleResponse",
                    "Authorization is %s\n", (authorization != NULL) ?
                    authorization : "NULL");
              */

              free_valid(uristr);

              if (authorization) {
                /* Set header here */
                if (m->p.status == 401) {
                  /* replace the header if already exists */
                  HeaderReplaceOrAppend(&c->h_orq, "Authorization", 
                                        authorization); 
                }
                else {
                  /* replace the header if already exists */
                  /* There could be multiple proxy-authorizations.
                   * Should we use HeaderAppend() XXX */
                  HeaderAppend(&c->h_orq, 
                                        "Proxy-Authorization", 
                                        authorization); 
                }
                free_valid(authorization);

                /* 
                 * outgoing request. Invoke the request thread policy for
                 * the new INVITE 
                 */
                c->state = SIPCALL_OUTGOINGCALL;

                request_t *r;
                body_t body;
                body.content_length = 0;
                if (c->self_sdp == NULL) {
                  debug("CallhandleRequest", "No SDP specified\n");
                }
                else {
                  /* We are sending the self session always */
                  body.content_type = "application/sdp";
                  body.content = c->self_sdp;
                }
                
                r = FormatRequest("INVITE", c->type, c->af, 
                    c, ++c->lastcseq_tx, c->self_sdp ? &body : NULL);
                

                /* Should store the outgoing request here */
                HTTP_Duplicate(&c->p, &r->p);

                debug("CallHandleResponse", "Calling CreatePolicyThread\n");
                CallCreatePolicyThread(r, c->remote_uri, c->proxy);
              }
              else {
                e = -1;
              }
            }

            HTTP_AuthenticateFree(&auth);
          }
          else {
            debug("SipCallHandleResponse", "No call object\n");
          }


          if (m->body.content_type != NULL &&
               m->body.content != NULL) {
            /* Inform about any provisional body to the higher layer */
            parser_header_t *h = RFC822_Header(m->p.headers, "Content-Disposition");
            char *disposition = NULL;
            char *handling = NULL;
            if (h && h->elements->element) {
              disposition = h->elements->value;
              if (disposition && !*disposition) {
                /* disposition is "" */
                disposition = NULL;
              }

              handling = RFC822_Attribute1(h->elements->parameters, "handling");
              if (handling && !*handling) {
                /* Handling is "" */
                handling = NULL;
              }
            }
            SIPMessageBody msgbody(m->body.content,
                                   m->body.content_type, 
                                   m->body.content_length,
                                   disposition, 
                                   handling
                                  );

            if (c->callobj) {
              pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
              c->callobj->OnNewMessageBody(msgbody);
              pthread_mutex_lock(&c->mutex);
            }
            else {
              debug("CallHandleResponse", "No call object\n");
            }
          }

          if ((e != 0) && c->callobj) {
            pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
            if (c->callobj->OnCallRejected() == 0) {
              c->callobj->mark_done = 1;
              delete c->callobj;
              c->callobj = NULL;
            }
            pthread_mutex_lock(&c->mutex); 
            if (c->callobj == NULL) {
              CallDone(c);
            }
          }
          else {
            debug("CallHandleResponse", "No call object\n");
          }
        }
      }
      else {
        debug("CallHandleResponse", "Message in unexpected state\n");
      }
      break;

    case SIPCALL_INCOMINGCALL: /* waiting for API message */
      debug("CallHandleResponse", "received %d in INCOMING state\n", m->p.status);
      if (from_api && m->p.status >= 100 && m->p.status < 700) {
        if (rp == NULL) {
          debug("CallHandleResponse", "No original request\n");
        }
        else {
          if (m->p.status < 200) {
            /* Send the provisional response */
            MessageEnqueueResponseNew(&rp->mqueue, 
              NULL, nobody_handled, m->p.status, m->p.reason,
              rp, &c->h_ors, c->body.content ? &c->body : NULL);
            pthread_cond_broadcast(&rp->cond);
          }
          else if (m->p.status < 300) {
            c->state = SIPCALL_INCOMINGCALL_ACK;
            /* Send the succssful response to remote */
            body_t body;
            int sdp_mismatch = 0;
            char *sdp_mismatch_str = NULL;

            memset(&body, 0, sizeof(body));

            if (c->self_sdp == NULL && c->body.content == NULL) {
              /* Format a new default SDP */
              debug("CallHandleResponse", "No SDP is set, using default\n");
              if (c->callobj) {
                pthread_mutex_unlock(&c->mutex); /* to avoid deadlock */
                c->callobj->SetSelfSessionDescription(
                  SIPEndpoint::GetDefaultSessionDescription());
                pthread_mutex_lock(&c->mutex);
              }
              else {
                debug("CallHandleResponse", "No call object, can not set default\n");
              }
            }

            if (c->self_sdp) {
              body.content_type = "application/sdp";
              char * sdp = 
                CallMatchSessionWithSource(c->remote_sdp, c->self_sdp, 
                                                   &sdp_mismatch_str);
              /* Note that we do not update the local.
               * SDP here */
              body.content = sdp;
              body.content_length = ((sdp != NULL) ? strlen(sdp) : 0);
              /* Do not free sdp free_valid(sdp); */
              if (body.content == NULL) {
                /* sdp_mismatch indicates that the
                 * common subset can not be found.
                 * so we should respond with 4xx
                 * indicating the error.
                 */
                sdp_mismatch = 1;
                debug("CallHandleResponse", "An SDP mismatch\n");
              }
            }
            else if (c->body.content) {
              /* Set to user defined content */
              debug("Using user define content %s\n",
                (c->body.content_type ? c->body.content_type : 
                 (char *) "<no-mime-type>"));

              body = c->body;
            }
            else {
              body.content_type = NULL;
              body.content = NULL;
              body.content_length = 0;
            }

            /* Check for Accept header if present and send accordingly */
            if (!sdp_mismatch) { 
              if (CallCompareContentType(body.content_type, &rp->p) == true) { 
                /* Add the Contact header */
                IptelAddress contactaddr = CallExtractContact(c,false);
                /* XXX set contact as < > */
                HeaderFree(&c->h_ors); /* XXX is this correct to free h_ors here*/
                HeaderAppend(&c->h_ors, "Contact", 
                             (char *)(const char *) contactaddr);

                MessageEnqueueResponseNew(&rp->mqueue, 
                  NULL, nobody_handled, m->p.status, m->p.reason,
                  rp, &c->h_ors, body.content_type ? &body : NULL);
              }
              else {
                /* Should we always accept application/sdp ? */
                HeaderAppend(&c->h_ors, "Accept", body.content_type ?
                  body.content_type : (char *) "application/sdp" /* this is default */);
                DString str;
                DStringInit(&str);
                DStringPrintf(&str, "You do not accept %s !", body.content_type ?
                  body.content_type : "application/sdp");

                /* Should it be 406 or 415 !!! XXX. Should we add
                 * Other headers like Accept-Encoding and Accept-Language.
                 */
                MessageEnqueueResponseNew(&rp->mqueue,
                  NULL, nobody_handled, 415, DStringValue(&str), rp, &c->h_ors, NULL);
                DStringFree(&str);

                /* Inform the API about the call close */
                c->state = SIPCALL_IDLE;

                if (c->callobj) {
                  pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
                  if (c->callobj->OnHangup() == 0) {
                    c->callobj->mark_done = 1;
                    delete c->callobj;
                    c->callobj = NULL;
                  }
                  pthread_mutex_lock(&c->mutex); 
                  if (c->callobj == NULL) {
                    CallDone(c);
                  }
                }
                else {
                  debug("CallHandleResponse", "No call object\n");
                }              
              }
            }
            else {
                /* Should we always accept application/sdp ? */
                HeaderAppend(&c->h_ors, "Accept", "application/sdp");
                if (sdp_mismatch_str) {
                  /* If the warning are given add them. */
                  DString str;
                  DStringInit(&str);
                  DStringPrintf(&str, "\"%s\"", sdp_mismatch_str);
                  HeaderAppend(&c->h_ors, "Warning", DStringValue(&str));
                  DStringFree(&str);
                }

                /* Should it be 488 or 606!!! 
                 * I think, this is local to this UA hence 488.
                 * XXX should we add the warning as the body here
                 * in response.
                 */
                MessageEnqueueResponseNew(&rp->mqueue,
                  NULL, nobody_handled, 488, NULL, rp, &c->h_ors, NULL);

                /* Inform the API about the call close */
                c->state = SIPCALL_IDLE;

                if (c->callobj) {
                  pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
                  if (c->callobj->OnHangup() == 0) {
                    c->callobj->mark_done = 1;
                    delete c->callobj;
                    c->callobj = NULL;
                  }
                  pthread_mutex_lock(&c->mutex); 
                  if (c->callobj == NULL) {
                    CallDone(c);
                  }
                }
                else {
                  debug("CallHandleResponse", "No call object\n");
                }              
            }

            free_valid(sdp_mismatch_str);
            free_valid(body.content);

            pthread_cond_broadcast(&rp->cond);

            /*
             * We should do retransmission here. XXX.
             * the proxy code does not do retransmission for
             * this INVITE response 2xx.
             */
          }
          else {
            /* This is a rejecttion,
            * check if the existing error code is present,
            * e.g., for 401 then use that.
            */
            debug("CallHandleResponse", "Status is %d\n", rp->status);
            if (rp->status >= 400) {
              m->p.status = rp->status;
              strassign(&m->p.reason, rp->reason);
              HeaderMerge(&c->h_ors, &rp->headers);
            }

            c->state = SIPCALL_IDLE;

            /* Send the failure response to remote
             * Set the body if neeed */
            MessageEnqueueResponseNew(&rp->mqueue, 
              NULL, nobody_handled, m->p.status, m->p.reason,
              rp, &c->h_ors, c->body.content_type ? &c->body : NULL);
            pthread_cond_broadcast(&rp->cond);

            /*
             * Should we reset the c->body here ?
             * ANyway since this is terminating thecall,
             * does not matter.
             */
          }
        }
      }
      else {
        debug("CallHandleResponse", "Message in unexpected state, ignored\n");
      }
      break;

    case SIPCALL_INCOMINGCALL_ACK: /* waiting for ACK from remote */
      debug("CallHandleResponse", "received %d in INCOMING_ACK state\n", m->p.status);
      debug("CallHandleResponse", "Message in unexpected state, ignored\n");
      break;

    case SIPCALL_ACTIVE:
      /* Ignore any response in active state */
      debug("CallHandleResponse", "received %d in ACTIVE state\n", m->p.status);
      //BBB
      /* You should send the ACK directly to
       * the contact address specified. XXX ! */
      if (m->p.status < 300) {
	/* Set the unauth count to 0 */
	c->unauth_count = 0;
	
	/* Send ACK only for 2xx responses,
	 * client.c takes care of rest */
	if (rp != NULL) {
	  URI_Free(&rp->p.URI); 
	  rp->p.URI = URI_Copy(c->remote_uri);
	
	  debug("CallHandleResponse", "Enqueued ACK into tracsation\n");
	  MessageEnqueueRequestNew(&rp->mqueue, NULL,
	     nobody_handled, "ACK", rp, /* no body_t */ NULL, c->route);
	  pthread_cond_broadcast(&rp->cond);
	}
	else {
	  /* Should we create a new request here XXX, call OnError
	   * instead of OnHangup() */
	  debug("CallHandleResponse", "Can not send ACK, no request\n");
	  
	  if (c->callobj) {
	    pthread_mutex_unlock(&c->mutex); /* This is important to avoid deadlock */
	    if (c->callobj->OnHangup() == 0) {
	      c->callobj->mark_done = 1;
	      delete c->callobj;
	      c->callobj = NULL;
	    }
	    pthread_mutex_lock(&c->mutex);
	    if (c->callobj == NULL) {
	      CallDone(c);
	    }
	  }
	}
      }
      
      break;

    case SIPCALL_TERMINATING: /* call is being terminated */
      debug("CallHandleResponse", "received %d in TERMINATING state\n", m->p.status);
      /* Remove the request */
      /* set the state to IDLE irrespective or response received */
      c->state = SIPCALL_IDLE;
      break;

    default:
      error("CallHandleResponse", "surprising state %d\n", c->state);
      break;
  }

  if ( rp != NULL) {
    pthread_mutex_unlock(&rp->mutex);
  }

  m->handled = finished_handling;

  return 0;
}

int SIPCall::OnAuthorizationNeeded(const char *method,
                                   const char *realm,
                                   char *& user,
                                   char *& pass)
{
  if (userid != NULL && passwd != NULL) {
    user = strdup(userid);
    pass = strdup(passwd);
    return 0;
  }
  else {
    debug("OnAuthorizationNeeded", "No userid or password\n");
    return -1;
  }
}

int SIPCall::OnProxyAuthorizationNeeded(const char *method,
                                   const char *realm,
                                   char *& user,
                                   char *& pass)
{
  if (userid != NULL && passwd != NULL) {
    user = strdup(userid);
    pass = strdup(passwd);
    return 0;
  }
  else {
    debug("OnProxyAuthorizationNeeded", "No userid or password\n");
    return -1;
  }
}

