/* sipsdp.cpp -  SIP SDP implementation.
 * (c) 1999-2001 Columbia University. All rights reserved.
*/
extern "C" {
#include <string.h>       /* strdup */
#include <stdlib.h>       /* malloc */
#include <sys/time.h>     /* struct timeval */
#include <sys/types.h>    /* uint, ... */
#include <sys/socket.h>   /* socket */
#include <netinet/in.h>   /* AF_INET, ... */
#include <arpa/inet.h>    /* inet_addr, inet_ntoa */
#include <unistd.h>       /* gethostname */
}
#include <time.h>         /* time(), ... */

#include "iptelapi.h"     /* IptelAddress,Session etc */
#include "sipapi.h"       /* SIPCall, Endpoint etc */

extern "C" {
#include "strdupn.h"
#include "dstring.h"
#include "error.h"
}

/* Application chat is treated as text */
#define INTEROPERATE_WITH_SIPC 1

/* Memory management */
#define AllocateObject(type,count)    \
    (((count) == 1) ? (new type) : (new type[count]));
#define DeallocateObject(obj)  if (obj) delete (obj);

/* Default RTP ports start from here */
static unsigned short default_rtp_port = 13000;

SessionDescription *SIPEndpoint::GetDefaultSessionDescription()
{
  if (SIPEndpoint::default_session == NULL) {
    SessionDescription *session;
  
    session = /* Default Session */
      AllocateObject(SessionDescription,1);

    MediaInfo sessionRx;

    sessionRx.SetDataType(MediaInfo::Audio);
    sessionRx.AddFormat(0 /* G711 U law 64 K */);
    sessionRx.SetDirection(MediaInfo::Receive);
    sessionRx.SetRxAddress(IptelIPAddress(SIPLibrary::HostInAddr4));
    sessionRx.SetRxPort(default_rtp_port);
    session->AddSession(sessionRx);

    /* We will add two session descriptions - the first one
     * is added before. IPv6 is added, only if our IPv6 
     * support is usable.
    */
    if (AFIIPv6Active()) {
      sessionRx.SetAddressType(MediaInfo::AddressIP6);
      sessionRx.SetRxAddress(IptelIPAddress(SIPLibrary::HostInAddr6));
      sessionRx.SetRxPort(default_rtp_port);
      session->AddSession(sessionRx);
    }

    default_rtp_port += 2;
    
    return session;
  }
  else {
    return SIPEndpoint::default_session->Clone();
  }
}


void SIPEndpoint::SetDefaultSessionDescription(SessionDescription & s)
{
  if (SIPEndpoint::default_session) {
    delete SIPEndpoint::default_session;
  }
  SIPEndpoint::default_session = s.Clone();
}

char *SIPCall::FormatSDP(SessionDescription & sessionList, int *length)
{
  int i, j;
  char *value;
  char dotted[256];

  char addr4[] = "IP4";
  char addr6[] = "IP6";
  char *addrptr; 
  
  DString sdp;
  DStringInit(&sdp);

  /* Set the globals here */
  if (sessionList.headers["v"] == NULL) {
    sessionList.headers += GenericHeaderValue("v", "0");
  }
  if (sessionList.headers["o"] == NULL) {
    unsigned long sdpVersion = 0;
    int af;
    sdpVersion = rand();

    DString oline;
    DStringInit(&oline);
    

    /* find if there is atleast one IPv4. if it is not, default to IPv6 */
    MediaInfo session(MediaInfo::UnknownDirection, MediaInfo::UnknownData,
                      MediaInfo::UnknownTransport);
    
    if (sessionList.FindSessionIndex(session) < 0) {
      /* there is none with IPv4 */
      af = AF_INET6;
      addrptr = addr6;
      inet_ntop(af, &SIPLibrary::HostInAddr6, dotted, sizeof(dotted));
    }
    else {
      af = AF_INET;
      addrptr = addr4;
      inet_ntop(af, &SIPLibrary::HostInAddr4, dotted, sizeof(dotted));
    }
    
    DStringPrintf(&oline, "%s %lu %lu IN %s %s",
      (af == AF_INET) ? SIPLibrary::HostName4 : SIPLibrary::HostName6,
      (unsigned long) sessionList.GetSessionId(),
      (unsigned long) sdpVersion, addrptr, dotted);
    
    sessionList.headers += GenericHeaderValue("o", DStringValue(&oline));
    DStringFree(&oline);
  }
  if (sessionList.headers["s"] == NULL) {
    sessionList.headers += GenericHeaderValue("s", "SIP Library call");
  }
  if (sessionList.headers["t"] == NULL) {
    DString timeheader;
    DStringInit(&timeheader);
    /*
     * t= line has NTP time for start of session.
     * Add 2208988800 to Unix time to get NTP.
     * This number is given in SDP RFC 
     */
    DStringPrintf(&timeheader, "%lu 0", (unsigned long) time(0) +
                  2208988800LU);
    sessionList.headers += GenericHeaderValue("t", DStringValue(&timeheader));
    DStringFree(&timeheader);
    /* default for SIP is t=<localtime> 0 */
  }
  
  DStringPrintf(&sdp,"%s=%s\r\n", "v", (const char *)
                sessionList.headers["v"]);
  DStringPrintf(&sdp,"%s=%s\r\n", "o", (const char *)
                sessionList.headers["o"]);
  DStringPrintf(&sdp,"%s=%s\r\n", "s", (const char *)
                sessionList.headers["s"]);

  const char *iline = (const char *) sessionList.headers["i"];
  const char *uline = (const char *) sessionList.headers["u"];
  const char *eline = (const char *) sessionList.headers["e"];

  if (iline) {
    DStringPrintf(&sdp, "%s=%s\r\n", "i", iline);
  }
  if (uline) {
    DStringPrintf(&sdp, "%s=%s\r\n", "u", uline);
  }
  if (eline) {
    DStringPrintf(&sdp, "%s=%s\r\n", "e", eline);
  }

  bool oAddressAdded = false;
  IptelIPAddress mediaAddress;
  bool oSingleMediaTransport = true;

  for (i=0; i<sessionList.GetMediaCount(); i++) {
    MediaInfo & media = sessionList[i];
    if (!mediaAddress.IsValid()) {
      /* First media info */
      mediaAddress = media.GetRxAddress();
      mediaAddress.SetPort(0);
    }
    else if (!(mediaAddress == media.GetRxAddress())) {
      /* Media info is different for this */
      oSingleMediaTransport = false;
    }
  }

  if (oSingleMediaTransport) {
    /* All media transport addresses are same,
     * put c= line before any m= line 
     */
    if (mediaAddress.IsIP4()) {
      in_addr temp;
      temp = (in_addr)mediaAddress;
      inet_ntop(AF_INET, &temp, dotted, sizeof(dotted));
      addrptr = addr4;
    } 
    else {
      in6_addr temp;
      temp = (in6_addr)mediaAddress;
      inet_ntop(AF_INET6, &temp, dotted, sizeof(dotted));
      addrptr = addr6;
    }

    DStringPrintf(&sdp,"c=IN %s %s", addrptr, dotted);
    if (mediaAddress.IsMulticast()) {
      DStringPrintf(&sdp, "/%d", 127 /* default is 127 */);
    }
    DStringAppend(&sdp,"\r\n", -1);
    oAddressAdded = true;
  }

  DStringPrintf(&sdp,"%s=%s\r\n", "t", (const char *)
                sessionList.headers["t"]);

  for (i=0; i<sessionList.GetMediaCount(); i++) {
    char *p;
    MediaInfo & media = sessionList[i];

    MediaInfo::Direction dir = media.GetDirection();

    if ((dir == MediaInfo::Receive) ||
         (dir == MediaInfo::TransmitAndReceive)) {

      MediaInfo::DataType datatype = media.GetDataType();
      MediaInfo::TransportType transport = media.GetTransportType();

      switch (datatype) {
      case MediaInfo::Audio:
        DStringPrintf(&sdp, "m=audio");
        break;
      case MediaInfo::Video:
        DStringPrintf(&sdp, "m=video");
        break;
      case MediaInfo::Data: /* This is not defined by RFC2327 SDP */
        DStringPrintf(&sdp, "m=data");
        break;
      case MediaInfo::Application:

#ifdef INTEROPERATE_WITH_SIPC
        DStringPrintf(&sdp, "m=text");
#else
        DStringPrintf(&sdp, "m=application");
#endif
        break;
      case MediaInfo::Control:
        DStringPrintf(&sdp, "m=control");
        break;
      default:
        /* Any new application can be added */
        p = (char *) (const char *) media.parameters["datatype"];
        DStringPrintf(&sdp, "m=%s", ((p && *p) ? p : "application"));
        break;
      }

      /* Set port number */
      DStringPrintf(&sdp, " %u ", (unsigned int) media.GetRxPort());

      /* Set the transport */
      switch(transport) {
      case MediaInfo::RTP_AVP_Transport:
        DStringPrintf(&sdp, "RTP/AVP");
        break;
      case MediaInfo::TCP_Transport:
        DStringPrintf(&sdp, "tcp");
        break;
      case MediaInfo::UDP_Transport:
        DStringPrintf(&sdp, "udp");
        break;
      case MediaInfo::H323_Transport:
        DStringPrintf(&sdp, "H.323");
        break;
      default:
        /* Default is UDP transport */
        p = (char *) (const char *) media.parameters["transport"];
        DStringPrintf(&sdp, "%s", ((p && *p) ? p : "udp"));
        break;
      }

      /* Set the format list */
      for (j=0; j<media.GetFormatCount(); j++) {
        FormatInfo & format = media[j];
        if ((datatype == MediaInfo::Audio) ||
             (datatype == MediaInfo::Video)) {
          DStringPrintf(&sdp," %d", format.GetPayloadType());
        }
        else {
          const char *name = format.GetFormatName();
          if (name != NULL) {
            /* Encode the application name, here */
            DStringPrintf(&sdp, " %s", name);
          }
          else {
            /* Just encode the payloadtype */
            DStringPrintf(&sdp, " %d", format.GetPayloadType());
          }
        }
      }
      if (media.GetFormatCount() <= 0) {
        /* There is no format specified. This
         * is not allowed in SDP. Set 0 for the time being.
         * This needs to be modified later. XXX */
        DStringPrintf(&sdp, " 0");
      }

      DStringAppend(&sdp, "\r\n",-1);

      /* Set the dynamic payload type list */
      if ((datatype ==MediaInfo::Audio) ||
           (datatype == MediaInfo::Video)) {
        for (j=0; j<media.GetFormatCount(); j++) {
          FormatInfo & format = media[j];
          int payloadtype = format.GetPayloadType();
          /* >=96 for dynamic payload types */
          /*
           * Always include the definition; XXX is this correct?
           */
          if (1 /* payloadtype >= 96 */) {
            const char *name = format.GetFormatName();
            int channelcount = format.GetChannelCount();
            int clockrate = format.GetClockRate();
            if (name != NULL) {
              DStringPrintf(&sdp, "a=rtpmap:%d %s/%d",
                            payloadtype,
                            name, clockrate);
              if (channelcount > 1) {
                DStringPrintf(&sdp, "/%d", channelcount);
              }
              DStringAppend(&sdp, "\r\n",-1);
            }
          }
        }
      }


      /* Add the address if not already added */
      if (!oAddressAdded) {
        IptelIPAddress ipaddr;
        ipaddr = media.GetRxAddress();
        if (ipaddr.IsValid() && 
            media.GetNetworkType() == MediaInfo::NetworkIN){
    
          if (media.GetAddressType() == MediaInfo::AddressIP6) {
            struct in6_addr temp; /* the anonymous variable is needed to avoid
                                   * a c++ warning in some compilers. */
            addrptr = addr6;
            temp = (in6_addr)ipaddr;
            inet_ntop(AF_INET6, &temp, dotted, sizeof(dotted));
          }
          else {
            struct in_addr temp;
            addrptr = addr4;
            temp = (in_addr)ipaddr;
            inet_ntop(AF_INET, &temp, dotted, sizeof(dotted));
          }
          
          DStringPrintf(&sdp,"c=IN %s %s", addrptr, dotted);
            
          if (ipaddr.IsMulticast()) {
            DStringPrintf(&sdp, "/%d", media.GetRxTtl());
          }

          DStringAppend(&sdp,"\r\n", -1);
        }
        else {
          //cerr << "ipaddress is invalid " << ipaddr << endl;
        }
      }
    }
  }

  /* Return the char array */
  value = strdup(DStringValue(&sdp));
  
    
  /* Update *length */
  if (length) {
    *length = DStringLength(&sdp);
  }
  DStringFree(&sdp); /*  Free DString */
  return value;
}

SessionDescription *SIPCall::DeformatSDP(const char *sdp,int length)
{
  if (sdp == NULL) {
    debug("DeformatSDP", "SDP is NULL\n");
    /* No change */
    return NULL;
  }

  SessionDescription *remotesession;
  remotesession = AllocateObject(SessionDescription,1);

  char *t = NULL;
  char *lasts, *line;
  int last_media_index = -1;

  /* Allocate new string, to avoid overiting the parameter */
  if (length <= 0) {
    /* Length is invalid, assume a null terminated string */
    t = strdup(sdp);
  }
  else {
    t = (char *)malloc(sizeof(char) * (length+1));
    memcpy(t, sdp, length);
    t[length] = 0;
  }

  char *original_t = t;

  lasts = NULL;

  IptelIPAddress globalIpAddress;
  MediaInfo media;
  bool oAddMedia = false;
  media.SetAddressType(MediaInfo::UnknownAddress);

  while ((line = strtok_r(t, "\r\n", &lasts)) != NULL) {
    t = NULL; /* This is very important */
    if (!strncmp(line, "c=", 2)) {
      /* c=IN IP4 a.b.c.d/ttl/numofaddress */
      char *l2 = NULL;
      char *ptr;
      char *net_str = NULL;
      char *addr_str = NULL;

      line += 2;

      net_str = strtok_r(line, " ", &l2); /* ignore IN */
      addr_str = strtok_r(NULL, " ", &l2); /* ignore IP4/IP6 */

      if (strcmp(net_str, "IN") != 0) {
        media.SetNetworkType(MediaInfo::UnknownNetwork);
        media.parameters += GenericHeaderValue("network", net_str);
      }
      if (strcmp(addr_str, "IP4") == 0) {
        media.SetAddressType(MediaInfo::AddressIP4);
        media.parameters += GenericHeaderValue("addresstype", addr_str);
      }
      if (strcmp(addr_str, "IP6") == 0) {
        media.SetAddressType(MediaInfo::AddressIP6);
        media.parameters += GenericHeaderValue("addresstype", addr_str);
      }
      ptr = strtok_r(NULL, " ", &l2); /* get the address */
      if (ptr != NULL) {
        char *addr, *ttlstr;
        l2 = NULL;
        addr = strtok_r(ptr, "/", &l2); /* get IP address */
        ttlstr = strtok_r(NULL, "/ ", &l2); /* get TTL */
        /* Ignore numofaddresses */
        IptelIPAddress ipaddress = addr;
       
        if (last_media_index != -1) {
          /* Set the media specific IP address */
          unsigned short port = media.GetRxPort();
          media.SetRxAddress(addr);
          media.SetRxPort(port);
          if (ttlstr != NULL) {
            media.SetRxTtl(atoi(ttlstr));
          }
        }
        else {
          /* Set the globalIpAddress, c= line before any media */
          globalIpAddress = ipaddress;
        }
      }
    } 
    else if (!strncmp(line, "m=", 2)) {
      /* Add previous m= line first */
      if (oAddMedia) {
        if (!media.GetRxAddress().IsValid()) {
          IptelIPAddress ipaddr = globalIpAddress;
          ipaddr.SetPort(media.GetRxPort());
          media.SetRxAddress(ipaddr);
        }
        /* If the remote sends us a v6 sdp and we dont
         * have that capability, ignore it.
        */
        if (media.GetRxAddress().IsIP4()  ||
          (media.GetRxAddress().IsIP6() && AFIIPv6Active())) {
          remotesession->AddSession(media);  
        }
        
        media = MediaInfo(); /* Allocate new */
        oAddMedia = false;
      }

      /* m=audio port RTP/AVP 0 4 8 */
      char *l2 = NULL;
      char *ptr;
      char *mediastr, *portstr, *transportstr;

      line += 2;

      mediastr = strtok_r(line, " ", &l2); /* audio/video */
      if (mediastr) {
        if (!strcasecmp(mediastr, "audio")) {
          media.SetDataType(MediaInfo::Audio);
        }
        else if (!strcasecmp(mediastr, "video")) {
          media.SetDataType(MediaInfo::Video);
        }
        else if (!strcasecmp(mediastr, "application")) {
          media.SetDataType(MediaInfo::Application);
        }
        else if (!strcasecmp(mediastr, "text")) {
          debug("DeformatSDP", "Adding text\n");
          media.SetDataType(MediaInfo::Application);
        }
        else if (!strcasecmp(mediastr, "data")) {
          media.SetDataType(MediaInfo::Data);
        }
        else if (!strcasecmp(mediastr, "Control")) {
          media.SetDataType(MediaInfo::Control);
        }
        else {
          /* Default is Unknown data */
          media.SetDataType(MediaInfo::UnknownData);
          media.parameters += GenericHeaderValue("datatype", mediastr);
        }

        portstr = strtok_r(NULL, " ", &l2); /* port */
        media.SetRxPort(portstr ? atoi(portstr) : 0);

        transportstr = strtok_r(NULL, " ", &l2); /* RTP/AVP, udp... */
        if (transportstr) {
          if (!strcasecmp(transportstr, "RTP/AVP") ||
               !strcasecmp(transportstr, "RTP")) {
            media.SetTransportType(MediaInfo::RTP_AVP_Transport);
          }
          else if (!strcasecmp(transportstr, "tcp")) {
            media.SetTransportType(MediaInfo::TCP_Transport);
          } 
          else if (!strcasecmp(transportstr, "udp")) {
            media.SetTransportType(MediaInfo::UDP_Transport);
          }
          else if (!strcasecmp(transportstr, "H.323")) {
            media.SetTransportType(MediaInfo::H323_Transport);
          }
          else {
            /* Default is Unknwon Transport */
            media.SetTransportType(MediaInfo::UnknownTransport);
            media.parameters += GenericHeaderValue("transport", transportstr);
          }
        }

        while ((ptr = strtok_r(NULL, " ", &l2)) != NULL &&
                *ptr != '\0') {
          MediaInfo::DataType datatype = media.GetDataType();
          if ((datatype == MediaInfo::Audio) ||
               (datatype == MediaInfo::Video)) {
            /* get the payload type */
            FormatInfo format;
            format.SetPayloadType(atoi(ptr));
            media.AddFormat(format);
          }
          else {
            /* Get the application, data, controlname */
            FormatInfo format;
            format.SetPayloadType(-1);
            format.SetFormatName(ptr);
            media.AddFormat(format);
          }
        }

        oAddMedia = true;
      }

    }
    else if (! strncmp(line, "a=rtpmap:", 9)) {
      /* a=rtpmap:96 pcmu/8000 , Dynamic payload type...*/
      /* XXX note that a=rtpmap: can occur any where in the
      * SDP body, do not assume that m= line
      * is followed by all relevant a=rtpmap lines...
      * Implement this using a table with mapping of all
      * dynamic payload types.
      * Update all the formats in the end using this table.
      */
      char *l2=NULL;
      char *ptstr, *descr;

      line += 9;
      ptstr = strtok_r(line, " ", &l2);
      descr = strtok_r(NULL, " ", &l2);
      if (oAddMedia) {
        /* Search for this payload type in existing list */
        int ptint = ptstr ? atoi(ptstr) : -1;
        bool oFound = false;
        for (int idx=0; idx < media.GetFormatCount(); idx++) {
          if (ptint == (int)media[idx]) {
            oFound = true;
            break;
          }
        }
        if (oFound) {
          char *cnlstr, *clockstr, *descriptstr;
          char *l5 = NULL;
          descriptstr = strtok_r(descr, "/", &l5);
          clockstr = strtok_r(NULL, "/", &l5);
          cnlstr = strtok_r(NULL, "/ ", &l5);

          media.RemoveFormat(ptint);
          FormatInfo format(ptint, descriptstr, 
                            clockstr ? atol(clockstr) : 8000,
                            cnlstr ? atoi(cnlstr) : 1);

          media.AddFormat(format);
        }
      }
    }
    else {
      /* ignore, or just stack it up... */
      if (oAddMedia == false) {
        /* global attributes */
        remotesession->headers += GenericHeaderValue(line, '=');
      }
      else {
        /* media specific attributes */
        media.headers += GenericHeaderValue(line, '=');
      }
    }
  }

  if (oAddMedia) {
    if (!media.GetRxAddress().IsValid()) {
      IptelIPAddress ipaddr = globalIpAddress;
      ipaddr.SetPort(media.GetRxPort());
      media.SetRxAddress(ipaddr);
    }
    /* If the remote sends us a v6 sdp and we dont
     * have that capability, ignore it.
    */
    if (media.GetRxAddress().IsIP4()  ||
       (media.GetRxAddress().IsIP6() && AFIIPv6Active())) {
      remotesession->AddSession(media);  
      oAddMedia = false;
    }
  }

  if (original_t != NULL) {
    free(original_t);
  }

  return remotesession;
}
