/*
 *  authenticate.c
 *  
 *  Used for authenticate the SIP client
 *
 *  static global variable:
 *    dict_h nonceHandle;
 *
 *  Functions for external usage:
 *    http_authenticate_t authenticatea  *authMethod, 
 *                                       http_authorization_t auth,
 * 
 *  Hash Functions:
 *    static ht_val auth_ht_value(dict_obj o);
 *    static int auth_ht_comp(dict_obj obj1, dict_obj obj2);
 * 
 *  Functions for Basic Authentication:
 *    static int authenticateBasicPwdMatch(char *userID,char *passwd, 
 *    static int authenticateBasicMatch(char *userID, char *passwd);
 *    int authenticateBasic(char *userID, char *passwd);
 *
 *  Function for Digest Authentication:
 *    static char *authenticateDigestNonce(char *privateKey);
 *    static char *authenticateDigestNonceMatch(char *nonce);
 *    static char *authenticateDigestResponse(http_authorization_t auth,
 *                                    char *password);
 *    static char *authenticateDigest(http_authorization_t auth);
 *
 *  Xiaotao Wu, Henning Schulzrinne, Tarun Kapoor
 *  Last modified: 12/01/00
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/utsname.h>
#include <netdb.h> /* MAXHOSTNAMELEN */
#include <pthread.h>

#include "sysdep.h" /* Define the system dependencies */
#include "authenticate.h"
#include "dict.h"
#include "ht.h"
#include "global.h"
#include "md5.h"
#include "base64.h"
#include "http.h"
#include "strdupn.h"
#include "config.h"
#include "error.h"
#include "querydb.h"

#define STALETIME 60
#define HT_ENTRY 1000
#define BUCKET_ENTRY 10

static dict_h nonceHandle = NULL;

/* Define to print out debugging information about digest authentication */
/* #define DEBUG_DIGEST */

/*
 * Used to get the value of the string for hashing
 * The value is calculated as the summary of ASCII value of each 
 * character in the name string
 */
static ht_val auth_ht_value(dict_obj o)
{
  char *s;
  ht_val r=0;

  s = (char *)o;
  while (*s != ':') {
    r += *s;
    ++s;
  }

  return r;
} /* auth_ht_value */


/*
 *  Used to compare the value in hash table
 *  We need to search the record based on the name
 *  If the name match the string, we return match.
 *  The further work for matching the password will be done outside
 *  hashtable so that we can distinguish the different errors.
 *
 *  return -1 : The first object's name part less than the second one.
 *  return 0  : The first object's name part equal the second one.
 *  return 1  : The first object's name part greater than the second one.
 *
 *  suppose the userID is unique in the file 
 *
 */
static int auth_ht_comp(dict_obj obj1, dict_obj obj2)
{
  char *s, *t;

  s = (char *)obj1;
  t = (char *)obj2;

  while ((*s != ':') && (*t != ':')) {
    if (*s > *t)
      return 1;
    if (*s < *t)
      return -1;
    ++s;
    ++t;
  }

  if (*s != ':')
    return 1;
  if (*t != ':')
    return -1;
  return 0;
} /* auth_ht_comp() */

/*
 * authenticateBasicPwdMatch()
 * Verify that the password 'compareStr' matches the hashed password 'hash',
 * given the specified userid, realm and algorithm.
 * Return an appropriate auth_status.
 */
static auth_status authenticateBasicPwdMatch(const char *userid,
                                             const char *hash,
                                             const char *realm,
                                             const char *algorithm,
                                             const char *passwd)
{
  DString compare_val;
  char *compareStr;
  auth_status ret;

  if (strcompare(algorithm, "MD5", 1) != 0) {
    /* Unsupported algorithm */
    return auth_wrong_algorithm;
  }

  DStringInit(&compare_val);

  DStringPrintf(&compare_val, "%s:%s:%s", userid, realm, passwd);

  compareStr = MD5(DStringValue(&compare_val));

  DStringZero(&compare_val);
  DStringFree(&compare_val);

  if (strcompare(compareStr, hash, 0) == 0) {
    ret = auth_success;
  }
  else {
    ret = auth_wrong_password;
  }

  free_valid(compareStr);
  return ret;
} /* authenticateBasicPwdMatch */


/*
 * authenticateBasicMatch()
 * See whether there the user entry 'user' matches the userid and passwd
 * in 'auth'.
 * Return an appropriate auth_status.
 */
static auth_status authenticateBasicMatch(const put_table_entry *user,
                                          http_authorization_t auth)
{
  /* Defensive */
  if (user == NULL || auth.userid == NULL)
    return auth_no_user;
  if (auth.password == NULL)
    return auth_wrong_password;

  return authenticateBasicPwdMatch(auth.userid, user->hash_value,
                                   user->realm, user->algorithm,
                                   auth.password);
} /* authenticateBasicMatch() */


/*
 * authenticateBasic()
 * Do a Basic authentication on the user.
 */
static auth_status authenticateBasic(const put_table_entry *user,
                                     http_authorization_t auth)
{
  return authenticateBasicMatch(user, auth);
} /* authenticateBasic */


/*
 * Used to generate the nonce value
 * The nonce value will be:
 * MD5(Time-Stamp:PrivateKey)
 * The Time-Stamp and the nonce value will be stored in hash table
 * return the nonce string.
 */
static char *authenticateDigestNonce(const char *privateKey)
{
  char *timestamp;
  char *dstr;
  char *nonce;
  char *s;
  struct ht_args argsp =
     {HT_ENTRY, BUCKET_ENTRY, auth_ht_value, auth_ht_value};

  timestamp = (char *)malloc(20*sizeof(char));
  sprintf(timestamp, "%d", (unsigned)time(NULL));
  dstr = (char *)malloc(strlen(timestamp) + strlen(privateKey) + 2);
  strcpy(dstr, timestamp);
  strcat(dstr, ":");
  strcat(dstr, privateKey);
  nonce = MD5(dstr);
  free(dstr);

  s = malloc(strlen(nonce) + strlen(timestamp) + 2);
  sprintf(s, "%s:", nonce);
  strcat(s, timestamp);

  /* create hash table and insert the record */
  if (nonceHandle == NULL) {
    nonceHandle = ht_create(auth_ht_comp, auth_ht_comp, 
                            DICT_UNIQUE_KEYS, (int *)0, &argsp);
  }
  ht_insert(nonceHandle, (dict_obj)s);

  free(timestamp);
  return nonce;
} /* authenticateDigestNonce() */


/*
 * authenticateDigestNonceMatch()
 * Match nonce in database.
 * Returns auth_no_user if the nonce doesn't exist.  A bit dubious, but
 * whatever.
 */
static auth_status authenticateDigestNonceMatch(char *nonce)
{
  char *nonce_new;
  char *r = NULL;
  unsigned timestamp;

  nonce_new = (char *)malloc(strlen(nonce) + 2);
  strcpy(nonce_new, nonce);
  strcat(nonce_new, ":");
  if (nonceHandle != NULL)
    r = (char *)ht_search(nonceHandle, (dict_obj)nonce_new);
  free(nonce_new);

  if (r == NULL)
    return auth_no_user;

  timestamp = atoi((char *)(strchr(r, ':') + 1));
  if ((unsigned)time(NULL) - timestamp > STALETIME)
    return auth_nonce_stale;

  return auth_success;
} /* authenticateDigestNonceMatch */


/* Wrapper for a string so that it won't be a NULL pointer */
#define NONNULL(x) ((x) != NULL ? (x) : "")

/*
 * authenticateDigestResponse()
 * Generate a correct response-digest, based on the auth info.
 * Caller must free the returned string.
 */
static char *authenticateDigestResponse(http_authorization_t auth, 
                                        const char *HA1)
{
  DString A2, KD;
  char *HA2, *HKD;

  DStringInit(&A2);
  DStringInit(&KD);

  DStringPrintf(&A2, "%s:%s",
		NONNULL(auth.http_method), NONNULL(auth.digest_uri));
  HA2 = MD5(DStringValue(&A2));

#ifdef DEBUG_DIGEST
  debug("authenticateDigestResponse",
        "HA1 = \"%s\", A2 = \"%s\", HA2 = \"%s\"\n",
        HA1, DStringValue(&A2), HA2);
#endif

  DStringFree(&A2);
  
  DStringPrintf(&KD, "%s:%s:%s", HA1, NONNULL(auth.nonce), HA2);
  HKD = MD5(DStringValue(&KD));

#ifdef DEBUG_DIGEST
  debug("authenticateDigestResponse",
        "KD = \"%s\", HKD = \"%s\"\n", DStringValue(&KD), HKD);
#endif

  free(HA2);
  DStringFree(&KD);

  return HKD; 
} /* authenticateDigestResponse */


/*
 * authenticateDigest()
 * Perform digest authentication.
 */
static auth_status authenticateDigest(const put_table_entry *user,
                                      http_authorization_t auth)
{
  auth_status ret;
  char *resp;

#ifdef DEBUG_DIGEST
  debug("authenticateDigest", "auth.userid == %s\n",
        auth.userid != NULL ? auth.userid : "(NULL)");
#endif

  if (auth.userid == NULL) {
    return auth_no_user;
  }
 
  if (user == NULL) {
    return auth_no_user;
  }

  if (strcompare(user->algorithm, auth.algorithm, 1) != 0 &&
      /* "MD5" is the default algorithm */
      !(auth.algorithm == NULL &&
        strcompare(user->algorithm, "MD5", 1) == 0)) {
    return auth_wrong_algorithm;
  }

  if (strcompare(user->realm, auth.realm, 0) != 0) {
    return auth_wrong_realm;
  }

  /* Get the value the response string is supposed to have */
  resp = authenticateDigestResponse(auth, user->hash_value);
  
  if (strcmp(resp, auth.response_digest) != 0) {
    /* Wrong password */
    ret = auth_wrong_password;
  }
  else if (authenticateDigestNonceMatch(auth.nonce) != 0) {
    ret = auth_nonce_stale;
  }
  else {
    ret = auth_success;
  }
  free_valid(resp);
  
  return ret;
} /* authenticateDigest */


/*
 * authenticate()
 * Used to do authentication, based on the auth.method to decide
 * which authentication method will be used: Basic or Digest
 */
auth_status authenticate(http_authentication_method_t authMethod,
                         const put_table_entry *user,
                         http_authorization_t authorization) 
{
  if (authorization.method != authMethod) {
    return auth_wrong_method;
  }
  if (user == NULL) {
    return auth_no_user;
  }
  if (authMethod == auth_basic) {
    return authenticateBasic(user, authorization);
  }
  else if (authMethod == auth_digest) {
    int m = authenticateDigest(user, authorization);
    if (m != auth_success) {
      debug("authenticate", "authenticateDigest failed, m = %d\n", m);
    }
    return m;
  }
  else {
    /* Confused? Shouldn't get here for other auth methods.*/
    return auth_internal_error;
  }
} /* authenticate */


/*
 * build_basic_auth_challenge()
 * Build a challenge for Basic authentication for the user 'user'.
 * (The only necessary fields in a Basic authentcation are the method and the
 * realm).
 * Return false on failure.
 */
int build_basic_auth_challenge(http_authenticate_t *authen,
                               const put_table_entry *user)
{
  if (user->realm == NULL)
    return 0;

  authen->method = auth_basic;
  authen->realm = strdup(user->realm);

  return 1;
} /* build_basic_auth_challenge */


/*
 * build_digest_auth_challenge()
 * Build a challenge for Digest authentication for the user userid.
 * Return false on failure.
 */
int build_digest_auth_challenge(http_authenticate_t *authen,
                                const char *opaque,
                                int wasStale,
                                const char *privateKey,
                                const put_table_entry *user)
{
  if (user->realm == NULL || user->algorithm == NULL) {
    return 0;
  }

  authen->method    = auth_digest;
  authen->realm     = strdup(user->realm);
  authen->algorithm = strdup(user->algorithm);
  authen->domain    = strdup("");
  authen->nonce     = authenticateDigestNonce(privateKey);
  authen->opaque    = strdup(opaque);
  authen->stale     = strdup(wasStale ? "TRUE" : "FALSE");

  return 1;
} /* build_digest_auth_challenge */


/*
 * A NONCE HANDLING SYSTEM
 *
 * gNonceHead list is used to add the sipd-generated nonces.  they are added
 * at the beginning of the list by the AddNonce().  evtTime= time() +
 * theLifeTime. time() is called after the mutex is acquired. Therefore the
 * list is ordered; this helps in the deletion of the expired nonces. If a
 * nonce is expired the rest of them to the end of the list are
 * expired. Thus deletion can happen asynchronously and the mutex can be
 * released early.
 * 
 * gCallIdCseqDateNonceHead is used for nonces sent by the clients which are
 * the concatenation of CallId CSeq and Date header values The evtTime=
 * theTimeOccurred and are added by order of theTimeOccurred in the
 * list. Use the NonceFoundElseAddTimeOrdered() with this list.  Therefore
 * the list is ordered; this helps in the deletion of the expired nonces.
 * If a nonce is expired the rest of them to the end of the list are
 * expired.  Thus deletion can happen asynchronously and the mutex can be
 * released early.  If the nonce is found in the list then it must be
 * rejected.
 */


NonceHead_t gNonceHead = {NULL, NULL, PTHREAD_MUTEX_INITIALIZER};
NonceHead_t gCallIdCseqDateNonceHead = {NULL, NULL, PTHREAD_MUTEX_INITIALIZER};


/*
 * NonceFoundElseAddTimeOrdered()
 *
 * Use the following function for gCallIdCseqNonceHead ONLY
 * The evtTime is the time of the date header 
 * If it finds the nonce it returns a non-zero value, othewise a zero value
 * and adds the nonce in the list ordered by theTimeOccurred
 * *theData supplied becomes property of the list and should not be freed by
 * the caller.
 * theClockSkew is used to determine expired nonces and free them. 
 */
int NonceFoundElseAddTimeOrdered(NonceHead_t *theNonceHead,
                                 char *theData,
                                 time_t theTimeOccurred,
                                 time_t theClockSkew )
{
  Nonce_t *aNoncePtr, *tmpPtr;

  int expired = 0;
  int retValue = 1; /* >0 nonce found */
  time_t timeNow;
  
  if (!theNonceHead) return -1;

  pthread_mutex_lock(&(theNonceHead->nonces_lock));

  timeNow = time(NULL);

  aNoncePtr = theNonceHead->first;
  
  while (aNoncePtr) {
    if (timeNow - theClockSkew > aNoncePtr->evtTime) {
      expired = 1;
      break;
    }
    if (aNoncePtr->evtTime >= theTimeOccurred) {
      if (aNoncePtr->evtTime > theTimeOccurred) 
	break;
      else 
	if (!strcmp(aNoncePtr->nonceData, theData))
	  goto munlock_ret;
    }
    else
      break;
    aNoncePtr = aNoncePtr->next;
  }
  
  /* If you are here, then you have to add it */
  retValue = 0;

  tmpPtr = malloc(sizeof(Nonce_t));
  if (tmpPtr == NULL) 
    goto munlock_ret; 
  
  /* Although we have not found it, we return 1 due to the memory failure so
   * that the caller function rejects the nonce. Otherwise, when we have not
   * enough memory we are vulnerable to replay attacks. At the same time the
   * caller does not have to distinguish error or normal return values.
   */

  /* If you are here, then you have to add it */
  retValue = 0;

  tmpPtr->nonceData = theData;
  tmpPtr->evtTime = theTimeOccurred;

  if ( theNonceHead->first == NULL ) {    
    theNonceHead->first = tmpPtr;
    theNonceHead->last = tmpPtr;
    tmpPtr->next = NULL;
    tmpPtr->prev = NULL;
  }
  else if (aNoncePtr) {
    tmpPtr->next = aNoncePtr;
    tmpPtr->prev = aNoncePtr->prev;
    aNoncePtr->prev = tmpPtr;
    if (tmpPtr->prev == NULL) 
      theNonceHead->first = tmpPtr;
    else
      tmpPtr->prev->next = tmpPtr;
  }
  else {
    /* Add it at the end */
    tmpPtr->next = NULL;
    tmpPtr->next = theNonceHead->last;
    theNonceHead->last->next = tmpPtr;
    theNonceHead->last = tmpPtr;
  }

  if (expired) {
    theNonceHead->last = aNoncePtr->prev;
    if (aNoncePtr->prev) 
      aNoncePtr->prev->next = NULL;
    else {
      theNonceHead->first = NULL;
    }

    pthread_mutex_unlock(&(theNonceHead->nonces_lock));
    FreeNonceList(aNoncePtr);
    /* You can later spawn a thread and return earlier */
    return retValue; 
    /* Not found therefore added */
    /* You cannot have found it and expired == 1 */
  }

munlock_ret:
  pthread_mutex_unlock(&(theNonceHead->nonces_lock));
  return retValue;
} /* NonceFoundElseAddTimeOrdered */


/*
 * Use the following functions for gNonceHead
 * The evtTime is the expiration time
*/

/*
 * AddNonce()
 * Adds the nonce in the list specified by *theNonceHead.
 * *theData supplied becomes property of the list and should not be 
 * freed by the caller. In case of failure it returns a non-zero error.
 * The nonce added will be valid for time specified by theLifeTime.
 */
int AddNonce(NonceHead_t *theNonceHead, char *theData, time_t theLifeTime)
{
  Nonce_t *aNoncePtr;

  if (!theNonceHead) return -1;
  if (!theData) return -1;

  aNoncePtr = malloc(sizeof(Nonce_t));
  if (!aNoncePtr) return -1;
  aNoncePtr->nonceData = theData;
  
  pthread_mutex_lock(&(theNonceHead->nonces_lock));

  aNoncePtr->evtTime = theLifeTime + time(NULL);
  aNoncePtr->next = theNonceHead->first;

  if (theNonceHead->first) {
    theNonceHead->first->prev = aNoncePtr;
  }
  else {
    theNonceHead->last = aNoncePtr;
  }

  theNonceHead->first = aNoncePtr;
  aNoncePtr->prev = NULL;

  pthread_mutex_unlock(&(theNonceHead->nonces_lock));
  return 0;
} /* AddNonce */


/*
 * FreeNonceList()
 * Simply delete a linked listed of Nonce_t and the nonce data
 * Attention: It does not acquire the mutex, Therefore make sure that 
 * you have detached the *theFirstNonce list from NonceHead_t
 */
void FreeNonceList(Nonce_t *theFirstNonce)
{
  Nonce_t *tmpPtr;
  
  while (theFirstNonce) {
    tmpPtr = theFirstNonce->next;
    if (theFirstNonce->nonceData) free(theFirstNonce->nonceData);
    free(theFirstNonce);
    theFirstNonce = tmpPtr;
  }
} /* FreeNonceList */


/* 
 * NonceFound()
 * If the nonce is found unexpired in the list specified by *theNonceHead
 * a zero error code is returned. All expired nonces are removed from the list
 * upon return.
 */
int NonceFound(NonceHead_t *theNonceHead, char *theData, int theRemoveFlag)
{
  Nonce_t *aNoncePtr, *aPrevNoncePtr;
  int found = 0;
  time_t timeNow;

  if (!theNonceHead) return -1;

  pthread_mutex_lock(&(theNonceHead->nonces_lock));
  
  timeNow = time(NULL);

  aNoncePtr = theNonceHead->first;

  while (aNoncePtr) {
    if (timeNow > aNoncePtr->evtTime) goto expired_label; 
    if (!strcmp( aNoncePtr->nonceData, theData)) {
      found = 1;
      if (theRemoveFlag) {
	if (aNoncePtr == theNonceHead->last)
	  theNonceHead->last = aNoncePtr->prev;
	if (aNoncePtr == theNonceHead->first) {
	  theNonceHead->first = aNoncePtr->next;
	  if(aNoncePtr->next) aNoncePtr->next->prev = NULL;
	}
        else {
	  aNoncePtr->prev->next = aNoncePtr->next;
	  if (aNoncePtr->next) aNoncePtr->next->prev = aNoncePtr->prev;
	}
	free(aNoncePtr->nonceData);
	free(aNoncePtr);
      }
      break;
    }
    aPrevNoncePtr = aNoncePtr;
    aNoncePtr = aNoncePtr->next;
  }
  
  pthread_mutex_unlock(&(theNonceHead->nonces_lock));
  return found;

 expired_label:
  theNonceHead->last = aNoncePtr->prev;
  if (aNoncePtr->prev) 
    aNoncePtr->prev->next = NULL;
  else {
    theNonceHead->first = NULL;
  }

  pthread_mutex_unlock(&(theNonceHead->nonces_lock));  
  FreeNonceList(aNoncePtr);
  /*
   * The rest are expired also.
   * You can later spawn a thread and return earlier
   */
  return found;
} /* NonceFound */



