/*
 * authenticateSIP.c - Authenticate SIP requests
 */

/* Copyright 1999-2001 Columbia University.  All rights reserved. */

#include <stdio.h>
#include <stdlib.h>     /* calloc() */
#include <assert.h>
#include <unistd.h>     /* close() */
#include <string.h>
#include "request-sip.h"
#include "sip.h"
#include "strdupn.h"
#include "sysdep.h"
#include "rfc822.h"
#include "http.h"
#include "authenticateSIP.h"
#include "config.h"

/* #define NON_FOLDED_AUTH_HEADER */

/*
 * setup_local_or_proxy_auth()
 * Based on the value of do_proxy_auth, set response_header,
 * challenge_status, and r->authenticate appropriately.
 */
static void setup_local_or_proxy_auth(request_t *r, int do_proxy_auth,
                                      const char **response_header,
                                      int *challenge_status)
{
  if (do_proxy_auth) {
    *response_header = "Proxy-Authenticate";
    *challenge_status = 407;
    r->authorization = HTTP_ProxyAuthorization(&r->p);
  }
  else {
    *response_header = "WWW-Authenticate";
    *challenge_status = 401;
    r->authorization = HTTP_Authorization(&r->p);
  }
} /* setup_local_or_proxy_auth */
 

/*
 * validate_auth_info()
 * Validate that the authentication info provided in r->authorization 
 * a) exists, and b) is acceptable for r->source_user
 * If it doesn't, set the status to 'challenge_status', and return false.
 * For malformed requests, set the status to 400, and return < 0.
 */
static int validate_auth_info(request_t *r, int challenge_status)
{
  if (r->authorization.method == auth_unknown) {
    /* No authorization header in request */ 
    ResponseSetStatus(r, challenge_status, NULL);
    return 0;
  }

  if (r->authorization.userid == NULL) {
    ResponseSetStatus(r, 400, "Missing User ID in Authorization");
    return -1;
  }
  
  /* XXX: we currently just declare that r->authorization.userid
   * has to be the same as r->source_user->user.  Eventually there should
   * probably be an explicit additional field in the database for this, since
   * we don't want such a strict one-to-one mapping.
   */
  if (strcmp(r->source_user->user, r->authorization.userid) != 0) {
    DString reason;

    DStringInit(&reason);
    DStringPrintf(&reason, "Must authenticate with username %s",
                  r->source_user->user);
    ResponseSetStatus(r, challenge_status, DStringValue(&reason));

    DStringFree(&reason);

    return 0;
  }
  
  return 1;
} /* validate_auth_info */


/*
 * build_basic_response_header()
 * Build a basic auth challenge, and insert it as an appropriate response
 * header (WWW-Authenticate or Proxy-Authenticate) in the response.
 * Return false if build_basic_auth_challenge failed (which means that
 * userid doesn't exist in the database).
 */
static int build_basic_response_header(request_t *r,
                                       const char *response_header)
{
  int error;

  error = build_basic_auth_challenge(&r->authenticate, r->source_user);
  if (error != 0) {
    /* only print if successful build_auth_challenge */
    HeaderPrintf(&r->headers, 0, response_header, 
                 "Basic realm=\"%s\"",
                 r->authenticate.realm);
  }

  return error;
} /* build_basic_response_header */


/* Some defines to make the below cleaner */
#ifdef NON_FOLDED_AUTH_HEADER
#define REAL_SEPARATOR ", "
#else
#define REAL_SEPARATOR ",\r\n\t"
#endif
#define SEPARATOR (pos++ == 0 ? "" : REAL_SEPARATOR)
#define NONNULL(str) ((str) != NULL && (str)[0] != '\0')
#define ADDFIELD(h, field, pattern) \
	  do { \
	    if (NONNULL(r->authenticate.field)) { \
	      DStringPrintf((h), "%s" pattern, SEPARATOR, \
			    r->authenticate.field); \
	    } \
	  } while (0)
#define ADDQUOTED(h, field) ADDFIELD(h, field, #field "=\"%s\"")
#define ADDUNQUOTED(h, field) ADDFIELD(h, field, #field "=%s")
     
/*
 * build_digest_response_header()
 * Build a digest auth challenge, and insert it as an appropriate response
 * header (WWW-Authenticate or Proxy-Authenticate) in the response.
 * Return false if build_digest_auth_challenge failed (which means that
 * userid doesn't exist in the database).
 */
static int build_digest_response_header(request_t *r,
                                        const char *response_header,
					auth_status ret)
{
  int error;

  error = build_digest_auth_challenge(&r->authenticate,
                                      "" /* ??? What should "Opaque" be? */,
                                      (ret == auth_nonce_stale),
                                      config.PrivateKey,
                                      r->source_user);
  if (error != 0) {
    DString header;
    int pos;

    DStringInit(&header);
    pos = 0;

    DStringAppend(&header, "Digest ", -1);

    ADDQUOTED(&header, realm);
    ADDQUOTED(&header, domain);
    ADDQUOTED(&header, nonce);
    ADDQUOTED(&header, opaque);
    ADDUNQUOTED(&header, stale);
    ADDUNQUOTED(&header, algorithm);

    HeaderAppend(&r->headers, response_header, DStringValue(&header));
    DStringFree(&header);
  }
  return error;
} /* build_digest_response_header */


/* 
 * AuthSIP_BasicDigest()
 * Check that the request 'r' contains valid Basic or Digest authentication
 * information, based on the setting of Config.AuthMethod, for the user
 * named by 'userid' (or Proxy-authentication if do_proxy_auth is true).  If
 * so, return true.  If not, set r's status to the appropriate status code,
 * insert an appropriate *-Authenticate header, and return a value
 * indicating the reason for the failure.
 * If bad_pw_okay is true, then if authenticate() returns
 * auth_wrong_password, don't set status or add the header, but
 * still return that auth status.  The caller is going to handle it
 * specially.
 */
static auth_status AuthSIP_BasicDigest(request_t *r, 
                                       int do_proxy_auth, int build_headers,
                                       int bad_pw_okay)
{ 
  int error;
  auth_status ret = auth_internal_error;
  const char *response_header;
  int challenge_status;
  
  setup_local_or_proxy_auth(r, do_proxy_auth, &response_header,
                            &challenge_status);

  error = validate_auth_info(r, challenge_status);

  if (error < 0) {
    if (build_headers) {
      ResponseBuildDefault(r);
    }
    return auth_malformed_auth;
  }
   
  if (error == 0) {
    goto authorize;
  }

  /* Perform authentication. */
  ret = authenticate(config.AuthMethod, r->source_user, r->authorization);

  if (ret != auth_success &&
      !(bad_pw_okay && ret == auth_wrong_password)) {
    /* Authentication failed. */
    ResponseSetStatus(r, challenge_status, NULL);

  authorize:
    if (build_headers) {
      ResponseBuildDefault(r);
    }
    /* Build appropriate response header */
    if ((config.AuthMethod == auth_basic &&
	 !build_basic_response_header(r, response_header)) ||
	(config.AuthMethod == auth_digest &&
	 !build_digest_response_header(r, response_header, ret))) {
      DString reason;
      
      DStringInit(&reason);
      DStringPrintf(&reason, "Unknown username %s for authentication",
		    r->source_user->user);
      ResponseSetStatus(r, 403, DStringValue(&reason));

      DStringFree(&reason);

      ret = auth_no_user;
    }
  }

  return ret;
} /* AuthSIP_BasicDigest */


/*
 * AuthenticateSIP()
 * We want to verify that the request 'r' contains valid authentication
 * information for the request's source address (r->source_user).
 * Based on Config.AuthMethod, dispatch to the appropriate authentication
 * function.
 * If do_proxy_auth is true, proxy-authentication is done.
 * If build_headers is true, the response's default headers will be built
 * before returning.
 * If bad_pw_okay is true, then if authenticate() returns
 * auth_wrong_password, don't set status or add the header, but
 * still return that auth status.  The caller is going to handle it
 * specially.
 */
auth_status AuthenticateSIP(request_t *r, int do_proxy_auth, int build_headers,
                            int bad_pw_okay)
{
  if (config.AuthMethod == auth_none) {
    /* If server does not require authentication, always return success */
    return auth_success;
  }
  else if (r->source_user == NULL) {
    return auth_no_user; /* ??? Is this redundant? */
  }
  else if (config.AuthMethod == auth_basic ||
           config.AuthMethod == auth_digest) {
    return AuthSIP_BasicDigest(r, do_proxy_auth, build_headers,
                               bad_pw_okay);
  }
  else {
    if (build_headers) {
      ResponseBuildDefault(r);
    }
    ResponseSetStatus(r, 500, "Invalid authentication method configured");
    error("AuthenticateSIP",
          "config.AuthMethod configured with unknown authentication method %d\n",
          config.AuthMethod);

    return auth_internal_error;
  }
} /* AuthenticateSIP */

