/*
 * Canonicalize.c
 * Given a URI, return a canonicalized version of the same URI,
 * suitable for database lookups.
 */

/* Copyright 2001 Columbia University.  All rights reserved. */

#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>
#include <regex.h>
#include <stdlib.h>
#include <string.h>

#include "error.h"
#include "strdupn.h"

#include "config.h"
#include "sip.h"

/* the canonicalizer functions, libcanon.so */
#include "canonicalizer_interface.h" 

#include "canonicalize.h"

static int use_canonicalizer = 0;

static pthread_mutex_t canon_pipe_lock = PTHREAD_MUTEX_INITIALIZER;
#define LOCK()   pthread_mutex_lock(&canon_pipe_lock)
#define UNLOCK() pthread_mutex_unlock(&canon_pipe_lock)

/* Forward function declarations */
static int canonicalize_terminate_internal(void);

static char** tokenize(const char *string, int *num_tokens) 
{
 
  char** tokens;
  char *token_arr[12], *current_token = NULL;
  char *tosplit = strdup(string), *lasts;
  int maxtokens = 12, curtokens = 0, i;
  char *tosplit_orig = tosplit;


  if(!string || !num_tokens) {
    error("tokenize", "Bad call. Parameters null\n");
    return NULL;
  }
  
  while(curtokens < maxtokens &&
        (current_token = strtok_r(tosplit," \t", &lasts))) {
    tosplit = NULL;
    token_arr[curtokens++] = current_token;
  }
  
  if(curtokens == maxtokens && current_token)
    error("tokenize","Too many tokens found in canonicalize options");

  tokens = (char**)calloc(curtokens+2, sizeof(char*));
  
  tokens[0] = strdup("canonicalize");

  for (i=0; i < curtokens; i++) {
    tokens[i+1] = strdup(token_arr[i]);
  }

  *num_tokens = curtokens+1;    /* argc[argv] = NULL */

  free(tosplit_orig);
  
  return tokens;
} /* tokenize */


/*
 * free_tokens()
 * free a token list returned by tokenize.
 */
static void free_tokens(char** tokens, int num_tokens)
{
  int i;

  for (i = 0; i < num_tokens; i++) {
    free(tokens[i]);
  }

  free(tokens);
} /* free_tokens */


/*
 * canonicalize_init()
 * Initialize the canonicalizer, which loads as a shared-dynamically
 * loaded library, if one is defined by the configuration.  
 * In case of error, return non-zero.
 */
int canonicalize_init(void)
{
  int ret = 0;

  char** argv;
  int argc;

  LOCK();
  
  if (use_canonicalizer || !config.Canonicalize)
    goto done;
  
  debug("canonicalize_init", "Invoking canonicalizer with options %s\n", 
	config.Canonicalize);
  
  /* split the canonicalizer options into an array, so that it can
   * use parse_opts.
   */
  if ((argv = tokenize(config.Canonicalize, &argc)) != NULL) {
    if(!config.database) 
      debug("canonicalize_init","Warning: config.database null\n");
    
    if (InitCanonicalize2(argc, argv, config.database) != 0) 
      ret = 1;
    else
      use_canonicalizer = 1;
  }

  free_tokens(argv, argc);
  
done:    
  UNLOCK();

  return ret;
  
} /* canonicalize_init */


/*
 * canonicalize_terminate_internal()
 * Terminate the canonicalizer library.  Return non-zero if something's wrong.
 * Internal function: don't lock/unlock.
 */
static int canonicalize_terminate_internal(void)
{
  if (use_canonicalizer) {
    
    CloseCanonicalize();
    
    use_canonicalizer = 0;
  }

  return 0;

} /* canonicalize_terminate_internal */


/*
 * canonicalize_terminate()
 * Terminate the canonicalize program.  Return non-zero if something's wrong.
 * Public function: do locking.
 */
int canonicalize_terminate(void)
{
  int ret;

  LOCK();

  ret = canonicalize_terminate_internal();
  
  UNLOCK();
  
  return ret;
} /* canonicalize_terminate */

/*
 * do_canon_proc()
 * Call the canonicalize_uri function (in the canon shared library).
 * 
 * Return its result.
 * 
 */
static int do_canon_proc(uri_t *uri, canon_result *result)
{
  if (!use_canonicalizer) {
    result->u.unique = (uri_t *) malloc(sizeof(uri_t));
    if (result->u.unique == NULL) {
      result->status = canon_error;
      result->u.error = NULL;
      return 1;
    }
    result->status = canon_unique;
    *result->u.unique = URI_Copy(*uri);
    return 0;
  }

  return  canonicalize_uri(uri, result);

} /* do_canon_proc */


/*
 * canonicalize_hostname()
 * Canonicalize the hostname part of a URI.  Return true if the hostname
 * part corresponds to a 'local' host, i.e. one described by 'domain',
 * and thus the host part of the URI was modified.
 * Note that this function doesn't care about URI schemes; the calling
 * function presumably should.
 */
static int canonicalize_hostname(uri_t *uri)
{
  int result;

  if (!uri->host || uri->host[0] == '\0')
    return 0;

  result = regexec(config.domain, uri->host, 0, NULL, 0);

  if (result == REG_NOMATCH) {
    return 0;
  }

  if (result != 0) {
    char buf[1024];
    regerror(result, config.domain, buf, sizeof(buf));
    errno = 0;
    error("canonicalize_hostname",
          "Error matching hostname \"%s\": %s\n",
          uri->host, buf);
    return 0;
  }

  strassign(&uri->host, config.ServerName);

  return 1;
} /* canonicalize_hostname */


/*
 * canonicalize()
 * Canonicalize a URL.  Return one of:
 * - a canonicalized form of the same URL
 * - an array of nameaddrs, indicating an ambiguous canonicalization
 * - a string representing an internal error
 * If the host was remote, set the status to canon_remote rather than
 * canon_unique. 
 * Return non-zero on error, with a string in result->u.error if possible.
 */
int canonicalize(const uri_t *orig_uri, canon_result *result)
{
  uri_t *uri;
  int is_sip;
  int ret;

  uri = (uri_t *) malloc(sizeof(uri_t));
  if (uri == NULL) {
    result->status = canon_error;
    result->u.error = NULL;
    return 0;
  }

  *uri = URI_Copy(*orig_uri);

  if (strcasecmp(uri->scheme, "sip") == 0) {
    is_sip = 1;
  }
  else if (strcasecmp(uri->scheme, "tel") == 0) {
    is_sip = 0;
  }
  else {
    /* Don't transform anything other than sip and tel URIs */
    result->status = canon_unique;
    result->u.unique = uri;
    return 0;
  }

  if (is_sip) {
    if (!canonicalize_hostname(uri)) {
      /* Remote host */
      result->status = canon_remote;
      result->u.unique = uri;
      return 0;
    }
  }

  debug("canonicalize","calling do_canon_proc...\n");
  ret = do_canon_proc(uri, result);

  URI_Free(uri);
  free(uri);

  return ret;
} /* canonicalize */


/*
 * free_canon_result()
 * Free the result of a canonicalization.  Don't free the canon_result
 * itself, though.
 */
extern void free_canon_result(canon_result *result)
{
  int i;

  switch (result->status) {
  case canon_unique:
  case canon_remote:
    URI_Free(result->u.unique);
    free_valid(result->u.unique);
    result->u.unique = NULL;
    break;

  case canon_multiple:
    for (i = 0; i < result->u.multiple.nameaddr_count; i++) {
      URI_Free(&result->u.multiple.nameaddrs[i].addr);
      free_valid(result->u.multiple.nameaddrs[i].name);
    }
    free_valid(result->u.multiple.nameaddrs);
    result->u.multiple.nameaddrs = NULL;
    break;

  case canon_error:
    free_valid(result->u.error);
    break;

  default:
    error("free_canon_result",
          "Unknown result status %d\n", result->status);
    break;
  }
}
