/* Failure Policy: a transaction policy to handle "generic" failed requests */
/* Copyright 1999 - 2001, Columbia University.  All rights reserved. */

#include <stdlib.h>

#include "sysdep.h"
#include "policy.h"
#include "config.h"

/*
 * Note: failure_policy should not be chosen as the transaction policy if
 * there is also a user policy.  It doesn't handle subsequent requests or
 * responses.
 */

/*
 * failure_init()
 * Initialize a failure policy.  Remember the status value.
 */
static transaction_policy_ret failure_init(request_t *r,
                                           const policy_info_t *info,
                                           int status)
{
  r->transaction_policy_info = malloc(sizeof(int));
  *((int*)(r->transaction_policy_info)) = status;

  return transaction_policy_ok;
} /* failure_init */


/*
 * failure_handle_initial_request()
 * Called when a request which failed for some reason first arrives.
 * Return the status code.
 */
static transaction_policy_ret failure_handle_initial_request(request_t *r) {
  int status = *((int*)(r->transaction_policy_info));

  /* Make sure we're not "failing" with a success or intermediate response */
  if (status < 300)
    return transaction_policy_error;

  policy_send_new_response(r, status, NULL);

  /* Nothing else needs to be executed for the transaction */
  policy_done(r, transaction_policy);

  return transaction_policy_ok;
} /* failure_handle_initial_request */


/*
 * failure_handle_subsequent_request()
 * Handle a subsequent request after a failure.  This can only happen
 * if a user policy didn't immediately send a final response.  Since
 * there shouldn't be a user policy, return an error.
 */
static transaction_policy_ret failure_handle_subsequent_request(request_t *r,
                                                    message_t *m)
{
  error("failure_handle_request", "Shouldn't be called");
  return transaction_policy_error;
} /* failure_handle_subsequent_request */


/*
 * failure_handle_response()
 * Handle a response after a failure.  This can only happen if a user policy
 * proxied the request.  Since there shouldn't be a user policy, return an
 * error. 
 */
static transaction_policy_ret failure_handle_response(request_t *r,
                                                      message_t *m,
                                                      branch_t *b)
{
  error("failure_handle_response", "Shouldn't be called");
  return transaction_policy_error;
} /* failure_handle_response */


/* 
 * failure_timeout_expired()
 * Handle a timeout for this policy.  It never sets a timeout, so return
 * an error.
 */
static transaction_policy_ret failure_timeout_expired(request_t *r)
{
  error("failure_timeout_expired", "Shouldn't be called");
  return transaction_policy_error;
} /* failure_timeout_expired */
     

/*
 * failure_cleanup()
 * Free the stored status value.
 */
static transaction_policy_ret failure_cleanup(request_t *r)
{
  free (r->transaction_policy_info);
  return transaction_policy_ok;
} /* failure_cleanup */


transaction_policy_t failure_policy =
{
  "failure",
  &failure_init,
  &failure_handle_initial_request,
  &failure_handle_subsequent_request,
  &failure_handle_response,
  &failure_timeout_expired,
  &failure_cleanup
};
