/*
license.c - handles license validations for CINEMA components.

Copyright 1998-2001 by Columbia University; all rights reserved 
*/

#ifdef __linux__
#define _GNU_SOURCE /* Needed for strptime() declaration from <time.h> */
#endif

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <netdb.h>      /* MAXHOSTNAMELEN */
#include <sys/types.h>
#include <sys/param.h>  /* MAXHOSTNAMELEN */
#include <unistd.h>
#include <netinet/in.h> /* struct in_addr */
#include <arpa/inet.h>  /* inet_addr() */
#include <sys/socket.h>

#include "uri.h"         /* uri_t */
#include "md5.h"
#include "dbapi.h"       /* db access functions */
#include "error.h"

#include "license.h"

#include "sysdep.h"

#ifdef DO_LICENSING

static char *hostname;  /* the hostname for which  we check for licenses */

struct reg_info {       /* components of a license entry */
  char *domain;
  char *exp_str;
  time_t t_expire;
  char *md5;
};


/*
 * noisy_getcanonicalhostname()
 * Get and return the canonical hostname of this machine.
 * Not thread-safe -- only intended for startup time.
 * This is a copy of getcanonicalhostname() from sipd-config.c, but a lot
 * more chatty about what's going on.
 */
static char *noisy_getcanonicalhostname(void)
{
  char buf[MAXHOSTNAMELEN];
  struct hostent *hep;
  struct in_addr addr;
  char *addr_str;

  addr.s_addr = INADDR_ANY;

  gethostname(buf, MAXHOSTNAMELEN);
  
  hep = gethostbyname(buf);
  if (hep == NULL) {
    error("getcanonicalhostname", 
          "gethostbyname failed on %s, h_errno = %d\n",
          buf, h_errno);
    return NULL;
  }

  memcpy(&addr, hep->h_addr, sizeof(addr));

  hep = gethostbyaddr((char *)&(addr.s_addr), sizeof(addr.s_addr), AF_INET);
  if (hep == NULL) {
    error("getcanonicalhostname", 
          "gethostbyaddr failed on %s, h_errno = %d\n",
          inet_ntoa(addr), h_errno);
    return NULL;
  }

  addr_str = inet_ntoa(addr);
  
  if (hep->h_name == NULL) {
    fprintf(stderr,
            "WARNING: No forward hostname lookup was found for "
            "the IP address %s.  Please check your name resolution "
            "and DNS setup.\n Using \"%s\" as the local hostname "
            "for licensing.\n",
            addr_str, addr_str);
    return strdup(addr_str);
  }
  else if (strchr(hep->h_name, '.') == NULL) {
    fprintf(stderr, 
            "WARNING: The resolved local hostname \"%s\" does not "
            "appear to be a fully-qualified domain name.  Please check "
            "your name resolution and DNS setup.\n Using numeric "
            "address \"%s\" as the local hostname for licensing.\n",
            hep->h_name, addr_str);
    return strdup(addr_str);
  }
  else {
    return strdup(hep->h_name);
  }
} /* getcanonicalhostname */


/* check_reg_info()
 * Confirm that the registration information corresponds
 * to this host, and is up-to-date.  Returns false if not.
 */
static int check_reg_info(struct reg_info *reg_info)
{
  time_t t;
  char *domain_start;

  if (reg_info->t_expire != 0) {
    time(&t);
    if (t > reg_info->t_expire) {
      fprintf(stderr,
              "check_license: expiration date %s (%ld) is in the past.\n",
              reg_info->exp_str, (long)reg_info->t_expire);
      return 0;
    }
  }

  domain_start = hostname + strlen(hostname) - strlen(reg_info->domain);
  if (strcmp(domain_start, reg_info->domain) != 0) {
    fprintf(stderr,
            "check_license: The local hostname \"%s\" does not"
            "match the license domain \"%s\".\n",
            hostname, reg_info->domain);
    return 0;
  }
  
  return 1;
} /* check_registration */


/* check_md5()
 * Validate the the MD5 hash of the registration information.
 */
static int check_md5(const char* internal_code, struct reg_info *reg_info)
{
  int ret = 0, n;
  char buf[256], *md5;

#ifdef HAVE_SNPRINTF
  n = snprintf(buf, sizeof(buf), "%s:%s:%s", reg_info->domain,
               reg_info->exp_str, internal_code);
#else
  if (strlen(reg_info->domain) + strlen(reg_info->exp_str) +
      strlen(internal_code) >= sizeof(buf)) {
    fprintf(stderr, "check_license: snprintf error or overflow\n");
  }
  n = sprintf(buf, "%s:%s:%s", reg_info->domain,
              reg_info->exp_str, internal_code);
#endif
  
  if (n < 0 || n >= sizeof(buf)) {
    fprintf(stderr, "check_license: snprintf error or overflow\n");
  }
  md5 = MD5(buf);
  if (strcmp(md5, reg_info->md5) == 0) {
    ret = 1;
  }
  else {
    fprintf(stderr,
            "check_license: hash string \"%s\" is not valid "
            "(it does not authorize the domain \"%s\" and"
            " expiration date \"%s\").\n",
            reg_info->md5,
            reg_info->domain, reg_info->exp_str);
  }
  free(md5); /* this is malloc'ed storage */
  return ret;
}


#ifdef USE_MYSQL
/* 
 * registration_valid_database
 *  Use the database as the source of license information.
 *
 * Returns 1 if the running system has a valid registration code.
 */
static int registration_valid_database(const cinema_license_info *p_linfo)
{
  struct reg_info reg_info;
  ResultTable query_result;
  int rows, i;
  
  if (!p_linfo->license_uri || !p_linfo->table_name) {
    error("check_license", "Invalid parameters to check_license\n");
    return -1;
  }
  
  debug("check_license", "Reading license data from %s\n",  
        p_linfo->license_uri);
  
  /* Check whether db connection has been established. if yes, 
   * read. or open the connection and then read
   */
  if (!isConnectionEstablished()) {
    /* Old-style database configuration */
    uri_t table_uri = URI_Parse(p_linfo->license_uri);

    if (strcasecmp(table_uri.scheme, "sql") == 0 ||
        strcasecmp(table_uri.scheme, "fastsql") == 0) {
      char *password;
      char *path = "sip";

      password = table_uri.password;
      if ( password != NULL && !strcmp(password,"NULL")) {
        password = NULL;
      }
      if (table_uri.path) {
        path = table_uri.path;
      }
      if (initializeConnection(table_uri.user, password, path,
                               table_uri.host, table_uri.port) != 0) {
        fatal("ConfigUser", 
              "Could not open database %s\n", p_linfo->license_uri);
      }
    }
    else {
      fatal("registration_valid",
            "Unknown database scheme %s in UserConfig pseudo-URL %s\n",
            table_uri.scheme, p_linfo->license_uri);
    }
    URI_Free(&table_uri);
  }

  
  query_result = query(p_linfo->table_name, NULL, NULL, 
                       "domain, expiry_date, hash_value",
                       &rows);
  
  if (rows <= 0 || !query_result) {
    error("registration_valid",
          "No license entries were obtainable from config file."
          "Check your database\n");
    return -1;
  }

  for (i=0; i<rows; i++) {
    /* for each registration entry obtained from db, check againt our 
     * name/ip. all query_result fields will be non-null.
     */
    reg_info.domain = query_result[i][0];
    reg_info.exp_str = query_result[i][1];

    if (strcmp(reg_info.exp_str, "-") == 0) {
      reg_info.t_expire = 0;
    }
    else {
      /* Some strptime() implementations (glibc, FreeBSD) don't initialize
         all the fields of struct tm, so set it to zero first. */
      struct tm tm_expire;
      memset(&tm_expire, '\0', sizeof(tm_expire));
      if (strptime(reg_info.exp_str, "%Y-%m-%d", &tm_expire) == NULL) {
        fprintf(stderr, 
                "check_license: Invalid expiration date \"%s\".\n",
                reg_info.exp_str);
        return -1;
      }
    
      reg_info.t_expire = mktime(&tm_expire);
      if (reg_info.t_expire == (time_t)-1) {
        fprintf(stderr, "check_license: Invalid expiration date \"%s\".\n",
                reg_info.exp_str);
        return -1;
      }
    } /* else */

    reg_info.md5 = query_result[i][2];

    /* call check_md5 and check_reg_info. if both return true, we are done.
     */
    if (check_md5(p_linfo->internal_code, &reg_info) && 
        check_reg_info(&reg_info)) {
      freeResultTable(query_result, rows, 3); 
      return 1;
    }
  } /* for */
  
  /* none matches, free and return false. */
  freeResultTable(query_result, rows, 3); 
  return -1;
} /* registration_valid_database */

#endif /* USE_MYSQL */


/*
 * parse_registration()
 * Parses a registration string: of the form "domain:YYYY-MM-DD:md5" or
 * "domain:-:md5" for unlimited-time licenses.
 * Return false if it could not be parsed, otherwise fills in reg_info.
 * reg_info will contain pointers into registration string, which must be
 * writeable.
 * Not thread-safe; should only be called at program startup.
 */
static int parse_registration(char *registration, struct reg_info *reg_info)
{
  char *lasts;
  struct tm tm_expire;

  reg_info->domain   = strtok_r(registration, ":", &lasts);
  reg_info->exp_str  = strtok_r(NULL, ":", &lasts);
  reg_info->md5      = strtok_r(NULL, ":\n", &lasts);

  if (reg_info->domain == NULL || reg_info->exp_str == NULL || 
      reg_info->md5 == NULL) {
    fprintf(stderr,
            "check_license: Malformed license entry"
            "(not enough fields).\n");
    return 0;
  }
  
  if (strcmp(reg_info->exp_str, "-") == 0) {
    reg_info->t_expire = 0;
  }
  else {
    /* Some strptime() implementations (glibc, FreeBSD) don't initialize
       all the fields of struct tm, so set it to zero first. */
    memset(&tm_expire, '\0', sizeof(tm_expire));
    if (strptime(reg_info->exp_str, "%Y-%m-%d", &tm_expire) == NULL) {
      fprintf(stderr, "check_license: Invalid expiration date \"%s\".\n",
              reg_info->exp_str);
      return 0;
    }
    
    reg_info->t_expire = mktime(&tm_expire);
    if (reg_info->t_expire == (time_t)-1) {
      fprintf(stderr, "check_license: Invalid expiration date \"%s\".\n",
              reg_info->exp_str);
      return 0;
    }
  }
  return 1;
} /* parse_registration */


/* 
 * registration_valid_file
 *  Look for license entries from a file.
 *
 * Returns 1 if the running system has a valid registration code.
 */
static int registration_valid_file(const cinema_license_info *p_linfo)
{
  struct reg_info reg_info;
  char code[BUFSIZ];
  FILE *regs;
  int lineno;

  char filename[BUFSIZ];
  
  /* skip file:// and take the rest */
  strncpy(filename, p_linfo->license_uri+7, sizeof(filename));

  debug("check_license", "Reading license file %s\n", filename);

  if ((regs = fopen(filename, "r")) == NULL) { 
    error("check_license", "Cannot open license file %s\n", 
          filename);
    return -1;
  }

  lineno = 0;
  
  while (fgets(code, BUFSIZ, regs) != NULL) {
    lineno++;
    if (parse_registration(code, &reg_info) &&
        check_md5(p_linfo->internal_code, &reg_info) &&
        check_reg_info(&reg_info)) {
      fclose(regs);
      return 1;
    }
  }
  fclose(regs);
  return -1;
} /* registration_valid */


/* 
 * check_registration()
 * Checks for binary registration
 * if match, return 0, else -1.
 */
int check_registration(const cinema_license_info *p_linfo)
{
  int retval = 0;

  hostname = noisy_getcanonicalhostname();
  if (hostname == NULL) {
    error("check_registration", "getcanonicalhostname");
  }
  
  if (!p_linfo || !p_linfo->license_uri || !p_linfo->internal_code) {
    error("check_registration", "One or more parameters null\n");
    return -1;
  }
  
  /* Check whether the uri is for file or database. If it is for
   * database, and MYSQL is not there, flag an error.
   */
  if (strstr(p_linfo->license_uri, "file://") == p_linfo->license_uri) {
    retval = registration_valid_file(p_linfo);
  }
  else {
#ifdef USE_MYSQL
    if ((strstr(p_linfo->license_uri, "sql://") == p_linfo->license_uri) ||
       (strstr(p_linfo->license_uri, "fastsql://") == p_linfo->license_uri)) {
      retval = registration_valid_database(p_linfo);
    }
#else
    fprintf(stderr, "Dont know where to look for license information\n");
#endif
  }

  if (retval <= 0) {
    fprintf(stderr, "\nNo valid licenses were found.\n"
            "You can request a new license from the web page\n"
            "<http://www.cs.columbia.edu/~hgs/license/>.\n");
    return -1;
  }
  
  /* everything ok, continue */
  free(hostname);
  hostname = NULL;
  return 0;
} /* check_registration */

#else /* ! DO_LICENSING */

/* 
 * check_registration()
 * Dummy registration check if we configured with licensing turned off.
 */
int check_registration(const cinema_license_info *p_linfo)
{
 debug("check_registration", "Your application has not been compiled with"
        "licensing enabled\n");
 return -1;
} /* check_registration */

#endif


