/*
 * message.c: manipulate messages and message queues.
 * Copyright 1998-1999 by Columbia University; all rights reserved 
 */

#include <stdlib.h>
#include <string.h>      /* strdup */
#include "request-sip.h"
#include "message.h"
#include "strdupn.h"

/*
 * MessageCreate()
 * Create a new message, with real values for all the fields.
 */
message_t *MessageCreate(message_type_t type, parser_t *p,
                    body_t *body, const struct sockaddr_storage* sin, 
                    branch_t *b, handled_by_t handled)
{
  message_t *new = (message_t *) calloc(1, sizeof(message_t));
  if (new == NULL) {
    error("MessageCreate", "Couldn't allocate new memory space");
    return NULL;
  }

  HTTP_Duplicate(&new->p, p);

  strassign(&new->body.content_type, body->content_type);
  new->body.content_length = body->content_length;
  /* XXX nulls in body.  (This is far from the only offender on this
   * point.) */
  strassign(&new->body.content, body->content);

  memcpy(&new->source, sin, sizeof(new->source));
  new->handled = handled;
  new->branch = b;
  new->next = NULL;
  new->type = type;
  
  return new;
} /* MessageCreate */


/*
 * MessageCreatePseudoResponse()
 * Create a pseudo-response -- a pseudo-message which has only a response
 * status and nothing else.
 */
message_t *MessageCreatePseudoResponse(branch_t *b, handled_by_t handled,
                                       int status)
{
  message_t *new = (message_t *) calloc(1, sizeof(message_t));
  if (new == NULL) {
    error("MessageCreatePseudoResponse", 
          "Couldn't allocate new memory space");
    return NULL;
  }

  new->type = message_pseudo_message;
  new->handled = handled;
  new->branch = b;
  new->p.type = http_response;
  new->p.status = status;

  return new;
} /* MessageCreatePseudoResponse */


/*
 * MessageCreateTrivial()
 * Create a trivial message -- its only information is its message
 * type.
 */
message_t *MessageCreateTrivial(message_type_t type,  handled_by_t handled)
{
  message_t *new = (message_t *) calloc(1, sizeof(message_t));
  if (new == NULL) {
    error("MessageCreateTrivial", "Couldn't allocate new memory space");
    return NULL;
  }

  new->type = type;
  new->handled = handled;

  return new;
} /* MessageCreateTrivial */


/*
 * MessageDuplicate()
 * Duplicate a message.
 */
message_t *MessageDuplicate(message_t *m)
{
  message_t *new = (message_t *) calloc(1, sizeof(message_t));
  if (new == NULL) {
    error("MessageDuplicate", "Couldn't allocate new memory space");
    return NULL;
  }

  HTTP_Duplicate(&new->p, &m->p);

  strassign(&new->body.content_type, m->body.content_type);
  new->body.content_length = m->body.content_length;
  /* XXX nulls in body.  (This is far from the only offender on this
   * point.) */
  strassign(&new->body.content, m->body.content);

  memcpy(&new->source, &m->source, sizeof(new->source));
  new->handled = m->handled;
  new->branch = m->branch;
  new->next = NULL;

  return new;
} /* MessageDuplicate */


/*
 * MessageDelete()
 * Delete an mqueue message.
 */
void MessageDelete(message_t *m)
{
  HTTP_Free(&m->p);
  free_valid(m->body.content_type);
  free_valid(m->body.content);
  free_valid(m->user_policy_info);
  free_valid(m->transaction_policy_info);
  free_valid(m);
} /* MessageDelete */


/*
 * GetRequestAsMessage()
 * Copy the original request of a request_t to a 'message_t' structure.
 */
message_t *GetRequestAsMessage(request_t *r)
{
  message_t *new = (message_t *) calloc(1, sizeof(message_t));
  if (new == NULL) {
    error("GetRequestAsMessage", "Couldn't allocate new memory space");
    return NULL;
  }

  HTTP_Duplicate(&new->p, &r->p);

  strassign(&new->body.content_type, r->request_body.content_type);
  new->body.content_length = r->request_body.content_length;
  /* XXX nulls in body.  (This is far from the only offender on this
   * point.) */
  strassign(&new->body.content, r->request_body.content);

  memcpy(&new->source, &r->source_sin, sizeof(new->source));
  new->handled = nobody_handled;
  new->branch = NULL;
  new->next = NULL;
  return new;
} /* GetRequestAsMessage */


/*
 * MessageReallyEnqueue()
 * Actually enqueue a message on a list.  The message structure
 * must be allocated by the caller.
 */
void MessageReallyEnqueue(message_t **list, message_t *new)
{
  message_t *m;

  if (*list == NULL)
    *list = new;
  else {
    m = *list;
    while (m->next != NULL) {
      m = m->next;
    }
    m->next = new;
  }
}


/*
 * MessageEnqueue()
 * Add a new message to the end of a message queue.
 * Make sure you have a lock on the structure that owns the list.
 */
void MessageEnqueue(message_t **list, message_type_t type, parser_t *p,
                    body_t *body, const struct sockaddr_storage *sin, 
                    branch_t *b, handled_by_t handled)
{
  message_t *new;
  new = MessageCreate(type, p, body, sin, b, handled);
  if (new) {
    MessageReallyEnqueue(list, new);
  }
}


/*
 * MessageEnqueuePseudoResponse()
 * Enqueue a pseudo-response -- a message which has only a response code
 * and nothing else.
 */
void MessageEnqueuePseudoResponse(message_t **list, branch_t *b,
                                  handled_by_t handled, int status)
{
  message_t *new;
  new = MessageCreatePseudoResponse(b, handled, status);

  if (new) {
    MessageReallyEnqueue(list, new);
  }
} /* MessageEnqueuePseudoResponse */

/*
 * MessageEnqueueApiRequest()
 * Enqueue a request from API to call thread - only the method is
 * valid.
 */
void MessageEnqueueApiRequest(message_t **list, const char *method,
                              handled_by_t handled)
{
  message_t *new;
  new = MessageCreateTrivial(message_api_message, handled);
  if ( new ) {
    new->type = message_api_message;
    new->p.method = strdup(method ? method : "UNKNOWN");
    new->p.type = http_request;
    MessageReallyEnqueue(list, new);
  }
}

/*
 * MessageEnqueueApiResponse()
 * Enqueue a response from API to call thread - only the status and reason.
 */
void MessageEnqueueApiResponse(message_t **list, int status, 
                               const char *reason,
                              handled_by_t handled)
{
  message_t *new;
  new = MessageCreateTrivial(message_api_message, handled);
  if ( new ) {
    new->type = message_api_message;
    new->p.status = status;
    new->p.reason = reason ? strdup(reason) : NULL;
    new->p.type = http_response;
    MessageReallyEnqueue(list, new);
  }
}

/*
 * MessageEnqueuePseudoRequest()
 * Enqueue a request (ACK) from callto request - only the method is
 * valid. This is typically called for reINVITE or ACK, or maybe CANCEL.
 */
void MessageEnqueuePseudoRequest(message_t **list, const char *method,
                                handled_by_t handled, body_t *body)
{
  message_t *new;
  new = MessageCreateTrivial(message_api_message, handled);
  if ( new ) {
    new->type = message_pseudo_message;
    new->p.method = strdup(method ? method : "UNKNOWN");
    new->p.type = http_request;

    if ( body != NULL && body->content != NULL &&
      body->content_type != NULL ) {
      strassign(&new->body.content_type, body->content_type);
      new->body.content_length = body->content_length;
      /* XXX nulls in body.  (This is far from the only offender on this
       * point.) */
      strassign(&new->body.content, body->content);
    }
    
    MessageReallyEnqueue(list, new);
  }
}

/*
 * MessageTrivialEnqueue()
 * Enqueue a "fake" message -- its only information is its message
 * type.
 */
void MessageTrivialEnqueue(message_t **list, message_type_t type,
                           handled_by_t handled)
{
  message_t *new;

  new = MessageCreateTrivial(type, handled);
  if (new) {
    MessageReallyEnqueue(list, new);
  }
} /* MessageTrivialEnqueue */


/*
 * GetNextMessage()
 * Get the next message on the specified message queue.
 */
message_t *GetNextMessage(struct message_t **mlist)
{
  message_t *m;
  /* XXX do I care about the "handled_by" values here? */
  if (*mlist == NULL)
    return NULL;

  m = *mlist;
  *mlist = m->next;
  return m;
} /* GetNextMessage */


