/* policy.h - definition and calls for a policy handler. */

#ifndef POLICY_H_
#define POLICY_H_

/* Information passed from execute_policy() to the individual policy
   callback. */

#include "request-sip.h"

typedef struct policy_info_t policy_info_t;

/* Values a user policy callback function can return */
typedef enum {user_policy_ok, user_policy_defer, user_policy_error}
user_policy_ret;

/* Values a transaction policy callback function can return */
typedef enum {transaction_policy_ok, transaction_policy_error}
transaction_policy_ret;

/* Types of policies. */
typedef enum {unknown_policy, transaction_policy, user_policy} policy_type;

/* The definition of a user policy */
typedef struct user_policy_t {
  char *name;
  user_policy_ret (*init)(request_t *r, const policy_info_t *info,
                          int status);
  user_policy_ret (*handle_initial_request)(request_t *r);
  user_policy_ret (*handle_subsequent_request)(request_t *r,
                                               message_t *m);
  user_policy_ret (*handle_response)(request_t *r, message_t *m,
                                     branch_t *branch);
  user_policy_ret (*timeout_expired)(request_t *r);
  user_policy_ret (*cleanup)(request_t *r);
} user_policy_t;


/* The definition of a transaction policy */
/* This is currently the same as a user policy, except for the return type,
 * since it can't defer. */
typedef struct transaction_policy_t {
  char *name;
  transaction_policy_ret (*init)(request_t *r, const policy_info_t *info,
                                 int status);
  transaction_policy_ret (*handle_initial_request)(request_t *r);
  transaction_policy_ret (*handle_subsequent_request)(request_t *r,
                                                      message_t *m);
  transaction_policy_ret (*handle_response)(request_t *r, message_t *m,
                                     branch_t *branch);
  transaction_policy_ret (*timeout_expired)(request_t *r);
  transaction_policy_ret (*cleanup)(request_t *r);
} transaction_policy_t;


/* Functions policies can call */
void policy_send_new_response(request_t *r, int status, char *reason);
void policy_send_modified_new_response(request_t *r, int status, char *reason,
                                       headers_t *new_headers,
                                       headers_t *del_headers,
                                       body_t *body);
void policy_forward_response(request_t *r, message_t *m);
void policy_forward_user_response(request_t *r, message_t *m);
void policy_proxy_original_request_to(request_t *r, uri_t *uri,
                                      message_t *m, void *policy_info,
                                      policy_type type);
void policy_proxy_subsequent_request_to(request_t *r, branch_t *b,
                                        message_t *m);
void policy_cancel_branch(request_t *r, branch_t *b);
void policy_set_timeout(request_t *r, double delay, policy_type type);
void policy_done(request_t *r, policy_type type);

/* Execute a policy */
void execute_policy(request_t *r, const policy_info_t *info, int status);

/* Libsip's policies */
extern transaction_policy_t failure_policy;

#endif /* POLICY_H_ */
