/*
 * policy-core.c - execute a policy handler.
 *
 * This executes the low-level core logic of a SIP transaction.  The
 * high-level core logic -- anything which isn't inevitable in the
 * protocol -- goes in a policy handler instead, which this calls.
 *
 * Copyright 1998-2001 by Columbia University; all rights reserved 
 */

#include <sys/errno.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>

#include "policy.h"
#include "message.h"
#include "timespec.h"
#include "strdupn.h"
#include "pthread-util.h"
#include "config.h"
#include "sysdep.h"
#include "log_interface.h"

#include "rr_support.h"

static int  WaitAndHandleNextMessages(request_t *r);
static void CalculateNextTimeout(request_t *r, struct timespec **next_expire,
                                 policy_type *next_expire_policy);
static int  ExecuteTimeout(request_t *r, struct timespec *next_expire,
                           policy_type next_expire_policy);
static int  PolicyHandleMessage(request_t *r, message_t *m);
static void CheckForFinalResponse(request_t *r);
static void CleanupMessageQueue(request_t *r);
static void CleanupPolicyMachine(request_t *r);
static void SendComplaint(request_t *r);


/*
 * EXECUTE_IN_USER_AND_TRANSACTION()
 * MACRO: Execute a function of a user policy, if it exists.  If it returns
 * policy_defer, or if it doesn't exist (is NULL), execute the function in
 * the transaction policy.  If either one returns policy_error, execute BAIL.
 * If a policy is done, don't invoke its callback functions.
 * XXX: if user policy isn't done but transaction policy is, and the user
 * policy defers, we drop the request.  Is this right?
 */
#define EXECUTE_IN_USER_AND_TRANSACTION(R, FUNCTION, BAIL) \
  do { \
    user_policy_ret ret = user_policy_defer; \
    if ((R)->user_policy != NULL && !(R)->user_policy_done) { \
      debug("EXECUTE_IN_USER_AND_TRANSACTION", \
            "r=0x%p: executing user policy function %s->%s\n", \
            (R), (R)->user_policy->name, #FUNCTION); \
      if ((ret = (R)->user_policy->FUNCTION ) != user_policy_ok && \
          ret != user_policy_defer) { \
        BAIL; \
      } \
    } \
    if (ret == user_policy_defer && !(R)->transaction_policy_done) { \
      debug("EXECUTE_IN_USER_AND_TRANSACTION", \
            "r=0x%p: executing transaction policy function %s->%s\n", \
            (R), (R)->transaction_policy->name, #FUNCTION); \
      if (((R)->transaction_policy->FUNCTION ) != transaction_policy_ok) { \
        BAIL; \
      } \
    } \
  } while(0)
/* EXECUTE_IN_USER_AND_TRANSACTION */

/*
 * execute_policy()
 * This is the core code which executes a policy.  Call the policy's init
 * function, then wait for events and dispatch them to the policy.  Finish
 * when the policy calls policy_done().
 */
void execute_policy(request_t *r, const policy_info_t *info, int status)
{
  debug("execute_policy",
        "Started, r=0x%p, info=0x%p, status=%d, policy=%s/%s\n",
        r, info, status,
        r->user_policy != NULL ? r->user_policy->name : "NULL",
        r->transaction_policy->name);

  /* Neither policy has been invoked yet.  However, if the user policy is
   * NULL, mark it 'done'. */
  if (r->user_policy == NULL) {
    r->user_policy_done = 1;
  }
  else {
    r->user_policy_done = 0;
  }

  r->transaction_policy_done = 0;

  if (r->transaction_policy->init(r, info, status) != transaction_policy_ok) {
    goto cleanup_no_user;
  }

  if (r->user_policy != NULL) {
    if (r->user_policy->init(r, info, status) != user_policy_ok) {
      goto cleanup;
    }
  }

  EXECUTE_IN_USER_AND_TRANSACTION(r, handle_initial_request(r), goto cleanup);

  CheckForFinalResponse(r);

  while (!r->done) {
    if (!WaitAndHandleNextMessages(r)) {
      break;
    }
  }

cleanup:  /* Jump here when any policy function returns *_policy_error,
           * or a value not in the enum. */

  /* Clean up policy-local data */
  if (r->user_policy != NULL) {
    r->user_policy->cleanup(r);
  }

cleanup_no_user: /* Jump here if transaction_policy->init() returned 
                    an error. */

  r->transaction_policy->cleanup(r);

  /* If we haven't sent a final response by this point, something's
   * definitely wrong. */
  if (!r->sent_final) {
    SendComplaint(r);
  }

  CleanupPolicyMachine(r);
} /* execute_policy */


/*
 * WaitAndHandleNextMessage()
 * Core of the execute_policy() loop.  Figure out the time until the next
 * timeout (either policy timeout, or the transaction timeout), if there is
 * one, then wait for that time, then perform the appropriate actions.
 * Return false on failure, otherwise true.
 */
static int WaitAndHandleNextMessages(request_t *r)
{
  struct message_t *m;
  int thr_err;
  struct timespec *next_expire;
  policy_type next_expire_policy;

  CalculateNextTimeout(r, &next_expire, &next_expire_policy);
  
  if (next_expire != NULL)
    thr_err = pthread_cond_timedwait(&r->cond, &r->mutex, next_expire);
  else
    thr_err = pthread_cond_wait(&r->cond, &r->mutex);
  debug("execute_policy", "r=0x%p: wait exiting\n", r);
  
  if (thr_err) {
    if (thr_err != ETIME && thr_err != ETIMEDOUT) {
      error("WaitAndHandleNextMessage",
            "Error waiting for policy events: %s (%d)",
            strerror(thr_err), thr_err);
      /* Critical error in pthread_cond_[timed]wait.
       * Abort the transaction. */
      return 0;
    }
    else {
      if (!ExecuteTimeout(r, next_expire, next_expire_policy)) {
        return 0;
      }
      CheckForFinalResponse(r);
    }
  }
  
  /* Walk the message queue, and send requests not already handled */
  /* We could keep a pointer to the last unhandled request,
   * but I don't expect there to be that many. */
  /* If the final timeout has triggered, don't bother running the loop. */ 
  for (m = r->mqueue; m != NULL && !r->done; m = m->next) {      
    if (!PolicyHandleMessage(r, m)) {
      return 0;
    } 
    CheckForFinalResponse(r);
  }
  
  CleanupMessageQueue(r);

  return 1;
} /* WaitAndHandleNextMessage */


/*
 * QualifiedTimespecCompare()
 * Compare two timespecs, returning <0, 0, >0 depending on which one is
 * greater.
 * If a timespec is 0, or if its 'done' is true, consider it infinite.
 * If two timespecs are both infinite, they are equal.
 */
static int QualifiedTimespecCompare(struct timespec a, int a_done,
                                    struct timespec b, int b_done)
{
  if (a_done || a.tv_sec == 0) {
    if (b_done || b.tv_sec == 0) {
      return 0;
    }
    return 1;
  }
  if (b_done || b.tv_sec == 0) {
    return -1;
  }
  return timespec_compare(a, b);
} /* QualifiedTimespecCompare */


/*
 * CalculateNextTimeout()
 * Calcluate the next expiration time for either of the request's policies,
 * or for the final transaction timeout.
 */
static void CalculateNextTimeout(request_t *r, struct timespec **next_expire,
                                 policy_type *next_expire_policy)
{
  /* Note: this logic is structured so that if all three timeouts are equal,
   * the final timeout is picked.  This is only likely to happen if all
   * three timeouts are infinite. */
  if (QualifiedTimespecCompare(r->transaction_policy_timeout,
                               r->transaction_policy_done,
                               r->user_policy_timeout,
                               r->user_policy_done) < 0) {
    if (QualifiedTimespecCompare(r->transaction_policy_timeout,
                                 r->transaction_policy_done,
                                 r->final_timeout,
                                 0 /* FALSE */) < 0) {
      *next_expire = &r->transaction_policy_timeout;
      *next_expire_policy = transaction_policy;
    }
    else {
      *next_expire = &r->final_timeout;
      *next_expire_policy = unknown_policy;
    }
  }
  else {
    if (QualifiedTimespecCompare(r->user_policy_timeout,
                                 r->user_policy_done,
                                 r->final_timeout,
                                 0 /* FALSE */) < 0) {
      *next_expire = &r->user_policy_timeout;
      *next_expire_policy = user_policy;
    }
    else {
      *next_expire = &r->final_timeout;
      *next_expire_policy = unknown_policy;
    }
  }

  if ((*next_expire)->tv_sec == 0) {
    /* The least timeout was infinite, so mark this as a NULL timeout instead.
     * Note that this should do the right thing even if some timeouts are
     * infinite because they're 'done' rather than because they're 0, since
     * the final timeout is always preferred in case of equality, and the
     * final timeout is never 'done'. */
    *next_expire = NULL;
    *next_expire_policy = unknown_policy;
  }
} /* CalculateNextTimeout */


/*
 * ExecuteTimeout()
 * A timeout has triggered -- pthread_cond_timedwait returned ETIME/ETIMEDOUT.
 * Execute the appropriate policy's timeout_expired() function.
 * Return false on error, else true.
 */
static int ExecuteTimeout(request_t *r, struct timespec *next_expire,
                          policy_type next_expire_policy)
{
  if (next_expire == NULL) {
    /* Shouldn't get here */
    error("ExecuteTimeout", "r=0x%p: next_expire == NULL", r);
    return 0;
  }
          
  /* The timeout expired.  Call the appropriate policy's callback.*/
  /* Don't trigger another timeout unless one is set by the callback. */
  next_expire->tv_sec = 0;
  if (next_expire_policy == user_policy) {
    if (r->user_policy->timeout_expired(r) != user_policy_ok) {
      return 0;
    }
  }
  else if (next_expire_policy == transaction_policy) {
    if (r->transaction_policy->timeout_expired(r) !=
        transaction_policy_ok) {
      return 0;
    }
  }
  else if (next_expire_policy == unknown_policy) {
    /* The final timeout has triggered. */
    r->done = 1;
  }

  return 1;
} /* ExecuteTimeout */


/*
 * PolicyHandleMessage()
 * Handle a message on the message queue.  Dispatch it to the appropriate
 * policy callbacks.
 * Return false on error, else true. 
 */
static int PolicyHandleMessage(request_t *r, message_t *m)
{
  if (m->handled >= policy_handled) {
    /* This message is being stored by the policy for future use,
     * and doesn't need to be executed again. */
    return 1;
  }

  if (m->type != message_real_message && 
      m->type != message_pseudo_message) {
    /* This message shouldn't be on the request (policy) queue.  Delete it. */
    error("PolicyHandleMessage",
          "r=0x%p: Surprising message type %d on mqueue", r, m->type);
    m->handled = finished_handling;
    return 1;
  }

  if (m->p.type == http_request) {
    EXECUTE_IN_USER_AND_TRANSACTION(r, handle_subsequent_request(r, m),
                                    return 0);
  }
  else if (m->p.type == http_response) {
    /* For messages from call thread to userapi policy, m->branch is NULL
     * because there is no actual branch. */
    if (m->branch && m->p.status >= 200 &&
        ClientSentFinal(m->branch) && !m->branch->marked_done) {

      /* If this branch is done (i.e. it's received a final response),
       * decrement branches_outstanding if we haven't done so before
       * for this branch. This is done only for final responses. */

      debug("PolicyHandleMessage",
            "r=0x%p: final response (%d) from branch %d\n",
            r, m->p.status, m->branch->branch);
      r->branches_outstanding--;
      m->branch->marked_done = 1;
    }

    EXECUTE_IN_USER_AND_TRANSACTION(r, handle_response(r, m, m->branch),
                                    return 0);
  }
  else {
    error("PolicyHandleMessage",
          "Surprising message parse type %d on mqueue", m->p.type);
    m->handled = finished_handling;
  }

  return 1;
} /* PolicyHandleMessage */


/*
 * CheckForFinalResponse()
 * Check whether the transaction has just sent a final response.  If so,
 * set the request's final timeout to config.timeout from now, and cancel
 * any outstanding branches.
 */
static void CheckForFinalResponse(request_t *r)
{
  struct timeval now;
  int i;

  if (!r->sent_final || r->final_timeout.tv_sec != 0) {
    /* Either we haven't yet sent a final response, or this function's
     * already been called for this transaction. */
    return;
  }

  gettimeofday(&now, NULL);
  r->final_timeout = timespec_delay(now, (double)config.timeout);

  for (i = 0; i < r->branch_count; i++) {
    if (!ClientSentFinal(&r->branches[i])) {
      debug("CheckForFinalResponse", "r=0x%p: cancelling branch %d\n",
            r, r->branches[i].branch);
      policy_cancel_branch(r, &r->branches[i]);
    }
  }
} /* CheckForFinalResponse */


/*
 * CleanupMessageQueue()
 * Remove all nodes marked finished_handling from the request's
 * message queue.
 */
static void CleanupMessageQueue(request_t *r)
{
  struct message_t *m, *m_next, **prev_next;

  prev_next = &r->mqueue;
  for (m = r->mqueue; m != NULL; m = m_next) {
    m_next = m->next;
    if (m->handled == finished_handling) {
      *prev_next = m->next;
      MessageDelete(m);
    } else {
      prev_next = &m->next;
    }
  }
} /* CleanupMessageQueue */


/*
 * CleanupPolicyMachine()
 * Clean up all the state machinery associated with a transaction and policy.
 * Tell each branch to shut down, wait for the branch threads to terminate,
 * then call RequestFree().
 */
static void CleanupPolicyMachine(request_t *r)
{
  pthread_t *tid;
  int i, n;

  /* Alert the branches to shut down */
  for (i = 0; i < r->branch_count; i++) {
    if (!r->branches[i].shutdown) {
      MessageTrivialEnqueue(&r->branches[i].mlist, message_shutdown_client,
                            nobody_handled);
      pthread_cond_broadcast(&r->branches[i].response_received);
    }
  }

  /* Build a list of branch threads for WaitForMultipleObjects */
  n = 0;
  if (r->branch_count > 0) {
    tid = (pthread_t *) calloc(r->branch_count, sizeof(pthread_t));
    if (tid == NULL) {
      error("CleanupPolicyMachine", "Couldn't allocate tid");
    }
    else {
      for (i = 0; i < r->branch_count; i++) {
        if (r->branches[i].tid != 0) {
          tid[n++] = r->branches[i].tid;
        }
      }
    }
  }
  else {
    tid = NULL;
  }

  /* Need to free mutex before waiting for branches, or we deadlock. */
  /* NOTE: we have to make sure that RequestFree is *only* called from
   * here for transactions that have policies, or we have a race condition
   * which can lead to threads reading from freed memory. */
  pthread_mutex_unlock(&r->mutex);

  /* Wait for branches to complete */
  debug("CleanupPolicyMachine", "waiting for %d branch%s\n", n,
        (n != 1) ? "es" : "");

  if (n > 0) {
    WaitForMultipleObjects(n, tid, TRUE, INFINITE);
  }

  debug("CleanupPolicyMachine", "r=%p: policy %s/%s finished\n", r, 
        r->user_policy != NULL ? r->user_policy->name : "NULL",
        r->transaction_policy->name);

  free_valid(tid);
  RequestFree(r);
} /* CleanupPolicyMachine */


/*
 * SendComplaint()
 * Something went wrong in a policy.  If we haven't already sent a final
 * response, send a 500 error message. 
 */
static void SendComplaint(request_t *r)
{
  if (!r->sent_final) {
    policy_send_new_response(r, 500, "Internal Policy Error");
  }
} /* SendComplaint */


/* Functions a policy can call. */

/*
 * policy_done()
 * A policy calls this when it is finished with the transaction.  It tells
 * the core to clean things up.
 */
void policy_done(request_t *r, policy_type type)
{
  if (type == user_policy) {
    r->user_policy_done = 1;
  }
  else if (type == transaction_policy) {
    r->transaction_policy_done = 1;
  }
  else {
    error("policy_done", "r=0x%p: unknown policy type %d", r, type);
  }

  if (r->user_policy_done && r->transaction_policy_done && !r->sent_final) {
    SendComplaint(r);
  }
} /* policy_done */


/* 
 * policy_set_timeout()
 * A policy calls this to have the timeout callback called when the
 * specified time elapses.
 */
void policy_set_timeout(request_t *r, double delay, policy_type type)
{
  struct timeval now;
  struct timespec timeout;

  gettimeofday(&now, NULL);
  timeout = timespec_delay(now, delay);

  if (type == user_policy) {
    r->user_policy_timeout = timeout;
  }
  else if (type == transaction_policy) {
    r->transaction_policy_timeout = timeout;
  }
  else {
    error("policy_set_timeout", "r=0x%p: unknown policy type %d", r, type);
  }
} /* policy_set_timeout */
 


/*
 * policy_send_modified_new_response()
 * A policy calls this when it wants to send a locally-generated response
 * back to originating client.
 * status should usually not be 200-class, since this server isn't equipped
 * to behave as a UAC.  (XXX policy faking UAC?  Possible for CGI.)
 * reason may be NULL, in which case the standard reason will be filled in.
 * delta_headers contains headers to be added to, removed from, or changed
 * in the default set of response headers.
 * This response will be remembered until something else changes it.
 * Retransmissions of the request will receive retransmissions of the
 * most recent response.
 */
void policy_send_modified_new_response(request_t *r, int status, char *reason,
                                       headers_t *new_headers,
                                       headers_t *del_headers,
                                       body_t *body)
{
  int retransmit;
  char *method;

  method = HTTP_Method(&r->p);
  /* For new responses to INVITE, we retransmit all final responses. */
  retransmit = (status >= 200 && strcmp(method, "INVITE") == 0);
  free_valid(method);

  ResponseSetStatus(r, status, reason);
  ResponseBuildDefault(r);

  if (new_headers != NULL) {
    HeaderMerge(&r->headers, new_headers);
  }
  if (del_headers != NULL) {
    HeaderDeleteSet(&r->headers, del_headers);
  }

  /* Order here is important.  ResponseSetBody sets or overrides
   * Content-Type and Content-Length. */
  if (body != NULL) {
    ResponseSetBody(r, body->content, body->content_type,
                    body->content_length);
  }
  else {
    ResponseSetBody(r, NULL, NULL, 0);
  }

  if (status >= 200) {
    r->sent_final = 1;
    if (status < 300) {
      r->sent_200 = 1;
    }
  }

  ResponseSend(r, retransmit, 0);

  if (r->sent_final) {
    if (LogData) {
      (*LogData)(r);
    }
  }
} /* policy_send_modified_new_response */


/*
 * policy_send_new_response()
 * A wrapper around policy_send_modified_new_response for the
 * simple case when the caller doesn't want to modify the response.
 */
void policy_send_new_response(request_t *r, int status, char *reason)
{
  policy_send_modified_new_response(r, status, reason, NULL, NULL, NULL);
} /* policy_send_new_response */


/*
 * policy_forward_response()
 * Forward a received response upstream.
 * Set various flags as appropriate.
 * Generally the same characteristics apply here as apply in
 * send_new_response.
 */
void policy_forward_response(request_t *r, message_t *m)
{
  int retransmit;
  char *method;

  method = HTTP_Method(&r->p);

  /* For forwarded responses to INVITE, we retransmit 300-600's, but not
     200's; those are ACK'd end-to-end.
   */
  retransmit = (m->p.status >= 300 && strcmp(method, "INVITE") == 0);


  if (m->p.status >= 200) {
    r->sent_final = 1;
    if (m->p.status < 300) {
      r->sent_200 = 1;
    }
  }

  HeaderFree(&r->headers);

  /* RecordRoute transformation goes here..
   */
  if (config.RecordRoute && strcmp(method,"INVITE") == 0) {
    ModifyRRElement(m, r->hash);
  }

  free_valid(method);
  
  HeaderCopyParsedHeaders(&r->headers, m->p.headers);
  ResponseSetStatus(r, m->p.status, m->p.reason);
  ResponseSetBody(r, m->body.content, m->body.content_type,
                  m->body.content_length);
  /* Strip top Via field.  N.B.: we send our reply to the parent
   * address that originated the request, whether or not it matches
   * the next Via field.  XXX?
   */
  HeaderDeleteFirst(&r->headers, "Via");
  ResponseSend(r, retransmit, 0);

  if (r->sent_final) {
    if (LogData) {
      (*LogData)(r);
    }
  }
} /* policy_forward_response */


/*
 * policy_forward_user_response()
 * Forward a received response upstream.
 * Set various flags as appropriate.
 * Generally the same characteristics apply here as apply in
 * send_new_response.
 * It is assumed that the call thread has already
 * updated the headers as needed.
 */
void policy_forward_user_response(request_t *r, message_t *m)
{
  int retransmit;
  char *method;

  method = HTTP_Method(&r->p);

  /* For forwarded responses to INVITE, we retransmit all responses.
   * This is the only difference between this function and
   * policy_forward_respose(). Should these two be merged ? XXX.
   */
  /* KKK should we have all responses to INVITE retransmitted */
  retransmit = (/*m->p.status >= 300 && */ strcmp(method, "INVITE") == 0);
  free_valid(method);

  if (m->p.status >= 200) {
    r->sent_final = 1;
    if (m->p.status < 300) {
      r->sent_200 = 1;
    }
  }

  /* Do not rewrite the r->headers */
  //HeaderFree(&r->headers);
  //HeaderCopyParsedHeaders(&r->headers, m->p.headers);
  ResponseSetStatus(r, m->p.status, m->p.reason);
  /* Body is already set 
  ResponseSetBody(r, m->body.content, m->body.content_type,
                  m->body.content_length);
  */
  /* Do not stript the top via */
  //HeaderDeleteFirst(&r->headers, "Via");
  ResponseSend(r, retransmit, 0);

  if (r->sent_final) {
    if (LogData) {
      (*LogData)(r);
    }
  }
} /* policy_forward_user_response */


/*
 * increase_branch_space()
 * Internal function to increase the branch space for a request
 * Increases the branch space by at least one; grows the space used
 * exponentially for efficiency reasons.
 * Returns non-zero on failure.
 */
static int increase_branch_space(request_t *r)
{
#if 0
  /* XXX This doesn't work!  We pass the pointer to the branch_t to the
   * client; if we subsequently realloc(), the client's pointer points to
   * freed memory, and we blow up real good.  Sigh. */
  int old_branch_space = r->branch_space;

  /* This should only execute exactly once in normal circumstances,
   * but put it in a while loop for paranoia's sake. */
  while (r->branch_space <= r->branch_count) {
    if (r->branch_space == 0) {
      r->branch_space = 1;
    }
    else {
      r->branch_space *= 2;
    }
  }

  r->branches = (branch_t *) realloc(r->branches,
                                     r->branch_space * sizeof(branch_t));
  if (r->branches == NULL) {
    return 1;
  }

  /* Zero the memory for the newly-allocated branches (only) */
  memset(&r->branches[old_branch_space], '\0',
         (r->branch_space - old_branch_space) * sizeof(branch_t));

  return 0;
#else
  /* XXX Quick and dirty alternative. */
  if (r->branch_count == MAX_BRANCHES) {
    return 1;
  }
  if (r->branch_space == 0) {
    r->branches = calloc(MAX_BRANCHES, sizeof(branch_t));
    r->branch_space = MAX_BRANCHES;
  }
  return 0;
#endif
} /* increase_branch_space */


/* policy_proxy_original_request_to()
 * Proxy the original request of the transaction to the specified
 * URI.  This starts the client branch, and updates the request
 * data structures.
 */
void policy_proxy_original_request_to(request_t *r, uri_t *uri,
                                      message_t *m, void *policy_info,
                                      policy_type type)
{
  int branch_num;
  int thr_err;

  if (r->branch_space <= r->branch_count) {
    if (increase_branch_space(r) != 0) {
      error("policy_proxy_original_request_to",
            "couldn't increase branch space");
      return;
    }
  }

  branch_num = r->branch_count;
  r->branch_count++;

  r->branches[branch_num].branch      = branch_num;
  r->branches[branch_num].parent      = r;
  r->branches[branch_num].uri         = URI_Copy(*uri);
  r->branches[branch_num].orig        = MessageDuplicate(m);

  r->branches[branch_num].sent_200    = 0;
  r->branches[branch_num].sent_final  = 0;

  if (type == user_policy) {
    r->branches[branch_num].user_policy_info = policy_info;
    r->branches[branch_num].transaction_policy_info = NULL;
  }
  else if (type == transaction_policy) {
    r->branches[branch_num].user_policy_info = NULL;
    r->branches[branch_num].transaction_policy_info = policy_info;
  }
  else {
    error("policy_proxy_original_request_to", "r=0x%p: unknown policy type %d",
          r, type);
  }

  pthread_cond_init(&r->branches[branch_num].response_received, NULL);

  debug("policy_proxy_original_request_to",
        "Request %p: starting client %d, branch = 0x%p\n",
        r, branch_num, &r->branches[branch_num]);
        
  debug("pthread_create", "Creating client thread\n");
  /* NB: we do not create this thread detached, since we wait for it
   * when execute_policy() returns. */
  if ((thr_err = pthread_create(&r->branches[branch_num].tid, NULL, 
                                client, (void *)&r->branches[branch_num]))) {
    error("pthread_proxy_original_requst_to", "branch thread %d: %s (%d)\n",
          branch_num, strerror(thr_err), thr_err);
  } else {
    r->branches_outstanding++;
  }
} /* policy_proxy_original_request_to */


/* policy_proxy_subsequent_request_to()
 * Proxy a subsequent request (ACK, typically) on an existing branch.
 */
void policy_proxy_subsequent_request_to(request_t *r, branch_t *b,
                                        message_t *m)
{
  MessageEnqueue(&b->mlist, message_real_message, &m->p, &m->body,
                 &m->source, b, policy_handled);
  pthread_cond_broadcast(&b->response_received);
} /* policy_proxy_subsequent_request_to */


/* policy_cancel_branch()
 * Tell an existing branch to cancel itself.
 */
void policy_cancel_branch(request_t *r, branch_t *b)
{
  MessageTrivialEnqueue(&b->mlist, message_cancel_branch, policy_handled);
  pthread_cond_broadcast(&b->response_received);
} /* policy_proxy_subsequent_request_to */


