#ifndef REQUEST_SIP_H
#define REQUEST_SIP_H

#include <pthread.h>
#include <limits.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "rfc822.h"
#include "http.h"
#include "sip.h"
#include "uri.h"
#include "dstring.h"
#include "error.h"
#include "afi.h"                        /* AFI functions */
#include "headers.h"
#include "request-common.h"

#include "dbaccess.h"
#include "canonicalizer_interface.h"

/* A client branch. */
typedef struct branch_t {
  int branch;                /* branch index; -1 if none */
  struct request_t *parent;  /* parent request */
  pthread_t tid;             /* thread id for branch */
  uri_t uri;                 /* URI to try */

  struct message_t *orig;    /* Original request for this branch */
  struct message_t *mlist;   /* Message queue for this branch */
  int sent_200;              /* Have we sent a 200 response to parent? */
  int sent_final;            /* Have we sent a final response to parent? */

  pthread_cond_t response_received; /* alert to messages on the queue */  

  int marked_done;           /* has the parent noticed that we finished? */
  int shutdown;              /* Have we shut down our state machine? */

  int sent_something;        /* do we have something interesting to
                              * wake the parent up about? */

  void *user_policy_info;    /* User-policy-specific info for the branch */
  void *transaction_policy_info; /* Transaction-policy-specific info for the
                              * branch */

} branch_t;

/* The maximum number of branches for a request */
#define MAX_BRANCHES 32

/* This has to be included down here, after branch_t is declared */
#include "message.h"

extern void *client(void *arg);
extern int client_stateless(request_t *r, uri_t *uri, message_t *m);

/* List of pending requests. */
/* The name of this structure is somewhat of a misnomer, for historical
 * reasons (because sipd evolved from rtspd).  This actually contains
 * a) the state for a transaction, b) the initial request which created the
 * transaction, and c) the current response to the transaction.
 * 
 * NOTE: The top part of this structure (until "COMMON ENDS HERE") is a
 * shared structure defined in libcine/request-common.h as request_common_t.
 * DO NOT CHANGE THE TOP PART OF THIS STRUCTURE WITHOUT CHANGING THAT ONE AS
 * WELL! 
 */
struct request_t {
  /* Basic request information */
  double time_received;               /* time request was received */

  /* Where is the response going to? */
  int type;                           /* SOCK_DGRAM, SOCK_STREAM */
  int socket;                         /* socket id for response */
  struct sockaddr_storage source_sin; /* socket address request came from */
  struct sockaddr_storage sin;        /* socket address for response */

  /* Information about the request */
  parser_t p;                         /* header for this request */
  body_t request_body;                /* The body of the request; */

  http_authorization_t authorization; /* Authorization header */

  /* Information about the response */
  int status;
  char *reason;
  headers_t headers;                  /* Headers of response */
  body_t response_body;               /* Body to send with respnse */

  void *server;    /* server object associated with this request */
  /* COMMON ENDS HERE */

  /* SIP-specific information about the request */
  char *version;                      /* request version */

  /* SIP request headers */
  sip_cseq_t cseq;                    /* sequence number of request */
  char *callID;                       /* the call ID of the request */
  sip_addr_t to, from;                /* addresses */
  sip_via_t via;                      /* top-most Via header */
  time_t expires;                     /* Expires header */
  sip_contact_t *contact;             /* SIP Contact headers */
  char *user_agent;                   /* SIP User-Agent */

  unsigned long hash;                 /* Hash of identifying headers */

  int acked;                          /* request has been ACKed */

  /* SIP-specific Information about the response .*/
  http_authenticate_t authenticate;   /* SIP WWW-Authenticate (response) */

  /* Transaction state: */
  /* Locking. */
  pthread_cond_t cond;               /* condition variable */
  pthread_mutex_t mutex;             /* mutex for access */

  /* User information.  Used when necessary. */
  put_table_entry *source_user;      /* The originator of the transaction */
  put_table_entry *destination_user; /* The (logical) destination of the
                                      * transaction */

  /* the canonicalized request uri. check the status flag before
   * accessing the unique pointer for the canonicalized uri.
  */
  canon_result canonicalized_uri;

  /* Communication with subsequent requests and responses */
  message_t *mqueue;                  /* Message queue for this transaction */
  int sent_final;                     /* Whether we've sent a final
                                         response to the original request */
  int sent_200;                       /* Whether we've sent a 200 response
                                         to to the original request. */

  /* State for the policy core */
  int done;                           /* Whether the policy handler is
                                         done with this transaction. */
  struct timespec final_timeout;      /* Final timeout for the transaction */

  /* Transaction & user policies */
  struct transaction_policy_t *transaction_policy;
                                      /* Low-level policy to use for this
                                         transaction */
  struct timespec transaction_policy_timeout;
                                      /* Transaction policy timeout, or 0 */
  void *transaction_policy_info;      /* Info specific to a transaction
                                         policy */
  int transaction_policy_done;        /* Whether the transaction policy is done
                                         with this transaction. */

  struct user_policy_t *user_policy;  /* High-level policy to use for this
                                         transaction.  May be NULL. */
  struct timespec user_policy_timeout;/* User policy timeout, or 0 */
  void *user_policy_info;             /* Info specific to a user policy */
  int user_policy_done;               /* Whether the user policy is done
                                         with this transaction. */

  /* clients (proxy branches) */
  branch_t *branches;
  int branch_count;                   /* no of parallel searches (branches) */
  int branch_space;                   /* number of branch_t's allocated */
  int branches_outstanding;           /* current branches awaiting responses */

  void *call;                         /* call object associated with request*/
  struct sockaddr_storage remote_sin; /* needed for user agent in libsip++  */
  uri_t   outbound_proxy;             /* Address of the outbound proxy
                                         if any */ 
  int     preferred_type;             /* SOCK_DGRAM or SOCK_STREAM or
                                         0 as a preferred socket type
                                         for outgoing request,
                                         typically used by a user
                                         agent */

  struct request_t *next;             /* next request in queue */ 
};

/*
 * Init functions that must be called once during program startup.
 */
extern void InitLibSip();

/* Kills all libsip threads (in an unclean fashion). should be
 * used with extreme caution, only when abrupt libsip termination is needed. 
*/
extern void KillLibSipThreads();      

/* generate and handle requests */

/* sin must be non null */
extern request_t *RequestCreate(int type, int socket, 
                                const struct sockaddr_storage* sin);

extern void ResponseReceived(request_t *p, int responses);
extern void RequestFree(request_t *r);
extern void RequestDelete(request_t *r);
extern request_t *RequestSearch(request_t *r, unsigned long hash);
extern request_t *RequestFind(parser_t *p, unsigned long hash);
extern int RequestAdd(request_t *r);
extern void RequestProcess(request_t *r);

/* generate and handle responses */
extern const char *GetStandardReason(int status);
extern void ResponseSetBody(request_t *r, const char *body,
                            const char *content_type,
                            int content_length);
extern void ResponseSetStatus(request_t *r, int status, const char *reason);
extern int  ResponseBuildDefault(request_t *r);
extern int  ResponseCopy(request_t *dest, request_t *src);
extern int  ResponseSend(request_t *r, int retransmit, int cleanup);
extern void *ResponseProcessThread(void* ptr);

/* query stuff about clients */
extern int ClientSent200(branch_t *branch);
extern int ClientSentFinal(branch_t *branch);

/* used for api also */
extern unsigned long ComputeRequestHash(request_t *r);

/* This is an RTSP/SIP difference */
#define MISSING_CONTENT_LENGTH_MEANS_INFINITE

#endif
