/*
response.c - generate responses

Copyright 1998-1999 by Columbia University; all rights reserved 
*/

#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>    /* malloc() */
#include <sys/stat.h>  /* fstat() */
#include <time.h>      /* cftime(), time(), gmtime_r() */
#include <string.h>    /* strcasecmp() */
#include <math.h>      /* floor() */
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "log_interface.h"       /* (*LogData)() */
#include "config.h"
#include "dstring.h"   /* DString... */
#include "request-sip.h"
#include "headers.h"
#include "strdupn.h"   /* strassign() */
#include "timespec.h"
#include "sysdep.h"
#include "msgflow.h"
#include "libsip.h"


#include "sipsnmp_interface.h"

/*
 * Reasons why an errno is a "normal" message failure.  I.e., it's
 * failure to reach a host, rather than internal failure.
 */
#define ERRNO_NORMAL_FAILURE(err) \
   ((err) == ECONNREFUSED || \
    (err) == ECONNRESET || \
    (err) == ETIMEDOUT || \
    (err) == EHOSTUNREACH || \
    (err) == ENETUNREACH)


static char *m[700];

/*
 * ResponseInit()
 * Initialize the default reason phrases for different reason
 * status code.
 */
void ResponseInit(void)
{
  m[100] = "Trying";
  m[180] = "Ringing";
  m[181] = "Call Is Being Forwarded";
  m[182] = "Queued";
  m[183] = "Session Progress";
  m[200] = "OK";
  m[201] = "Created";
  m[300] = "Multiple Choices";
  m[301] = "Moved Permanently";
  m[302] = "Moved Temporarily";
  m[305] = "Use Proxy";
  m[380] = "Alternative Service";
  m[400] = "Bad Request";
  m[401] = "Unauthorized";
  m[402] = "Payment Required";
  m[403] = "Forbidden";
  m[404] = "Not Found";
  m[405] = "Method Not Allowed";
  m[406] = "Not Acceptable";
  m[407] = "Proxy Authentication Required";
  m[408] = "Request Time-out";
  m[409] = "Conflict";
  m[410] = "Gone";
  m[411] = "Length Required";
  m[412] = "Precondition Failed";
  m[413] = "Request Entity Too Large";
  m[414] = "Request-URI Too Large";
  m[415] = "Unsupported Media Type";
  m[420] = "Bad Extension";
  m[480] = "Temporarily unavailable";
  m[481] = "Invalid Call-ID";
  m[482] = "Loop detected";
  m[483] = "Too many hops";
  m[484] = "Address incomplete";
  m[485] = "Ambiguous";
  m[486] = "Busy here";
  m[487] = "Request cancelled";
  m[488] = "Not acceptable here";
  m[500] = "Internal Server Error";
  m[501] = "Not Implemented";
  m[502] = "Bad Gateway";
  m[503] = "Service Unavailable";
  m[504] = "Gateway Time-out";
  m[505] = "SIP Version not supported";
  m[600] = "Busy";
  m[603] = "Decline";
  m[604] = "Does Not Exist Anywhere";
  m[606] = "Not Acceptable";
} /* ResponseInit */


/*
 * GetStandardReason()
 * Get the standard reason phrase for this status code
 */
const char *GetStandardReason(int status)
{
  if (status < 100 || status >= 700) {
    return "Unknown status";
  }
  if (m[status] != NULL) {
    return m[status];
  }
  else {
    /* Round to the nearest x00 status. */
    status = ((int)(status / 100)) * 100;
    return m[status];
  }
} /* GetStandardReason */


/*
* Send response to network.
* N.B. the fact that response-sending (send() or sendto()) is atomic
* is important here, since several threads can write to the same socket.
* Don't break it up!
*/
static int ResponseSendSocket(request_t *r, DString msg)
{
  int n = 0;       /* bytes read or written */
  int retval = 0;  /* return value */


  if ( r->socket != (-1) ) { /* Needed to catch error */
    /* Valid socket */

    debug("ResponseSendSocket", "Sending %d bytes to socket %d via %s\n%s\n",
          msg.length, r->socket, r->type == SOCK_DGRAM ? "UDP" : "TCP",
          DStringValue(&msg));

    DRAW_MESSAGE_FLOW(
    {
      DString msg;
      DStringInit(&msg);
      DStringPrintf(&msg, "%d %s", r->status, 
        r->reason == NULL ? GetStandardReason(r->status) :
        r->reason);
      DrawMessageFlow("SIPAPI","This","Remote", DStringValue(&msg));
      DStringFree(&msg);
    });

    /* For UDP, we send the response back to the address indicated in the
     Via header, via a new socket. */
    if (r->type == SOCK_DGRAM) {
      if (r->via.sin.ss_family == AF_UNSPEC) {
        error("ResponseSendSocket", "Missing Via");
        retval = -1;      
      } else {

        if (AFIGetPort(&r->via.sin) == 0) { /* Default port is 5060 */
          AFISetPort(&r->via.sin, 5060);
        }
        debug2("ResponseSendSocket", 
               (struct sockaddr*)&r->via.sin, "is remote address\n");

        n = sendto(r->socket, msg.string, msg.length, 0, 
                  (struct sockaddr *)&r->via.sin, 
                  (r->via.sin.ss_family == AF_INET6) ? 
                  sizeof(struct sockaddr_in6) : 
                  sizeof(struct sockaddr_in));
      }
    }
    else {
      n = send(r->socket, msg.string, msg.length, 0);
    }
    if (n < 0) {
      error("ResponseSendSocket", "type=%d", r->type);
      retval = -1;
    }
 
  }
  else {
    /* Socket is NULL */
    error("ResponseSendSocket","socket is -1\n");
  }

  if (SnmpProcessMessage) {
    (*SnmpProcessMessage)(r, SNMP_TYPE_RES_OUT, NULL);
  }

  return retval;
} /* ResponseSendSocket */


/* 
 * ResponseBuildDefault()
 * Given a request, create the standard response headers for it and add
 * them to the header structure.
*/
int ResponseBuildDefault(request_t *r)
{
  parser_header_t *h;  /* request header */
  parser_element_t *e;

  /* If there are previous headers here, delete them. */
  HeaderFree(&r->headers);

  /* Echo Via headers. */
  if ((h = RFC822_Header(r->p.headers, "Via")) && h->elements){
    for (e = h->elements; e; e = e->next) {
      HeaderAppend(&r->headers, "Via", e->element);
    }
  }

  /* Echo Record-Route. */
  if ((h = RFC822_Header(r->p.headers, "Record-Route")) && h->elements){
    for (e = h->elements; e; e = e->next) {
      HeaderAppend(&r->headers, "Record-Route", e->element);
    }
  }

  /* Copy From, To, Timestamp, Call-ID from request. */
  /* TODO: Add To tags. */
  if ((h = RFC822_Header(r->p.headers, "From")) && h->elements)
    HeaderAppend(&r->headers, "From", h->elements->element);
  if ((h = RFC822_Header(r->p.headers, "To")) && h->elements)
    HeaderAppend(&r->headers, "To", h->elements->element);
  if ((h = RFC822_Header(r->p.headers, "Timestamp")) && h->elements)
    HeaderAppend(&r->headers, "Timestamp", h->elements->element);
  HeaderAppend(&r->headers, "Call-ID", r->callID);

  HeaderPrintf(&r->headers, 0, "Cseq", "%lu %s", (long)r->cseq.seq, 
    r->cseq.method ? r->cseq.method : "");
  HeaderCftime(&r->headers, "Date", "%a, %d %b %Y %H:%M:%S GMT", 0, 0);
  if ( server_name != NULL ) {
    HeaderPrintf(&r->headers, 0, "Server", "%s/%s", server_name,
                 VERSION);
  }
  
  ResponseSetBody(r, NULL, NULL, 0);
  return 0;
} /* ResponseBuildDefault */


/*
* ResponseSetBody()
* Set body and body content properties for request or response. If
* content_length is < 0, use length of body.
*/
void ResponseSetBody(request_t *r, const char *body,
                     const char *content_type, int content_length)
{
  strassign(&r->response_body.content,      body);
  strassign(&r->response_body.content_type, content_type);
  r->response_body.content_length = content_length;

  /* Content-Type, Content-Length */
  if (r->response_body.content_length < 0)
    r->response_body.content_length =
      r->response_body.content ? strlen(r->response_body.content) : 0;
  if (content_type && r->response_body.content_length > 0) {
    HeaderReplaceOrAppend(&r->headers, "Content-Type",
                          r->response_body.content_type);
  }
  else {
    HeaderDelete(&r->headers, "Content-Type");
  }
  HeaderDelete(&r->headers, "Content-Length");
  HeaderPrintf(&r->headers, 1, "Content-Length", "%d",
    r->response_body.content_length);
} /* ResponseSetBody */


/*
* ResponseSetStatus()
* Set the status and reason phrase.
*/
void ResponseSetStatus(request_t *r, int status, const char *reason)
{
  r->status = status;
  strassign(&r->reason, reason);
} /* ResponseSetStatus */


/*
 * ResponseCopy()
 * dest is a request which duplicates src, except (possibly) for its
 * Via: headers.  Generate appropriate Via headers for dest, then copy
 * all the other attributes of the response (status, reason, headers, body).
 */
int ResponseCopy(request_t *dest, request_t *src)
{
  parser_header_t *h;  /* request header */
  parser_element_t *e;

  HeaderFree(&dest->headers);

  dest->status = src->status;
  strassign(&dest->reason, src->reason);

  /* Dest echos its own Via headers. */
  if ((h = RFC822_Header(dest->p.headers, "Via")) && h->elements){
    for (e = h->elements; e; e = e->next) {
      HeaderAppend(&dest->headers, "Via", e->element);
    }
  }

  HeaderCopyAllBut(&dest->headers, &src->headers, "Via");

  ResponseSetBody(dest, src->response_body.content,
    src->response_body.content_type,
    src->response_body.content_length);

  return 0;
} /* ResponseCopy */


/*
* ResponseSend()
* Generate a response message. If 'retransmit' is true, retransmit until
* ACKed, if an INVITE request.
* If 'cleanup' is true, this will release the request's mutex and
* free the request. 
* N.B.: this thread must have the request's mutex locked!
*/
int ResponseSend(request_t *r, int retransmit, int cleanup)
{
  DString msg;            /* to make this MT-safe, new string for each call */
  int retval = 0;         /* return value */
  int repeat = 7;         /* # times to send response (or ACKed): 7 */
  float delay = 0.5;      /* initial retransmit delay, in seconds */
  struct timeval now;     /* current time */
  struct timespec next;   /* next response transmission */
  char *method;           /* The method of the request */

  DStringInit(&msg);

  /* Status line = SIP-Version SP Status-Code SP seq-no SP Reason-Phrase CRLF */
  DStringPrintf(&msg, "SIP/%s %d %s\r\n",
    r->version ? r->version : "2.0",  r->status, 
    r->reason ? r->reason : GetStandardReason(r->status));

  HeaderMessage(&msg, &r->headers);
  DStringAppend(&msg, "\r\n", -1);
  DStringAppend(&msg, r->response_body.content,
    r->response_body.content_length);

  if ( r->socket != -1 ) {
    /*
     * No need to print the message here as it is printed in
     * ResponseSendSocket() 
     */
    debug("ResponseSend", "r=%p, r->socket = %d\n",
        r, r->socket);
  }

  method = HTTP_Method(&r->p);
  /* Send non-INVITE and 1xx responses only once. */
  if (method == NULL || strcmp(method, "INVITE") != 0
      || r->status < 200 || !retransmit) {
    repeat = 1;
    delay = 0.;
  }

  /* 
   * Retransmit final responses. Can't use usleep() in a threaded
   * program, so we use pthread_cond_timewait.  When an ACK or
   * CANCEL arrives, this thread will be woken up.
   */
  gettimeofday(&now, NULL);
  for (; !r->acked && repeat > 0 && retval >= 0; repeat--, delay *= 2) {
    int c;
    
    retval = ResponseSendSocket(r, msg);
    if (repeat > 1 && retval >= 0) {
      next = timespec_delay(now, delay);
      do {
        c = pthread_cond_timedwait(&r->cond, &r->mutex, &next);
      } while (!r->acked && c != ETIMEDOUT);
    }
  }

  /* If we're cleaning up the response, take this opportunity to
   * log it. */
  if (cleanup) {
    if (LogData) {
      (*LogData)(r);
    }
  }

  /* If final retransmitted response, and we're cleaning up here, hold onto
   * results for 30 seconds, in case a retransmit shows up. */
  if (r->status >= 200 && retransmit && cleanup) {
    next = timespec_delay(now, delay + 30);
    pthread_cond_timedwait(&r->cond, &r->mutex, &next);
  }

  /*
   * For now, we free the headers after each response even though they
   * get re-used from the provisional response.  This avoids tricky
   * definitions of 'permanent' and 'temporary' headers.  Strictly
   * speaking, each response should also use the 'Timestamp' and 'Via'
   * values of the last idempotent request.
   */
  DStringFree(&msg);
  free_valid(method);
  if (cleanup) {
    pthread_mutex_unlock(&r->mutex);
    RequestFree(r);
  }
  return retval;
} /* ResponseSend */

/*
 * ResponseForward()
 * Forward the response statelessly to the second via header.
 * The first Via header is removed from the response.
 */
void ResponseForward(request_t *r)
{
  sip_via_t via;   /* Second via */
  DString reqmsg;  /* Response message */
  char *version;   /* version 2.0 */
  char *protocol;  /* protocol, SIP */
  headers_t hdrs;  /* list of response headers */
  int n; 

  /* Get the second via header */
  memset(&via, 0, sizeof(via));
  via = SIP_Via(&r->p, 2, 1, NULL, 0);

  debug2("ResponseForward", (struct sockaddr*)&via.sin, 
         "Response(0x%p) has a second Via\n", r);
        

  DStringInit(&reqmsg);

  /* Perform initial transmission */
  version  = HTTP_Version(&r->p);
  protocol = HTTP_Protocol(&r->p);
  
  hdrs.headers = NULL;
  hdrs.last = NULL;

  /* Print the response line */  
  DStringPrintf(&reqmsg, "%s/%s %u %s\r\n", 
                protocol, version, r->p.status, r->p.reason ?
                r->p.reason : "<No reason>");
      
  /* If the response to be proxied is missing a Content-Length
   * header field, add one.  Needed for UDP -> TCP proxying. */
  if (RFC822_Header(r->p.headers, "Content-Length") == NULL) {
    HeaderPrintf(&hdrs, 0, "Content-Length", "%d", 
                 r->request_body.content_length);
  }
      
  /* This is a kludge; I should write a HeaderCopyAllParsedHeadersBut()
   * function for efficiency.*/
  HeaderCopyParsedHeaders(&hdrs, r->p.headers);
  HeaderDeleteFirst(&hdrs, "Via");

  HeaderMessage(&reqmsg, &hdrs);
  DStringAppend(&reqmsg, "\r\n", 2);
  DStringAppend(&reqmsg, r->request_body.content, 
                r->request_body.content_length);
  
  free_valid(version);
  free_valid(protocol);
  HeaderFree(&hdrs);

  debug("client_stateless", "Sending out response \n%s\n",
        reqmsg.string);
  if ((n = sendto(r->socket, reqmsg.string, reqmsg.length, 0,
                  (struct sockaddr *) &via.sin, sizeof(via.sin))) < 0) {
    /* If there was a problem, print out an error message. */
    if (!(ERRNO_NORMAL_FAILURE(errno))) {
      error("ResponseForward", "Error sending proxied response\n");
    }
    else {
      debug("ResponseForward",
            "Normal error sending proxied response: %s (%d)\n",
            strerror(errno), errno);
    }
  }

  free_valid(via.hostname);
  free_valid(via.branch.str);
  DStringFree(&reqmsg);
} /* ResponseForward */

/*
 * ResponseProcessThread()
 * Process an incoming SIP response.
 * This is the thread which actually does the processing work.
 */
void *ResponseProcessThread(void* ptr)
{
  request_t *r = (request_t *) ptr;
  request_t *rp;               /* corresponding outgoing request */
  int branch_id = -1;          /* numeric branch identifier */
  parser_header_t *h;
  char *protocol;

  if (SnmpProcessMessage) {
    sipsnmp_extra_data extra_data;
    sip_cseq_t cseq;    /* we need the method for which this is a response */
    cseq = SIP_CSeq(&r->p);
    strncpy(extra_data.request_type, cseq.method, 
            sizeof(extra_data.request_type));
    (*SnmpProcessMessage)(r, SNMP_TYPE_RES_IN, &extra_data);
  }

  protocol = HTTP_Protocol(&r->p);
  if (strcmp(protocol, "SIP") != 0) {
    /* Response isn't SIP */
    goto fail;
  }
  if (!SIP_CallID(&r->p)) {
    /* Response doesn't have a CallID */
    goto fail;
  }
  r->cseq         = SIP_CSeq(&r->p);
  r->callID       = strdup(SIP_CallID(&r->p));
  if (r->source_sin.ss_family == AF_INET6) {
    r->via          = SIP_Via(&r->p, 1, 1, config.ProxyName6, config.port);
  }
  else {
    r->via          = SIP_Via(&r->p, 1, 1, config.ProxyName, config.port);
  }
  r->to           = SIP_To(&r->p);
  r->from         = SIP_From(&r->p);

  if (AFIIsAnyHost((struct sockaddr*)&r->via.sin)) {
    /* Missing or invalid top Via header. */
    goto fail;
  }

  if ((h = RFC822_Header(r->p.headers, "Content-Type")))
    r->request_body.content_type = h->elements->value;

  DRAW_MESSAGE_FLOW(
  {
    char msg[100];
    sprintf(msg,"%d %s",r->p.status,GetStandardReason(r->p.status));
    DrawMessageFlow("SIPAPI","Remote","This", msg);
  });

  /* Find the parent request that this response corresponds to. */
  rp = RequestSearch(r, r->via.branch.hash);
  debug("ResponseProcessThread", "Response rp=%p r=%p\n", rp, r);

  if (rp) {
    /* 
     * We do not use negative branch id.
     * An invalid negative value if set to 0 for robustness
     */
    if ( r->via.branch.id < 0 ) {
      debug("ResponseProcessThread", "Response:0x%p Branchid(%u)<0\n", 
            r, r->via.branch.id);
      r->via.branch.id = 0;
    }

    if (r->via.branch.id >= 0 && r->via.branch.id < rp->branch_count
        && r->via.islocal) {
      /*
       * This response is for a branch of ours.  We enqueue the response in
       * the branch's queue, and signal it.
       */
      branch_id = r->via.branch.id;

      MessageEnqueue(&rp->branches[branch_id].mlist, message_real_message,
                     &r->p, &r->request_body, &r->source_sin,
                     &rp->branches[branch_id], nobody_handled);
      debug("ResponseProcessThread",
           "Request %p, branch %d: Signalling response_received, status = %d\n",
           rp, branch_id, r->p.status);
      pthread_cond_broadcast(&rp->branches[branch_id].response_received);
    }
    
    pthread_mutex_unlock(&rp->mutex);
  }
  else {
    debug("ResponseProcessThread",
          "Response r=%p: no matching request\n", r);

    if ( config.TryStateless ) {
      /* Forward the response to the Via header */
      /* Make sure that the branch id is 0 since we do not use
       * branch id for stateless proxy */
      
      if ( r->via.branch.id == 0 &&
           r->via.branch.hash == 0 &&
           r->via.islocal) {
        debug("ResponseProcessThread",
              "stateless response\n");
        ResponseForward(r);
      }
      else {
        /* Ignore the unsolicited response */
        debug("ResponseProcessThread", "Ignoring; branch=%u-%lu, islocal=%d\n",
              r->via.branch.id,   r->via.branch.hash,
              r->via.islocal);
      }
    }
  }

fail:
  /* For whatever reason, we couldn't figure out anything to do
   * with this response, so just ignore it.
   */
  RequestFree(r);
  free(protocol);
  return NULL;
} /* ResponseProcessThread */
