/*
sip.c - Parser for SIP-specific headers

Copyright 1998, 1999, 2000 by Columbia University; all rights reserved 
*/

#include <stdlib.h>  /* atoi() */
#include <string.h>  /* strchr(), strncpy() */
#include <ctype.h>   /* isspace(), isalpha() */
#include <stdio.h>   /* sscanf() */
#include <time.h>    /* struct tm */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#ifndef HAVE_TIMEGM
#include "timegm.h"
#endif

#include "rfc822.h"
#include "http.h"
#include "sip.h"
#include "host2ip.h"
#include "shortform.h"
#include "config.h"
#include "error.h"
#include "strdupn.h"
#include "sysdep.h"


/* Global const values */
const int default_sip_port = 5060;

static const uri_nv_t sip_default_params[] = {
  {"user", "ip"},
  {"transport", "udp"},
  {"ttl", "1"},
  {"method", "INVITE"},
  {NULL, NULL}
};

const uri_defaults_t sip_uri_defaults[] = {
  {"sip", 5060, sip_default_params},
  {NULL, 0, NULL}
};


/* A table used by libcine for this protocol */
const shortform_map Proto_Shortform_Map[] =
{
  {"Content-Type", "c"},
  {"Content-Encoding", "e"},
  {"From", "f"},
  {"Call-ID", "i"},
  {"Supported", "k"},
  {"Contact", "m"},
  {"Content-Length", "l"},
  {"Subject", "s"},
  {"To", "t"},
  {"Via", "v"},
  {NULL, NULL}
};


/*
 * SIP_NameAddr()
 * Break up name-addr (To, From, Contact, Route) addresses:
 * '[name] <url>' or 'url'.  Return false if the argument isn't
 * of that form.  DStrings must already be initialized.
 * addr may have surrounding whitespace and/or <> remaining; URI_Parse
 * can handle this.
 */
int SIP_NameAddr(char *v, DString *name, DString *addr)
{
  char *s, *uri_start = 0;
  int backslash = 0;

  if (v) {
    /* "name" <...> */
    if (*v == '"') {
      for (s = v+1; *s; s++) {
        if (*s == '\\') {
          backslash = 1;
        }
        else if (*s != '"' || backslash) {
          DStringAppend(name, s, 1);
          backslash = 0;
        }
        /* reached end of quoted string (") */
        else {
          break;
        }
      }
      DStringAppend(addr, s+1, -1);
    }
    /* name <...> */
    else if ((uri_start = strchr(v, '<')) != NULL) {
      /* remove any trailing white space after name */
      for (s = uri_start-1; (s > v) && isspace((int)*s); s--) ;
      DStringAppend(name, v, s - v + 1);
      DStringAppend(addr, uri_start, -1);
    }
    /* URL only */
    else {
      DStringAppend(addr, v, -1);
    }
  }
  return 1;
} /* SIP_NameAddr */


/*
 * Parse name-addr (To, From, Contact, Route) addresses:
 * '[name] <url>' or 'url'.
 */
sip_addr_t sip_addr(char *v)
{
  sip_addr_t a;

  DString name, addr;

  if (!v) 
    goto badaddr;

  DStringInit(&name);
  DStringInit(&addr);
  if (!SIP_NameAddr(v, &name, &addr))
    goto badaddr;

  a.uri = URI_Parse(DStringValue(&addr));
  a.display_name = DStringValue(&name);

  DStringFree(&addr);
  /* "name" remains, sip_addr_t.display_name is dynamic */

  return a;

badaddr:
  memset(&a.uri, 0, sizeof(a.uri));
  a.display_name = NULL;

  return a;
} /* sip_addr */


/*
* Parse HTTP-date or delta-seconds.
* If valid, return it, else 0.
*/
static time_t date_delta(char *v)
{
  time_t t = 0;

  if (v) {
    if (isalpha((int)v[0])) {
      struct tm tm;

      if (Parse_HTTP_Date(v, &tm) == NULL) {
        debug("date_delta", "Could not parse the date\n");
      }
      else {
        t = timegm(&tm);
        if (t == -1) t = 0;
      }
    }
    else {
      t = atoi(v);
      if (t > 0) t += time(0);      
      else t = 1;
    }
  }
  return t;
} /* date_delta */


/*
* Call-ID: local-id@host
*/
char *SIP_CallID(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Call-ID");

  if (h && h->elements)
    return h->elements->value;
  else
    return NULL; 
} /* SIP_CallID */


/*
 * SIP_CSeq()
 * CSeq: 1234 INVITE
 * Return CSeq description.
*/
sip_cseq_t SIP_CSeq(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "CSeq");
  sip_cseq_t cseq;

  strcpy(cseq.method, "");
  cseq.seq = 0;
  if (h && h->elements->value) {
    sscanf(h->elements->value, 
           "%u %20s", (unsigned int *)&cseq.seq, cseq.method);
  }
  return cseq;
} /* SIP_CSeq */


/*
 * SIP_Contact()
 * Contact: John Doe <sip:foo@bar;user=phone> ;q=0.5 ;action=proxy ;expires=3
 * Create an array of Contact headers. Last Contact has NULL for uri.scheme
*/
sip_contact_t *SIP_Contact(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Contact");
  parser_element_t *e;
  parser_parameter_t *pp;
  sip_addr_t addr;
  int n = 0;
  sip_contact_t *loc = NULL;
  char *v;

  /* Allocate storage. */
  if (h) {
    for (e = h->elements; e; e = e->next) n++;
    loc = (sip_contact_t *)calloc(n+1, sizeof(sip_contact_t)); 

    for (e = h->elements, n = 0; e && e->value; e = e->next, n++) {
      pp = e->parameters;
      addr = sip_addr(e->value);

      loc[n].uri = addr.uri;
      loc[n].nameaddr = SIP_BuildNameAddr(addr.display_name, &addr.uri);

      free_valid(addr.display_name);

      v = RFC822_Attribute1(pp, "q");
      loc[n].q = v ? atof(v) : 1.0;

      v = RFC822_Attribute1(pp, "action");
      if (!v) loc[n].action = L_unspecified;
      else if (strcasecmp(v, "proxy")    == 0) loc[n].action = L_proxy; 
      else if (strcasecmp(v, "redirect") == 0) loc[n].action = L_redirect;
      else loc[n].action = L_unspecified;
      /* XXX: should L_unspecified be config.defaultreg instead? */

      v = RFC822_Attribute1(pp, "expires");
      loc[n].expires = date_delta(v);

      v = RFC822_Attribute1(pp, "methods");
      if (v) strassign(&loc[n].methods, v);
    }
  }
  return loc;
} /* SIP_Contact */


/*
 * SIP_BuildContact()
 * Print a sip_contact_t to a string.
 * The caller must free the string.
 */
char *SIP_BuildContact(const sip_contact_t *loc)
{
  DString locstr;

  DStringInit(&locstr);

  /* Assuming the nameaddr field was constructed by SIP_BuildNameAddr(), it
   * should always be in bracketed form. */
  DStringAppend(&locstr, loc->nameaddr, -1);
  
  if (loc->expires == 1) {
    /* A special case returned by date_delta, above */
    DStringAppend(&locstr, "; expires=\"0\"", -1);
  }
  else if (loc->expires != 0) {
    struct tm tm;
    gmtime_r(&loc->expires, &tm);

    DStringAppend(&locstr, "; expires=\"", -1);
    DStringStrftime(&locstr, "%a, %b %d %Y %H:%M:%S GMT", &tm);
    DStringAppend(&locstr, "\"", -1);
  }

  if (loc->action == L_proxy) {
    DStringAppend(&locstr, "; action=proxy", -1);
  }
  else if (loc->action == L_redirect) {
    DStringAppend(&locstr, "; action=redirect", -1);
  }

  /* XXX: should q == 0.0 be printed? */
  if (loc->q >= 0.0 && loc->q <= 1.0) {
    DStringPrintf(&locstr, "; q=%.2f", loc->q);
  }

  return DStringValue(&locstr);
} /* SIP_BuildContact */


/*
 * SIP_FreeContact()
 * Free a sip_contact_t structure
 */
void SIP_FreeContact(sip_contact_t *loc)
{
  int n;

  if (!loc) return;

  for (n = 0; loc[n].uri.scheme != NULL; n++) {
    URI_Free(&loc[n].uri);
    free_valid(loc[n].nameaddr);
    free_valid(loc[n].comment);
    free_valid(loc[n].methods);
  }
  free(loc);
} /* SIP_FreeContact */


/*
* SIP_Expires()
* Expires: 45 or
* Expires: Wed, Oct 15, 1997 16:30:25 GMT
* Return 0 if no 'Expires' header and 1 if 'Expires: 0'
*/
time_t SIP_Expires(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Expires");

  return h ? date_delta(h->elements->value) : 0;
} /* SIP_Expires */


/*
* SIP_From()
* From: John Foo <sip:foo@bar.com>
* F:    John Foo <sip:foo@bar.com>
*/
sip_addr_t SIP_From(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "From");
  return sip_addr(h->elements->value);
} /* SIP_From */


/*
* SIP_To()
* To: John Foo <sip:foo@bar.com>
* T:  Jonn Foo <sip:foo@bar.com>
*/
sip_addr_t SIP_To(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "To");
  return sip_addr(h->elements->value);
} /* SIP_To */

/*
* SIP_Subject()
* Subject: This is an example subject
* S:  This is an example subject
*/
const char *SIP_Subject(parser_t *p)
{
  parser_header_t *h = RFC822_Header(p->headers, "Subject");

  if (h) { 
    return h->elements->value;
  }
  else {
    return NULL;
  }
} /* SIP_To */


/*
 * SIP_Has_Mandatory_Headers()
 * Make sure To, From, CSeq, and Call-ID are present.  Return false if not.
 * Via is also a mandatory header, but we can be more lenient about that.
 */
int SIP_Has_Mandatory_Headers(parser_t *p)
{
  return (RFC822_Header(p->headers, "To") != NULL &&
          RFC822_Header(p->headers, "From") != NULL &&
          RFC822_Header(p->headers, "CSeq") != NULL &&
          RFC822_Header(p->headers, "Call-ID") != NULL);
} /* SIP_Has_Mandatory_Headers */


/*
 * SIP_Via()
 * Via: SIP/2.0/UDP host:5060
 * Retrieve 'order'th Via header (1...).  Allow UDP to be omitted and
 * allow spaces between SIP, 2.0, and UDP and other components.  Return
 * Via description.  If 'resolve' is true, resolve host name to IP
 * address. If 'localhost' is not NULL, it sets the 'islocal' flag in
 * the return value if the Via header is indeed the local host. Set the
 * 'branch' structure's branch identifiers -- just the token if !islocal,
 * otherwise the whole thing.
 * XXX Should use maddr.
 */
sip_via_t SIP_Via(parser_t *p, int order, int resolve, char *localhost,
                  unsigned short localport)
{
  parser_element_t *e = RFC822_Element_Ordered(p->headers, "Via", order);
  sip_via_t via;
  int port;
  char host[256] = { 0 };
  char *branch   = NULL;
  int numeric    = 0;           /* is Via address a numeric address? */

  memset(&via, 0, sizeof(via));
  via.type       = SOCK_DGRAM;
  AFISetPort(&via.sin, default_sip_port);

  if (e) {
    char *value = e->value;

    /* protocol (must be "SIP" followed by /) */
    value = &value[strspn(value, " \t\r\n")];
    if (strncasecmp(value, "SIP", 3) != 0) {
      errno = 0;
      error("SIP_Via", "Invalid protocol %s", value);
      return via;
    }
    value += 3;
    value = &value[strspn(value, " \t\r\n")];
    if (value[0] != '/') {
      errno = 0;
      error("SIP_Via", "Invalid syntax %s", value);
      return via;
    }
    value++;
    value = &value[strspn(value, " \t\r\n")];

    /* version (must be "2.0") */
    if (strncmp(value, "2.0", 3) != 0) {
      errno = 0;
      error("SIP_Via", "Invalid version %s", value);
      return via;
    }
    value += 3;
    value = &value[strspn(value, " \t\r\n")];

    /* protocol ("UDP" or "TCP") */
    if (value[0] == '/') {
      value++;
      value = &value[strspn(value, " \t\r\n")];
      if (strncasecmp(value, "UDP", 3) == 0) {
        value += 3;
        via.type = SOCK_DGRAM;
      } else if (strncasecmp(value, "TCP", 3) == 0) {
        value += 3;
        via.type = SOCK_STREAM;
      } else {
        errno = 0;
        error("SIP_Via", "Invalid transport %s", value);
        return via;
      }
    }
    value = &value[strspn(value, " \t\r\n")];

    /* parse host */
    /* IPv6 addresses can be either a host name (which is
     * simple) or can be 'colon' addresses. the colon
     * address can have 8 sets of 4 digits each, 
     * and 7 colons and will be wrapped by two
     * square brackets.
     */
    if (value[0] == '[') { /* IPv6 */
      char *save = value;
      ++value;
      sscanf(value, "%255[0-9a-fA-F:.]", host);
      value += strlen(host);
      if (value[0] != ']') {
        errno = 0;
        error("SIP_Via", "Invalid IPv6 address %s\n", save);
        return via;
      }
      ++value;  /* skip the trailing ] */
      
      /* translate the ipv6 address into a 16-byte notation.
       * AFITranslateNumeric will take care of setting the family.
      */
      if (AFITranslateNumeric(&via.sin, host) != 0) {
        errno = 0;
        error("SIP_Via", "Bad address %s\n", host);
        return via;
      }
      numeric = 1;
    }
    else {
     struct in_addr addr;
     sscanf(value, "%255[A-Za-z0-9.-]", host);
     value += strlen(host);
     /* If it is a IPv4 address then family = AF_INET. Note that 
      * there is also no need to call AFITranslateNumeric as 
      * inet_addr suffices for IPv4.
      */
     addr.s_addr = inet_addr(host);
     if (addr.s_addr != -1) {
       struct sockaddr_in *p4 = (struct sockaddr_in*) &via.sin;
       p4->sin_family = AF_INET;
       p4->sin_addr = addr;
       numeric = 1;
     }
    }

    if (resolve && !numeric) { /* address is not numeric, resolve == 1.
                                * dont specify via.sin.ss_family. 
                                * this will cause AFIGetAddrByName to choose 
                                * the best address.
                                */
      if (AFIGetAddrByName(&via.sin, host) != 0) {
        errno = 0;
        error("SIP_Via", "Invalid host %s\n", host);
        return via;
      }
    }
    
    if (host[0]) {
      strassign(&via.hostname, host);
    }

    value = &value[strspn(value, " \t\r\n")];

    /* port */
    if (value[0] == ':') {
      value++;
      value = &value[strspn(value, " \t\r\n")];
      sscanf(value, "%d", &port);
      AFISetPort(&via.sin, port);
    }

    if (localhost) {
      if (localhost[0] == '[') { /* IPv6 dotted address */
        via.islocal = (strstr(&localhost[1], host) == &localhost[1]);
      }
      else {
        via.islocal = (strcasecmp(localhost, host) == 0);
      }
    }
    else via.islocal = 0;  

    if (via.islocal && localport != 0) {
      /*
       * Check if the local port number is same as via port;
       * If not this is not local request.
       */
      if (AFIGetPort(&via.sin) != localport) {
        via.islocal = 0;
      }
    }

    via.branch.hash = 0;
    via.branch.id = -1;

    branch = RFC822_Attribute1(e->parameters, "branch");
    strassign(&via.branch.str, branch);

    if (branch) {
      if (via.islocal) {
        /* Try to interpret branch tag as hash - branch ID pair */
        if (sscanf(branch, "%lu-%d", &via.branch.hash, &via.branch.id) < 2) {
          /* sscanf failed, reset */
          via.branch.hash = 0;
          via.branch.id = -1;
        }
      }
    }
  }
  return via;
} /* SIP_Via */


/*
 * SIP_ViaEqual()
 * Check whether 'via1' and 'via2' describe the same via header.  Used
 * to detect request merging.
 */
int SIP_ViaEqual (const sip_via_t *via1, const sip_via_t *via2)
{
  if (via1->type != via2->type) {
    return 0;
  }

  if (AFIComparePort(&via1->sin, &via2->sin) == 0) {
    return 0;
  }

  if (strcompare(via1->branch.str, via2->branch.str, 0) != 0) {
    return 0;
  }

  if (strcompare(via1->hostname, via2->hostname, 1) != 0) {
    return 0;
  }

  /* XXX: maddr, ttl */

  return 1;
} /* SIP_ViaEqual */


/*
 * SIP_BuildNameAddr()
 * Given a name and address, print the nameaddr to a string.
 * It is the caller's responsibility to free the string.
 *
 * This is used to build Contact headers.
 */
char *SIP_BuildNameAddr(const char *name, const uri_t *uri)
{
  DString nameaddr, addrstr;

  DStringInit(&nameaddr);
  DStringInit(&addrstr);

  if (name && *name) {
    /* Display name not null or empty */
    DStringAppend(&nameaddr, "\"", -1);
    DStringAppendQuoted(&nameaddr, name, "\\\"");
    DStringAppend(&nameaddr, "\" ", -1);
  }

  URI_SPrint(uri, &addrstr);

  DStringAppend(&nameaddr, "<", -1);
  DStringAppendD(&nameaddr, &addrstr);
  DStringAppend(&nameaddr, ">", -1);

  DStringFree(&addrstr);

  return DStringValue(&nameaddr);
} /* SIP_BuildNameAddr */
